
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <fstream>
#include <string>
#include <list>

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "sct_parselib/parselib.h"
#include "sct_parselib/tokens_to_types.h"
#include "sct_parselib/rule.h"
#include "sct_parselib/parser.h"
#include "sct_parselib/sstreamwrap.h"

#include "gtksct.h"

#define	GLADE_XML_FILE_PATH		"gtksct.glade"
#define	GLADE_XML_TOP_WIN		"gtksct_window"
#define	GLADE_XML_RULES_CLIST		"rules_clist"
#define	GLADE_XML_SYSCALLNAME_ENTRY	"syscallname_entry"
#define	GLADE_XML_RULEID_ENTRY		"ruleid_entry"
#define	GLADE_XML_WHEN_COMBO		"when_combo"
#define	GLADE_XML_FILTER_ENTRY		"filter_entry"
#define	GLADE_XML_ACTION_COMBO		"action_combo"

const char DEFAULT_CONFIG_FILE_NAME[] = "../sct_config/sct_example.conf";

struct gtksct_data {
  GladeXML* win_glade;
  GtkWidget* win;
  GtkWidget* rules_clist;
  GtkWidget* syscallname_entry;
  GtkWidget* ruleid_entry;
  GtkWidget* when_combo;
  GtkWidget* filter_entry;
  GtkWidget* action_combo;
  list<rule>* rules;
};

/* forward references. */
gboolean gtksct_on_delete_window_event_cb(GtkWidget* widget,
				      	  GdkEvent* event,
				      	  gpointer data);
void gtksct_on_rules_clist_select_row_cb(GtkCList* clist,
					 gint row,
					 gint column,
					 GdkEventButton* event,
					 gpointer data);
gboolean gtksct_add_rules_to_gui(gtksct_t self);
void gtksct_show_rule_data(gtksct_t self, rule* rule);
bool parse_load_config_file(list<rule>* rules);
bool load_config_file(const char * config_file_name,
		      sstreamwrap& data);
bool parse_config_file(sstreamwrap& data,
		       list<rule>* rules, 
		       const std::string& config_file_name);
rule* gtksct_find_rule_by_name(gtksct_t self, const char* rule_name);

/* temporary global debug variable - until the library's ridden of it. */
int debug = 0;

/*
 * Public functions.
 */

gtksct_t gtksct_new()
{
  gtksct_t self = NULL;

  self = (gtksct_t)malloc(sizeof(struct gtksct_data));
  if (!self) {
    fprintf(stderr, "gtksct_new: cannot alloc 'self'\n");
    goto gtksct_new_ret;
  }

  self->rules = new list<rule>;
  if (!self->rules) {
    fprintf(stderr, "gtksct_new: cannot alloc 'self->rules'\n");
    goto gtksct_new_err_rules;
  }

  if (!parse_load_config_file(self->rules)) {
    goto gtksct_new_err_parseconf;
  }

  self->win_glade =
    glade_xml_new(GLADE_XML_FILE_PATH, GLADE_XML_TOP_WIN);
  if (!self->win_glade) {
    fprintf(stderr,
	    "gtksct_new: failed loading glade file '%s', window '%s'\n",
	    GLADE_XML_FILE_PATH, GLADE_XML_TOP_WIN);
    goto gtksct_new_err_glade;
  }

  self->win = glade_xml_get_widget(self->win_glade, GLADE_XML_TOP_WIN);
  self->rules_clist =
    glade_xml_get_widget(self->win_glade, GLADE_XML_RULES_CLIST);
  self->syscallname_entry =
    glade_xml_get_widget(self->win_glade, GLADE_XML_SYSCALLNAME_ENTRY);
  self->ruleid_entry =
    glade_xml_get_widget(self->win_glade, GLADE_XML_RULEID_ENTRY);
  self->when_combo =
    glade_xml_get_widget(self->win_glade, GLADE_XML_WHEN_COMBO);
  self->filter_entry =
    glade_xml_get_widget(self->win_glade, GLADE_XML_FILTER_ENTRY);
  self->action_combo =
    glade_xml_get_widget(self->win_glade, GLADE_XML_ACTION_COMBO);

  gtk_widget_set_sensitive(self->when_combo, FALSE);
  gtk_widget_set_sensitive(self->action_combo, FALSE);

  gtk_signal_connect(
          GTK_OBJECT(self->win),
          "delete_event",
          GTK_SIGNAL_FUNC(gtksct_on_delete_window_event_cb),
          self);
  gtk_signal_connect(
          GTK_OBJECT(self->rules_clist),
          "select_row",
          GTK_SIGNAL_FUNC(gtksct_on_rules_clist_select_row_cb),
          self);


  /* fill GUI with config file's data. */
  gtksct_add_rules_to_gui(self);

  gtk_widget_show(self->win);

  goto gtksct_new_ret;

 gtksct_new_err_glade:
 gtksct_new_err_parseconf:
  delete(self->rules);
 gtksct_new_err_rules:
  free(self);
  self = NULL;
 gtksct_new_ret:
  return self;
}

void gtksct_delete(gtksct_t self)
{
  if (self) {
    gtk_widget_destroy(self->win);
    gtk_object_unref(GTK_OBJECT(self->win_glade));

    free(self);
  }
}


/*
 * Private functions.
 */

/*
 * GTK Callback functions.
 */

/* process 'delete_event' signals for the top-level window, by deleting it. */
gboolean
gtksct_on_delete_window_event_cb(GtkWidget* widget,
				 GdkEvent* event,
				 gpointer data)
{
  gtksct_t self = (gtksct_t)data;

  gtksct_delete(self);

  gtk_main_quit();

  /* tell gtk not to destroy our top-levelwindow - we already did. */
  return TRUE;
}

/* process 'select_row' signals emitted for the rules_clist in the window. */
/* causes the rule's details to be shown on the right part of the window.  */
void
gtksct_on_rules_clist_select_row_cb(GtkCList* clist,
				    gint row,
				    gint column,
				    GdkEventButton* event,
				    gpointer data)
{
  gtksct_t self = (gtksct_t)data;
  gchar* rule_name = NULL;

  if (gtk_clist_get_text(GTK_CLIST(self->rules_clist), row, 0, &rule_name)) {
    rule* rule = gtksct_find_rule_by_name(self, rule_name);
    if (rule) {
      gtksct_show_rule_data(self, rule);
    }
  }
}


/*
 * GUI contents manipulation functions.
 */

/* fill GUI with config file's data. */
gboolean
gtksct_add_rules_to_gui(gtksct_t self)
{
  std::string rule_name;
  gchar* rule_name_for_clist[1];

  list<rule>::const_iterator it = self->rules->begin();
  for (; it != self->rules->end(); ++it) {
    rule_name = (*it).get_rule_name();
    rule_name_for_clist[0] = (char*)rule_name.c_str();
    gtk_clist_append(GTK_CLIST(self->rules_clist), rule_name_for_clist);
  }

  return TRUE;
}

void
gtksct_show_rule_data(gtksct_t self, rule* rule)
{
  char buf[40];
  const char* syscallname = rule->get_syscall_name().c_str();
  const char* filter = rule->get_filter_as_string().c_str();

  gtk_entry_set_text(GTK_ENTRY(self->syscallname_entry), syscallname);

  sprintf(buf, "%d", rule->get_rule_id());
  gtk_entry_set_text(GTK_ENTRY(self->ruleid_entry), buf);

  if (rule->get_rule_when() == RW_BEFORE)
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(self->when_combo)->entry), "before");
  else
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(self->when_combo)->entry), "after");

  /* stringify the filter. */
  gtk_entry_set_text(GTK_ENTRY(self->filter_entry), filter);

  /* TODO - check the action type, and stringify it. */
  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(self->action_combo)->entry),
		     sct_get_action_as_string(rule->get_action()));
}

/*
 * Config file handling functions.
 */

bool parse_load_config_file(list<rule>* rules)
{
  const char* config_file_name = DEFAULT_CONFIG_FILE_NAME;
  sstreamwrap data;

  if (!load_config_file(config_file_name, data)){
    fprintf(stderr,
	    "parse_load_config_file: failed to read the config file '%s'.\n", 
	    config_file_name);
    return 0;
  }

  if (!parse_config_file(data, rules, config_file_name)) {
    fprintf(stderr,
	    "parse_load_config_file: failed to parse the config file.\n");
    return 0;
  }

  return 1;
}

bool load_config_file(const char * config_file_name,
		      sstreamwrap& data)
{
    //first of all, make some sanity checks on the file

    //preferably, that user will be root, but we relax that requirement for
    //development
    struct stat stat_info;
    if (stat(config_file_name, &stat_info)==-1){
	fprintf(stderr,
		"load_config_file: stat '%s' failed, with error '%s'\n",
		config_file_name, strerror(errno));
	return false;
    }

    //the config file is a regular file
    if (!S_ISREG(stat_info.st_mode)){
	fprintf(stderr,
		"load_config_file: file '%s' is not a regular file\n", 
		config_file_name);
	return false;
    }

    //now, that we've verified the file is usable by us, let's load
    //its contents into a buffer.
    ifstream in(config_file_name);
    in>>data.rdbuf();
    return true;
}

bool 
parse_config_file(sstreamwrap& data,
		  list<rule>* rules,
		  const std::string& config_file_name)
{
    /* init ths parselib */ 
    if (sct_parselib_init()) {
	    fprintf(stderr, "Initializing the parser library failed\n"); 
	    return 0; 
    }

    //create a new parser
    parser p(data, config_file_name);
    //start parsing the config data
    return p.parse(*rules);
}


/*
 * Rule data manipulation function.
 */

/* find the rule with the given name, in our rules list. */
rule*
gtksct_find_rule_by_name(gtksct_t self, const char* rule_name)
{
  rule* a_rule = NULL;  /* assume failure. */
  const char* curr_rule_name = NULL;
  list<rule>::iterator it = self->rules->begin();

  for (; it != self->rules->end(); ++it) {
    curr_rule_name = (*it).get_rule_name().c_str();
    if (curr_rule_name && strcmp(curr_rule_name, rule_name) == 0) {
      a_rule = &(*it);
      break;
    }
  }

  return a_rule;
}
