/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/*
 * $Id: special_hijack.c,v 1.4 2002/09/14 19:35:36 mulix Exp $
 */

/* 
 * This file has the hijacking functions for special system calls, such 
 * as exit(), execve() or fork(). 
 */ 

#include "linux/sched.h"
#include "asm/uaccess.h"

asmlinkage static int stub_syscall_fork(struct pt_regs regs) 
{
	typedef int (*stub_func_fork)(struct pt_regs regs);
	typedef int (*stub_before_func_fork)(int* syscall_rc); 
	typedef int (*stub_after_func_fork)(int syscall_rc);
	stub_func_fork orig_syscall_func;
	stub_before_func_fork before_func;
	stub_after_func_fork after_func;
	sct_syscall_retval_policy retval; /* after calling the before */
					  /* func, should we continue */
					  /* and call the original    */
					  /* syscall?                 */
	int fake_syscall_retval = 0;  /* the 'before' function will */
			              /* fill this variable for us  */
				      /* if it wants to 'fail' the  */
				      /* syscall                    */
	int orig_syscall_retval = 0;  /* the original syscall retval */

	struct syscall_hijack_data* hdata = &hijack_data[2];

	atomic_inc(&hdata->invocations); 

	down(&hdata->sem);
	before_func = (stub_before_func_fork)hdata->before_hijack_func;
	orig_syscall_func = (stub_func_fork)hdata->orig_syscall;
	after_func = (stub_after_func_fork)hdata->after_hijack_func;
	up(&hdata->sem);

	retval = (before_func ? (*before_func)(&fake_syscall_retval)
		  : SCT_USE_ORIG_RETVAL);

	/* invoke the original syscall, if the hijack function returned 0. */
	if (retval == SCT_USE_ORIG_RETVAL) {
		orig_syscall_retval = (orig_syscall_func ? (*orig_syscall_func)(regs) : -ENOSYS);
		if (after_func) 
			(*after_func)(orig_syscall_retval);
	}

	down(&hdata->sem);
	if (atomic_read(&hdata->invocations) == 1 && hdata->defered_release_flag == 1) {
	    /* we couldn't reduce the module's use count when the syscall  */
	    /* was originally released (as marked by the 'defered' flag) - */
	    /* so we do so now.                                            */
	    hdata->defered_release_flag = 0;
	    MOD_DEC_USE_COUNT;
	}
	up(&hdata->sem);

	if (retval == SCT_USE_ORIG_RETVAL)
		retval = orig_syscall_retval; 
	else 
		retval = fake_syscall_retval;

	atomic_dec(&hdata->invocations); 

	return retval; 
}

/* stub function definitions for syscall 'execve' (11). */
/* my apologies in advance to the style police. */ 
asmlinkage static int stub_syscall_execve(struct pt_regs regs) 
{
	typedef int (*stub_before_func_11)(char*, char**, char**, int* syscall_rc); 
	typedef int (*stub_after_func_11)(int syscall_rc, char*, char**, char**);
	stub_before_func_11 before_func;
	stub_after_func_11 after_func;
	sct_syscall_retval_policy retval; /* after calling the before */
					  /* func, should we continue */
					  /* and call the original    */
					  /* syscall?                 */
	int fake_syscall_retval = 0;  /* the 'before' function will */
			              /* fill this variable for us  */
				      /* if it wants to 'fail' the  */
				      /* syscall                    */
	int orig_syscall_retval = 0;  /* the original syscall retval */

	struct syscall_hijack_data* hdata = &hijack_data[11];
	
	atomic_inc(&hdata->invocations); 

	down(&hdata->sem);
	before_func = (stub_before_func_11)hdata->before_hijack_func;
	after_func = (stub_after_func_11)hdata->after_hijack_func;
	up(&hdata->sem);
	
	retval = (before_func
		  ? (*before_func)((char*)regs.ebx, (char**)regs.ecx, 
				   (char**)regs.edx, &fake_syscall_retval)
		  : SCT_USE_ORIG_RETVAL);
	
	/* invoke the original syscall, if the hijack function returned 0. */
	if (retval == SCT_USE_ORIG_RETVAL) {
		int error;

		/* arch/i386/kerne/process.c: sys_execve() */ 
		char * filename;
		
		lock_kernel();
		filename = getname((char *) regs.ebx);
		error = PTR_ERR(filename);
		if (IS_ERR(filename))
			goto out; 

		error = do_execve(filename, (char **) regs.ecx, 
				  (char **) regs.edx, &regs);
		if (error == 0)
#ifdef PF_DTRACE
			current->flags &= ~PF_DTRACE;
#else
			current->flags &= ~PT_DTRACE;
#endif
	out: 
		orig_syscall_retval = error; 
		putname(filename); 
		unlock_kernel();

		if (after_func) {
#if 0
                        /* we have a problem here. 'filename' was originally in the OLD process */
                        /* address space, which the original execve syscall obliterated.  */
                        /* it is not found in the new process address space, so it must be  */
                        /* passed as a kernel space value which we save - but the tracking */
                        /* code firmly expects a user space ptr. therefore, disable it for now.  */
			(*after_func)(orig_syscall_retval, 				
				      /* filename, */ 
				      /* argv, */ 
				      /* envp */);
#endif /* 0 */  
			printk(KERN_NOTICE "sorry, after hooks for execve for i386 "
			       "not implemented yet\n"); 
		}
	}

	/* calling unconditional down() is never a good idea but we   */
	/* *must* grab this lock and lower the ref count, or we risk  */
	/* having it too high by one, leading to being unable to      */
	/* unload the module. also, none of the other holders of this */
	/* lock can block, so down() should be ok.                    */
	down(&hdata->sem);
	if (atomic_read(&hdata->invocations) == 1 && hdata->defered_release_flag == 1) {
	    /* we couldn't reduce the module's use count when the syscall  */
	    /* was originally released (as marked by the 'defered' flag) - */
	    /* so we do so now.                                            */
	    hdata->defered_release_flag = 0;
	    MOD_DEC_USE_COUNT;
	}
	up(&hdata->sem);
	
	
	if (retval == SCT_USE_ORIG_RETVAL)
		retval = orig_syscall_retval; 
	else 
		retval = fake_syscall_retval;

	atomic_dec(&hdata->invocations); 

	return retval; 
}
	
