/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/*
 * $Id: special_hijack.c,v 1.4 2002/09/14 19:35:37 mulix Exp $
 */

/* 
 * This file has the hijacking functions for special system calls, such 
 * as exit(), execve() or fork(). 
 */ 

#include "linux/sched.h"

/* stub function definitions for syscall 'fork' (2). */
/* my apologies in advance to the style police. */ 
asmlinkage static int
stub_syscall_fork(void) {
	typedef int (*stub_func_fork)(void);
	typedef int (*stub_before_func_fork)(int* syscall_rc); 
	typedef int (*stub_after_func_fork)(int syscall_rc);
	stub_func_fork orig_syscall_func;
	stub_before_func_fork before_func;
	stub_after_func_fork after_func;
	sct_syscall_retval_policy retval; /* after calling the before */
					  /* func, should we continue */
					  /* and call the original    */
					  /* syscall?                 */
	int fake_syscall_retval = 0;  /* the 'before' function will */
			              /* fill this variable for us  */
				      /* if it wants to 'fail' the  */
				      /* syscall                    */
	int orig_syscall_retval = 0;  /* the original syscall retval */

	struct syscall_hijack_data* hdata = &hijack_data[2];

	atomic_inc(&hdata->invocations); 

	down(&hdata->sem);
	before_func = (stub_before_func_fork)hdata->before_hijack_func;
	orig_syscall_func = (stub_func_fork)hdata->orig_syscall;
	after_func = (stub_after_func_fork)hdata->after_hijack_func;
	up(&hdata->sem);

	retval = (before_func ? (*before_func)(&fake_syscall_retval)
		  : SCT_USE_ORIG_RETVAL);

	/* invoke the original syscall, if the hijack function returned 0. */
	if (retval == SCT_USE_ORIG_RETVAL) {
		orig_syscall_retval = (orig_syscall_func ? (*orig_syscall_func)() : -ENOSYS);
		if (after_func) 
			(*after_func)(orig_syscall_retval);
	}

	down(&hdata->sem);
	if (atomic_read(&hdata->invocations) == 1 && hdata->defered_release_flag == 1) {
	    /* we couldn't reduce the module's use count when the syscall  */
	    /* was originally released (as marked by the 'defered' flag) - */
	    /* so we do so now.                                            */
	    hdata->defered_release_flag = 0;
	    MOD_DEC_USE_COUNT;
	}
	up(&hdata->sem);

	if (retval == SCT_USE_ORIG_RETVAL)
		retval = orig_syscall_retval; 
	else 
		retval = fake_syscall_retval;

	atomic_dec(&hdata->invocations); 

	return retval; 
}

/* stub function definitions for syscall 'execve' (11). */
/* my apologies in advance to the style police. */ 
asmlinkage static int
stub_syscall_execve(char* file, char** argv, char** env) {
	typedef int (*stub_func_execve)(char* file, char** argv, char** env);
	typedef int (*stub_before_func_execve)(char*, char**, char**, int* syscall_rc); 
	typedef int (*stub_after_func_execve)(int syscall_rc, char*, char**, char**);
	stub_func_execve orig_syscall_func;
	stub_before_func_execve before_func;
	stub_after_func_execve after_func;
	sct_syscall_retval_policy retval; /* after calling the before */
					  /* func, should we continue */
					  /* and call the original    */
					  /* syscall?                 */
	int fake_syscall_retval = 0;  /* the 'before' function will */
			              /* fill this variable for us  */
				      /* if it wants to 'fail' the  */
				      /* syscall                    */
	int orig_syscall_retval = 0;  /* the original syscall retval */
	
	struct syscall_hijack_data* hdata = &hijack_data[11];
	
	atomic_inc(&hdata->invocations); 

	down(&hdata->sem);
	before_func = (stub_before_func_execve)hdata->before_hijack_func;
	orig_syscall_func = (stub_func_execve)hdata->orig_syscall;
	after_func = (stub_after_func_execve)hdata->after_hijack_func;
	up(&hdata->sem);
	
	retval = (before_func
		  ? (*before_func)(file, argv, env, &fake_syscall_retval)
		  : SCT_USE_ORIG_RETVAL);
	
	/* invoke the original syscall, if the hijack function returned 0. */
	if (retval == SCT_USE_ORIG_RETVAL) {
		orig_syscall_retval = (orig_syscall_func ? 
				       (*orig_syscall_func)(file, argv, env) : -ENOSYS);
		
		if (after_func) {
#if 0
                        /* we have a problem here. 'file' was originally in the OLD process */
                        /* address space, which the original execve syscall obliterated.  */
                        /* it is not found in the new process address space, so it must be  */
                        /* passed as a kernel space value which we save - but the tracking */
                        /* code firmly expects a user space ptr. therefore, disable it for now.  */
			(*after_func)(orig_syscall_retval, file, argv, env);
#endif /* 0 */  
			printk(KERN_NOTICE "sorry, after hooks for execve for uml "
			       "not implemented yet\n"); 
		}
	}

	/* calling unconditional down() is never a good idea but we   */
	/* *must* grab this lock and lower the ref count, or we risk  */
	/* having it too high by one, leading to being unable to      */
	/* unload the module. also, none of the other holders of this */
	/* lock can block, so down() should be ok.                    */
	down(&hdata->sem);
	if (atomic_read(&hdata->invocations) == 1 && hdata->defered_release_flag == 1) {
	    /* we couldn't reduce the module's use count when the syscall  */
	    /* was originally released (as marked by the 'defered' flag) - */
	    /* so we do so now.                                            */
	    hdata->defered_release_flag = 0;
	    MOD_DEC_USE_COUNT;
	}
	up(&hdata->sem);
	
	
	if (retval == SCT_USE_ORIG_RETVAL)
		retval = orig_syscall_retval; 
	else 
		retval = fake_syscall_retval;

	atomic_dec(&hdata->invocations); 

	return retval; 
}
	
