/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il             */
/* License: GNU General Public License                              */

/*
 *$Id: pidtrace.h,v 1.3 2002/11/25 21:43:45 mulix Exp $
 */

/*
 * This code is used to track system call invocations in the hijacker 
 * module so that we'll be able to answer the question "which processes
 * are using this system call currently?". It is 
 * a) only useful when the module can be unloaded (debug builds)
 * b) impossible to get right in a module. If an application gets a 
 * sigsegv in the middle of a system call, it will die with our stub 
 * function still on the stack and will cause a small memory leak. 
 * 
 * Therefore, this code is only activated when SCT_MODULE_DEBUG is defined. 
 */ 

#ifndef _MODULE_HIJACK_PIDTRACE_H_
#define _MODULE_HIJACK_PIDTRACE_H_ 

#include "syscall_hijack_private.h" 

#define PROC_SYSCALLS_DEBUG "syscalls"

struct syscall_pid {
	struct list_head list; 
	pid_t pid; 
};

#undef SCT_MODULE_DEBUG 

#ifdef SCT_MODULE_DEBUG

static inline struct syscall_pid* setup_pid_trace(void)
{
        /* allocate and set up the pid_trace */ 
	struct syscall_pid* p; 
	if ((p = kmalloc(sizeof(*p), GFP_KERNEL))) {
		memset(p, 0, sizeof(*p)); 
	        p->pid = current->pid; 
	        INIT_LIST_HEAD(&p->list); 
	} else 
	        SCTDBG(DEB_HIJACK, "oom trying to allocate the "
		       "pid_trace struct!\n"); 

	return p; 
}

static inline void add_pid_trace(struct syscall_pid* p, 
				 struct syscall_hijack_data* hdata)
{
	assert(hdata); 
	
	if (!p)
		return; 

	list_add(&p->list, &hdata->pid_list); 
}

static inline void remove_pid_trace(struct syscall_pid* p, 
				   struct syscall_hijack_data* hdata)
{
	struct list_head* tmp; 

	assert(hdata); 

	if (!p) 
		return; 

	list_for_each(tmp, &hdata->pid_list) {
		struct syscall_pid* ptr = list_entry(tmp, struct syscall_pid, 
						     list); 
		if (ptr == p) {
			if (p->pid != ptr->pid)
				BUG(); 
			list_del(tmp); 
			kfree(p); 
			break; 
		}
	}
}

extern int syscalls_read_proc(char* page, char** start, off_t off, 
			      int count, int* eof, void* data); 

static inline int register_syscalls_proc_entry(struct proc_dir_entry* dir)
{
	struct proc_dir_entry* syscall; 

        /* register for "which process is blocking on which syscall" */ 
	if (!(syscall = create_proc_read_entry(PROC_SYSCALLS_DEBUG, 0644, dir, 
					       syscalls_read_proc, NULL))) {
		printk(KERN_WARNING "proc: creating %s failed\n", 
		       PROC_SYSCALLS_DEBUG); 
		return -1; 
	}
	syscall->owner = THIS_MODULE; 
	return 0; 
}
	
#else /* SCT_MODULE_DEBUG not defined */ 


static inline struct syscall_pid* setup_pid_trace(void)
{
	return NULL; 
}

static inline void add_pid_trace(struct syscall_pid* p, 
				 struct syscall_hijack_data* hdata)
{
}

static inline void remove_pid_trace(struct syscall_pid* p, 
				   struct syscall_hijack_data* hdata)
{
}

static inline int register_syscalls_proc_entry(struct proc_dir_entry* dir)
{
	return 0; 
}

#endif /* SCT_MODULE_DEBUG not defined */ 

#endif /* _MODULE_HIJACK_PIDTRACE_H_ */ 
