/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il             */
/* License: GNU General Public License                              */

/*
 *$Id: syscall_hijack.c,v 1.7 2002/09/19 16:22:25 mulix Exp $
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sys.h>
#include <linux/stddef.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/proc_fs.h> 
#include <linux/list.h> 
#include <linux/smp_lock.h> 
#include <asm/semaphore.h>

#include "hijack/syscall_hijack_public.h"
#include "common/module_interface.h" 
#include "common/sct_compat.h" 
#include "common/util.h"

#include "syscall_hijack_private.h" 
#include "pidtrace.h" 

MODULE_AUTHOR("guy keren <choo@actcom.co.il>, "
	      "Muli Ben-Yehuda <mulix@actcom.co.il>");

MODULE_DESCRIPTION("System Call Hijacking and Releasing, "
		   "for 'syscalltrack'");

MODULE_LICENSE("GPL");

MODULE_PARM(allow_unloading, "i"); 
MODULE_PARM_DESC(allow_unloading, "set to true to allow the module "
		 "to be unloaded"); 

static int allow_unloading = 1; /* defaults to 1 while debugging */ 

MODULE_PARM(debug_level, "l"); 
MODULE_PARM_DESC(debug_level, "set to bigger than 0 for debug message."); 

unsigned long debug_level = 1; /* default to 1 for debugging */

#define PROC_MODULE_NAME "sct_hijack" 

/* hijack data for all syscalls. */
static struct syscall_hijack_data hijack_data[NR_syscalls];

extern func_ptr* sys_call_table[NR_syscalls];

/* unfortunately, we need to be aware of 'sct_syscall_retval_policy',  */
/* which is currently defined in sct_rules. we want to have it defined */
/* someplace neutral, so we dont have a dependancy on sct_rules. then  */
/* again, this would require reworking the entire stub mechanism...    */
#include "sct_rules/rule_engine_public.h"

/* ok, include the autogenerated stuff... */
#include "syscall_hijack_autogen.c"
#include "archdep_syscalls.h"
#include "special_hijack.c"

/* hijack the given system call. returns '0' on success, */
/* '1' if the syscall is already hijacked, and           */
/* a negative int on error.                              */
int hijack_syscall(int syscall_id,
		   func_ptr function,
		   hijack_func_when_t when)
{
	int retval = -ENOSYS;
	int rc;
	int defered_flag = 0;
	struct syscall_hijack_data* hdata = NULL;

	if (syscall_id < 0 || syscall_id >= NR_syscalls) {
		printk(KERN_NOTICE "hijack_syscall: invalid syscall_id '%d'\n",
			syscall_id);
		return retval;
	}
    
	hdata = &hijack_data[syscall_id];

	if ((rc = down_interruptible(&hdata->sem)) < 0)
		return -ERESTARTSYS;

	
        /* remove the 'defered release flag' if it was set,   */
	/* but remember it was set, for later use.            */
	/* this flag might be set here if we marked this call */
	/* for defered release, and before the release could  */
	/* happen, we are re-hijacking it.                    */
	if (hdata->defered_release_flag) {
		defered_flag = 1;
		hdata->defered_release_flag = 0;
	}

	/* if the syscall has active invocations, and a hijck function */
	/* is still registered, then we cannot hijack it again - have  */
	/* to wait until its unregistered, or until it has no active   */
	/* invocations.                                                */
	if ((atomic_read(&hdata->invocations) > 0) && hdata->orig_syscall) {
		printk(KERN_NOTICE "hijack_syscall: syscall %d busy\n",
			syscall_id);
		retval = 1;
		goto hijack_syscall_release_lock;
	} 
    
	SCTDBG(DEB_HIJACK, "beginning syscall %d hijack\n", syscall_id);
	if (!stub_functions[syscall_id]) {
		printk(KERN_NOTICE "hijack_syscall: %d: unsupported syscall\n",
			syscall_id);
		retval = -ENOENT;
		goto hijack_syscall_release_lock;
	}
    
	if (!hdata->orig_syscall) { /* only hijack if we dont have it already hijacked */
		func_ptr orig_syscall;

		func_ptr* sys_call_table_addr = (func_ptr*)
			get_module_symbol(NULL, versioned_string(sys_call_table)); 
		if (!sys_call_table_addr) {
			printk(KERN_NOTICE "hijack_syscall: %d: couldn't locate "
				"sys_call_table\n", syscall_id);
			retval = -ENOENT;
			goto hijack_syscall_release_lock;
		}
	
		orig_syscall = sys_call_table_addr[syscall_id];
		
		/* if we had a defered release for this syscall earlier, */
		/* we didn't DEC the use count of the module, and thus   */
		/* shouldn't INC it now.                                 */
		if (!defered_flag) {
			MOD_INC_USE_COUNT;
		}
	
		/* put into the syscall table the stub function */
		sys_call_table_addr[syscall_id] =
			(func_ptr)stub_functions[syscall_id];

		/* put into the hijacking functions table the user */
		/* supplied function.				   */
		if (when == HFW_BEFORE)
			hdata->before_hijack_func = function;
		else
			hdata->after_hijack_func = function;
	
		/* put the original syscall into the orig syscalls table */
		hdata->orig_syscall = orig_syscall;
	
		retval = 0;
	}
	else if (when == HFW_BEFORE && !hdata->before_hijack_func) {
		hdata->before_hijack_func = function;
		retval = 0; 
	}
	else if (when == HFW_AFTER && !hdata->after_hijack_func) {
		hdata->after_hijack_func = function;
		retval = 0;
	}
	else { 
		printk(KERN_NOTICE "hijack_syscall: %d: already hijacked for %s\n",
		       syscall_id, when == HFW_BEFORE ? "'before'" : "'after'");
		retval = 1;
	}

 hijack_syscall_release_lock:
	up(&hdata->sem); 

	return retval;
}

int hijack_syscall_before(int syscall_id,
			  func_ptr function)
{
	return hijack_syscall(syscall_id, function, HFW_BEFORE);
}

int hijack_syscall_after(int syscall_id,
			 func_ptr function)
{
	return hijack_syscall(syscall_id, function, HFW_AFTER);
}


/* releases a previously hijacked system call. */
int release_syscall(int syscall_id, hijack_func_when_t when)
{
	int retval = -ENOSYS;

	struct syscall_hijack_data* hdata = NULL;

	if (syscall_id < 0 || syscall_id >= NR_syscalls) {
		printk(KERN_NOTICE "release_syscall: invalid syscall_id '%d'\n",
			syscall_id);
		return retval;
	}

	hdata = &hijack_data[syscall_id];

	if ((retval = down_interruptible(&hdata->sem)) < 0)
		return -ERESTARTSYS;
    
	retval = 0; 

	/* nothing to do if it's already released */ 
	if (!hdata->orig_syscall)
		goto done; 

	/* if we're asked to remove a 'before' function, but still got */
	/* an 'after' function, or vice versa, then just release that  */
	/* function, but keep the syscall itself hijacked.             */
	if (hdata->before_hijack_func && hdata->after_hijack_func) {
		if (when == HFW_BEFORE) {
			hdata->before_hijack_func = (func_ptr)NULL;
		}
		else if (when == HFW_AFTER) {
			hdata->after_hijack_func = (func_ptr)NULL;
		}
	} /* ok, only one of them is stil hijacked */
	else {
		func_ptr* sys_call_table_addr = (func_ptr*)
			get_module_symbol(NULL, versioned_string(sys_call_table));		
		if (sys_call_table_addr) {
			func_ptr orig_syscall = hdata->orig_syscall;
			
			sys_call_table_addr[syscall_id] = orig_syscall;
			hdata->orig_syscall = (func_ptr)NULL;
			hdata->before_hijack_func = (func_ptr)NULL;
			hdata->after_hijack_func = (func_ptr)NULL;
			if (atomic_read(&hdata->invocations) == 0) 
				MOD_DEC_USE_COUNT;
			else {
				if (atomic_read(&hdata->invocations) < 0)
					BUG(); 
				SCTDBG(DEB_HIJACK, "release_syscall: %d: syscall busy "
				       "(%d active invocations) - marking for defered "
				       "release\n", syscall_id, 
				       atomic_read(&hdata->invocations));
				/* mark the 'defered release flag' for this syscall. */
				hdata->defered_release_flag = 1;
			}
		}
	}
 done: 
	up(&hdata->sem); 
	
	return retval;
}

int release_syscall_before(int syscall_id)
{
	return release_syscall(syscall_id, HFW_BEFORE);
}

int release_syscall_after(int syscall_id)
{
	return release_syscall(syscall_id, HFW_AFTER);
}

int syscalls_read_proc(char* page, char** start, off_t off, 
		       int count, int* eof, void* data)
{
	/* iterate over all of the system calls, looking for those that */ 
	/* are still hijacked, and who's holding them */ 
	int i; 
	int written = 0; 
	struct syscall_hijack_data* hdata; 
	struct list_head* tmp; 
	int orig_count; 

	/* protect us from premature unloading */ 
	MOD_INC_USE_COUNT; 

	memset(page, 0, PAGE_SIZE); 
	if (count > PAGE_SIZE)
		count = PAGE_SIZE; 

	orig_count = count; 

	for (i = 0; i < NR_syscalls; ++i) {
		hdata = &hijack_data[i]; 

		if (down_interruptible(&hdata->sem)) {
			MOD_DEC_USE_COUNT; 
			return -ERESTARTSYS; 
		}

		if (!atomic_read(&hdata->invocations)) {
			if (!list_empty(&hdata->pid_list))
				BUG(); 
			up(&hdata->sem); 
			continue; 
		}
		
		/* ok, the semaphore is locked, let's do our thing */ 
		list_for_each(tmp, &hdata->pid_list) {
			struct syscall_pid* ptr = 
				list_entry(tmp, struct syscall_pid, list); 
			written = snprintf(page, count, "%d->%d\n", 
					   i, ptr->pid); 
			count -= written;
			page += written;

			if (count <= 0) {
				up(&hdata->sem); 
				goto done; 
			}
		}
		up(&hdata->sem); 
	}
 done: 

	MOD_DEC_USE_COUNT; 
	return (orig_count - count); 
}

static struct proc_dir_entry* dir; 
static void init_debug(void)
{
	struct proc_dir_entry* res; 

	if (!(dir = proc_mkdir(PROC_MODULE_NAME, NULL))) {
		printk(KERN_WARNING "proc_mkdir failed\n"); 
		return; 
	}

	dir->owner = THIS_MODULE; 
	
	if (!(res = create_proc_entry("debug", 0644, dir))) {
		printk(KERN_WARNING "create proc entry 'debug' failed\n"); 
		goto out1; 
	}

	res->write_proc = debug_write_proc; 
	res->owner = THIS_MODULE; 

	if (register_syscalls_proc_entry(dir))
		goto out2; 

	return; 

 out2: 
	remove_proc_entry("debug", dir); 
 out1: 
	remove_proc_entry(PROC_MODULE_NAME, NULL); 
}

static void cleanup_debug(void)
{
	remove_proc_entry("debug", dir); 
	remove_proc_entry(PROC_SYSCALLS_DEBUG, dir); 
	remove_proc_entry(PROC_MODULE_NAME, NULL); 
}

int init_module(void)
{
        int i;

	memset(hijack_data, 0, sizeof(hijack_data));
	
        for (i=0; i<NR_syscalls; i++) {
		init_MUTEX(&hijack_data[i].sem);
		INIT_LIST_HEAD(&hijack_data[i].pid_list); 
		atomic_set(&hijack_data[i].invocations, 0); 
        }

	init_debug(); 

	init_hijack_syscalls(); 

	/* only allow unloading this module if the user explicitly said so */
	if (!allow_unloading)
		MOD_INC_USE_COUNT;

        return 0;
}

void cleanup_module(void)
{
	cleanup_debug(); 

        /* do nothing - this module cannot be unloaded (unless debugging). */
	if (!allow_unloading)
		printk(KERN_NOTICE "Error - syscall_hijack module "
		       "should NOT be unloaded!!");
}

EXPORT_SYMBOL(hijack_syscall_before);
EXPORT_SYMBOL(hijack_syscall_after);
EXPORT_SYMBOL(release_syscall_before); 
EXPORT_SYMBOL(release_syscall_after); 
