/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2003 Muli Ben-Yehuda, mulix@mulix.org         */ 
/* License: GNU General Public License                              */

/*
 * $Id: ctrl_dev.c,v 1.8 2003/01/31 17:57:25 mulix Exp $
 */ 

#include <linux/config.h>
#include <linux/version.h>      /* module things */
#include <linux/module.h>       /* module things */
#include <linux/kernel.h>       /* printk() */
#include <linux/slab.h>         /* kmalloc() */
#include <linux/fs.h>           /* everything... */
#include <linux/errno.h>        /* error codes */
#include <linux/types.h>        /* size_t */
#include <linux/proc_fs.h>      /* /proc stuff */ 
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <asm/atomic.h>         /* test_and_set_bit() */ 
#include <asm/uaccess.h>        /* copy_to/from_user */

#include "sct_rules/rule_engine_public.h"
#include "sct_rules/membuf.h"
#include "common/module_interface.h"
#include "common/util.h"
#include "common/sct_compat.h"

#include "update_rules.h"
#include "device_file.h"
#include "ctrl_dev.h"

#define SCT_CTRL_MAJOR 0   /* dynamic major by default */

/* the major number we use */
static int ctrl_major = SCT_CTRL_MAJOR;

/* 1 is the device file is in use, 0 if we're free at the moment */ 
volatile long ctrl_dev_in_use; 

sct_ctrl_dev_t ctrl_dev; 

/* forward declarations */ 
void handle_request_packet(sct_ctrl_dev_t* dev);

void build_response_packet(struct sct_membuf* respbuf, unsigned int packet_id, 
			   unsigned int command, int result_code, 
			   size_t datasz, unsigned char* data); 

/* does this request buffer contain a whole request packet, or not yet? */
int req_has_whole_packet(struct sct_membuf* req);

#define INITIAL_REQ_SIZE 32 * 1024

extern struct file_operations ctrl_fops; 

/* utility functions */ 

static inline size_t get_size(dev_req_t* req)
{
	return req->size; 
}

static inline void set_size(dev_req_t* req)
{
	req->size = sizeof(req->size) + 
		sizeof(req->id) + 
		sizeof(req->cmd) +
		sizeof(req->result) + 
		sizeof(req->data_sz) + 
		req->data_sz + 
		sizeof(req->magic);
}

void build_response_packet(struct sct_membuf* respbuf, unsigned int packet_id, 
			   unsigned int command, int result_code, 
			   size_t datasz, unsigned char* data)
{
	dev_req_t resp; 

	resp.size = 0; /* calculated last */ 
	resp.id = packet_id; 
	resp.cmd = command | REPLY; 
	resp.result = result_code; 
	resp.data_sz = datasz; 
	resp.data = data; 
	resp.magic = DEV_FILE_MAGIC_COOKIE; 

	set_size(&resp); 

	SCTDBG(DEB_DEV_FILE, "build_response_packet: returning a packet "
	       "of size %d, buffer available space is %d\n", resp.size, 
	       sct_membuf_free_for_write(respbuf)); 
	
	/* TODO: dont prepare the response in advance - 
	   feed it to uspace in chunks, as the reads come in. */ 
	
	/* does the buffer have enough free space for our response? */ 
	if (sct_membuf_free_for_write(respbuf) < get_size(&resp)) {
		int rc; 
		/* we need to enlarge the respbuf */ 
		if ((rc = sct_membuf_enlarge(respbuf, get_size(&resp))) < 0) {
			/* we probably ran OOM */ 
			printk(KERN_NOTICE "build_response_packet could "
			       "not grow the respbuf, error is %d", rc); 
			return; 
		}
	}
	
	/* we are guaranteed to have enough space */ 
	sct_membuf_add_chunk(respbuf, &resp.size, sizeof(resp.size)); 
	sct_membuf_add_chunk(respbuf, &resp.id, sizeof(resp.id)); 
	sct_membuf_add_chunk(respbuf, &resp.cmd, sizeof(resp.cmd)); 
	sct_membuf_add_chunk(respbuf, &resp.result, sizeof(resp.result)); 
	sct_membuf_add_chunk(respbuf, &resp.data_sz, sizeof(resp.data_sz)); 
	sct_membuf_add_chunk(respbuf, resp.data, resp.data_sz); 
	sct_membuf_add_chunk(respbuf, &resp.magic, sizeof(resp.magic)); 

	SCTDBG(DEB_DEV_FILE, "build_response_packet: returning a packet "
	       "of size %d, buffer serialized size is %d\n", resp.size, 
	       sct_membuf_available_for_read(respbuf)); 
}

void handle_update_packet(sct_ctrl_dev_t* dev, int id, int cmd)
{
	int rule_id;
	sct_rule_when_t when;
	int syscall_id, mux_func_id;
	sct_action_t* a = NULL;
	sct_filter_node_t* f = NULL;
	unsigned int rule_len; 
	int packet_rc; 
	struct sct_membuf* req = dev->req; 
	int retval; 
	int dummy; 
	struct sct_membuf* mbuf = NULL;

	/* we've already checked that it's a valid packet of the valid size, */ 
	/* so there's no need for the usual size checks here */ 
	/* skip the result field */ 
	sct_membuf_get_chunk(req, &dummy, sizeof(dummy)); 
	SCTDBG(DEB_DEV_FILE, "packet result is %d\n", dummy); 

	sct_membuf_get_chunk(req, &rule_len, sizeof(rule_len)); 
	SCTDBG(DEB_DEV_FILE, "packet rule len is %d\n", dummy); 

	/* are sanity checks necessary here? tracker_build_rule */
	/* should perform them for us */
	if ((retval = tracker_build_rule(req, 
					 &rule_id, &when, 
					 &syscall_id, &mux_func_id, 
					 &a, &f)) < 0) {
		SCTDBG(DEB_DEV_FILE, "handle_request_packet: "
		       "tracker_build_rule returned %d\n", retval); 
		packet_rc = CTL_TRACKER_EBADTRACKINGRULE;
		goto err_build_rule; 
	}
	
	if ((retval = sanity_check_user_input(syscall_id, mux_func_id, when)) < 0) {
		/* not good enough, but all sanity_check_user_input will */
		/* return to us is -EINVAL on error. it needs serious */
		/* surgery */
		SCTDBG(DEB_DEV_FILE, "handle_request_packet: sanity_check "
		       "returned %d\n",  retval); 
		packet_rc = CTL_TRACKER_EBADTRACKINGRULE;
		goto err_failed_sanity_check;
	}
	
	SCTDBG(DEB_RULES, "handle_request_packet: about to modify rules!\n");
	
	/* modify rules */
	switch (cmd) {
	case(CTL_TRACKER_CMD_ADD_RULE):
		retval = tracker_add_rule(rule_id, when,
					  syscall_id, mux_func_id,
					  f, a);
		a = NULL; f = NULL; /* mark that we consumed the action and the filter */ 
		break;
	
	case(CTL_TRACKER_CMD_DEL_RULE):
		retval = tracker_del_rule(rule_id, when,
					  syscall_id, mux_func_id);
		break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL):
		retval = tracker_del_all_rules_for_syscall(syscall_id,
							   mux_func_id);
		break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES):
		retval = tracker_del_all_module_rules();
		break;
	
	case(CTL_TRACKER_CMD_PRINT_ALL_RULES):
		retval = tracker_print_all_rules();
		break;

	case(CTL_TRACKER_CMD_GET_ALL_RULE_COUNT): 
		if (!(mbuf = sct_membuf_create(sizeof(int), NULL, NULL, 
					       SCT_MEMBUF_EXACT_MATCH))) {
			retval = -ENOMEM; 
			break; 
		}
		retval = tracker_get_all_rule_count(mbuf); 
		break; 
	
	default:
		printk(KERN_NOTICE "device_file.c:handle_packet(): invalid or unsupported "
		       "type of request - %d\n", cmd);
		packet_rc = CTL_TRACKER_EINVAL;
		goto err_invalid_request;
	}

	switch (retval) {
	case(-EINTR):
		packet_rc = CTL_TRACKER_EINTR;
		break;
	case(-ENOMEM):
		packet_rc = CTL_TRACKER_ENOMEM;
		break;
	default: /* catch all */
		/* TODO: any reason we don't return -ERRNO and be done with it? */ 
		if (retval < 0)
			packet_rc = CTL_TRACKER_EINVAL; 
		else
			/* this has the good side effect that if retval is 0, */ 
			/* packet_rc become CTL_TRACKER_SUCCESS */ 
			packet_rc = retval; 
	}

	/* build response packet */
 err_invalid_request:
 err_failed_sanity_check:
	if (a)
		sct_free_action(a); 
	if (f)
		sct_free_filter_node(f); 
 err_build_rule:
	build_response_packet(dev->resp, id, cmd, packet_rc,
			      mbuf ? sct_membuf_available_for_read(mbuf) : 0, 
			      mbuf ? sct_membuf_get_data(mbuf) : NULL);
	if (mbuf)
		sct_membuf_destroy(mbuf);
	/* and we're done! */
}

void handle_get_packet(sct_ctrl_dev_t* dev, unsigned int id, unsigned int cmd)
{
	int retval = -1, packet_rc = CTL_TRACKER_EINVAL; 
	struct sct_membuf* mbuf = NULL; 
	
	switch(cmd) {
	case CTL_TRACKER_CMD_GET_ALL_RULES:
		if (!(mbuf = sct_membuf_create(1024 * 32, NULL, NULL, 
					       SCT_MEMBUF_EXACT_MATCH))) {
			retval = -ENOMEM; 
			goto done; 
		}
		retval = tracker_get_all_rules(mbuf); 
		break; 
	default: 
		; /* nothing */ 
	}

	switch (retval) {
	case(-EINTR):
	        packet_rc = CTL_TRACKER_EINTR;
		break;
	case(-ENOMEM):
		packet_rc = CTL_TRACKER_ENOMEM;
		break;
	case (0):
		packet_rc = CTL_TRACKER_SUCCESS;
		break;
	default: /* catch all */
		packet_rc = CTL_TRACKER_EINVAL;
	}
 done:
	build_response_packet(dev->resp, id, cmd, packet_rc, 
			      sct_membuf_available_for_read(mbuf), 
			      sct_membuf_get_data(mbuf)); 
	
	if (mbuf) 
		sct_membuf_destroy(mbuf); 
}

void handle_request_packet(sct_ctrl_dev_t* dev)
{
	int dummy, packet_id, cmd;
	struct sct_membuf* req = dev->req; 
	
	/* parse the user space supplied buffer */
	/* we are guaranteesd the packet is valid */
	/* when we get called, since req_has_whole_packet */
	/* has validated it */
	/* no checking, since we know the mbuf is big enough */ 
	sct_membuf_get_chunk(req, &dummy, sizeof(dummy)); /* skip the size field */
	SCTDBG(DEB_DEV_FILE, "packet size is %d\n", dummy); 
	sct_membuf_get_chunk(req, &packet_id, sizeof(packet_id)); /* id field */ 
	SCTDBG(DEB_DEV_FILE, "packet id is %d\n", packet_id); 
	sct_membuf_get_chunk(req, &cmd, sizeof(cmd)); /* the type of request field */
	SCTDBG(DEB_DEV_FILE, "packet command is %d\n", cmd); 

	/* now, what kind of command is it? */ 
	switch (cmd) {
	case(CTL_TRACKER_CMD_ADD_RULE):
	case(CTL_TRACKER_CMD_DEL_RULE):
	case(CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL):
	case(CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES):	
	case(CTL_TRACKER_CMD_PRINT_ALL_RULES):
	case(CTL_TRACKER_CMD_GET_ALL_RULE_COUNT):
		handle_update_packet(dev, packet_id, cmd); 
		break;
	case(CTL_TRACKER_CMD_GET_ALL_RULES):
		handle_get_packet(dev, packet_id, cmd); 
		break; 
	default:
		printk(KERN_NOTICE "device_file.c:handle_packet(): invalid or unsupported "
		       "type of request - %d\n", cmd);
		build_response_packet(dev->resp, packet_id, cmd, CTL_TRACKER_EINVAL, 0, NULL); 
	}	
	/* the response packet was built by one of the packt handlers or by us - return */ 
}

/* does this request buffer contain a whole request packet, or not yet? */
int req_has_whole_packet(struct sct_membuf* req)
{
	unsigned int req_sz;
	unsigned int cookie = DEV_FILE_MAGIC_COOKIE;
	unsigned int packet_cookie;
	unsigned char* cookie_loc; 
	unsigned char* data; 
	size_t datasz; 
    
	if (!req)
		return -EINVAL;

	/* get the size of this request packet, which is the first 'int' in */
	/* the data buffer */
	data = sct_membuf_get_data(req); 
	datasz = sct_membuf_available_for_read(req); 

	if (datasz < sizeof(req_sz))
		return -EINVAL; 

	memcpy(&req_sz, data, sizeof(req_sz));
	
	/* now, sanity checks on req_sz */

	/* we dont have quite enough data yet */
	if (req_sz > datasz) {
		return -EINVAL; /* not a full packet yet */
	}
	
	/* ok, so req_sz must be smaller or equal to the data_size. cool! */
    
	/* let's check the last int from the end, if it's the magic cookie */
	/* we expect */


	if (debug_level & DEB_DEV_FILE) {
		SCTDBG(DEB_DEV_FILE, "request packet dump: "); 
		for (cookie_loc = data; cookie_loc < data + req_sz; cookie_loc += sizeof(int))
			printk("%0x ", *(unsigned int*)cookie_loc);
		printk("\n"); 
	}
	
	cookie_loc = &data[req_sz - sizeof(cookie)]; 

	memcpy(&packet_cookie, cookie_loc, sizeof(cookie));
	if (packet_cookie != cookie) {
		printk(KERN_NOTICE "req_has_whole_packet got invalid cookie - "
		       "wanted %0x, got %0x\n", cookie, packet_cookie); 
		return -EINVAL; 
	}
	
	/* whee, success */
	return 0;
}

int init_ctrl_dev_file(sct_ctrl_dev_t** pdev)
{
	int retval;
	
	sct_ctrl_dev_t* dev; 
	
	dev = *pdev = &ctrl_dev; 

	/*
	 * Register your major, and accept a dynamic number
	 */
	retval = register_chrdev(ctrl_major, "sct_ctrl", &ctrl_fops);
	if (retval < 0)
		goto unset_dev; 
	
	if (ctrl_major == 0) 
		ctrl_major = retval; /* dynamic */
	
	sema_init(&dev->sem, 1);

	retval = -ENOMEM; 
	if (!(dev->req = sct_membuf_create(INITIAL_REQ_SIZE, sct_cp_from_user, NULL,
					   SCT_MEMBUF_EXACT_MATCH)))
		goto unregister_chrdev;  

	if (!(dev->resp = sct_membuf_create(INITIAL_REQ_SIZE, NULL, sct_cp_to_user, 
					    SCT_MEMBUF_EXACT_MATCH))) 
		goto destroy_request; 
	
	return 0; 

 destroy_request:
	sct_membuf_destroy(dev->req); 

 unregister_chrdev:
	unregister_chrdev(ctrl_major, "sct_ctrl");
	
 unset_dev:
	*pdev = NULL; 

	return retval; 
}

void destroy_ctrl_dev_file(sct_ctrl_dev_t* dev)
{
	unregister_chrdev(ctrl_major, "sct_ctrl");

	sct_membuf_destroy(dev->req); 
	dev->req = NULL; 

	sct_membuf_destroy(dev->resp); 
	dev->resp = NULL; 
}

/*
 * Open and close
 */
static int ctrl_open (struct inode *inode, struct file *filp)
{
	/* only one process make talk to us. this is ok for the configuration   */
	/* device file, but we'll have to redo this to support multiple readers */
	/* for the logging device file */ 
	if (test_and_set_bit(0, &ctrl_dev_in_use))
		return -EBUSY; /* already open */ 

	if (!capable(CAP_SYS_PTRACE)) {
		clear_bit(0, &ctrl_dev_in_use); 
		return -EPERM;
	}
	
	MOD_INC_USE_COUNT;

	/* and use filp->private_data to point to the device data */
	filp->private_data = &ctrl_dev;
	
	return 0;          /* success */
}

static int ctrl_release (struct inode *inode, struct file *filp)
{
	sct_ctrl_dev_t* dev = (sct_ctrl_dev_t*)filp->private_data; 

	if (!ctrl_dev_in_use)
		BUG(); 

	sct_membuf_clear(dev->req); 
	sct_membuf_clear(dev->resp); 
	
	clear_bit(0, &ctrl_dev_in_use); 

	MOD_DEC_USE_COUNT;

	return 0;
}


/*
 * Data management: read and write
 */

static ssize_t ctrl_write(struct file *filp, const char *buf, size_t count,
			  loff_t *f_pos)
{
	sct_ctrl_dev_t *dev = (sct_ctrl_dev_t*)filp->private_data;
	ssize_t retval; 

	if (down_interruptible (&dev->sem))
		return -ERESTARTSYS;

	if ((retval = sct_membuf_add_chunk(dev->req, buf, count)) < 0) {
		goto add_chunk_error;
	}
	
	if (!req_has_whole_packet(dev->req)) {
		SCTDBG(DEB_DEV_FILE, "whoohoo! got a whole packet!\n"); 
		handle_request_packet(dev);
	}
	
	up (&dev->sem);    
	return retval; 
	
 add_chunk_error:

	SCTDBG(DEB_DEV_FILE, "req_add_chunk returned '%d'\n", retval); 
	up (&dev->sem);
	return retval;
}

static ssize_t ctrl_read(struct file *filp, char *buf, size_t count,
			 loff_t *f_pos)
{
	sct_ctrl_dev_t *dev = (sct_ctrl_dev_t*)filp->private_data;
	ssize_t retval = 0;
	size_t available; 
	
	if (down_interruptible (&dev->sem))
		return -ERESTARTSYS;

	available = sct_membuf_available_for_read(dev->resp); 
	if (count > available)
		count = available; 

	retval = sct_membuf_get_chunk(dev->resp, buf, count);
	
	up (&dev->sem);

	return retval; 
}

/*
 * The fops
 */

struct file_operations ctrl_fops = {
	MODULE_OWNER(THIS_MODULE)
	read: ctrl_read,
	write: ctrl_write, 
	open: ctrl_open,
	release: ctrl_release
};

