/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/*
 * $Id: log_dev.c,v 1.10 2003/01/29 15:47:24 mulix Exp $
 */ 

#include <linux/config.h>
#include <linux/version.h>      /* module things */
#include <linux/module.h>       /* module things */
#include <linux/kernel.h>       /* printk() */
#include <linux/errno.h>        /* error codes */
#include <linux/types.h>        /* size_t */
#include <linux/proc_fs.h>      /* /proc stuff */ 
#include <asm/atomic.h>         /* test_and_set_bit() */ 
#include <asm/uaccess.h>        /* copy_to/from_user */

#include "common/module_interface.h"
#include "common/sct_compat.h"
#include "common/util.h"

#include "log_dev.h"
#include "sct_rules_private.h"

#define SCT_LOG_MAJOR 0   /* dynamic major by default */

#define SCT_RING_BUFFER_SIZE 32 * 1024

/* the major number we use */
static int log_major = SCT_LOG_MAJOR;

sct_log_dev_t log_dev; 

spinlock_t loguser_chain_lock = SPIN_LOCK_UNLOCKED;

extern struct file_operations log_fops; 

int init_log_dev_file(sct_log_dev_t** pdev)
{
	int retval;
	
	sct_log_dev_t* dev; 

	dev = *pdev = &log_dev;

	if ((retval = register_chrdev(log_major, "sct_log", &log_fops)) < 0)
		goto unset_dev; 

	if (log_major == 0) 
		log_major = retval; /* dynamic */

	retval = -ENOMEM; 

	/* create first loguser entry. */
	if (!(dev->loguser_head = create_loguser()))
		goto unregister_chrdev;

	dev->num_logusers = 0;

	retval = -EINVAL; 
	if ((retval = sct_register_loguser_head(dev->loguser_head)))
		goto deallocate_loguser;
	
	retval = 0; 	

	init_waitqueue_head (&dev->wait);

	return retval; 

 deallocate_loguser:
	sct_loguser_destroy(dev->loguser_head);

 unregister_chrdev:
	unregister_chrdev(log_major, "sct_log");

 unset_dev:
	*pdev = NULL; 

	return retval; 
}

void destroy_log_dev_file(sct_log_dev_t* dev)
{
	sct_loguser_t* lb;
	sct_loguser_t* nxt;

	if (!dev)
		BUG(); 

	/* in case it was registered */ 
	sct_deregister_loguser_head();
	
	lb = dev->loguser_head;
	while (lb)
	{
		nxt = sct_loguser_get_next(lb);

		destroy_loguser(lb);

		lb = nxt;
	}
	
        dev->loguser_head = NULL;
	if (dev->num_logusers != 0)
		BUG();
	unregister_chrdev(log_major, "sct_log");
}

/*
 * Open and close
 */
static int log_open(struct inode *inode, struct file *filp)
{
	sct_loguser_t* loguser;
	sct_loguser_t* last;

	if (!capable(CAP_SYS_PTRACE)) {
		return -EPERM;
	}
	
	MOD_INC_USE_COUNT;

	spin_lock(&loguser_chain_lock);
	if (log_dev.num_logusers == 0)
	{
		log_dev.num_logusers++;
		spin_unlock (&loguser_chain_lock);
                /* no one is using the dev_file yet */
		filp->private_data = log_dev.loguser_head;
	} else {
		spin_unlock (&loguser_chain_lock);

                /* there's at least one open reader. 
                   creating a new loguser for this one. */
		if (!(loguser = create_loguser()))
			goto no_mem;

		spin_lock (&loguser_chain_lock);
		if (!(last = get_last_loguser()))
			BUG();
		sct_loguser_set_next (last, loguser);
		log_dev.num_logusers ++;
		spin_unlock(&loguser_chain_lock);

		filp->private_data = loguser;
	}

	return 0;          /* success */

 no_mem:
	MOD_DEC_USE_COUNT;
	return -ENOMEM;
}

static int log_release(struct inode *inode, struct file *filp)
{
	sct_loguser_t* prev;
	sct_loguser_t* cur = filp->private_data;
	sct_loguser_t* next;
	char* lf;

	if (!log_dev.num_logusers)
		BUG(); 

	spin_lock (&loguser_chain_lock);
	if (log_dev.num_logusers == 1) {
		log_dev.num_logusers--;
		spin_unlock(&loguser_chain_lock);

	        /*when no one is reading from the dev file, use default log format */
		lf = sct_loguser_change_log_format(cur, RW_BEFORE, NULL); 
		if (lf)
			kfree(lf);
		lf = sct_loguser_change_log_format(cur, RW_AFTER, NULL); 
		if (lf)
			kfree(lf);
	}
	else {
		if (cur == log_dev.loguser_head) {
			log_dev.loguser_head = sct_loguser_get_next(log_dev.loguser_head);
			sct_update_loguser_head(log_dev.loguser_head);
		}
		else {
			next = sct_loguser_get_next(cur);
			if (!(prev = get_prev_loguser(cur)))
				BUG();
			sct_loguser_set_next(prev,next);
		}

		log_dev.num_logusers--;
		spin_unlock(&loguser_chain_lock);

		destroy_loguser(cur);
	}

	MOD_DEC_USE_COUNT;

	return 0;

}

/* some of the read code is adapted from systrace.c, at */ 
/* http://www-sor.inria.fr/~patarin/systrace.c */ 
static int log_read(struct file *filp, char *buf, size_t count,
		    loff_t *f_pos)
{
	sct_loguser_t *loguser = (sct_loguser_t*)filp->private_data;
	struct sct_membuf *membuf = sct_loguser_get_membuf(loguser);

	ssize_t ret; 

	ret = -ERESTARTSYS;

	if (lock_loguser_interruptible(loguser))
		goto out_nolock;

	/* nothing to read */
	while (!sct_membuf_available_for_read(membuf)) {
		ret = -EAGAIN;
		if (filp->f_flags & O_NONBLOCK)
			goto out;
		unlock_loguser(loguser);
		interruptible_sleep_on(&log_dev.wait);
		ret = -ERESTARTSYS;
		if (lock_loguser_interruptible(loguser))
			goto out_nolock;

		ret = -ERESTARTSYS;
		if (signal_pending(current)) 	/* a signal arrived */
			goto out; 		/* tell the fs layer to handle it */
		/* otherwise loop */
	}

	ret = sct_membuf_get_chunk(membuf, buf, count);

 out:
	unlock_loguser(loguser);

 out_nolock:
	return ret;
}

static int log_write(struct file *filp, const char *buf, size_t count,
		     loff_t *f_pos)
{
	return -EBADF; 
}

int change_log_format(sct_loguser_t* loguser, sct_rule_when_t when, char *arg)
{
	int len,ret=0;
	char *lf, *old_frmt;
	if (arg) {
		if (!(len = strlen_user(arg))) {
			ret = -EINVAL;
			goto out_no_alloc;
		}

		if (!(lf = mem_alloc (len + 1))) {
			ret = -ENOMEM;
			goto out_no_alloc;
		}

		if (copy_from_user(lf, arg, len)) {
			ret = -EINVAL;
			goto out;
		}
	}
	else {
		lf = DEFAULT_LOG_FORMAT;
	}

	old_frmt = sct_loguser_change_log_format (loguser, when, lf);
			
	if (old_frmt) {
		mem_free (old_frmt);
	}

	return 0;

out:
        mem_free(lf);
out_no_alloc:
	return ret;
}


static int log_ioctl(struct inode *inode, struct file *filp,
		     unsigned int cmd, unsigned long arg)
{
	int ret;
	sct_loguser_t* loguser = (sct_loguser_t*)filp->private_data;
	struct sct_membuf* membuf = sct_loguser_get_membuf(loguser);

	SCTDBG(DEB_DEV_FILE, "log_ioctl called with cmd %d\n", cmd); 
	
	ret = -ERESTARTSYS;
	if (lock_loguser_interruptible(loguser))
		goto out_nolock;

	ret = 0;
	
	switch (cmd) {
	case CTL_IOC_DROP_OLD: 
		sct_membuf_drop_old(membuf); 
		break; 

	case CTL_IOC_DROP_NEW:
		sct_membuf_drop_new(membuf); 
		break; 

	case CTL_IOC_FLUSH: 
		sct_membuf_clear(membuf);
		break;

	case CTL_IOC_ENABLE_DEV: 
		/* only enable if we're not already enabled */ 
		if (!sct_get_loguser_head())
			sct_register_loguser_head(log_dev.loguser_head);
		break; 
		
	case CTL_IOC_DISABLE_DEV:
		sct_deregister_loguser_head(); 
		break; 

	case CTL_IOC_RESIZE:
		ret = sct_membuf_resize(membuf, arg);
		break;

	case CTL_IOC_BEFORE_LOG_FORMAT:
		ret = change_log_format (loguser, RW_BEFORE, (char*)arg);
		break;

	case CTL_IOC_AFTER_LOG_FORMAT:
		ret = change_log_format (loguser, RW_AFTER, (char*)arg);
		break;

	case CTL_IOC_MAX_REC_SIZE: 
		ret = sct_loguser_set_max_record_len(loguser, arg); 
		break; 

	default:
		ret =  -EINVAL;
		break;
	}

	unlock_loguser(loguser);

 out_nolock:
	return ret;
}

/*
 * The fops
 */

struct file_operations log_fops = {
	MODULE_OWNER(THIS_MODULE)
	.open = log_open,
	.release = log_release, 
	.read = log_read, 
	.write = log_write, 
	.ioctl = log_ioctl
};

void wake_up_logger(void)
{
	wake_up_interruptible(&log_dev.wait);
}

sct_loguser_t* create_loguser(void)
{
	struct sct_membuf* membuf;
	sct_loguser_t* loguser;

	if (!(membuf = sct_membuf_create(SCT_RING_BUFFER_SIZE, NULL, 
					 sct_cp_to_user, SCT_MEMBUF_DROP_OLD)))
                goto no_mem;

	if (!(loguser = sct_loguser_create(membuf,NULL,NULL)))
                goto free_membuf;

	return loguser;

 free_membuf:
	sct_membuf_destroy(membuf);
 no_mem:
	return NULL;
}

void destroy_loguser(sct_loguser_t* loguser)
{
	if (!loguser) 
		return;

	if (!sct_loguser_get_membuf(loguser))
		BUG();
	sct_membuf_destroy(sct_loguser_get_membuf(loguser));

	if (sct_loguser_get_log_format(loguser, RW_BEFORE))
		kfree(sct_loguser_get_log_format(loguser, RW_BEFORE));

	if (sct_loguser_get_log_format(loguser, RW_AFTER))
		kfree(sct_loguser_get_log_format(loguser, RW_AFTER));

	sct_loguser_destroy(loguser);
}

sct_loguser_t* get_last_loguser(void)
{
	sct_loguser_t* ret = log_dev.loguser_head;
	while (sct_loguser_get_next(ret) != NULL)
		ret = sct_loguser_get_next(ret);
	return ret;
}

sct_loguser_t* get_prev_loguser(sct_loguser_t* cur)
{
	sct_loguser_t* ret = log_dev.loguser_head;
	while (sct_loguser_get_next(ret) != cur) {
		if (sct_loguser_get_next(ret) == NULL)
			return NULL; /* not found */
		ret = sct_loguser_get_next(ret);
	}
	return ret;
}
