/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il             */
/* License: GNU General Public License                              */

/*
 *$Id: module_main.c,v 1.15 2003/01/31 14:08:43 mulix Exp $
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/config.h>
#include <linux/unistd.h>
#include <linux/sys.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/sysctl.h>
#include <linux/limits.h>
#include <linux/sched.h>
#include <linux/in.h>
#include <linux/proc_fs.h> 
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/ipc.h>

#include "sct_rules/filter_vars_public.h"
#include "sct_rules/rule_engine_public.h"
#include "hijack/syscall_hijack_public.h"
#include "common/module_interface.h"
#include "common/util.h"
#include "common/sct_compat.h"

#include "sct_rules_private.h"
#include "tracker_sysctl.h"
#include "logger_helpers.h"
#include "device_file.h"
#include "update_rules.h"

#include "version.h" 

MODULE_AUTHOR("guy keren <choo@actcom.co.il>, "
	      "Muli Ben-Yehuda <mulix@actcom.co.il>, "
	      "Gilad Ben-Yossef <gilad@benyossef.com>");

MODULE_DESCRIPTION("System Call Filter Evaluation and Userspace "
		   "Interface, for 'syscalltrack'");

MODULE_LICENSE("GPL");

unsigned long debug_level = 1; /* default to 1 for debugging */

MODULE_PARM(debug_level, "l"); 
MODULE_PARM_DESC(debug_level, "set to bigger than 0 for debug message."); 

#define PROC_MODULE_NAME "sct_rules"

/*
 * Forward declarations.
 */

/* the dispatch function for all sysctl handling code */
int tracker_sysctl_from_user(ctl_table *table, int *name, 
			     sysctl_namelen_t namelen,
			     void *oldval, size_t *oldlenp,
			     void *newval, size_t newlen, void **context);

/* the dispatch function for all sysctl handling code */
int tracker_sysctl_to_user(ctl_table *table, int *name, 
			   sysctl_namelen_t namelen,
			   void *oldval, size_t *oldlenp,
			   void *newval, size_t newlen, void **context);

/* initialize this mux function for this syscall_id */
void init_mux_function(int syscall_id,
		       int mux_func_id,
		       func_ptr track_before_mux_func,
		       func_ptr track_after_mux_func);

/* cleanup this mux function */ 
void cleanup_mux_function(int syscall_id, int mux_func_id); 
/*
 * Static data structures.
 */

/* defined in sct_rules_private.h */
struct syscall_track_data track_data[NR_syscalls];

/* defined in sct_rules_private.h */
struct mux_data_t mux_data[NR_MUX_SYSCALLS];

/* data used to register sysctl functions for user-mode access to the module. */
static ctl_table tracker_table[] = {
	{CTL_TRACKER_CMD_ADD_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl_from_user},
	{CTL_TRACKER_CMD_DEL_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl_from_user},
	{CTL_TRACKER_CMD_ENABLE_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl_from_user},
	{CTL_TRACKER_CMD_DISABLE_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl_from_user},
	{CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl_from_user},
	{CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl_from_user},
	{CTL_TRACKER_CMD_PRINT_ALL_RULES, NULL, NULL, 1024, 
	 0600, NULL, NULL, &tracker_sysctl_from_user}, 
	{CTL_TRACKER_CMD_GET_ALL_RULES, NULL, NULL, 1024, 
	 0600, NULL, NULL, &tracker_sysctl_to_user}, 
	{CTL_TRACKER_CMD_GET_ALL_RULE_COUNT, NULL, NULL, 1024, 
	 0600, NULL, NULL, &tracker_sysctl_from_user}, 
	{0}
};

static ctl_table tracker_top_table[] = {
	{CTL_TRACKER, NULL, NULL, 0, 0500, tracker_table},
	{0}
};


static struct ctl_table_header* tracker_sysctl_header = NULL;

#include "syscall_tracker_autogen.c"
#include "archdep_tracker.h" 
#include "special_tracker.c"

/* sysctl handling routine. allows adding/deleting/enabling/disabling rules. */
/* we use this multiplexer for actions which get input from the user */ 
int tracker_sysctl_from_user(ctl_table *table,
			     int *name, sysctl_namelen_t namelen,
			     void *oldval, size_t *oldlenp,
			     void *newval, size_t newlen, void **context)
{
	int rule_id;
	sct_rule_when_t when;
	int syscall_id;
	sct_action_t* a;
	sct_filter_node_t* f;
	int retval = 0;
	unsigned char* buf = NULL;
	int mux_func_id = 0;
	int command; 
	struct sct_membuf* m; 

	if (!capable(CAP_SYS_PTRACE))
		return -EPERM; 

	if (newlen < 0 || namelen != 1) {
		retval = -EINVAL;
		goto err_out;
	}

	/* the user's command */ 
	command = name[0]; 

	buf = kmalloc(newlen, GFP_KERNEL);
	if (!buf){
		retval = -ENOMEM;
		printk(KERN_NOTICE "tracker_sysctl_from_user: buffer allocation failed\n");
		goto err_oom;
	}
	if (copy_from_user(buf, newval, newlen)) {
		retval = -EFAULT; 
		printk(KERN_NOTICE "tracker_sysctl_from_user: failed to copy from user\n");
		goto err_copy_from_user;
	}

	if (!(m = sct_membuf_create_with_data(buf, newlen, NULL, NULL))) { 
		retval = -ENOMEM; 
		printk(KERN_NOTICE "tracker_sysctl_from_user: failed to copy from user\n");
		goto err_create_membuf; 
	}

	/* build a rule from the user supplied buffer */
	if ((retval = tracker_build_rule(m, &rule_id, &when, 
					 &syscall_id, &mux_func_id, 
					 &a, &f)) < 0)
		goto err_build_rule; 
    
	if ((retval = sanity_check_user_input(syscall_id, mux_func_id, 
					      when)) < 0) {
		goto err_failed_sanity_check;
	}

	/* which operation did the user request? */
	switch (command) {
        case(CTL_TRACKER_CMD_ADD_RULE):
		retval = tracker_sysctl_add_rule(rule_id, when,
						 syscall_id, mux_func_id,
						 f, a);
		a = NULL; f = NULL; /* mark that we consumed the action and filter */ 
		break;
	
        case(CTL_TRACKER_CMD_DEL_RULE):
		retval = tracker_sysctl_del_rule(rule_id, when,
						 syscall_id, mux_func_id);
		break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL):
		retval = tracker_sysctl_del_all_rules_for_syscall(syscall_id,
								  mux_func_id);
		break;

        case(CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES):
		retval = tracker_sysctl_del_all_module_rules();
		break;

        case(CTL_TRACKER_CMD_PRINT_ALL_RULES):
		retval = tracker_sysctl_print_all_rules();
		break;

	case(CTL_TRACKER_CMD_GET_ALL_RULE_COUNT):
		retval = tracker_sysctl_get_all_rule_count(oldval, oldlenp); 
		break; 

        default:
		printk(KERN_NOTICE "tracker_sysctl_from_user: invalid or unsupported sysctl "
		       "operation %d\n", command);
		retval = -EINVAL;
	}
	/* all is well - we can return peacefully. */
	
 err_failed_sanity_check:
	if (a)
		sct_free_action(a); 
	if (f)
		sct_free_filter_node(f); 
 err_build_rule:
	sct_membuf_destroy(m); 
 err_create_membuf: 
 err_copy_from_user:
	kfree(buf);
 err_oom:
 err_out:
	return retval;
}

/* sysctl handling routine. allows returning all rules to uspace rules. */
/* we use this multiplexer for actions which return complex output to */ 
/* the user */ 
int tracker_sysctl_to_user(ctl_table *table,
			   int *name, sysctl_namelen_t namelen,
			   void *oldval, size_t *oldlenp,
			   void *newval, size_t newlen, void **context)
{
	int retval; 
	int i; 
	int command;

	if (!capable(CAP_SYS_PTRACE))
		return -EPERM; 

	SCTDBG(DEB_SYSCTL, "tracker_sysctl_to_user: namelen = '%d', " 
	       "newlen = '%d'\n", (int)namelen, newlen);
	for (i = 0; i < namelen; i++) {
		SCTDBG(DEB_SYSCTL, "tracker_sysctl_to_user: name[%d] "
		       "= %d\n", i, name[i]);
	}

	/* the user's command */ 
	command = name[0]; 
	switch(command) {
	case CTL_TRACKER_CMD_GET_ALL_RULES:
		retval = tracker_sysctl_get_all_rules(oldval, oldlenp); 
		break; 
	default:
		printk(KERN_NOTICE "tracker_sysctl_to_user: "
		       "invalid or unsupported sysctl operation %d\n", 
		       command);
		retval = -EINVAL;
		break; 
	}
	return retval;
}


/* init muxed syscalls data, for the given syscall id. */
void init_mux_syscall(int syscall_id,
		      func_ptr syscall_track_before_func,
		      func_ptr syscall_track_after_func,
		      int num_muxed_functions)
{
	int j, id;
	struct mux_data_t* mux; 

	if ((id = mux_data_index(syscall_id)) < 0) {
		printk(KERN_WARNING "init_mux_syscall(): syscall '%d' is "
		       "NOT a mux syscall!\n", syscall_id);
		return;
	}
	mux = &mux_data[id]; 

	/* all access to mux functions are 1 based (the first mux function */ 
	/* is numbered 1 and resides in array cell 1, not 0), so make sure */ 
	/* we have enough room for them */ 
	num_muxed_functions++; 

	/* set the multiplexer tracking stubs */ 
	before_track_functions[syscall_id] = syscall_track_before_func;
	after_track_functions[syscall_id] = syscall_track_after_func;

	mux->mux_tracker = syscall_track_before_func;
	mux->num_muxed = num_muxed_functions;

	mux->muxed_functions =
		(func_ptr*)kmalloc(num_muxed_functions*sizeof(func_ptr), GFP_KERNEL);
	if (mux->muxed_functions == NULL)
		goto init_mux_syscall_cleanup_1;
    
	mux->before_rules =
		(sct_rule_list_t**)kmalloc(num_muxed_functions*sizeof(sct_rule_list_t*),
				       GFP_KERNEL);
	if (mux->before_rules == NULL)
		goto init_mux_syscall_cleanup_2;

	mux->after_rules =
		(sct_rule_list_t**)kmalloc(num_muxed_functions*sizeof(sct_rule_list_t*),
				       GFP_KERNEL);
	if (mux->after_rules == NULL)
		goto init_mux_syscall_cleanup_3;

	for (j=0; j < num_muxed_functions; j++) {
		mux->muxed_functions[j] = (func_ptr)NULL;
		mux->before_rules[j] = (sct_rule_list_t*)NULL;
		mux->after_rules[j] = (sct_rule_list_t*)NULL;
	}

	/* success, we can return now */
	return;

 init_mux_syscall_cleanup_3:
	kfree(mux->before_rules);

 init_mux_syscall_cleanup_2:
	kfree(mux->muxed_functions);

 init_mux_syscall_cleanup_1:
	before_track_functions[syscall_id] = (func_ptr)NULL;
	after_track_functions[syscall_id] = (func_ptr)NULL;
	mux->mux_tracker = (func_ptr)NULL;
	mux->num_muxed = 0;
}

void cleanup_mux_syscall(int syscall_id, int num_muxed_functions)
{
	int id;
	
	if ((id = mux_data_index(syscall_id)) < 0) {
		printk(KERN_WARNING "cleanup_mux_syscall(): syscall '%d' is "
		       "NOT s mux syscall!\n", syscall_id);
		return;
	}

	kfree(mux_data[id].muxed_functions); 
	kfree(mux_data[id].before_rules); 
	kfree(mux_data[id].after_rules); 
}

void init_mux_function(int syscall_id,
		       int mux_func_id,
		       func_ptr track_before_mux_func,
		       func_ptr track_after_mux_func)
{
	int id;

	if ((id = mux_data_index(syscall_id)) < 0) {
		printk(KERN_WARNING "init_mux_function(): syscall '%d' is "
		       "NOT s mux syscall!\n", syscall_id);
		return;
	}

	mux_data[id].muxed_functions[mux_func_id] = track_before_mux_func;
	mux_data[id].before_rules[mux_func_id] = sct_create_tracking_rule_list();
	mux_data[id].after_rules[mux_func_id] = sct_create_tracking_rule_list();
}

void cleanup_mux_function(int syscall_id, int mux_func_id)
{
	int id;

	if ((id = mux_data_index(syscall_id)) < 0) {
		printk(KERN_WARNING "cleanup_mux_function(): syscall '%d' is "
		       "NOT s mux syscall!\n", syscall_id);
		return;
	}
	sct_delete_tracking_rule_list(mux_data[id].before_rules[mux_func_id]); 
	sct_delete_tracking_rule_list(mux_data[id].after_rules[mux_func_id]); 
}

/* function used to perform logging of structs. */
int struct_logger_callback(char spec_char, struct sct_membuf* mbuf, 
			   unsigned int max_rec_len, 
			   void* kern_struct_ptr, 
			   void* struct_logger_data_ptr, void* cb_data)
{
	int rc = 0;
	int dyn_struct_type;
	struct struct_logger_data* logger_data; 
	struct struct_data* struct_data; 

	logger_data = (struct struct_logger_data*)struct_logger_data_ptr;

	SCTDBG(DEB_STRUCTS, "called, with kern_struct_ptr == %p, "
	       "struct_logger_data_ptr == %p, cb_data == %p\n", 
	       kern_struct_ptr, struct_logger_data_ptr, cb_data); 

	SCTDBG(DEB_STRUCTS, "logger_data = %p, logger_data.struct_type = %u, "
	       "logger_data.param_index = %d\n",  logger_data, 
	       logger_data ? logger_data->struct_type : 0, 
	       logger_data ? logger_data->param_index : 0); 

	struct_data = (logger_data ? get_struct_data(logger_data->struct_type) : NULL);

	SCTDBG(DEB_STRUCTS, "struct_data = %p\n", struct_data); 

	if (get_var_indexed_attr(VT_PARAM, logger_data->param_index,
				 VAN_DYN_TYPE, &dyn_struct_type)) {
		struct struct_data* dyn_struct_data =
			get_struct_data(dyn_struct_type);
		SCTDBG(DEB_STRUCTS, "logger_data->param_index is a VT_DYN_TYPE!\n"); 
		if (dyn_struct_data) {
			if (dyn_struct_data->struct_size > struct_data->struct_size) {
				printk(KERN_NOTICE "sct: struct_logger_callback: bad cast "
				       "from struct type '%d' (size %d) to "
				       "struct type '%d' (size %d). using "
				       "static struct type.\n",
				       logger_data->struct_type,
				       struct_data->struct_size,
				       dyn_struct_type,
				       dyn_struct_data->struct_size);
			}
			else {
				struct_data = dyn_struct_data;
			}
		}
	}
	
	SCTDBG(DEB_STRUCTS, "about to call the struct specific callback, "
		"struct_data = %p\n", struct_data); 

	if (struct_data) {
		SCTDBG(DEB_STRUCTS, "the logger is %p, the data is %p\n", 
		       struct_data->logger, kern_struct_ptr); 
		rc = (*struct_data->logger)(kern_struct_ptr, mbuf, 
					    max_rec_len);
	}
	
	return rc;
}

void init_callback_functions(void) 
{
	reset_all_vars();
	set_persistent_callback(VT_PROC_UID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_EUID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_SUID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_GID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_EGID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_SGID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_PID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_COMM, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_PPID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_PCOMM, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_ANCESTORS, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_CWD, proc_current_callback, NULL);

	set_logger_callback('S', struct_logger_callback, NULL);
	set_logger_callback('U', logger_helper_mux, NULL);
}

void cleanup_callback_functions(void) {
	; /* nothing */ 
}

/* forward reference for a special function to handle 'socketcall' syscall, */
/* for 'before' rules. */
asmlinkage long sct_sys_socketcall(int call, unsigned long *args); 

/* forward referenec for a special function to handle 'socketcall' syscall, */
/* 'after' rules. */
asmlinkage long sct_sys_socketcall_after(int syscall_rc, int call, unsigned long *args);

/* forward referenec for a special function to handle 'ipc' syscall, */
/* 'before' rules. */
asmlinkage int sct_sys_ipc (uint call, int first, int second, int third, void *ptr, long fifth);

/* forward referenec for a special function to handle 'ipc' syscall, */
/* 'after' rules. */
asmlinkage int sct_sys_ipc_after (int syscall_rc, uint call, int first, int second, int third, void *ptr, long fifth);

void init_mux_data(void)
{
	int i; 
	for (i = 0; i < NR_MUX_SYSCALLS; ++i) {
		mux_data[i].mux_tracker = (func_ptr)NULL;
		mux_data[i].num_muxed = 0;
		mux_data[i].muxed_functions = (func_ptr*)NULL;
		mux_data[i].demux_functions = (func_ptr*)NULL;
		mux_data[i].before_rules = (sct_rule_list_t**)NULL;
		mux_data[i].after_rules = (sct_rule_list_t**)NULL;
	}
}

void cleanup_mux_data(void)
{
	; /* nothing */ 
}

static void init_tracking_rule_lists(void)
{
	int i; 

	/* syscall 1 (exit) is a special case, it cannot have any "post" rules */ 
	/* this should pprobably be done via a generic mechanism of some sort, */
	/* if other syscalls or other archs need it.*/ 
	/* NOTE: we ignore entry 0, which is not used on linux */ 
	track_data[1].before_rules = sct_create_tracking_rule_list(); 
	track_data[1].after_rules = NULL; 

	/* and now, the rest */ 
	for (i = 2; i < NR_syscalls; ++i) {
		track_data[i].before_rules = sct_create_tracking_rule_list();
		track_data[i].after_rules = sct_create_tracking_rule_list();
	}	
}

static void cleanup_tracking_rule_lists(void)
{
	int i;
	for (i=0; i<NR_syscalls; ++i) {
		sct_delete_tracking_rule_list(track_data[i].before_rules);
		sct_delete_tracking_rule_list(track_data[i].after_rules);
	}
}

struct proc_dir_entry* dir; 
static void init_debug(void)
{
	struct proc_dir_entry* res; 

	if (!(dir = proc_mkdir(PROC_MODULE_NAME, NULL))) {
		printk(KERN_WARNING "proc_mkdir failed\n"); 
		return; 
	}
	
	dir->owner = THIS_MODULE; 
	
	if (!(res = create_proc_entry("debug", 0644, dir))) {
		printk("create_proc_entry 'debug' failed\n"); 
		goto out1; 
	}

	res->write_proc = debug_write_proc; 
	res->owner = THIS_MODULE; 

	return; 
	
 out1: 
	remove_proc_entry(PROC_MODULE_NAME, NULL); 
}

static void cleanup_debug(void)
{
	remove_proc_entry("debug", dir); 
	remove_proc_entry(PROC_MODULE_NAME, NULL); 
}

int init_module (void)
{
	int result;
	
	if ((result = init_dev_file()))
		return result;
	
	init_tracking_rule_lists(); 

	SCTDBG(DEB_INIT, "initializing mux data\n"); 

	init_mux_data(); 

	SCTDBG(DEB_INIT, "initializing mux syscall socketcall\n"); 

	/* special handling for 'socketcall' syscall - */
	/* register a special tracker function.        */
	init_mux_syscall(102, /* syscall number */
			 (func_ptr)sct_sys_socketcall, 
			 (func_ptr)sct_sys_socketcall_after, 
			 18 /* number of functions */);

	SCTDBG(DEB_INIT, "initializing mux syscall ipc\n"); 

	/* special handling for 'ipc' syscall - */
	/* register a special tracker function.        */
	init_mux_syscall(117, /* syscall number */
			 (func_ptr)sct_sys_ipc,
			 (func_ptr)sct_sys_ipc_after,
			 11 /* number of functions */);
	SCTDBG(DEB_INIT, "initializing mux functions\n"); 

	init_mux_functions();
		
	SCTDBG(DEB_INIT, "initializing struct data\n"); 

	/* init calculation of data used to handle struct syscall parameters. */
	init_struct_data();

	init_callback_functions();

	init_special_syscalls(); 

	/* register sysctl handling function. */
 	tracker_sysctl_header = register_sysctl_table(tracker_top_table, 0);
	SCTDBG(DEB_SYSCTL, "tracker_sysctl_from_user_header - %p\n", 
	       tracker_sysctl_header); 

	init_debug(); 

	printk(KERN_INFO "%s loaded. Happy syscalltracking!\n", VERSION_STRING); 

	return 0;
}

void cleanup_module (void)
{
	SCTDBG(DEB_INIT, "cleaning up syscall filter module\n");

	cleanup_debug(); 

	if (tracker_sysctl_header)
		unregister_sysctl_table(tracker_sysctl_header);

	cleanup_special_syscalls(); 

	cleanup_callback_functions(); 

	cleanup_struct_data(); 

	cleanup_mux_functions(); 

	cleanup_mux_syscall(102, /* syscall_id */ 
			    18 /* num functions */); 

	cleanup_mux_syscall(117, /* syscall_id */
			    11 /* num functions */);

	cleanup_mux_data(); 

	cleanup_tracking_rule_lists(); 

	destroy_dev_file(); 
}



/***********************************************************************/    
/***********************************************************************/    

/* special handling for the 'socketcall' syscall, and its sub-elements. */

/* special function, used to track 'socketcall'. */
/* Argument list sizes for sys_socketall */
#define AL(x) ((x) * sizeof(unsigned long))
static unsigned char nargs[18]={AL(0),AL(3),AL(3),AL(3),AL(2),AL(3),
				AL(3),AL(3),AL(4),AL(4),AL(4),AL(6),
				AL(6),AL(2),AL(5),AL(5),AL(3),AL(3)};
#undef AL


/*
 *	System call vectors. 
 *
 *	Argument checking cleaned up. Saved 20% in size.
 *  This function doesn't need to set the kernel lock because
 *  it is set by the callees. 
 */

asmlinkage long sct_sys_socketcall(int call, unsigned long *args)
{
        /*printk(KERN_INFO "sct_sys_socketcall - call '%d'\n", call);*/

	unsigned long a[6];
	unsigned long a0,a1;
	int err;
	int dummy; 

	if(call<1||call>SYS_RECVMSG)
		return -EINVAL;

	/* copy_from_user should be SMP safe. */
	if (copy_from_user(a, args, nargs[call]))
		return -EFAULT;
		
	a0=a[0];
	a1=a[1];
	
	switch(call) 
	{
	case SYS_SOCKET:
		err = track_before_syscall_102_1(a0,a1,a[2], &dummy);
		break;
	case SYS_BIND:
		err = track_before_syscall_102_2(a0,(struct sockaddr *)a1, 
						 a[2], &dummy);
		break;
	case SYS_CONNECT:
		err = track_before_syscall_102_3(a0, (struct sockaddr *)a1, 
						 a[2], &dummy);
		break;
	case SYS_LISTEN:
		err = track_before_syscall_102_4(a0,a1, &dummy);
		break;
	case SYS_ACCEPT:
		err = track_before_syscall_102_5(a0,(struct sockaddr *)a1, 
						 (int *)a[2], &dummy);
		break;
	case SYS_GETSOCKNAME:
		err = track_before_syscall_102_6(a0,(struct sockaddr *)a1, 
						 (int *)a[2], &dummy);
		break;
	case SYS_GETPEERNAME:
		err = track_before_syscall_102_7(a0, (struct sockaddr *)a1, 
						 (int *)a[2], &dummy); 
		break;
	case SYS_SOCKETPAIR:
		err = track_before_syscall_102_8(a0,a1, a[2], (int *)a[3], 
						 &dummy);
		break;
	case SYS_SEND:
		err = track_before_syscall_102_9(a0, (void *)a1, a[2], a[3], 
						 &dummy); 
		break;
	case SYS_RECV:
		err = track_before_syscall_102_10(a0, (void *)a1, a[2], a[3],
						  &dummy);
		break;
	case SYS_SENDTO:
		err = track_before_syscall_102_11(a0,(void *)a1, a[2], a[3],
						  (struct sockaddr *)a[4], 
						  a[5], &dummy);
		break;
	case SYS_RECVFROM:
		err = track_before_syscall_102_12(a0, (void *)a1, a[2], a[3],
						  (struct sockaddr *)a[4], 
						  (int *)a[5], &dummy);
		break;
	case SYS_SHUTDOWN:
		err = track_before_syscall_102_13(a0,a1, &dummy);
		break;
	case SYS_SETSOCKOPT:
		err = track_before_syscall_102_14(a0, a1, a[2], (char *)a[3], 
						  a[4], &dummy);
		break;
	case SYS_GETSOCKOPT:
		err = track_before_syscall_102_15(a0, a1, a[2], (char *)a[3], 
						  (int *)a[4], &dummy); 
		break;
	case SYS_SENDMSG:
		err = track_before_syscall_102_16(a0, (struct msghdr *) a1, 
						  a[2], &dummy);
		break;
	case SYS_RECVMSG:
		err = track_before_syscall_102_17(a0, (struct msghdr *) a1, 
						  a[2], &dummy);
		break;
	default:
		err = SCT_USE_ORIG_RETVAL; 
		break;
	}
	return err;
}

asmlinkage long sct_sys_socketcall_after(int syscall_rc, int call, 
					 unsigned long *args)
{
	unsigned long a[6];
	unsigned long a0,a1;
	int err;

	/*printk(KERN_INFO "sct_sys_socketcall_after - call '%d'\n", call);*/
	
	if(call<1||call>SYS_RECVMSG)
		return -EINVAL;

	/* copy_from_user should be SMP safe. */
	if (copy_from_user(a, args, nargs[call]))
		return -EFAULT;
		
	a0=a[0];
	a1=a[1];
	
	switch(call) 
	{
	case SYS_SOCKET:
		err = track_after_syscall_102_1(syscall_rc, a0,a1,a[2]);
		break;
	case SYS_BIND:
		err = track_after_syscall_102_2(syscall_rc, a0,
						(struct sockaddr *)a1, a[2]);
		break;
	case SYS_CONNECT:
		err = track_after_syscall_102_3(syscall_rc, a0, 
						(struct sockaddr *)a1, 
						a[2]);
		break;
	case SYS_LISTEN:
		err = track_after_syscall_102_4(syscall_rc, a0,a1);
		break;
	case SYS_ACCEPT:
		err = track_after_syscall_102_5(syscall_rc, a0,
						(struct sockaddr *)a1, 
						(int *)a[2]);
		break;
	case SYS_GETSOCKNAME:
		err = track_after_syscall_102_6(syscall_rc, a0,
						(struct sockaddr *)a1, 
						(int *)a[2]);
		break;
	case SYS_GETPEERNAME:
		err = track_after_syscall_102_7(syscall_rc, a0, 
						(struct sockaddr *)a1, 
						(int *)a[2]); 
		break;
	case SYS_SOCKETPAIR:
		err = track_after_syscall_102_8(syscall_rc, a0,a1, a[2], 
						(int *)a[3]);
		break;
	case SYS_SEND:
		err = track_after_syscall_102_9(syscall_rc, a0, (void *)a1, 
						a[2], a[3]);
		break;
	case SYS_RECV:
		err = track_after_syscall_102_10(syscall_rc, a0, (void *)a1, 
						 a[2], a[3]);
		break;
	case SYS_SENDTO:
		err = track_after_syscall_102_11(syscall_rc, a0,(void *)a1, 
						 a[2], a[3],
						 (struct sockaddr *)a[4], a[5]);
		break;
	case SYS_RECVFROM:
		err = track_after_syscall_102_12(syscall_rc, a0, (void *)a1, 
						 a[2], a[3],
						 (struct sockaddr *)a[4], 
						 (int *)a[5]);
		break;
	case SYS_SHUTDOWN:
		err = track_after_syscall_102_13(syscall_rc, a0,a1);
		break;
	case SYS_SETSOCKOPT:
		err = track_after_syscall_102_14(syscall_rc, a0, a1, a[2], 
						 (char *)a[3], a[4]);
		break;
	case SYS_GETSOCKOPT:
		err = track_after_syscall_102_15(syscall_rc, a0, a1, a[2], 
						 (char *)a[3], 
						 (int *)a[4]); 
		break;
	case SYS_SENDMSG:
		err = track_after_syscall_102_16(syscall_rc, a0, 
						 (struct msghdr *) a1, a[2]);
		break;
	case SYS_RECVMSG:
		err = track_after_syscall_102_17(syscall_rc, a0, 
						 (struct msghdr *) a1, a[2]);
		break;
	default:
		err = SCT_USE_ORIG_RETVAL; 
		break;
	}
	return err;
}


/* end of special section for dealing with socketcall. */
/************************************************************************/
/************************************************************************/



/***********************************************************************/
/***********************************************************************/

/* special handling for the 'ipc' syscall, and its sub-elements. */

/* copied from sys_ipc():
 *
 * sys_ipc() is the de-multiplexer for the SysV IPC calls..
 *
 * This is really horribly ugly.
 */

asmlinkage int sct_sys_ipc (uint call, int first, int second,
			int third, void *ptr, long fifth)
{
	int version;
	int dummy;

	version = call >> 16; /* hack for backward compatibility */
	call &= 0xffff;

	switch (call) {
	case SEMOP:
		return track_before_syscall_117_1(first, ptr, second, &dummy);
	case SEMGET:
		return  track_before_syscall_117_2 (first, second, third, &dummy);
#if 0
	case SEMCTL: {
		union semun fourth;
		if (!ptr)
			return -EINVAL;
		if (get_user(fourth.__pad, (void **) ptr))
			return -EFAULT;
		return track_before_syscall_117_3 (first, second, third, fourth, &dummy);
	}
#endif /* 0 */ 
	case MSGSND:
		return track_before_syscall_117_4 (first, (struct msgbuf *) ptr,
 				   second, third, &dummy);
	case MSGRCV:
		switch (version) {
		case 0: {
			struct ipc_kludge tmp;
			if (!ptr)
				return -EINVAL;

			if (copy_from_user(&tmp,
					   (struct ipc_kludge *) ptr,
					   sizeof (tmp)))
				return -EFAULT;
			return track_before_syscall_117_5 (first, tmp.msgp, second,
					   tmp.msgtyp, third, &dummy);
		}
		default:
			return track_before_syscall_117_5 (first,
					   (struct msgbuf *) ptr,
					   second, fifth, third, &dummy);
		}
	case MSGGET:
		return track_before_syscall_117_6 ((key_t) first, second, &dummy);
	case MSGCTL:
		return track_before_syscall_117_7 (first, second, (struct msqid_ds *) ptr, &dummy);
	case SHMAT:
		switch (version) {
		default: {
			ulong raddr;
			return track_before_syscall_117_8 (first, (char *) ptr, second, &raddr, &dummy);
		}
		case 1:	/* iBCS2 emulator entry point */
			if (!segment_eq(get_fs(), get_ds()))
				return -EINVAL;
			return track_before_syscall_117_8 (first, (char *) ptr, second, (ulong *) third, &dummy);
		}
	case SHMDT:
		return track_before_syscall_117_9 ((char *)ptr, &dummy);
	case SHMGET:
		return track_before_syscall_117_10 (first, second, third, &dummy);
	case SHMCTL:
		return track_before_syscall_117_11 (first, second,
				   (struct shmid_ds *) ptr, &dummy);

	default:
		return SCT_USE_ORIG_RETVAL;
	}
}

/* stolen from sys_ipc():
 *
 * sys_ipc() is the de-multiplexer for the SysV IPC calls..
 *
 * This is really horribly ugly.
 */
asmlinkage int sct_sys_ipc_after (int syscall_rc, uint call, int first, int second,
			int third, void *ptr, long fifth)
{
	int version;

	version = call >> 16; /* hack for backward compatibility */
	call &= 0xffff;

	switch (call) {
	case SEMOP:
		return track_after_syscall_117_1(syscall_rc, first, ptr, second);
	case SEMGET:
		return track_after_syscall_117_2(syscall_rc, first, second, third);
#if 0 
	case SEMCTL: {
		union semun fourth;
		if (!ptr)
			return -EINVAL;
		if (get_user(fourth.__pad, (void **) ptr))
			return -EFAULT;
		return track_after_syscall_117_3 (syscall_rc, first, second, third, (void *)&fourth);
	}
#endif /* 0 */ 
	case MSGSND:
		return track_after_syscall_117_4(syscall_rc, first, (struct msgbuf *) ptr,
				   second, third);
	case MSGRCV:
		switch (version) {
		case 0: {
			struct ipc_kludge tmp;
			if (!ptr)
				return -EINVAL;

			if (copy_from_user(&tmp,
					   (struct ipc_kludge *) ptr,
					   sizeof (tmp)))
				return -EFAULT;
			return track_after_syscall_117_5 (syscall_rc, first, tmp.msgp, second,
					   tmp.msgtyp, third);
		}
		default:
			return track_after_syscall_117_5 (syscall_rc, first,
					   (struct msgbuf *) ptr,
					   second, fifth, third);
		}
	case MSGGET:
		return track_after_syscall_117_6(syscall_rc, (key_t) first, second);
	case MSGCTL:
		return track_after_syscall_117_7(syscall_rc, first, second, (struct msqid_ds *) ptr);
	case SHMAT:
		switch (version) {
		default: {
			ulong raddr;
			track_after_syscall_117_8 (syscall_rc, first, (char *) ptr, second, &raddr);
		}
		case 1:	/* iBCS2 emulator entry point */
			if (!segment_eq(get_fs(), get_ds()))
				return -EINVAL;
			return track_after_syscall_117_8 (syscall_rc, first, (char *) ptr, second, (ulong *) third);
		}
	case SHMDT:
		return track_after_syscall_117_9(syscall_rc, (char *)ptr);
	case SHMGET:
		return track_after_syscall_117_10 (syscall_rc, first, second, third);
	case SHMCTL:
		return track_after_syscall_117_11 (syscall_rc, first, second,
				   (struct shmid_ds *) ptr);

	default:
		return SCT_USE_ORIG_RETVAL;
	}
}


/* end of special section for dealing with socketcall. */
/************************************************************************/
/************************************************************************/
