/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/*
 * $Id: special_tracker.c,v 1.5 2002/12/08 21:46:40 mulix Exp $
 */

/* 
 * This file has the tracking functions for special system calls, such 
 * as exit(), execve() or fork(). 
 */ 

/* tracking function for syscall 'exit' (1). */
asmlinkage static long
track_before_exit(int error_code) {
	sct_action_t* action = NULL;
	int rule_id = -1;
	int dummy; 

	/* lock the semaphore */ 
	down(&tracker_sem);

	/* reset the vars */ 
	reset_all_vars();

	/* we've got one variable, the error code */ 
	set_var_indexed_callback(VT_PARAM, 0, global_param_callback, 
				 DT_INT, (void*)&error_code);

	/* did we match a rule? */ 
	action = sct_filter_syscall(track_data[1].before_rules,
				    1, &rule_id);

	if (action) {
		/* we did! perform the action */ 
		(void)sct_tracker_perform_action(action, RW_BEFORE, rule_id, 1,
						 &dummy, "exit", "i" , 
						 error_code); 
		wake_up_logger(); 
	}
	/* release the semaphore */ 
	up(&tracker_sem);

	/* doesn't mean anything, silence gcc */ 
	return 0; 
}

/* tracking function for syscall 'fork' (2). */
asmlinkage static int
track_before_syscall_2(int* syscall_rc) {
	sct_syscall_retval_policy retval = SCT_USE_ORIG_RETVAL; 


	{
		sct_action_t* action = NULL;
		int rule_id = -1;
	
		down(&tracker_sem);
 		reset_all_vars();
		action = sct_filter_syscall(track_data[2].before_rules,
					2,
					&rule_id);
		if (action) {
			retval = sct_tracker_perform_action(
			action, RW_BEFORE, rule_id, 2,
			syscall_rc, "fork",
			"v" ); 
			wake_up_logger(); 
		}
		up(&tracker_sem);
	}


	return retval;
}

/* tracking function for syscall 'fork' (2). */
asmlinkage static int
track_after_syscall_2(int syscall_rc) {
         int retval = 0; 



	{
		sct_action_t* action = NULL;
		int rule_id = -1;
		int dummy; 


		down(&tracker_sem);
		reset_all_vars();
		set_var_callback(VT_RETVAL, global_retval_callback, (void*)syscall_rc);
		action = sct_filter_syscall(track_data[2].after_rules,
					2,
					&rule_id);
		if (action) {
			sct_tracker_perform_action(
			action, RW_AFTER, rule_id, 2, &dummy, "fork",
			"v" );
			wake_up_logger(); 
		}
		up(&tracker_sem);
	}


	return retval; 
}

/* tracking function for syscall 'execve' (11). */
asmlinkage static int
track_before_syscall_11(const char* fname, char* const argv[], 
			char* const envp[], int* syscall_rc) {
	sct_syscall_retval_policy retval = SCT_USE_ORIG_RETVAL; 

	{
		sct_action_t* action = NULL;
		int rule_id = -1;
	
		down(&tracker_sem);
 		reset_all_vars();
		set_var_indexed_callback(VT_PARAM, 0, global_param_callback, DT_STRING, (void*)fname);
		action = sct_filter_syscall(track_data[11].before_rules,
					11,
					&rule_id);
		if (action) {
			retval = sct_tracker_perform_action(
			action, RW_BEFORE, rule_id, 11,
			syscall_rc, "execve",
			"P p p", 0, argv, envp); 
			wake_up_logger(); 
		}
		up(&tracker_sem);
	}


	return retval;
}

/* tracking function for syscall 'execve' (11). */
asmlinkage static int
track_after_syscall_11(int syscall_rc, const char* fname, 
		       char* const argv[], char* const envp[]) {
	int retval = 0; 
	
	{
		sct_action_t* action = NULL;
		int rule_id = -1;
		int dummy; 


		down(&tracker_sem);
		reset_all_vars();
		set_var_indexed_callback(VT_PARAM, 0, global_param_callback, DT_STRING, (void*)fname);
		set_var_callback(VT_RETVAL, global_retval_callback, (void*)syscall_rc);
		action = sct_filter_syscall(track_data[11].after_rules,
					11,
					&rule_id);
		if (action) {
			sct_tracker_perform_action(
			action, RW_BEFORE, rule_id, 11, &dummy, "execve",
			"P p p", 0, argv, envp);
			wake_up_logger(); 
		}
		up(&tracker_sem);
	}


	return retval; 
}

/* tracking function for syscall 'bdflush' (134). */
asmlinkage static long
track_before_bdflush(int func, long code) {
	sct_action_t* action = NULL;
	int rule_id = -1;
	int dummy; 

	/* lock the semaphore */ 
	down(&tracker_sem);

	/* reset the vars */ 
	reset_all_vars();

	set_var_indexed_callback(VT_PARAM, 0, global_param_callback, 
				 DT_INT, (void*)&func);

	set_var_indexed_callback(VT_PARAM, 1, global_param_callback, 
				 DT_LONG, (void*)&code);

	/* did we match a rule? */ 
	action = sct_filter_syscall(track_data[134].before_rules,
				    134, &rule_id);

	if (action) {
		/* we did! perform the action */ 
		(void)sct_tracker_perform_action(action, RW_BEFORE, rule_id, 134,
						 &dummy, "bdflush", "i l", 
						 func, code); 
		wake_up_logger(); 
	}
	/* release the semaphore */ 
	up(&tracker_sem);

	/* doesn't mean anything, silence gcc */ 
	return 0; 
}

/* tracking function for syscall 'vfork' (190). */
asmlinkage static int track_before_syscall_vfork(int* syscall_rc) 
{
	sct_syscall_retval_policy retval = SCT_USE_ORIG_RETVAL; 
	sct_action_t* action = NULL;
	int rule_id = -1;
	
	down(&tracker_sem);
	reset_all_vars();
	action = sct_filter_syscall(track_data[190].before_rules,
				    190,
				    &rule_id);
	if (action) {
		retval = sct_tracker_perform_action(
			action, RW_BEFORE, rule_id, 190,
			syscall_rc, "fork",
			"v" ); 
		wake_up_logger(); 
	}
	up(&tracker_sem);

	return retval;
}

/* tracking function for syscall 'vfork' (190). */
asmlinkage static int track_after_syscall_vfork(int syscall_rc) 
{
	int retval = 0; 

	sct_action_t* action = NULL;
	int rule_id = -1;
	int dummy; 

	down(&tracker_sem);
	reset_all_vars();
	set_var_callback(VT_RETVAL, global_retval_callback, (void*)syscall_rc);
	action = sct_filter_syscall(track_data[190].after_rules,
				    190,
				    &rule_id);
	if (action) {
		sct_tracker_perform_action(
			action, RW_AFTER, rule_id, 190, &dummy, "vfork",
			"v" );
		wake_up_logger(); 
	}
	up(&tracker_sem);

	return retval; 
}

void init_special_syscalls(void)
{
	before_track_functions[1] = (func_ptr)track_before_exit; 
	before_track_functions[2] = (func_ptr)track_before_syscall_2; 
	after_track_functions[2] = (func_ptr)track_after_syscall_2; 
	before_track_functions[11] = (func_ptr)track_before_syscall_11; 
	after_track_functions[11] = (func_ptr)track_after_syscall_11; 
	before_track_functions[134] = (func_ptr)track_before_bdflush; 
	before_track_functions[190] = (func_ptr)track_before_syscall_vfork; 
	after_track_functions[190] = (func_ptr)track_after_syscall_vfork; 

	init_arch_special_syscalls(); 
}

void cleanup_special_syscalls(void)
{
	; /* nothing */ 
}
