/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 * $Id: main.cpp,v 1.50 2003/02/01 17:45:52 mulix Exp $
 */

/* 
 * This is the main entry point into the system call tracker 
 * configuration utility
 */

#include <iostream>
#include <fstream>
#include <list>
#include <cstdio>
#include <map>
#include <string>

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h> //for stat()
#include <sys/types.h> //for uid_t
#include <getopt.h> //for getopt 

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "sct_rules/rule_engine_public.h"
#include "sct_parselib/rule.h"
#include "sct_parselib/logger.h"
#include "sct_parselib/parser.h"
#include "sct_parselib/syscalls.h"
#include "sct_parselib/tokens_to_types.h"
#include "sct_parselib/sstreamwrap.h"
#include "sct_parselib/parselib.h" 

#include "version.h" 

using namespace std;

const char* progname; 

/* forward declarations */
bool load_config_file(const char * config_file_name,
		      sstreamwrap& data);
bool parse_config_file(sstreamwrap& data,
		       list<rule>& rules, 
		       const string& config_file_name);

#ifdef SCT_CONFIG_DEBUG
int debug = 1;
#else /* !defined SCT_CONFIG_DEBUG */ 
int debug = 0;
#endif

typedef std::pair<bool, string> command; 
typedef std::map<const string, command> cmdmap; 

void usage(const char* name)
{
	char usage[] = 
		"%s (%s) is a configuration and control utility for the system call tracker.\n"
		"\n"
		"Usage: %s [OPTIONS]... [COMMAND]\n"
		"\n"
		"Options: \n"
		"  -v, --verbose                give verbose output.\n"
		"  -c, --control device_file    use the device file for control\n"
		"  -c, --control sysctl         use sysctl for control\n"
		"  -d, --device_file            use the device file for control\n"
		"  -s, --sysctl                 use sysctl for control\n"
		"\n"
		"Commands: \n"
		"  print                print all currently registered rules to syslog\n"
		"  delete               delete all currently registered rules\n"
		"  upload [file]        read the file 'file' if specified,\n"
		"                       or the default configuration file otherwise, \n"
		"                       and upload the rules in it to the kernel module\n"
		"  check [file]         check the validity of the configuration file\n"
		"                       'file' if specified or the default configuration\n"
		"                       file otherwise\n"
		"  download             get all of the currently registered rules and print\n"
		"                       them to stdout\n"
		"  count                get the number of currently registered rules\n";

	fprintf(stdout, usage, name, VERSION_STRING, name); 
		
	exit(EXIT_FAILURE);
}

static int string_to_impl(const char* str, sct_ctrl_impl_t* impl)
{
	if (!strcmp(optarg,"device_file"))
		*impl = SCT_CTRL_DEVICE_FILE; 
	else if (!strcmp(optarg,"sysctl"))
		*impl = SCT_CTRL_SYSCTL; 
	else
		return 0; 
    
	return 1; 
}

void parse_options(char** argv, int argc, 
		   cmdmap& cmds, bool& verbose, 
		   sct_ctrl_impl_t* impl)
{
	int option_index = 0;
	verbose = false;

	while(1){ 
		/* structure with all recognised options for sct_config */
		static struct option long_options[] = {
			{"verbose", 0, NULL, 'v'},  
			{"control", 1, NULL, 'c'},  
			{"device_file", 0, NULL, 'd'}, 
			{"sysctl", 0, NULL, 's'}, 
			{0, 0, 0, 0}
		};
		int c;
		c = getopt_long (argc, argv, "+vc:ds", long_options, &option_index);

		switch (c) {
		case 0: /* long option specified */ 
			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
			break;
		case 'v':
			verbose = true; 
			break; 
		case 'c':
			if (!string_to_impl(optarg, impl))
				usage(progname); 
			break; 
		case 'd':
			*impl = SCT_CTRL_DEVICE_FILE; 
			break; 
		case 's':
			*impl = SCT_CTRL_SYSCTL; 
			break; 
		case '?': /* unrecognised option */
			/* fall through */
		default:
			c = -1;
			break;
		}
		if(c == -1) 
			break;  /* no more options */
	}

	argc -= optind; 
	if (argc <= 0) { /* no command specified */
		usage(progname); 
	}

	argv += optind; 
	bool found_command = false; 
	while (argv[0]) {
		string cmd = argv[0]; 
		string param = ""; 
		if (cmds.find(cmd) != cmds.end()) {
			/* special handling for commands with parameters */ 
			if ((cmd == "check") || (cmd == "upload")) {
				if (--argc <= 0)
					usage(progname);

				param =  argv[1]; 
				argv++;
			}

			cmds[cmd] = make_pair<bool, string>(true, param); 
			found_command = true; 
		}
		argv++; argc--; 
	}
	
	if (!found_command)
		usage(progname);
}

static void pretty_print_one_rule(syscall_rules_t* rules)
{
	/* get the syscall name */ 
	std::string name = syscalls::syscall_name(rules->syscall_id); 

	/* get the when */ 
	std::string when = (rules->when == RW_BEFORE ? "before" : "after"); 

	fprintf(stdout, "rule {\n"); 
	fprintf(stdout, "        syscall_name = %s\n", name.c_str()); 
	fprintf(stdout, "        when = %s\n", when.c_str()); 
	fprintf(stdout, "        {\n"); 
	/* pretty print the tracking rules (filter + action) for this sycall */ 
	sct_print_tracking_rules(rules->list, 16); 
	fprintf(stdout, "        }\n"); 
	fprintf(stdout, "}\n\n"); 
}

static void pretty_print_rule_list(syscall_rules_t* rules)
{
	assert(rules); 

	for ( ; rules; rules = rules->next) {
		/* only print rule lists that actually have any rules */ 
		if (sct_get_tracking_rule_count(rules->list))
			pretty_print_one_rule(rules); 
	}
}

static void free_rule_list(syscall_rules_t* rules)

{
	int first = 1; /* we dont free the head of the list */ 

	assert(rules); 

	for ( ; rules; rules = rules->next) {
		sct_delete_tracking_rule_list(rules->list); 
		rules->list = NULL;
		if (!first)
			free(rules); 
		first = 0; 
	}
}

static int get_module_rules(sct_ctrl_t ctrl)
{
	int rc; 
	syscall_rules_t rules; 

	rc = sct_ctrl_get_all_rules(ctrl, &rules); 
	if (rc == -1) {
		logger::log(logger::LOG_LEVEL_ERROR, 
			    "Error: failed getting rules.\n"); 
		return 0; 
	}
	if (rc == 0) { /* empty list */ 
		fprintf(stdout, "There are no module rules defined\n"); 
		return 1; 
	}
	/* goodie! we have the list of rules! */ 
	/* now let's print them */ 
	pretty_print_rule_list(&rules); 
	free_rule_list(&rules); 

	return 1; 
}

static int print_module_rules(sct_ctrl_t ctrl)
{
	int rc = sct_ctrl_print_all_rules(ctrl);
	if (!rc){
		logger::log(logger::LOG_LEVEL_ERROR,
			    "Error: failed printing all rules.\n");
	}

	return rc; 
}

static int delete_all_rules(sct_ctrl_t ctrl)
{
	int rc = sct_ctrl_del_all_rules(ctrl);
	if (!rc){
		logger::log(logger::LOG_LEVEL_ERROR,
			    "Error: failed deleting all previous module rules.\n");
	}

	return rc;
}

static int get_module_rules_count(sct_ctrl_t ctrl)
{
	int rc = sct_ctrl_get_all_rule_count(ctrl); 
	if (rc == -1) {
		logger::log(logger::LOG_LEVEL_ERROR, 
			    "Error: failed getting the rule count.\n"); 
		return 0; 
	}
	printf("The module currently has %d rules registered\n", rc); 
	return 1; 
}

void make_cmd(cmdmap& cmds, const string& cmd)
{
	cmds[cmd] =  make_pair<bool, string>(false, ""); 
}

static void init_cmds_map(cmdmap& cmds)
{
	make_cmd(cmds, "check");
	make_cmd(cmds, "print"); 
	make_cmd(cmds, "count"); 
	make_cmd(cmds, "download"); 
	make_cmd(cmds, "upload"); 
	make_cmd(cmds, "delete"); 
}

static bool is_set(cmdmap& cmds, const string& val)
{
	if (cmds.find(val) == cmds.end())
		return false; 

	return cmds[val].first; 
}

int main(int argc, char ** argv)
{
	cmdmap cmds; 
	sct_ctrl_impl_t impl = SCT_CTRL_DEFAULT; 
	bool verbose; 
	int rc;

	progname = argv[0];

	if (argc < 2)
		usage(progname);

	init_cmds_map(cmds); 
	
	parse_options(argv, argc, cmds, verbose, &impl);
	
	if (verbose)
		logger::set_verbosity_level(logger::LOG_LEVEL_INFO);

	/* init the control library */
	sct_ctrl_t sct_ctrl;

	if (!is_set(cmds, "check")) {
		sct_ctrl = sct_ctrl_init(impl);
		if (!sct_ctrl){
			logger::log(logger::LOG_LEVEL_ERROR,
				    "sct_config failed to initialize the "
				    "ctrl library:\n"
				    "are the kernel modules loaded?\n");
			return 1;
		}
	}

	/* init ths parselib */ 
	if (sct_parselib_init()) {
		logger::log(logger::LOG_LEVEL_ERROR, 
			    "initializing the parselib failed\n"); 
		return 1; 
	}

	if (is_set(cmds, "count")) {
		rc = get_module_rules_count(sct_ctrl); 
		if (!rc)
			cout << "getting the count of module rules failed\n" << endl; 
		return (!rc); 
	}
     
	if (is_set(cmds, "print")) {
		rc = print_module_rules(sct_ctrl);
		if (rc)
			cout << "module rules printed to system log file" << endl;
		return (!rc);
	}

	if (is_set(cmds, "download")) {
		rc = get_module_rules(sct_ctrl); 
		if (!rc)
			cout << "getting module rules failed" <<endl; 
		return (!rc); 
	}
	
	if (is_set(cmds, "delete")) {
		rc = delete_all_rules(sct_ctrl); 
		if (rc)
			cout << "all module rules were deleted sucessfully" << endl;
		return (!rc);
	}
	
	/* it's either check or upload. let's go on */
	
	sstreamwrap data;
	
	string file_name = (is_set(cmds, "check") ? cmds["check"].second : 
			    cmds["upload"].second); 

	if (file_name == "-") {
		cin >> data.rdbuf();
	} else if (!load_config_file(file_name.c_str(), data)) {
		logger::log(logger::LOG_LEVEL_ERROR,
			    "Failed to read the config file '%s'. Quitting.\n", 
			    file_name.c_str());
		return 1;
	}
	//the list of rules
	list<rule> rules;

	if (!parse_config_file(data, rules, file_name.c_str())){
		logger::log(logger::LOG_LEVEL_ERROR,
			    "Failed to parse the config file. Quitting.\n");
		return 1;
	}

	if (is_set(cmds, "check") || is_set(cmds, "upload")) {
		cout << "parsed rule list:" << endl;
		list<rule>::const_iterator it = rules.begin();
		for (; it != rules.end(); ++it){
			cout << *it << endl;
		}
		if (is_set(cmds, "check"))
			return 0;
	}

	rc = delete_all_rules(sct_ctrl);
	if (!rc) {
		return 1;
	}

	list<rule>::const_iterator it = rules.begin();
	for (; it != rules.end(); ++it){
		sstreamwrap rule_data;
		rule_data << *it << '\0';
		logger::log(logger::LOG_LEVEL_INFO,
			    "handling rule:\n%s",
			    rule_data.str().c_str());

		sct_filter_node_t* filter = (*it).filter_node();

		sct_action_t* act =  (*it).get_action();

		rc = sct_ctrl_add_rule(sct_ctrl, (*it).get_rule_id(), 
				       (*it).get_rule_when(),
				       (*it).get_syscall_identifier(), 
				       filter, act, 1);
		if (rc < 0) {
			logger::log(logger::LOG_LEVEL_ERROR,
				    "Error: failed adding rule '%s'\n",
				    (*it).get_rule_name().c_str());
			return -1;
		}
		sct_free_action(act);
	}
	cout << "Successfully uploaded rules from file '"
	     << file_name << "'" << endl;
     
	return 0;
}


bool load_config_file(const char * config_file_name,
		      sstreamwrap& data)
{
	//first of all, make some sanity checks on the file

	//preferably, that user will be root, but we relax that requirement for
	//development
	struct stat stat_info;
	if (stat(config_file_name, &stat_info)==-1){
		logger::log(logger::LOG_LEVEL_ERROR,
			    "load_config_file: stat '%s' failed, with error '%s'\n",
			    config_file_name, strerror(errno));
		return false;
	}

	//the config file is a regular file
	if (!S_ISREG(stat_info.st_mode)){
		logger::log(logger::LOG_LEVEL_ERROR,
			    "load_config_file: file '%s' is not a regular file\n", 
			    config_file_name);
		return false;
	}
#if 0
	//the config file is owned by the current user
	if (stat_info.st_uid != geteuid()){
		logger::log(logger::LOG_LEVEL_ERROR,
			    "load_config_file: file is not owned by the current euid\n");
		return false;
	}

	//the config file is only writeable by the current user. 
	if ((stat_info.st_mode & (S_IRUSR | S_IWUSR)) != (S_IRUSR & S_IWUSR)){
		log("load_config_file: wrong permissions on config_file");
		log(config_file_name);
		return false;
	}
#endif

	//now, that we've verified the file is usable by us, let's load
	//its contents into a buffer.
	ifstream in(config_file_name);
	in>>data.rdbuf();
	return true;
}

bool parse_config_file(sstreamwrap& data,
		       list<rule>& rules,
		       const string& config_file_name)
{
	/* create a new parser */ 
	parser p(data, config_file_name);
	/* start parsing the config data */ 
	return p.parse(rules);
}
