/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU Lesser General Public License                       */

/*
 * $Id: dev_file_impl.c,v 1.20 2003/02/01 19:21:22 mulix Exp $
 */ 

/* this file implements the device file communication protocol between */
/* the kernel module and userspace. for protocol details, see          */
/* doc/sct_device_control_file_design.txt                              */ 

#include <assert.h>
#include <stdio.h>

#include "sct_rules/membuf.h"
#include "module/common/module_interface.h"

#include "sct_ctrl_lib.h"
#include "dev_file_impl.h"
#include "dev_file_private.h"

/* the ops struct for the device file */ 
sct_ctrl_ops_t dev_file_ops = {
	.init =                  NULL, 
	.cleanup =               NULL, 
	.add_rule =              dev_file_add_rule, 
	.del_rule =              dev_file_del_rule, 
	.enable_rule =           NULL,  
	.disable_rule =          NULL, 
	.del_all_syscall_rules = dev_file_del_all_syscall_rules, 
	.del_all_rules =         dev_file_del_all_rules, 
	.print_all_rules =       dev_file_print_all_rules, 
	.get_all_rule_count =    dev_file_get_all_rule_count, 
	.get_all_rules =         dev_file_get_all_rules
}; 

/* forward declaration */
static int dev_file_common(sct_ctrl_t ctrl, unsigned int cmd,
			   int rule_id, sct_rule_when_t when, int syscall_id, 
			   sct_filter_node_t* filter, sct_action_t* act, 
			   int enable_rule, void *oldval, size_t *oldlenp);

/* add a rule with the given rule id to the given syscall,    */
/* where the rule is made of this filter and this action      */
/* and 'enable rule' specifies if the rule should be          */
/* immediately enabled or not                                 */
/* return -1 on failure, the rule ID on success               */
int dev_file_add_rule(sct_ctrl_t ctrl, int rule_id, sct_rule_when_t when,
		      int syscall_id, 
		      sct_filter_node_t* filter, sct_action_t* act, 
		      int enable_rule)
{
	int rc = dev_file_common(ctrl, CTL_TRACKER_CMD_ADD_RULE, 
				 rule_id, when, syscall_id, 
				 filter, act, enable_rule,
				 NULL, NULL);
	if (rc < 0)
		return -1; 

	return rc; 
}

/* delete a rule with the given rule id to the given syscall, */
/* return 0 on failure, non 0 on success                      */
int dev_file_del_rule(sct_ctrl_t ctrl,
		      int rule_id,
		      sct_rule_when_t when,
		      int syscall_id)
{
	int rc = dev_file_common(ctrl, CTL_TRACKER_CMD_DEL_RULE,
				 rule_id, when, syscall_id, 
				 NULL /* filter */, NULL /* act */,
				 0 /* enable_rule */,
				 NULL, NULL);

	return (rc < 0 ? 0 : 1); 
}

/* enable a rule with the given rule id to the given syscall, */
/* return 0 on failure, non 0 on success                      */
int dev_file_enable_rule(sct_ctrl_t ctrl,
			 int rule_id,
			 sct_rule_when_t when,
			 int syscall_id)
{
	int rc = dev_file_common(ctrl, CTL_TRACKER_CMD_ENABLE_RULE, 
				 rule_id, when, syscall_id, 
				 NULL /* filter*/, NULL /* act */,
				 0 /* enable_rule */,
				 NULL, NULL); 
	return (rc < 0 ? 0 : 1); 
}

/* disable a rule with the given rule id to the given syscall, */
/* return 0 on failure, non 0 on success                       */
int dev_file_disable_rule(sct_ctrl_t ctrl,
			  int rule_id,
			  sct_rule_when_t when,
			  int syscall_id)
{ 
	int rc = dev_file_common(ctrl, CTL_TRACKER_CMD_DISABLE_RULE,
				 rule_id, when, syscall_id, 
				 NULL /* filter*/, NULL /* act */,
				 0 /* enable_rule */,
				 NULL, NULL);
	return (rc < 0 ? 0 : 1); 
}

/* delete all rules for this syscall, return 0 on failure, non */
/* 0 on succes                                                 */
int dev_file_del_all_syscall_rules(sct_ctrl_t ctrl, int syscall_id)
{
	int rc = dev_file_common(ctrl, 
				 CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL,
				 0 /* rule_id */, 
				 RW_BEFORE /* when */, 
				 syscall_id, 
				 NULL /* filter*/, NULL /* act */,
				 0 /* enable_rule */,
				 NULL, NULL);
	return (rc < 0 ? 0 : 1); 
}
/* delete all rules for all syscalls, return 0 on failure, non */
/* 0 on succes                                                 */
int dev_file_del_all_rules(sct_ctrl_t ctrl)
{
	int rc = dev_file_common(ctrl, 
				 CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES,
				 0 /* rule_id */, 
				 RW_BEFORE /* when */, 
				 0 /* syscall_id */, 
				 NULL /* filter*/, NULL /* act */,
				 0 /* enable_rule */,
				 NULL, NULL);
	return (rc < 0 ? 0 : 1); 
}

/* print the current rules, return 0 on failure, non */
/* 0 on succes                                       */
int dev_file_print_all_rules(sct_ctrl_t ctrl)
{
	int rc = dev_file_common(ctrl, 
				 CTL_TRACKER_CMD_PRINT_ALL_RULES,
				 0 /* rule_id */, 
				 RW_BEFORE /* when */, 
				 0 /* syscall_id */, 
				 NULL /* filter*/, NULL /* act */,
				 0 /* enable_rule */,
				 NULL, NULL);
	return (rc < 0 ? 0 : 1); 
}


/* the common dev file access routine. we return -1 on failure, */
/* >= 0 on success                                              */
static int dev_file_common(sct_ctrl_t ctrl, unsigned int cmd,
			   int rule_id, sct_rule_when_t when, int syscall_id, 
			   sct_filter_node_t* filter, sct_action_t* act, 
			   int enable_rule, void *oldval, size_t *oldlenp)
{
	/* build the rule, and pass it on the the protocol layer */
	int rc;
	int serial_len = 0;
	struct sct_membuf* m; 
	static const size_t initial_size = (16 * 1024); 

	assert(syscall_id >= 0);
	if (syscall_id < 0)
		return -1;

	if (!(m = sct_membuf_create(initial_size, NULL, NULL, 
				    SCT_MEMBUF_EXACT_MATCH))) {
		fprintf(stderr, "out of memory trying to allocate a membuf\n"); 
		return -1; 
	}
    
	serial_len = sct_serialize_tracking_rule(rule_id, when, syscall_id, act, 
						 filter, m); 
	if (serial_len <= 0){
		fprintf(stderr, "failed to serialize tracking rule, "
			"error '%d'\n", serial_len);
		sct_membuf_destroy(m); 
		return -1;
	}
    
	rc = dev_file_update_kernel(ctrl, cmd, rule_id, when, syscall_id, 
				    m, enable_rule, oldval, oldlenp);

	sct_membuf_destroy(m); 
    
	if (rc < 0) {
		fprintf(stderr, "updating the kernel failed. error: %s\n",
			resp_result_to_string(rc)); 
		return -1; 
	}
	return rc;
}

/* get all rule count. returns -1 on failure or the rule count */ 
int dev_file_get_all_rule_count(sct_ctrl_t ctrl)
{
	int sum;
	size_t sz = sizeof(sum);
	int rc = dev_file_common(ctrl, 
				 CTL_TRACKER_CMD_GET_ALL_RULE_COUNT,
				 0 /* rule_id */, 
				 RW_BEFORE /* when */, 
				 0 /* syscall_id */, 
				 NULL /* filter*/, NULL /* act */,
				 0 /* enable_rule */,
				 &sum, &sz);
	return (rc < 0 ? -1 : sum); 
}

/* get all of the rules. return -1 on error or the number of lists in the list */ 
int dev_file_get_all_rules(sct_ctrl_t ctrl, syscall_rules_t* p_rules)
{
	/* we can't use dev_file_common, since we want access to the kernel's reponse */ 
	dev_req_t req, resp; 
	int rc  = -1;  /* assume failure */ 
	struct sct_membuf* mbuf; 

	/* build the request */ 
	build_request(&req, CTL_TRACKER_CMD_GET_ALL_RULES, 
		      /* we don't use any of these parameters */ 
		      0, 0, 0, NULL, 0, 0); 
	
	if (!dev_file_do_request(ctrl, &req, &resp))
		goto done; 
	
	/* we got a response! */ 
	if (resp.result != CTL_TRACKER_SUCCESS) {
		fprintf(stderr, "getting the rule count failed, error: %s\n", 
			resp_result_to_string(resp.result)); 
		goto destroy_request; 
	}
	
	if (!(mbuf = sct_membuf_create_with_data(resp.data, resp.data_sz, 
						 NULL, NULL)))
		goto destroy_request; 
	
	rc = parse_rulesbuf(mbuf, p_rules); 
	if (rc == -1) {
		fprintf(stderr, "failed to parse rules buffer\n"); 
		goto destroy_request; 
	}

	sct_membuf_destroy(mbuf); 

 destroy_request:
	destroy_request(&req); 
	destroy_request(&resp); 
 done:
	return rc; 
}

