/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU Lesser General Public License                       */

/* 
 * $Id: sysctl_impl.c,v 1.18 2003/01/31 17:57:26 mulix Exp $
 */

#include <malloc.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* get the module's sysctl defines */
#include "module/common/module_interface.h"
#include "sct_rules/membuf.h"

#include "sct_ctrl_lib.h"
#include "sct_ctrl_lib_private.h"
#include "sysctl_impl.h"
#include "sysctl_private.h" 

sct_ctrl_ops_t sysctl_ops = {
	.init                  = NULL, 
	.cleanup               = NULL, 
	.add_rule              = sysctl_add_rule, 
	.del_rule              = sysctl_del_rule,
	.enable_rule           = NULL, 
	.disable_rule          = NULL, 
	.del_all_syscall_rules = sysctl_del_all_syscall_rules, 
	.del_all_rules         = sysctl_del_all_rules, 
	.print_all_rules       = sysctl_print_all_rules, 
	.get_all_rule_count    = sysctl_get_all_rule_count, 
	.get_all_rules         = sysctl_get_all_rules
}; 

static int sysctl_common(sct_ctrl_t ctrl, unsigned int sysctl_command, 
			 int rule_id, sct_rule_when_t when, int syscall_id, 
			 sct_filter_node_t* filter, sct_action_t* act,
			 int enable_rule UNUSED, void* oldval, size_t* oldlenp)
{
	/* we ignore enable_rule for now */
	int serial_len;
	int sysctl_cmd_struct[] = { CTL_TRACKER, sysctl_command };
	int rc; 
	struct sct_membuf* m; 
	const size_t initial_size = (16 * 1024); 
    
	assert(syscall_id >= 0);
    
	if (syscall_id < 0)
		return 0;

	if (!(m = sct_membuf_create(initial_size, NULL, NULL, SCT_MEMBUF_EXACT_MATCH))) {
		fprintf(stderr, "failed to create the membuf\n"); 
		return 0;
	}

	serial_len = sct_serialize_tracking_rule(rule_id, when, syscall_id, act, 
						 filter, m); 
    
	if (serial_len <= 0){
		fprintf(stderr, "failed to serialize tracking rule, "
			"error '%d'\n", serial_len);
		sct_membuf_destroy(m); 
		return 0;
	}

	rc = do_sysctl(m, sysctl_cmd_struct, 
		       sizeof(sysctl_cmd_struct)/ 
		       sizeof(sysctl_cmd_struct[0]),
		       oldval, oldlenp);

	sct_membuf_destroy(m); 

	if (rc < 0)
		perror("sysctl"); 

	return rc; 
}

		     

/* add rule */
int sysctl_add_rule(sct_ctrl_t ctrl, int rule_id, sct_rule_when_t when, 
		    int syscall_id, sct_filter_node_t* filter, 
		    sct_action_t* act, int enable_rule)
{
	int rc; 
	assert (filter && act); 
	if (!filter || !act)
		return 0;

	rc = sysctl_common(ctrl, CTL_TRACKER_CMD_ADD_RULE, 
			   rule_id, when, syscall_id, 
			   filter, act, enable_rule,
			   NULL, NULL);
	return rc; 
}

/* delete a rule */
int sysctl_del_rule(sct_ctrl_t ctrl, int rule_id, sct_rule_when_t when, int syscall_id)
{
	int rc = sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_RULE, 
			       rule_id, when, syscall_id, 
			       /* filter node */NULL, 
			       /* action */NULL, 0,
			       NULL, NULL);
	return (rc == -1?  0 : 1); 
}

/* enable a rule */
int sysctl_enable_rule(sct_ctrl_t ctrl UNUSED,
		       int rule_id UNUSED,
		       sct_rule_when_t when UNUSED,
		       int syscall_id UNUSED)
{
	fprintf(stderr, "eek! sct_ctrl_enable_rule is unimplented!\n");
	return 0;
}

/* disable a rule */
int sysctl_disable_rule(sct_ctrl_t ctrl UNUSED,
			int rule_id UNUSED,
			sct_rule_when_t when UNUSED,
			int syscall_id UNUSED)
{
	fprintf(stderr, "eek! sct_ctrl_disable_rule is unimplented!\n");
	return 0;
}

/* delete all rules for this syscall */
int sysctl_del_all_syscall_rules(sct_ctrl_t ctrl, int syscall_id)
{
	int rc = sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL, 
			       /* rule_id */ 0, /* when */ RW_BEFORE, syscall_id, 
			       /* filter node */NULL, 
			       /* action */NULL, 0,
			       NULL, NULL);
	return (rc == -1 ? 0 : 1); 
}

/* delete all rules for ALL syscalls */
int sysctl_del_all_rules(sct_ctrl_t ctrl)
{
	int rc = sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES, 
			       /* rule_id */ 0, /* when */ RW_BEFORE,
			       /* syscall_id */ 0,
			       /* filter node */NULL, 
			       /* action */NULL, 0,
			       NULL, NULL);
	return (rc == -1 ? 0 : 1); 
}

/* print the current rules */
int sysctl_print_all_rules(sct_ctrl_t ctrl)
{
	int rc = sysctl_common(ctrl, CTL_TRACKER_CMD_PRINT_ALL_RULES, 
			       /* rule_id */ 0, /* when */ RW_BEFORE,
			       /* syscall_id */ 0,
			       /* filter node */NULL, 
			       /* action */NULL, 0,
			       NULL, NULL);
	return (rc == -1 ? 0 : 1); 
}

/* get the count of all currently defined rules */ 
int sysctl_get_all_rule_count(sct_ctrl_t ctrl)
{
	int sum; 
	size_t sz = sizeof(sum);
	int rc = sysctl_common(ctrl, CTL_TRACKER_CMD_GET_ALL_RULE_COUNT, 
			       /* rule_id */ 0, /* when */ RW_BEFORE,
			       /* syscall_id */ 0,
			       /* filter node */NULL, 
			       /* action */NULL, 0,
			       &sum, &sz);
	return (rc == -1 ? -1 : sum);
}

/* get all of the current rules */
int sysctl_get_all_rules(sct_ctrl_t ctrl UNUSED, syscall_rules_t* p_rules)
{
	/* can't use sysctl_common, we need an out parameter */ 
	int sysctl_cmd_struct[] = { CTL_TRACKER, 
				    CTL_TRACKER_CMD_GET_ALL_RULES };
	
	/* 16 k should be enough until we implement retry on buf too small */ 
	unsigned char kernbuf[16 * 1024]; 
	size_t sbuf = sizeof(kernbuf); 
	
	struct sct_membuf* mbuf; 
	int rc; 

	rc = do_sysctl(NULL, sysctl_cmd_struct, sizeof(sysctl_cmd_struct) / 
		       sizeof(sysctl_cmd_struct[0]), kernbuf, &sbuf); 
	if (rc) { /* woops, sysctl failed. try to figure out why */ 
		if (errno == ENOSPC) {
			/* TODO: retry with a bigger buffer */ 
				printf("syscall_get_all_rules: buffer size %ld too small. \n", 
				       (unsigned long) sbuf); 
			} else {
				perror("sysctl");
			}
		return -1;
	}
	
	/* put the buffer we got back into a membuf */ 
	mbuf = sct_membuf_create_with_data(kernbuf, sbuf, NULL, NULL); 
	if (!mbuf)
		return -1; 
	
	rc = parse_rulesbuf(mbuf, p_rules); 
	if (rc == -1) {
		fprintf(stderr, "failed to parse rules buffer\n"); 
		return -1; 
	}

	sct_membuf_destroy(mbuf); 
	
	return rc; 
}
