/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/* 
 * $Id: filter_parser.cpp,v 1.7 2003/01/28 17:49:00 mulix Exp $
 */

#include <vector>
#include <string>
#include <iostream>
#include <ctype.h>
#include <assert.h>

/* for networking-related macros */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* for username/groupname related macros */
#include <grp.h>
#include <pwd.h>

/* for 'open'/'chmod' flags. */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "sct_rules/rule_engine_public.h"
#include "tokens_to_types.h"
#include "logger.h"
#include "filter_parser.h"
#include "validity_checks.h"
#include "parse_exception.h"

using namespace std; 

/* seperators between tokens */
#define WHITESPACE " \t\n\r"
#define SPECIAL_CHARS ".,()[]+-=<>~!&|"


filter_parser::filter_parser(const string& syscall_name)
    :strings_to_tokens(), macro_handlers(), str_toks(), started(false),
     _syscall_name(syscall_name)
{
    strings_to_tokens["in"]   = IN_OP;
    strings_to_tokens["=="]   = EQUAL;
    strings_to_tokens["<"]    = LESS_THEN;
    strings_to_tokens[">"]    = GREATER_THEN;
    strings_to_tokens["<="]   = LESS_OR_EQUAL;
    strings_to_tokens[">="]   = GREATER_OR_EQUAL;
    strings_to_tokens["!="]   = NOT_EQUAL;
    strings_to_tokens["+"]    = PLUS;
    strings_to_tokens["-"]    = MINUS;
    strings_to_tokens["<<"]   = SHIFT_LEFT;
    strings_to_tokens[">>"]   = SHIFT_RIGHT;
    strings_to_tokens["&"]    = BITWISE_AND;
    strings_to_tokens["|"]    = BITWISE_OR;
    strings_to_tokens["^"]    = BITWISE_XOR;
    strings_to_tokens["~"]    = BITWISE_NOT;
    strings_to_tokens["&&"]   = LOGICAL_AND;
    strings_to_tokens["||"]   = LOGICAL_OR;
    strings_to_tokens["!"]    = LOGICAL_NOT;
    strings_to_tokens["~="]   = PATTERN_MATCH;
    strings_to_tokens["=~"]   = PATTERN_MATCH;
    strings_to_tokens["("]    = LPAREN;
    strings_to_tokens[")"]    = RPAREN;
    strings_to_tokens["PPID"] = PPID;
    strings_to_tokens["PID"]  = PID;
    strings_to_tokens["UID"]  = UID;
    strings_to_tokens["EUID"] = EUID;
    strings_to_tokens["SUID"] = SUID;
    strings_to_tokens["GID"]  = GID;
    strings_to_tokens["EGID"] = EGID;
    strings_to_tokens["SGID"] = SGID;
    strings_to_tokens["COMM"] = COMM; 
    strings_to_tokens["PCOMM"] = PCOMM;
    strings_to_tokens["CWD"]  = CWD;
    strings_to_tokens["RETVAL"] = RETVAL; 
    strings_to_tokens["["]    = LBRACKET;
    strings_to_tokens["]"]    = RBRACKET;
    strings_to_tokens["PARAMS"] = PARAMS;
    strings_to_tokens["."]    = DOT;
    strings_to_tokens[","]    = COMMA;

    // built-in macros.
    register_macro("ipaddr", &filter_parser::calc_ipaddr);
    register_macro("htons", &filter_parser::calc_htons);
    register_macro("usernametoid", &filter_parser::calc_usernametoid);
    register_macro("groupnametoid", &filter_parser::calc_groupnametoid);

    // parameter-less built-in macros.

    // 'open' syscall constants.
    register_str_to_int_macro("O_RDONLY", O_RDONLY);
    register_str_to_int_macro("O_WRONLY", O_WRONLY);
    register_str_to_int_macro("O_RDWR", O_RDWR);
    register_str_to_int_macro("O_ACCMODE", O_ACCMODE);
    register_str_to_int_macro("O_CREAT", O_CREAT);
    register_str_to_int_macro("O_EXCL", O_EXCL);
    register_str_to_int_macro("O_NOCTTY", O_NOCTTY);
    register_str_to_int_macro("O_TRUNC", O_TRUNC);
    register_str_to_int_macro("O_APPEND", O_APPEND);
    register_str_to_int_macro("O_NONBLOCK", O_NONBLOCK);
    register_str_to_int_macro("O_NDELAY", O_NDELAY);
    register_str_to_int_macro("O_SYNC", O_SYNC);
    register_str_to_int_macro("O_FSYNC", O_FSYNC);
    register_str_to_int_macro("O_ASYNC", O_ASYNC);
    register_str_to_int_macro("O_DIRECT", O_DIRECT);
    register_str_to_int_macro("O_LARGEFILE", O_LARGEFILE);
    register_str_to_int_macro("O_DIRECTORY", O_DIRECTORY);
    register_str_to_int_macro("O_NOFOLLOW", O_NOFOLLOW);

    // 'chmod' syscall constants.
    register_str_to_int_macro("S_ISUID", S_ISUID);
    register_str_to_int_macro("S_ISGID", S_ISGID);
    register_str_to_int_macro("S_ISVTX", S_ISVTX);
    register_str_to_int_macro("S_IRUSR", S_IRUSR);
    register_str_to_int_macro("S_IWUSR", S_IWUSR);
    register_str_to_int_macro("S_IXUSR", S_IXUSR);
    register_str_to_int_macro("S_IRGRP", S_IRGRP);
    register_str_to_int_macro("S_IWGRP", S_IWGRP);
    register_str_to_int_macro("S_IXGRP", S_IXGRP);
    register_str_to_int_macro("S_IROTH", S_IROTH);
    register_str_to_int_macro("S_IWOTH", S_IWOTH);
    register_str_to_int_macro("S_IXOTH", S_IXOTH);
}


/* print this vector of tokens */
static void
print_str_tokens(const vector < string >& v)
{
    vector< string >::const_iterator it = v.begin();
    for ( ; it != v.end(); ++it){
	cerr<<"'"<<*it<<"' ";
    }
    cerr<<endl;
}

void
filter_parser::parse(sct_filter_node_t*& f, const string& data)
{
    string::size_type loc = 0;
    unsigned int in_paran = 0;
    unsigned int in_brackets = 0;

    while (loc < data.length()){
	/* skip whitespace */
 	loc = data.find_first_not_of(WHITESPACE, loc);
	if (loc == data.npos)
	    break;

	/* handle quoted strings. */
	if (data[loc] == '"') {
	    string::size_type end = loc+1;

	    while (end < data.length() && data[end] != '"') {
		/* skip escaped character. */
	        if (data[end] == '\\')
		    end++;
		end++;
	    }
	    if (end >= data.length() || data[end] != '"') {
		throw parse_exception("", _line_num, 
				      "unmatched double quotes char");
	    }
	    /* copy the stirng, including the double quotes. */
	    string literal = data.substr(loc, end - loc + 1); 
	    if (!literal.empty())
	        str_toks.push_back(literal);
	    loc = end+1;
	    continue;
	}

	/* is this a paranthesis? */
	if (data[loc] == '(') {
	    loc++;
	    str_toks.push_back("(");
	    in_paran++;
	    continue;
	}
	
	/* closing parantheasis */
	if (data[loc] == ')') {
	    if (!in_paran--)
		throw parse_exception("", _line_num, 
				      "unbalanced right parent ')'");
	    loc++;
	    str_toks.push_back(")");
	    continue;
	}

	/* left bracket */
	if (data[loc] == '[') {
	    loc++;
	    str_toks.push_back("[");
	    in_brackets++;
	    continue;
	}

	/* right bracket */
	if (data[loc] == ']') {
	    if (!in_brackets--)
		throw parse_exception("", _line_num, 
				      "unbalanced right bracket ']'");
	    loc++;
	    str_toks.push_back("]");
	    continue;
	}

	/* a dot? */
	if (data[loc] == '.') {
	    loc++;
	    str_toks.push_back(".");
	    continue;
	}
	    
	/* a comma? */
	if (data[loc] == ',') {
	    loc++;
	    str_toks.push_back(",");
	    continue;
	}
	    
	/* operators can be one or two chars. 
	   first check of the first + second chars are an op, 
	   if not, check just the first */
	string oper; 
	oper = data[loc];
	if (loc + 1 < data.length())
	    oper += data[loc + 1];
	if (is_operator(oper)) {
	    loc += 2;
	    str_toks.push_back(oper);
	    continue;
	}
	oper = data[loc];
	if (is_operator(oper)) {
	    loc++;
	    str_toks.push_back(oper);
	    continue;
	}

	/* get the literal */
 	string::size_type end = data.find_first_of(WHITESPACE SPECIAL_CHARS, loc+1);
	string literal = data.substr(loc, end - loc); 
	if (!literal.empty())
	    str_toks.push_back(literal);
	
	loc = end;
    }
    /* print the tokens */
    if (logger::get_verbosity_level() == logger::LOG_LEVEL_DEBUG) {
	logger::log(logger::LOG_LEVEL_DEBUG, "filter_parser::parse() "
		    "parsed these tokens:\n");
	print_str_tokens(str_toks);
	logger::log(logger::LOG_LEVEL_DEBUG, "\n");
    }

    get_token(curr_tok);
    
    f = expr_logical_or(false);

    if (curr_tok.t != END) {
	sstreamwrap buf;
	buf << "wrong filter expression - next token is '" 
	    << *curr_tok_str << "'" << '\0';

	throw parse_exception("", _line_num, buf.str());
    }
}

filter_parser::fl_token::fl_token():orig_text(), t(END)
{}

filter_parser::fl_token::fl_token(const fl_token& rhs)
{
    orig_text = rhs.orig_text;
    t = rhs.t;
    if (!copy_value_type(v, rhs.v))
	throw parse_exception("", 0, "internal error: copy_value_type() failed.");
}

filter_parser::fl_token&
filter_parser::fl_token::operator=(const fl_token& rhs)
{
    orig_text = rhs.orig_text;
    t = rhs.t;
    if (!copy_value_type(v, rhs.v))
	throw parse_exception("", 0, "internal error: copy_value_type() failed.");
    return *this;
}

filter_parser::fl_token::~fl_token()
{
    if (t == STRING && v.type == DT_STRING && v.val.str)
	delete[] v.val.str;
}

/* macro function registration. */
void
filter_parser::register_macro(const char* macro_name, macro_handler_func_t func)
{
    macro_handler_t* mh = new macro_handler_t;
    mh->macro = func;
    macro_handlers[macro_name]	= mh;
    strings_to_tokens[macro_name]    = BUILTIN_MACRO;
}

/* macro function registration. */
void
filter_parser::register_noparam_macro(const char* macro_name,
				      macro_handler_func_t func)
{
    macro_handler_t* mh = new macro_handler_t;
    mh->macro = func;
    macro_handlers[macro_name]	= mh;
    strings_to_tokens[macro_name]    = BUILTIN_PARAMLESS_MACRO;
}

/* 'string to int' macro function registration. */
void
filter_parser::register_str_to_int_macro(const char* macro_name, int value)
{
  str_to_int_macros[macro_name] = value;
  register_noparam_macro(macro_name, &filter_parser::calc_macro_str_to_int);
}

/* is this string a recognized operator? */
inline bool
filter_parser::is_operator(const string& str)
{
    return tokens::get_operator(str) != OT_LAST_OPERATOR;
}

/* build token builds a token from the *current* position
   of the curr_tok_str iterator */ 
void
filter_parser::build_token(str_tok_vec::const_iterator& pos, fl_token& tok) 
{
    /* check that we didnt hit the end of the tokens vector */

    if (pos == str_toks.end()) {
	tok.t = END;
	return;
    }

    tok.orig_text = *pos;

    map <string, token_t>::const_iterator it 
	= strings_to_tokens.find(*pos);
    if (it != strings_to_tokens.end()) {
	tok.t = (*it).second;
	return;
    }
    if (is_int((*pos).c_str())) {
	tok.t = NUMBER;
	tok.v.type = DT_INT; /* TODO: we need to know *which* number type, 
				but that will have to come later */
        /* Need to take care of 64/32bit things */
	long lval = strtol((*pos).c_str(), NULL, 0);
	if (lval <= INT_MIN || lval >= INT_MAX) {
	    sstreamwrap err;
	    err << "'" << (*pos) << "' - overflow or underflow";
	    throw parse_exception("", _line_num, err.str());
	}	
	tok.v.val.i = lval;
	return;
    }
    
    const string& this_string = *pos;
    if (this_string[0] == '"' && 
	this_string[this_string.length()-1] == '"'){
	tok.t = STRING;
	tok.v.type = DT_STRING;

	/* we play yucky games here with char ptrs since the value_type_t 
	   comes from sct_rules and doesnt know about cpp strings */
	char*& alias = tok.v.val.str;
	const string& tmp = this_string.substr(1, this_string.length() - 2);
	if (tmp.length()) {
	    alias = new char[tmp.length()+1];
	    strncpy(alias, tmp.c_str(), tmp.length() + 1);
	} else {
	    alias = NULL;
	}
	return;
    }

    /* TODO: handle other types */
    tok.t = LAST_TOKEN_TYPE;
}

/* advance the curr_tok_str iterator */
void 
filter_parser::get_token(fl_token& tok)
{
    if (!started) {
	curr_tok_str = str_toks.begin();
	started = true;
    } else {
	if (curr_tok_str != str_toks.end())
	    curr_tok_str++;
    }
    if (curr_tok_str == str_toks.end()) {
	tok.t = END; 
    } else {
	build_token(curr_tok_str, tok);
    }
}

void 
filter_parser::push_token()
{
    if (!started) {
	return;
    }
    curr_tok_str--;
}


/* peek at the next in line token */
void
filter_parser::peek_token(fl_token& tok)
{
    str_tok_vec::const_iterator foo;
    if (!started) {
	foo = str_toks.begin();
	build_token(foo, tok);
	/* DONT set started to true */
    } else {
	if (curr_tok_str != str_toks.end()) {
	    ++curr_tok_str;
	    if (curr_tok_str != str_toks.end()) {
		build_token(curr_tok_str, tok);
	    }
	    --curr_tok_str;
	}
    }
}

/* pop the next in line token, without carring for its value */
void
filter_parser::pop_token()
{
    if (!started) {
	curr_tok_str = str_toks.begin();
	started = true;
    } else {
	if (curr_tok_str != str_toks.end())
	    curr_tok_str++;
    }
}


sct_filter_node_t* 
filter_parser::prim(bool get)
{

    if (get)
	get_token(curr_tok);
    
    switch(curr_tok.t){
    case NUMBER:
    {
	int i = curr_tok.v.val.i;
	get_token(curr_tok);
	return sct_make_int_val(i);
    }
    case STRING:
    {
	assert(curr_tok.v.type == DT_STRING && curr_tok.v.val.str);
	sct_filter_node_t* e = sct_make_string_val(curr_tok.v.val.str);
	get_token(curr_tok);
	return e;
    }
    case LPAREN:
    {
	/* if we got an LPAREN but get is false, get the next token anyway */
	if (!get) 
	    get_token(curr_tok);
	sct_filter_node_t* e = expr_logical_or(get);
	if (curr_tok.t != RPAREN) {
	    /* the error message should depend on why we stopped */ 
	    /* if it's the end of data, complain about no RPAREN */ 
	    if (curr_tok.t == END) {
		throw parse_exception("", _line_num, "expression not "
				      "terminated by a right parenthasis");
	    } else { /* otherwise, complain about an unknown token */ 
		sstreamwrap err;
		err << "'" << curr_tok.orig_text << "' is unknown"; 
		throw parse_exception("", _line_num, err.str());
	    }
	}
	get_token(curr_tok);
	return e;
    }
    case UID:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_UID, tokens::get_data_type("uid_t"));
	get_token(curr_tok);
	return var;
    }
    case PPID:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_PPID, tokens::get_data_type("pid_t"));
	get_token(curr_tok);
	return var;
    }
    case PID:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_PID, tokens::get_data_type("pid_t"));
	get_token(curr_tok);
	return var;
    }
    case EUID:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_EUID, tokens::get_data_type("uid_t"));
	get_token(curr_tok);
	return var;
    }
    case SUID:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_SUID, tokens::get_data_type("uid_t"));
	get_token(curr_tok);
	return var;
    }
    case GID:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_GID, tokens::get_data_type("uid_t"));
	get_token(curr_tok);
	return var;
    }
    case EGID:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_EGID, tokens::get_data_type("uid_t"));
	get_token(curr_tok);
	return var;
    }
    case SGID:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_SGID, tokens::get_data_type("uid_t"));
	get_token(curr_tok);
	return var;
    }    
    case COMM:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_COMM, DT_STRING);
	get_token(curr_tok);
	return var;
    }    
    case PCOMM:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_PCOMM, DT_STRING);
	get_token(curr_tok);
	return var;
    }
    case CWD:
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_PROC_CWD, DT_STRING);
	get_token(curr_tok);
	return var;
    }
    case RETVAL: 
    {
	sct_filter_node_t* var = 
	    sct_make_var_val(VT_RETVAL, DT_LONG); 
	get_token(curr_tok); 
	return var; 
    }
    case PARAMS:	
    {
	sct_filter_node_t* var = parse_param();
	get_token(curr_tok);
	return var;
    }
    case BUILTIN_MACRO:	
    {
	sct_filter_node_t* constant =
		parse_builtin_macro(curr_tok.orig_text, true);
	get_token(curr_tok);
	return constant;
    }
    case BUILTIN_PARAMLESS_MACRO:	
    {
	sct_filter_node_t* constant =
		parse_builtin_macro(curr_tok.orig_text, false);
	get_token(curr_tok);
	return constant;
    }
    default:
    {
	sstreamwrap err;
	err << "'" << curr_tok.orig_text << "' is unknown";
	throw parse_exception("", _line_num, err.str());
    }
    }
}

sct_filter_node_t*
filter_parser::parse_param()
{
    get_token(curr_tok);
    if (curr_tok.t != LBRACKET)
	throw parse_exception("", _line_num, 
			      "PARAMS must be followed by a left bracket '['");
    get_token(curr_tok);
    if (curr_tok.t != NUMBER)
	throw parse_exception("", _line_num, 
			      "invalid index for PARAMS");
    /* the index is 1 based */
    assert(curr_tok.v.type == DT_INT); /* TODO: or any other number dt */
    int location = curr_tok.v.val.i;

    /* is this a struct casted to another struct? check if the index is */
    /*  followed by a dot.                                             */
    /* just peek, in case it's not a '.' */
    string struct_type = "";
    fl_token type_tmptok;
    peek_token(type_tmptok);
    if (type_tmptok.t == DOT) {
	pop_token(); /* pop the '.' */ 
	get_token(curr_tok); /* and here's the structure type name */
	/* UGLY HACK: strings are only parsed if they are quoted, so 
	   build_token doesnt know how to handle the structure field string 
	   and returns to us a LAST_TOKEN_TYPE. need to teach build_token() 
	   about structure fields. */
	assert(curr_tok.t == LAST_TOKEN_TYPE);
	struct_type = curr_tok.orig_text;
    } 
    
    get_token(curr_tok);
    if (curr_tok.t != RBRACKET) {
	sstreamwrap buf; 
	buf<<"invalid index or the index ('"<<location<<"') for PARAMS should "
	    "be followed by a ']'"<<'\0';
	throw parse_exception("", _line_num, buf.str());
    }
    
    /* is this a structure parameter? check if it's followed by a dot */
    /* just peek, in case it's not a '.' */
    string struct_field = "";
    fl_token tmptok;
    bool more_struct_fields = true;
    while (more_struct_fields) {
        peek_token(tmptok);
        if (tmptok.t == DOT) {
	    pop_token(); /* pop the '.' */ 
	    get_token(curr_tok); /* and here's the structure field name */
	    /* UGLY HACK: strings are only parsed if they are quoted, so 
	       build_token doesnt know how to handle the structure field string 
	       and returns to us a LAST_TOKEN_TYPE. need to teach build_token() 
	       about structure fields. */
	    assert(curr_tok.t == LAST_TOKEN_TYPE);
	    if (struct_field == "")
	        struct_field = curr_tok.orig_text;
	    else {
	        struct_field.append(".");
		struct_field.append(curr_tok.orig_text);
	    }
        } 
	else
	    more_struct_fields = false;
    }
    
    int field_index = -1;
    sct_data_type_t dt = DT_LAST_TYPE;
    int struct_type_index = -1;
    if (struct_type == "") {
        const syscalls::param* param =
	    syscalls::get_final_syscall_param(_syscall_name, location,
					      struct_field, &field_index);

        dt = (param ? param->type : DT_LAST_TYPE);

        if (dt == DT_LAST_TYPE) {
	    sstreamwrap buf;
	    buf << "invalid PARAM requested at PARAMS["<<location<<"]."; 
	    throw parse_exception("", _line_num, buf.str());
        }
    }
    else {
        /* handle type-casting of the struct, if required. */
	if (!syscalls::has_struct(struct_type)) {
	    sstreamwrap buf;
	    buf << "invalid cast operation to an unknwon struct type '"
		<< struct_type << "' requested at PARAMS["
		<< location << "].";
	    throw parse_exception("",
				  _line_num,
				  buf.str());
	}
	const syscalls::param* struct_param =
		syscalls::struct_param_at(struct_type,
					  struct_field,
					  &field_index);
        dt = (struct_param ? struct_param->type : DT_LAST_TYPE);
	if (!struct_param || dt == DT_LAST_TYPE) {
	    sstreamwrap buf;
	    buf << "struct type '" << struct_type
		<< " has no field named '" << struct_field
		<< "' requested at PARAMS[" << location << "].";
	    throw parse_exception("",
				  _line_num,
				  buf.str());
	}
	struct_type_index = syscalls::struct_type_to_index(struct_type);
    }

    /* finally, we make the filter node */
    sct_filter_node_t* rv;

    /* ok, is this a structure param? */
    if (field_index != -1) {
	rv =  sct_make_var_indexed_field_val(VT_PARAM, location -1, 
					 field_index, dt, struct_type_index);
	return rv;
    } else { /* no, just a regular param */
	rv = sct_make_var_indexed_val(VT_PARAM, location - 1, dt);
    }

    return rv;
}

sct_filter_node_t*
filter_parser::parse_builtin_macro(const string& macro_name_copy,
				   bool has_param)
{
    string macro_name = macro_name_copy;
    cout << "macro_name '" << macro_name << "'" << endl;
    fl_token macro_param;

    if (has_param) {
        get_token(curr_tok);
        if (curr_tok.t != LPAREN)
	    throw parse_exception("", get_line_num(), 
			          "macro must be followed by a left paren '('");
        get_token(curr_tok);
        if (curr_tok.t != NUMBER && curr_tok.t != STRING)
	    throw parse_exception("", get_line_num(), 
			          "macro parameter must be number or string");
        macro_param = curr_tok;

        get_token(curr_tok);
        if (curr_tok.t != RPAREN) {
	    sstreamwrap buf; 
	    buf<<"macro '"<<macro_name<<"': expecting right paren ')' after parameter"; 
	    throw parse_exception("", get_line_num(), buf.str());
        }
    }
    else {
	macro_param.orig_text = macro_name;
	macro_param.t = BUILTIN_PARAMLESS_MACRO;
    }
    
    macro_handlers_map::const_iterator mhm_it = macro_handlers.find(macro_name);
    if (mhm_it == macro_handlers.end()) {
	sstreamwrap buf; 
	buf<<"internal error: no handler for macro '"<<macro_name<<"'"; 
	throw parse_exception("", get_line_num(), buf.str());
    }
    macro_handler_t* mh = (*mhm_it).second;
    assert(mh);

    /* finally, we make the macro's value node */
    sct_filter_node_t* v = (this->*(mh->macro))(macro_param);

    return v;
}

sct_data_type_t
filter_parser::get_data_type_for_param(unsigned int param)
{
    int dummy;
    /* we dont support structs yet, ergo the "" location field */
    const syscalls::param* p = 
	syscalls::get_final_syscall_param(_syscall_name, param, "", &dummy);
    if (p)
	return p->type;
    
    return DT_LAST_TYPE;
}


sct_filter_node_t* 
filter_parser::term(bool get)
{
    if (get)
	get_token(curr_tok);
    
    sct_filter_node_t* left, *right;

    for (;;) {
	switch(curr_tok.t) {
	case (LOGICAL_NOT):
	    right = term(true);
	    left = sct_make_logical_not(right);
	    if (!left) {
		sstreamwrap err;
		err << "syntax error at expression ending in " << curr_tok.orig_text; 
		throw parse_exception("", _line_num, err.str());
	    }
	    return left;
	case (BITWISE_NOT):
	    right = term(true);
	    left = sct_make_bit_not(right);
	    if (!left) {
		sstreamwrap err;
		err << "syntax error at expression ending in " << curr_tok.orig_text;
		throw parse_exception("", _line_num, err.str());
	    }
	    return left;
	default:
	    /* prim will call get_token() as well, if get is true, so 
	       we want to return "our" token back to the fifo */
	    if (get)
		push_token();

	    return prim(get);
	}
    }
    }

sct_filter_node_t* 
filter_parser::expr(bool get)
{
    sct_filter_node_t* left = term(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (PLUS):
	    right = term(true);
	    left = sct_make_math_plus(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	case (MINUS):
	    right = term(true);
	    left = sct_make_math_minus(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}

sct_filter_node_t* 
filter_parser::shift(bool get)
{
    sct_filter_node_t* left = expr(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (SHIFT_LEFT):
	    right = expr(true);
	    left = sct_make_bit_left_shift(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	case (SHIFT_RIGHT):
	    right = expr(true);
	    left = sct_make_bit_right_shift(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}


sct_filter_node_t* 
filter_parser::order(bool get)
{
    sct_filter_node_t* left = shift(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (LESS_THEN):
	    right = shift(true);
	    left = sct_make_less_then(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	case (LESS_OR_EQUAL):
	    right = order(true);
	    left = sct_make_less_or_equal(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	case (GREATER_THEN):
	    right = shift(true);
	    left = sct_make_greater_then(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	case (GREATER_OR_EQUAL):
	    right = order(true);
	    left = sct_make_greater_or_equal(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}

sct_filter_node_t* 
filter_parser::equality(bool get)
{
    sct_filter_node_t* left = order(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (EQUAL):
	    right = order(true);
	    left = sct_make_equal(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	case (NOT_EQUAL):
	    right = order(true);
	    left = sct_make_not_equal(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	case (PATTERN_MATCH):
	    right = order(true);
	    left = sct_make_pattern_match(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	case (IN_OP):
	    right = vector(true);
	    left = sct_make_in_match(left, right);
	    if (!left) {
		sstreamwrap err;
		err << "syntax error at expressin ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}

sct_filter_node_t* 
filter_parser::expr_bitwise_and(bool get)
{
    sct_filter_node_t* left = equality(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (BITWISE_AND):
	    right = equality(true);
	    left = sct_make_bit_and(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}

sct_filter_node_t* 
filter_parser::expr_bitwise_xor(bool get)
{
    sct_filter_node_t* left = expr_bitwise_and(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (BITWISE_XOR):
	    right = expr_bitwise_and(true);
	    left = sct_make_bit_xor(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}


sct_filter_node_t* 
filter_parser::expr_bitwise_or(bool get)
{
    sct_filter_node_t* left = expr_bitwise_xor(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (BITWISE_OR):
	    right = expr_bitwise_xor(true);
	    left = sct_make_bit_or(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}


sct_filter_node_t* 
filter_parser::expr_logical_and(bool get)
{
    sct_filter_node_t* left = expr_bitwise_or(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (LOGICAL_AND):
	    right = expr_bitwise_or(true);
	    left = sct_make_logical_and(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}

sct_filter_node_t* 
filter_parser::expr_logical_or(bool get)
{
    sct_filter_node_t* left = expr_logical_and(get);
    sct_filter_node_t* right = NULL; 

    for (;;) {
	switch(curr_tok.t) {
	case (LOGICAL_OR):
	    right = expr_logical_and(true);
	    left = sct_make_logical_or(left, right);
	    if (!left) {
		sstreamwrap err;
		err<<"syntax error at expression ending in "<<curr_tok.orig_text<<'\0';
		throw parse_exception("", _line_num, err.str());
	    }
	    break;
	default:
	    return left;
	}
    }
}

#define MAX_VECTOR_LEN 100

sct_filter_node_t* 
filter_parser::vector(bool get)
{
    sct_filter_node_t* vector = NULL;
    sct_filter_node_t** vector_nodes = new sct_filter_node_t*[MAX_VECTOR_LEN];
    int num_vector_nodes = 0;
    bool done_vector_parse = false;

    get_token(curr_tok);
    if (curr_tok.t != LPAREN)
	throw parse_exception("", _line_num, 
			      "'in' must be followed by a vector's left parant '('");
    

    while (!done_vector_parse && num_vector_nodes < MAX_VECTOR_LEN) {
	sct_filter_node_t* node = expr(get);
	if (!node) {
	    sstreamwrap buf; 
	    buf<<"bad expression at index "<<num_vector_nodes+1
		<<" of vector"<<'\0';
	    throw parse_exception("",
				  _line_num,
				  buf.str());
	}
	vector_nodes[num_vector_nodes] = node;

	switch(curr_tok.t) {
	case (COMMA):
	    num_vector_nodes++;
	    break;
	case (RPAREN):
	    num_vector_nodes++;
	    done_vector_parse = true;
	    break;
	default:
	    {
	        sstreamwrap buf; 
	        buf<<"bad token '"<<curr_tok.orig_text<<"' after index "
		   <<num_vector_nodes+1
		   <<" of vector. expecting ',' or ')'"<<'\0';
	        throw parse_exception("",
				      _line_num,
				      buf.str());
	    }
	}
    }

    if (num_vector_nodes == MAX_VECTOR_LEN && !done_vector_parse) {
	sstreamwrap buf; 
	buf<<"too many elements in vector (limit is "<<MAX_VECTOR_LEN<<")"<<'\0';
	throw parse_exception("", _line_num, buf.str());
    }

    /* fill in the real vector. */
    vector = sct_make_filter_vector(num_vector_nodes);
    if (vector) {
	int i;

	for (i=0; i < num_vector_nodes; i++) {
	    sct_insert_filter_node(vector, i, vector_nodes[i]);
	}
    }

    // prepare the next token for parsing.
    get_token(curr_tok);

    return vector;
}


// macro calculating functions.

sct_filter_node_t*
filter_parser::calc_ipaddr(const fl_token& p1)
{
    sct_filter_node_t* v = NULL;
    if (p1.t != STRING || p1.v.type != DT_STRING) {
	sstreamwrap buf; 
	buf<<"macro 'ipaddr' expects a string parameter";
	throw parse_exception("", get_line_num(), buf.str());
    }
    string ip_str = p1.v.val.str;
    cout << "filter_parser::calc_ipaddr ip_str '" << ip_str << "'" << endl;
    unsigned int ip = inet_addr(ip_str.c_str());
    if (ip == INADDR_NONE) {
	sstreamwrap buf; 
	buf<<"macro 'ipaddr' expects an internet address parameter, not '" 
	   << ip_str << "'";
	throw parse_exception("", get_line_num(), buf.str());
    }
    cout << "filter_parser::calc_ipaddr ip '" << ip << "'" << endl;
    v = sct_make_uint_val(ip);

    return v;
}

sct_filter_node_t*
filter_parser::calc_htons(const fl_token& p1)
{
    sct_filter_node_t* v;
    if (p1.t != NUMBER || p1.v.type != DT_INT || p1.v.val.ui > USHRT_MAX) {
	sstreamwrap buf; 
	buf<<"macro 'htons' expects an 'unsigned short' parameter";
	throw parse_exception("", get_line_num(), buf.str());
    }
    unsigned short num = htons((unsigned short)p1.v.val.ui);
    v = sct_make_ushort_val(num);

    return v;
}

sct_filter_node_t*
filter_parser::calc_usernametoid(const fl_token& p1)
{
    sct_filter_node_t* v;
    if (p1.t != STRING || p1.v.type != DT_STRING) {
	sstreamwrap buf; 
	buf<<"macro 'usernametoid' expects a string parameter";
	throw parse_exception("", get_line_num(), buf.str());
    }    
    string name_str = p1.v.val.str;
    cout << "filter_parser::calc_usernametoid name_str '" << name_str << "'" << endl;
    struct passwd* pw = getpwnam(name_str.c_str());
    if (pw == NULL) {
	sstreamwrap buf; 
	buf << "macro 'usernametoid' expects a user name parameter, "
		"not '" << name_str << "'";
	throw parse_exception("", get_line_num(), buf.str());
    }
    v = sct_make_short_val(pw->pw_uid);
    
    return v;
}

sct_filter_node_t*
filter_parser::calc_groupnametoid(const fl_token& p1)
{
    sct_filter_node_t* v;
    if (p1.t != STRING || p1.v.type != DT_STRING) {
	sstreamwrap buf; 
	buf<<"macro 'groupnametoid' expects a string parameter";
	throw parse_exception("", get_line_num(), buf.str());
    }    
    string name_str = p1.v.val.str;
    cout << "filter_parser::calc_groupnametoid name_str '" << name_str << "'" << endl;
    struct group* gr = getgrnam(name_str.c_str());
    if (gr == NULL) {
	sstreamwrap buf; 
	buf<<"macro 'groupnametoid' expects a group name parameter, "
		"not '" << name_str << "'";
	throw parse_exception("", get_line_num(), buf.str());
    }
    v = sct_make_short_val(gr->gr_gid);
    
    return v;
}

sct_filter_node_t*
filter_parser::calc_macro_str_to_int(const fl_token& p1)
{
    sct_filter_node_t* v = NULL;

    str_to_int_macro_map::const_iterator it =
	str_to_int_macros.find(p1.orig_text);

    if (it != str_to_int_macros.end()) {
      v = sct_make_int_val((*it).second);
    }

    return v;
}
