/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/* 
 * $Id: syscalls.cpp,v 1.11 2002/11/23 21:35:30 mulix Exp $
 */ 

/* parse the syscalls.dat file and make it available */
/* to the parser, for validity checks                */  

#include <stdio.h>
#include <iostream>
#include <fstream>
#include <list>
#include <string>
#include <algorithm>

#include "syscalls.h"
#include "validity_checks.h"
#include "logger.h"
#include "tokens_to_types.h"
#include "parse_exception.h"
#include "sstreamwrap.h"

syscalls::syscalls_set syscalls::_sys;
syscalls::structs_set syscalls::_structs;

/* defined in main.cpp */
extern int debug;

/* line number. should be moved into the 'class syscalls' */ 
int lineno; 

/* file name, likewise */ 
std::string fname; 

syscalls::syscalls()
{
}

syscalls::~syscalls()
{
}

syscalls::param::param(const param& other)
{
	name = other.name;
	is_struct = other.is_struct;
	struct_type = other.struct_type;
	is_ptr = other.is_ptr;
	is_array = other.is_array;
	array_len = other.array_len;
	type = other.type;
}

syscalls::param& syscalls::param::operator=(const param& other)
{
	if (this != &other) {
		name = other.name;
		is_struct = other.is_struct;
		struct_type = other.struct_type;
		is_ptr = other.is_ptr;
		is_array = other.is_array;
		array_len = other.array_len;
		type = other.type;
	}
	return *this;
}

/* TODO: consolidate this with parser::log_error (move it to */ 
/* the logger class?) */ 
static void log_error(const exception_intf& e)
{
	sstreamwrap buf; 
	e.print(buf);
	buf << endl << ends; 
	
	logger::log(logger::LOG_LEVEL_ERROR, "%s", buf.str().c_str());
	exit(EXIT_FAILURE); 
}

/* get_syscall_token will get a token out of the string, from position */ 
/* 'start' to the first occurence of any of the delimiters in 'delim'. */ 
/* returns: the token. 'start' now points to the beginning of the next token */ 
/* (NOT including whitespace). */ 
/* if no token is found, a parse_exception is thrown. */
static string get_syscall_token(const string& str, string::size_type& start, 
				string::size_type& end, const char* delim)
{
	assert(delim);

	end = str.find_first_of(delim, start); 
	if (end == str.npos)
		throw parse_exception(fname, lineno, "could not find token"); 
	
	string token = str.substr(start, end - start);
    
	/* skip whitspace */
	start = str.find_first_not_of(" \t,", end + 1);

	return token;
}

static void parse_params(const string& line, string::size_type pos, 
			 syscalls::param_vec& params)
{
	string::size_type p1 = pos;
	string::size_type p2;
	std::string::size_type end; 
	string match, type, name; 
	bool is_struct, is_array;
	int array_len;
	do {
		try {
			/* get the matchable qualifier */
			match = get_syscall_token(line, p1, p2, " \t,)");
			if (match == "void")
				return; 
			
			/* get the data type */
			type = get_syscall_token(line, p1, p2, " \t,");
			
			/* get the param name */
			name = get_syscall_token(line, p1, p2, " \t,)");
		} catch (exception_intf& e) {
			log_error(e); 
			return ;
		}

#if 0
		cout << "parse_params: type '" << type.c_str() << "', name '"
		     << name.c_str() << "', match '" << match.c_str() << "'" << endl;
#endif

		/* insert the new param to the vector,
		   if it's matchable, give it its attributes
		   if not, mark it as an 'invalid' param */
		if (match != "matchable"){
			name = "INVALID";
		}
	
		/* check if this is a struct or a native type. */
		is_struct = false;
		is_array = false;
		array_len = -1;
		sct_data_type_t dt = DT_LAST_TYPE;
		string struct_type;
		const string struct_prefix = "struct_";
		const string const_struct_prefix = "const_struct_";
		const string array_prefix = "array[";
		const string ptr_prefix = "ptr_";
		const string pointer = "pointer"; 
		if (type.find(struct_prefix) == 0) {
			is_struct = true;
			struct_type = type.substr(struct_prefix.length(),
						  type.length() - 
						  struct_prefix.length());
		} else if (type.find(const_struct_prefix) == 0) {
			is_struct = true;
			struct_type = type.substr(const_struct_prefix.length(),
						  type.length() - 
						  const_struct_prefix.length());
		} else if (type.find(array_prefix) == 0) {
			is_array = true;
			std::string::size_type len_beg = type.find_first_of('[') + 1;
			std::string::size_type len_end = type.find_first_of(']', len_beg);
			if (len_end == type.npos)
				throw parse_exception(fname, lineno, "could not find "
						      "ending ']' for array"); 
			std::string array_len_str = type.substr(len_beg, len_end-len_beg);
			if (!is_int(array_len_str))
				throw parse_exception(fname, lineno, "array len is not "
						      "an integer"); 
			array_len = atoi(array_len_str.c_str());
			std::string real_type = type.substr(len_end+2);
			dt = tokens::get_data_type(real_type);
			if (dt == DT_LAST_TYPE)
				throw parse_exception(fname, lineno, "unknown array "
						      "members data type."); 
		} else if (type.find(ptr_prefix) == 0) {
			/* shout on matchable ptr_foo */ 
			if (match == "matchable") {
				sstreamwrap w; 
				w << "matchable ptr found! (name = " << name << " match = " 
				  << match << ends; 
				throw parse_exception(fname, lineno, w.str()); 
			}
		} else {
			dt = tokens::get_data_type(type);
		}
		
		syscalls::param* p = new syscalls::param; 
		p->name = name; p->is_struct = is_struct;
		p->is_ptr = false; /* we only get here for non matchable ptr params */
		p->is_array = is_array; p->array_len = array_len;
		if (is_struct) {
			p->struct_type = struct_type;
		} else {
			p->type = dt;
		}

		params.push_back(*p);
	
		end = line.find_first_not_of("\t\n\r ", p2); 
		if (end == std::string::npos)
			throw parse_exception(fname, lineno, "could not find "
					      "terminating ')'"); 
	} while (line[end] != ')');
}

static bool get_mux_id(const string& sid, int& syscall, int& func)
{
	if (sscanf(sid.c_str(), "%d.%d", &syscall, &func) != 2)
		return false; 
	return true; 
}

bool syscalls::is_syscall_line(const string& line)
{
	if (line.find("asis") == 0 || line.find("include") == 0 
	    || line.find("cpp") == 0)
		return false; 

	std::string::size_type colon1, colon2; 
	colon1 = line.find(':'); 
	colon2 = line.find(':', colon1 + 1); 
	if (colon1 != std::string::npos && colon2 != std::string::npos) 
		return true; 

	return false; 
}

static void strip_comments(std::string& line)
{
	if (line[0] == '#')
		line = ""; 
	std::string::size_type p = line.find_first_of("#"); 
	line = line.substr(0, p); 
}

bool syscalls::load(string file_name)
{
	syscalls::_sys.clear();
	syscalls::_structs.clear();
	lineno = -1; 
	fname = file_name; 
	
	ifstream file(file_name.c_str());
	
	char buf[2048];
	std::string prev_line, this_line; 
	while (file.getline(buf, sizeof(buf))) {
		std::string line(buf), sep(""); 
		if (prev_line.length() && this_line.length())
			sep = ' '; 
		
		++lineno; 
		prev_line += (sep + this_line);

		this_line = line; 

		/* strio comments */ 
		strip_comments(this_line); 

		if (this_line.length() == 0) {
			if (prev_line.length()) {
				handle_line(prev_line); 
				prev_line = ""; 
			}
			continue; 
		}
		if (!this_line.find('#')) {
			if (prev_line.length()) {
				handle_line(prev_line); 
				prev_line = ""; 
			}
			continue; 
		}
		if (!this_line.find("asis")) {
			if (prev_line.length()) {
				handle_line(prev_line); 
				prev_line = ""; 
			}
			continue; 
		}
		if (!this_line.find("include")) {
			if (prev_line.length()) {
				handle_line(prev_line); 
				prev_line = ""; 
			}
			continue; 
		}
		if (!this_line.find("cpp")) {
			if (prev_line.length()) {
				handle_line(prev_line); 
				prev_line = ""; 
			}
			continue; 
		}
		if (!this_line.find("struct ")) {
			if (prev_line.length()) {
				handle_line(prev_line); 
				prev_line = ""; 
			}
			continue; 
		}
		if (is_syscall_line(this_line)) {
			if (prev_line.length()) {
				handle_line(prev_line); 
				prev_line = ""; 
			}
			continue; 
		}
	}

	prev_line += " "; 
	prev_line += this_line; 
	if (prev_line.length() > 1)
		handle_line(prev_line); 

	if (debug) {
		logger::log(logger::LOG_LEVEL_DEBUG, "syscalls.dat:\n");
		print_structs();
		print_syscalls();
		logger::log(logger::LOG_LEVEL_DEBUG, "-----------------------\n");
	}
	
	return (!syscalls::_sys.empty());
}

void syscalls::handle_line(const string& line)
{
	try { 
		if (!line.find("struct "))
			handle_struct(line); 
		if (is_syscall_line(line))
			handle_syscall(line); 

	} catch (exception_intf& e) {
		log_error(e); 
	}
}

void syscalls::handle_struct(const string& line)
{
	int struct_type_index = 0;
	string::size_type p1 = 0, p2 = 0;

	if (line.find("struct "))
		throw parse_exception(fname, lineno, "handle_struct called "
				      "with a line which isn't a struct line!");

	/* handle 'struct' lines. */
	p1 = line.find_first_of(" \t");
	if (p1 == line.npos)
		throw parse_exception(fname, lineno, "malformed struct line"); 

	/* find struct type. */
	p1 = line.find_first_not_of(" \t", p1+1);
	if (p1 == line.npos)
		throw parse_exception(fname, lineno, "malformed struct line"); 
	string* struct_type = new string;
	try {
		(*struct_type) = get_syscall_token(line, p1, p2, " \t(");
	} catch (exception_intf& e){
		delete struct_type; 
		throw; 
	}
	/* we have a name, now let's get the parameters */
	/* TODO: we allocate these on the heap instead of 
	   on the stack because of a weird bug with gcc2.91.6.
	   if allocating on the stack, the assignment of 
	   id.params = params crashes. no idea why. */
	param_vec* params = new param_vec;
	try { 
		parse_params(line, p1, *params);
	} catch (exception_intf& e) {
		log_error(e); 
		delete struct_type; 
		delete params; 
		throw; 
	} 

	struct_info* info = new struct_info; 
	info->struct_type = *struct_type;
	info->struct_type_index = struct_type_index++;
	
	/* handle struct-within-struct by expanding the 'param_vec'    */
	/* to contain all the fields of the sub-param, in the matching */
	/* loaction.                                                   */
	param_vec* expanded_params = new param_vec;
	for (unsigned int i=0; i < params->size(); i++) {
		param& a_param = (*params)[i];
		if (a_param.is_struct) {
			/* locate the struct, and add all its fields. */
			if (!has_struct(a_param.struct_type)) {
				delete struct_type; 
				delete params; 
				delete info; 
				throw parse_exception(fname, lineno, "could not find struct"); 
			}
			param_vec& s_params = struct_params(a_param.struct_type);
			for (unsigned int j=0; j < s_params.size(); j++) {
				param s_param = s_params[j];
				string* old_name = new string(s_param.name);
				s_param.name = a_param.name + "." + *old_name;
				expanded_params->push_back(s_param);
			}
		}
		else {
			expanded_params->push_back(a_param);
		}
	}

	info->params = *expanded_params;
	
	_structs[*struct_type] = info;
}

void syscalls::handle_syscall(const string& line)
{
	string::size_type p1 = 0, p2 = 0;

	/* now start parsing out the syscall id */
	p1 = line.find(':');
	if (p1 == line.npos)
		throw parse_exception(fname, lineno, "malformed syscall line"); 

	p1 = line.find_first_not_of(" \t", p1+1);
	if (p1 == line.npos)
		throw parse_exception(fname, lineno, "malformed syscall line"); 

	string* sid = new string;
	try {
		*sid = get_syscall_token(line, p1, p2, ":");
	} catch (exception_intf& e){
		delete sid; 
		throw; 
	}
	int id = -1; 
	if (is_int(*sid)) {
		id = atoi((*sid).c_str());
	} else { // first try a mux id, then abort 
		int syscall, func; 
		if (get_mux_id(*sid, syscall, func)) {
			/*  check if the syscall id is a in the id.func format */ 
			/*  we use for mux syscalls */ 
			id = MAKE_MULTIPLEX_SYSCALL_ID(syscall, func); 
		} else {
			delete sid; 
			throw parse_exception(fname, lineno, "could not "
					      "find a syscall id"); 
		}
	}
	
	/* good, we got the id, now the name */
	string* name = new string; 
	try {
		*name = get_syscall_token(line, p1, p2, " \t(");
	} catch (exception_intf& e) {
		delete name; 
		delete sid; 
		throw; 
	}

	/* we have a name, now let's get the parameters */
	/* TODO: we allocate these on the heap instead of 
	   on the stack because of a weird bug with gcc2.91.6.
	   if allocating on the stack, the assignment of 
	   id.params = params crashes. no idea why. */
	param_vec* params = new param_vec;
	try { 
		parse_params(line, p1, *params);
	} catch (exception_intf& e) {
		log_error(e); 
		delete params; 
		delete name; 
		delete sid; 
		throw; 
	} 
	
	syscall_info* info = new syscall_info; 
	info->id = id; info->params = *params;
	
	_sys[*name] = info;
}

/* get the given parameter of this syscall */
/* returns NULL if no such syscall or loc  */
/* out of range. loc index is 1 based      */
const syscalls::param*
syscalls::syscall_param_at(const string& name, unsigned short int loc)
{
	/* is there such a syscall? */
	if (!has_syscall(name)) {
		return NULL;
	}

	const param_vec& params = syscall_params(name);
	if ((loc == 0) || (loc > params.size())) {
		return NULL;
	}

	const param& p = params[loc - 1];

	return &p;
}

/* get the given parameter of this struct */
/* also set the location of this field in */
/* the struct, into 'field_loc'.          */
/* returns NULL if no such struct or no   */
/* such field in struct.                  */
const syscalls::param*
syscalls::struct_param_at(const string& struct_type,
			  const string& field,
			  int* field_loc)
{
	/* is there such a struct type? */
	if (!has_struct(struct_type)) {
		return NULL;
	}

	const param_vec& params = struct_params(struct_type);

	int field_index = 0;
	param_vec::const_iterator it = params.begin();
	for ( ; it != params.end(); field_index++,it++) {
		if ((*it).name == field)
			break;
	}
	if (it == params.end()) {
		return NULL;
	}
	if (field_loc)
		*field_loc = field_index;
	const param& p = (*it);

	return &p;
}

typedef std::pair<std::string, syscalls::syscall_info*> syscall_pair;
typedef std::list<syscall_pair> syscall_info_list; 

static bool syscall_id_lt(syscall_pair& p1, syscall_pair& p2)
{
	int sid1, sid2, fid1, fid2; 
	sid1 = SYSCALL_PART_OF_SYSCALL_ID(p1.second->id); 
	sid2 = SYSCALL_PART_OF_SYSCALL_ID(p2.second->id); 
	if (sid1 < sid2)
		return true; 
	
	if (sid1 > sid2)
		return false; 
	
	fid1 = FUNC_PART_OF_SYSCALL_ID(p1.second->id); 
	fid2 = FUNC_PART_OF_SYSCALL_ID(p2.second->id); 
	if (fid1 < fid2)
		return true; 

	return false; 
}

/* we use this class to extract the info from the _sys map to a list */ 
class extract_syscall_info {
	syscall_info_list& lst; 
public:
	extract_syscall_info(syscall_info_list& x):lst(x){} 
	void operator()(const syscall_pair& i)
		{lst.push_back(i);}
};

void syscalls::print_syscalls()
{
	char sid[32]; 

	/* sort the syscalls into this list */ 
	syscall_info_list lst; 
	for_each(_sys.begin(), _sys.end(), extract_syscall_info(lst)); 
	lst.sort(syscall_id_lt); 

	syscall_info_list::const_iterator it = lst.begin();
	for ( ; it != lst.end(); ++it){
		const string& name = (*it).first;
		int id = (*it).second->id;
		sct_get_syscall_id_as_str(id, sid, sizeof(sid)); 
		const param_vec& params = (*it).second->params;
		param_vec::const_iterator p = params.begin(); 
		string params_string;
		for ( ; p != params.end(); ++p){
			params_string += ((*p).name + ":");
			if ((*p).is_struct) {
				params_string += "struct ";
				params_string += (*p).struct_type;
			}
			else {
				params_string += tokens::get_data_type_str((*p).type);
			}
			params_string += ", ";
		}
		/* remove trailing ', ', if we have it */
		if (params_string.size() >= 2)
			params_string[params_string.size() - 2] = '\0';
		logger::log(logger::LOG_LEVEL_DEBUG, 
			    "%s : %s(%s)\n", sid, name.c_str(), 
			    params_string.c_str());  
	}
}

void
syscalls::print_structs()
{
	syscalls::structs_set::const_iterator it = _structs.begin();
	for ( ; it != _structs.end(); ++it){
		const string& struct_type = (*it).second->struct_type;
		const param_vec& params = (*it).second->params;
		param_vec::const_iterator p = params.begin(); 
		string params_string;
		for ( ; p != params.end(); ++p){
			params_string += ((*p).name + ":" + 
					  tokens::get_data_type_str((*p).type)); 
			if ((*p).is_array) {
				params_string += '[';
				char buf[40];
				sprintf(buf, "%d", (*p).array_len);
				params_string += buf;
				params_string += ']';
			}
			params_string += ", ";
		}
		/* remove trailing ', ' */
		params_string[params_string.size() - 2] = '\0';
		logger::log(logger::LOG_LEVEL_DEBUG, 
			    "struct %s(%s)\n",
			    struct_type.c_str(), 
			    params_string.c_str());  
	}
}

bool 
syscalls::has_syscall(const string& name)
{
	return (syscalls::_sys.find(name) != _sys.end());
}

bool 
syscalls::has_struct(const string& struct_type)
{
	return (syscalls::_structs.find(struct_type) != _structs.end());
}

/* return the index of the struct with the given type, '-1' if not found. */
int
syscalls::struct_type_to_index(const string& type)
{
	int index = -1;

	structs_set::iterator it = syscalls::_structs.find(type);
	if (it != syscalls::_structs.end()) {
		index = (*it).second->struct_type_index;
	}

	return index;
}

int 
syscalls::syscall_id(const string& name)
{
	syscalls_set::const_iterator it = syscalls::_sys.find(name);
	if (it == syscalls::_sys.end())
		return -1;

	return (*it).second->id;
}

std::string syscalls::syscall_name(const unsigned int id)
{
	syscalls_set::const_iterator it = syscalls::_sys.begin(); 
	for (; it != syscalls::_sys.end(); ++it) {
		if ((*it).second->id == id)
			return (*it).first; 
	}
	
	return "(unknown)"; 
}


const syscalls::param*
syscalls::get_final_syscall_param(const string& sys_name, 
				  unsigned int location, 
				  const string& field, 
				  int* field_index)
{
	const syscalls::param* p = syscalls::syscall_param_at(sys_name, location);
	if (!p)
		return NULL;

	// handle structs
	if (p->is_struct) {
		const string& struct_type = p->struct_type;
		/* is there such a struct type? */
		if (!syscalls::has_struct(struct_type)) {
			return NULL;
		}

		/* no such parameter? */
		const syscalls::param* p2 =
			syscalls::struct_param_at(struct_type, field, field_index);

		return p2;
	}
	else
		if (field_index)
			*field_index = -1;

	return p;
}

