/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/* 
 * $Id: syscalls.h,v 1.6 2002/09/20 11:42:39 mulix Exp $
 */ 

/* parse the syscalls.dat file and make it available to the parser
   for validity checks */

#ifndef SYSCALLS_H_
#define SYSCALLS_H_

#include <string>
#include <map>
#include <vector>

#include "param_types.h"

using namespace std;

class syscalls{
 public: /* public typedefs */
	
	/* a parameter has a name and a type */
	struct param; 
		
	/* each syscall has parameters and an id */
	typedef vector<param> param_vec;
	typedef struct {
		unsigned int id; 
		param_vec params;
	} syscall_info;
	
	typedef struct {
		string struct_type;
		int struct_type_index;
		param_vec params;
	} struct_info;

	struct param { 
		param() {}
		param(const param& other); 
		param& operator=(const param& other); 
		
		string name;
		bool is_struct;
		string struct_type;
		bool is_ptr;
		bool is_array;
		int array_len;
		sct_data_type_t type;
	}; 

 public: /* public methods */
	syscalls();
	~syscalls();
	
	/* load the information from the syscalls.dat file */
	static bool load(string file_name);
	
	/* do we have a syscall with name 'name'? */
	static bool has_syscall(const string& name);
	
	/* do we have a struct with type 'type'? */
	static bool has_struct(const string& type);
	
	/* return the index of the struct with the given type, '-1' if not found. */
	static int struct_type_to_index(const string& type);
	
	/* what is the id of syscall 'name'? return -1 if we dont have
	   this syscall */ 
	static int syscall_id(const string& name);

	/* what is the name of this sycall? */ 
	/* O(n) search, use with caution. */ 
	static std::string syscall_name(const unsigned int id); 
	
	/* get the parameters vector for this syscall */
	static param_vec& syscall_params(const string& name)
		{return syscalls::_sys[name]->params;}
	
	/* get the given parameter of this syscall */
	/* returns NULL if no such syscall or loc  */
	/* out of range.                           */
	static const param* syscall_param_at(const string& name,
					     unsigned short int loc);
	
	/* get the parameters vector for this struct */
	static param_vec& struct_params(const string& struct_type)
		{return syscalls::_structs[struct_type]->params;}
	
	/* get the given parameter of this struct */
	/* also set the location of this field in */
	/* the struct, into 'field_loc'.          */
	/* returns NULL if no such struct or no   */
	/* such field in struct.                  */
	static const param* struct_param_at(const string& struct_type,
					    const string& field, int* field_loc);
	
	/* get the given parameter at this location */
	/* if it's a struct, get the relevant field */
	/* in the struct. returns a pointer to the  */
	/* param, or NULL if an error occures.      */
	/* returns the field's index in field_index,*/
	/* or -1 if an error occures.               */ 
	static const param* get_final_syscall_param(const string& sys_name, 
						    unsigned int location, 
						    const string& field, 
						    int* field_index);
	
	
	static void print_syscalls(void);
	static void print_structs(void);
	
	/* TODO: these should be private and accessible via a syscalls::iterator */ 
	/* we have a map of system calls, accessible by name */
	typedef map<string, syscall_info*> syscalls_set;
	typedef map<string, struct_info*> structs_set;
	static syscalls_set _sys;
	static structs_set _structs;
 private: 
	/* parse this logical line from the file */ 
	/* may be made of of several physical lines */ 
	static void handle_line(const string& line); 

	/* does this line describe a system call line? */ 
	static bool is_syscall_line(const std::string& line); 

	/* handle a struct line */ 
	static void handle_struct(const string& line); 

	/* handle a syscall line */ 
	static void handle_syscall(const string& line); 
};

#endif /* SYSCALLS_H_ */
