/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2003 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 * $Id: encode_decode.c,v 1.11 2003/02/01 19:21:22 mulix Exp $
 */
/*
 * Rule engine - primitive types encoding/decoding functions.
 */

#include "mem_manager_public.h"
#include "archdep.h"
#include "encode_decode_public.h"
#include "membuf.h" 

/* encode the given ptr with the given size     */
/* at the given pos, as long as the buffer is   */
/* not too small.                               */
static int do_encode(struct sct_membuf* m, void* ptr, unsigned char size)
{
	int rc; 

	assert(ptr && size); 
	
	if (!ptr || !size)
		return -EINVAL; 

	rc = sct_membuf_add_chunk(m, ptr, size); 
	return rc; 
}

/* decode the given ptr at the given pos,           */
/* as long as we are not past end of buf.           */
static int do_decode(struct sct_membuf* m, void* ptr,
		     unsigned char size)
{
	int rc; 

	assert(ptr && size);

	if (!ptr || !size)
		return -EINVAL;

	rc = sct_membuf_get_chunk(m, ptr, size); 

	if (rc != size)
		return -ENOSPC; 

	return 0; 
}



/* encode the given unsigned char at the given pos, */
/* as long as the buffer is not too small.          */
int encode_uchar(struct sct_membuf* m, unsigned char c)
{
	return do_encode(m, (void*)&c, sizeof(c));
}

/* decode the given unsigned char at the given pos, */
/* as long as we are not past end of buf.           */
int decode_uchar(struct sct_membuf* m, unsigned char* pc)
{
	return do_decode(m, pc, sizeof(*pc));
}

/* encode the given signed char at the given pos, */
/* as long as the buffer is not too small.          */
int encode_char(struct sct_membuf* m, char c)
{
	return do_encode(m, (void*)&c, sizeof(c));
}

/* decode the given signed char at the given pos, */
/* as long as we are not past end of buf.           */
int
decode_char(struct sct_membuf* m, char* pc)
{
	return do_decode(m, pc, sizeof(*pc));
}

/* encode the given int at the given pos,  */
/* as long as the buffer is not too small. */
int
encode_int(struct sct_membuf* m, int i)
{
	return do_encode(m, (void*)&i, sizeof(i));
}

/* decode the given int at the given pos, */
/* as long as we are not past end of buf. */
int
decode_int(struct sct_membuf* m, int* pi)
{
	return do_decode(m, pi, sizeof(*pi));
}

/* encode the given unsigned int at the given pos, */
/* as long as the buffer is not too small.         */
int
encode_uint(struct sct_membuf* m, unsigned int i)
{
	return do_encode(m, (void*)&i, sizeof(i));
}

/* decode the given unsigned int at the given pos, */
/* as long as we are not past end of buf.          */
int
decode_uint(struct sct_membuf* m, unsigned int* pi)
{
	return do_decode(m, pi, sizeof(*pi));
}

/* encode the given long at the given pos, */
/* as long as the buffer is not too small. */
int
encode_long(struct sct_membuf* m, long l)
{
	return do_encode(m, (void*)&l, sizeof(l));
}

/* decode the given long at the given pos, */
/* as long aswe are not past end of buf.   */
int
decode_long(struct sct_membuf* m, long* pl)
{
	return do_decode(m, pl, sizeof(*pl));
}

/* encode the given unsigned long at the given pos, */
/* as long as the buffer is not too small.          */
int
encode_ulong(struct sct_membuf* m, unsigned long l)
{
	return do_encode(m, (void*)&l, sizeof(l));
}

/* decode the given unsigned long at the given pos, */
/* as long aswe are not past end of buf.            */
int
decode_ulong(struct sct_membuf* m, unsigned long* pl)
{
	return do_decode(m, pl, sizeof(*pl));
}

/* encode the given long long at the given pos, */
/* as long as the buffer is not too small. */
int
encode_long_long(struct sct_membuf* m, long long l)
{
	return do_encode(m, (void*)&l, sizeof(l));
}

/* decode the given long long at the given pos, */
/* as long aswe are not past end of buf.   */
int
decode_long_long(struct sct_membuf* m, long long* pl)
{
	return do_decode(m, pl, sizeof(*pl));
}

/* encode the given short at the given pos, */
/* as long as the buffer is not too small.  */
int
encode_short(struct sct_membuf* m, short s)
{
	return do_encode(m, (void*)&s, sizeof(s));
}

/* decode the given short at the given pos, */
/* as long as we are not past end of buf.   */
int
decode_short(struct sct_membuf* m, short* ps)
{
	return do_decode(m, ps, sizeof(*ps));
}

/* encode the given unsigned short at the given pos, */
/* as long as the buffer is not too small.           */
int
encode_ushort(struct sct_membuf* m, unsigned short s)
{
	return do_encode(m, (void*)&s, sizeof(s));
}

/* decode the given unsigned short at the given pos, */
/* as long as we are not past end of buf.            */
int
decode_ushort(struct sct_membuf* m, unsigned short* ps)
{
	return do_decode(m, ps, sizeof(*ps));
}

/* encode the given float at the given pos, */
/* as long as the buffer is not too small.  */
int
encode_float(struct sct_membuf* m, float f)
{
	return do_encode(m, (void*)&f, sizeof(f));
}

/* decode the given float at the given pos, */
/* as long as we are nto past end of buf.   */
int decode_float(struct sct_membuf* m, float* pf)
{
	return do_decode(m, pf, sizeof(*pf));
}

/* encode the given string at the given pos, */
/* as long as the buffer is not too small.   */
int encode_string(struct sct_membuf* m, const char* str)
{
	/* handle NULL strings */
	int len = (str ? strlen(str) : 0);
	int rc; 
    
	/* encode the string's len, and then the string - */
	/* withut the string's delimiting '\0' character. */
	if ((rc = sct_membuf_add_chunk(m, &len, sizeof(len))) < 0)
		return rc; 

	/* don't encode a zero length strings */ 
	if (!len)
		return sizeof(len); 

	if ((rc = sct_membuf_add_chunk(m, str, len)) < 0)
		return rc; 

	return len + sizeof(len); 
}

/* decode the given string at the given pos, */
/* as long as we are not past end of buf.    */
int decode_string(struct sct_membuf* m, char** pstr)
{
	int rc = 0;
	int len;
    
	/* decode the string's len, and then the string -    */
	/* without the string's delimiting '\0' character.   */
	/* finally, append a '\0' char to the end of string. */
	if ((rc = decode_int(m, &len)) < 0)
		return rc; 

	/* handle NULL strings */
	if (len == 0) {
		*pstr = NULL;
		return 0;
	}

	(*pstr) = mem_alloc(len+1);
	if (!*pstr)
		return -ENOMEM; 

	rc = sct_membuf_get_chunk(m, *pstr, len); 
	if (rc < 0) {
		mem_free(*pstr); 
		*pstr = NULL; 
		return rc; 
	} 
	if (rc != len) {
		mem_free(*pstr); 
		*pstr = NULL; 
		return -ENOSPC; 
	}

	(*pstr)[len] = '\0'; 

	return 0; 
}

/* encode the given buffer 'in' at the given pos, */
/* as long as the destination buffer 'out' is not too small. */
int encode_buffer(struct sct_membuf* m, const void* in, 
	      unsigned int size)
{
	/* handle NULL buffers */
	int len = (in ? size : 0);
	int rc; 
    
	/* encode the buffer's len, and then the buffer */ 
	if ((rc = sct_membuf_add_chunk(m, &len, sizeof(len))) < 0)
		return rc; 

	/* don't encode a zero length buffer */ 
	if (!len)
		return sizeof(len); 

	if ((rc = sct_membuf_add_chunk(m, in, len)) < 0)
		return rc; 

	return len + sizeof(len); 
}

/* decode the given buffer '*pdst' at the given pos, */
/* as long as we are not past end of the buf 'src'.  */
int decode_buffer(struct sct_membuf* m, void** pdst, 
	      unsigned int* psize)
{
	int rc; 
	int len;
    
	/* decode the buffer's len, and then the buffer */ 
	if ((rc = decode_int(m, &len)) < 0)
		return rc; 

	/* handle NULL strings */
	if (len == 0) {
		*pdst = NULL;
		*psize = 0; 
		return 0;
	}

	(*pdst) = mem_alloc(len);
	if (!(*pdst))
		return -ENOMEM; 

	rc = sct_membuf_get_chunk(m, *pdst, len); 
	if (rc < 0) {
		mem_free(*pdst); 
		*pdst = NULL; 
		return rc; 
	}
	if (rc != len) { 
		mem_free(*pdst); 
		*pdst = NULL; 
		return -ENOSPC; 
	}
	
	return 0; 
}

/* encode the given ptr at the given pos, */
/* as long as the buffer is not too small.          */
int encode_ptr(struct sct_membuf* m, void* ptr)
{
	return do_encode(m, (void*)&ptr, sizeof(ptr));
}

/* decode the given ptr at the given pos, */
/* as long aswe are not past end of buf.            */
int decode_ptr(struct sct_membuf* m, void** pp)
{
	return do_decode(m, pp, sizeof(*pp));
}
