/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2003 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 * $Id: filter_in_out.c,v 1.23 2003/01/31 14:08:45 mulix Exp $
 */

/*
 * Rule engine - filter input/output and serialization routines.
 */

#include "mem_manager_public.h"
#include "encode_decode_public.h"
#include "rule_engine_private.h"


/* operators symbols, used for printing. */
const char* operators_symbols[] = {
	"==", /* OT_EQUAL */
	"<",  /* OT_LESS */
	">",  /* OT_GREATER */
	"<=", /* OT_LESS_EQUAL */
	">=", /* OT_GREATER_EQUAL */
	"!=", /* OT_NOT_EQUAL */
	"+",  /* OT_PLUS */
	"-",  /* OT_MINUS */
	"<<", /* OT_BIT_SHIFT_LEFT */
	">>", /* OT_BIT_SHIFT_RIGHT */
	"&",  /* OT_BIT_AND */
	"|",  /* OT_BIT_OR */
	"^",  /* OT_BIT_XOR */
	"~",  /* OT_BIT_NOT */
	"&&", /* OT_LOGIC_AND */
	"||", /* OT_LOGIC_OR */
	"!",  /* OT_LOGIC_NOT */
	"~=", /* OT_PATTERN_MATCH */
	"in", /* OT_IN_MATCH */
};

/*
 * Private functions.
 */

/* print spaces in the ammount given. */
void print_indent(unsigned int len)
{
	char buf[200];
	unsigned int bufsiz = sizeof(buf) - 1;

	len = (len > bufsiz) ? bufsiz : len;

	memset(buf, ' ', len);
	buf[len] = '\0';

	PPRINT(buf);
}

/* get a var as string */ 
const char* sct_get_var_type_as_str(sct_var_type_t var)
{
	switch (var) {
	case VT_UNDEF: return "undef"; 
	case VT_SYSCALL_ID: return "syscall id";
	case VT_PROC_UID: return "uid"; 
	case VT_PROC_EUID: return "uid"; 
	case VT_PROC_SUID: return "suid"; 
	case VT_PROC_GID: return "gid"; 
	case VT_PROC_EGID: return "egid"; 
	case VT_PROC_SGID: return "sgid"; 
	case VT_PROC_PID: return "pid"; 
	case VT_PROC_COMM: return "command"; 
	case VT_RETVAL: return "return value"; 
	case VT_PROC_PPID: return "parent pid"; 
	case VT_PROC_PCOMM: return "parent command"; 
	case VT_PROC_ANCESTORS: return "parent tree"; 
	case VT_PROC_CWD: return "cwd"; 
	case VT_PARAM: return "syscall parameter"; 
	case VT_LAST_TYPE: return "last type"; 
	default: return "unknown"; 
	}	
}

/* print a var node. */
static void sct_print_filter_node_var(struct tracker_filter_node* f, 
				      unsigned int indent)
{
	print_indent(indent);
	PPRINT("var '%s'", sct_get_var_type_as_str(f->u.var.type));
	PPRINT(" (index '%d'", f->u.var.var_index + 1);
	PPRINT(" field '%d'", f->u.var.field_index);
	PPRINT(" struct type '%d'", f->u.var.struct_type);
	PPRINT(" data type '%d')", f->data_type); 
	PPRINT("\n");
}

/* print a const node. */
static void sct_print_filter_node_const(struct tracker_filter_node* f, 
					unsigned int indent)
{
	print_indent(indent);
	switch (CONST_DATA_TYPE(f)) {
	case DT_BOOL:
		PPRINT("bool: %d\n", CONST_BOOL(f));
		break;
	case DT_CHAR:
		PPRINT("char: %c (%d)\n",
		       CONST_CHAR(f), (int)CONST_CHAR(f));
		break;
	case DT_INT:
		PPRINT("int: %d\n", CONST_INT(f));
		break;
	case DT_UINT:
		PPRINT("unsigned int: %u\n", CONST_UINT(f));
		break;
	case DT_SHORT:
		PPRINT("short: %hd\n", CONST_SHORT(f));
		break;
	case DT_USHORT:
		PPRINT("unsigned short: %hu\n", CONST_USHORT(f));
		break;
	case DT_LONG:
		PPRINT("long: %ld\n", CONST_LONG(f));
		break;
	case DT_ULONG:
		PPRINT("unsigned long: %lu\n", CONST_ULONG(f));
		break;
	case DT_FLOAT:
		PPRINT("float: %f\n", CONST_FLOAT(f));
		break;
	case DT_LONG_LONG: 
		PPRINT("long long: %lli\n", CONST_LONG_LONG(f));
		break;
	case DT_STRING:
	{
		const char* str = CONST_STRING(f);
		PPRINT("str: %s\n", (str ? str : "(nil)"));
		break; 
	}
	case DT_BUFFER:
	{
		char* out = sct_buffer_to_string(CONST_BUFFER(f), BUFFER_SIZE(f)); 
		PPRINT("buffer [size %d]: %s\n", BUFFER_SIZE(f), 
		       (out ? out : "(nil)"));
		if (out)
			mem_free(out); 
		break; 
	}
	case DT_PTR:
		PPRINT("pointer: %p\n", CONST_PTR(f));
		break;
	default:
		/* shouldn't happen. */
		PPRINT("(unknown type %d)\n", CONST_DATA_TYPE(f));
		break;
	}
}

/* print a compound node. */
static void sct_print_filter_node_compound(struct tracker_filter_node* f, 
					   unsigned int indent)
{
	const char* op = "(error)";

	if (f->u.compound.op >= 0 && f->u.compound.op < OT_LAST_OPERATOR)
		op = operators_symbols[f->u.compound.op];
	print_indent(indent);
	PPRINT("operator: '%s' (%d)\n", op, f->u.compound.op);
	if (f->u.compound.node1)
		do_print_filter_node(f->u.compound.node1, indent + INDENT_LEVEL);
	else {
		print_indent(indent);
		PPRINT("Empty\n");
	}
	if (f->u.compound.node2)
		do_print_filter_node(f->u.compound.node2, indent + INDENT_LEVEL);
	else {
		print_indent(indent);
		PPRINT("Empty\n");
	}
}

/* recursively print the given filter node, beginning */
/* with the given indentation level.                  */
void do_print_filter_node(struct tracker_filter_node* f, unsigned int indent)
{
	assert(f); 

	switch(f->type) {
	case NT_VAR:
		sct_print_filter_node_var(f, indent);
		break;
	case NT_CONST:
		sct_print_filter_node_const(f, indent);
		break;
	case NT_COMPOUND:
		sct_print_filter_node_compound(f, indent);
		break;
	case NT_VECTOR:
		sct_print_filter_node_vec(f, indent);
		break;
	default:
		/* shouldn't happen. */
		print_indent(indent);
		PPRINT("Error: invalid node type '%d'\n", f->type);
		break;
	}
}


/* print the given filter node (and all its sub-nodes) to output. */
/* output is stdout if in user-mode, or using printk if in kernel */
/* mode. the output is being formatted to reflext the hierarchy   */
/* of the expression stored in the tree - good for debugging.     */
void sct_print_filter_node(sct_filter_node_t* f, unsigned int indent)
{
	if (f)
		do_print_filter_node(f, indent);
	else {
		print_indent(indent); 
		PPRINT("Empty filter node\n");
	}
}


/*
 * Filter serialization/de-serialization functions.
 */


/*
 * Private functions.
 */

/* serialize a 'var' filter node. */
static int sct_serialize_filter_node_var(sct_filter_node_t* f, 
					 struct sct_membuf* m)
{
	int rc; 
	int sum = 0; 

	if ((rc = encode_char(m, NT_VAR)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_char(m, f->data_type)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_char(m, f->u.var.type)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_int(m, f->u.var.struct_type)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_int(m, f->u.var.var_index)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_int(m, f->u.var.field_index)) < 0)
		return rc; 

	sum += rc; 
	return sum; 
}

/* de-serialize a 'var' filter node. */
static int sct_deserialize_filter_node_var(sct_filter_node_t** pf, 
					   struct sct_membuf* m)
{
	char var_data_type;
	char var_type;
	int struct_type;
	int var_index;
	int var_field;

	int rc; 

	if ((rc = decode_char(m, &var_data_type)) < 0)
		return rc; 

	if ((rc = decode_char(m, &var_type)) < 0)
		return rc; 

	if ((rc = decode_int(m, &struct_type)) < 0)
		return rc; 

	if ((rc = decode_int(m, &var_index)) < 0)
		return rc; 

	if ((rc = decode_int(m, &var_field)) < 0)
		return rc; 

	if (!(*pf = sct_make_var_indexed_field_val(var_type, var_index, var_field,
						   var_data_type, struct_type)))
		return -EINVAL; 
	
	return 0; 
}

/* serialize a 'const' filter node. */
static int sct_serialize_filter_node_const(sct_filter_node_t* f, struct sct_membuf* m)
{
	int rc; 
	int sum = 0; 

	if ((rc = encode_char(m, NT_CONST)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_char(m, f->data_type)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_char(m, f->u.constant.type)) < 0)
		return rc; 

	sum += rc; 
	switch (f->u.constant.type) {
	case DT_BOOL:
		rc = encode_char(m, CONST_BOOL(f));
		break;
	case DT_CHAR:
		rc = encode_char(m, CONST_CHAR(f));
		break;
	case DT_SHORT:
		rc = encode_short(m, CONST_SHORT(f));
		break;
	case DT_USHORT:
		rc = encode_ushort(m, CONST_USHORT(f));
		break;
	case DT_INT:
		rc = encode_int(m, CONST_INT(f));
		break;
	case DT_UINT:
		rc = encode_uint(m, CONST_UINT(f));
		break;
	case DT_LONG:
		rc = encode_long(m, CONST_LONG(f));
		break;
	case DT_ULONG:
		rc = encode_ulong(m, CONST_ULONG(f));
		break;
	case DT_LONG_LONG: 
		rc = encode_long_long(m, CONST_LONG_LONG(f));
		break;
	case DT_FLOAT:
		rc = encode_float(m, CONST_FLOAT(f));
		break;
	case DT_STRING:
		rc = encode_string(m, CONST_STRING(f));
		break; 
	case DT_BUFFER:
		rc = encode_buffer(m, CONST_BUFFER(f), 
				   BUFFER_SIZE(f)); 
		break;
	case DT_PTR:
		rc = encode_ptr(m, CONST_PTR(f));
		break;
	default:
		/* shouldn't happen. */
		rc = -EINVAL;
		break;
	}


	if (rc < 0)
		return rc; 

	sum += rc; 
	return sum;
}

/* de-serialize a 'const' filter node. */
static int sct_deserialize_filter_node_const(sct_filter_node_t** pf, struct sct_membuf* m)
{
	char data_type;
	char constant_type;
	int rc; 

	if ((rc = decode_char(m, &data_type)) < 0)
		return rc; 

	if ((rc = decode_char(m, &constant_type)) < 0)
		return rc; 

	rc = -ENOMEM; 
	switch (constant_type) {
	case DT_BOOL:
	{
		char c;
		if ((rc = decode_char(m, &c)) < 0)
			return rc; 

		if (!(*pf = sct_make_bool_val(c)))
			return -ENOMEM; 

		break;
	}
	case DT_CHAR:
	{
		char c;
		if ((rc = decode_char(m, &c) < 0))
			return rc; 

		if (!(*pf = sct_make_char_val(c)))
			return -ENOMEM; 

		break;
	}	
	case DT_SHORT:
	{
		short s;
		if ((rc = decode_short(m, &s)) < 0)
			return rc; 
		
		if (!(*pf = sct_make_short_val(s)))
			return -ENOMEM; 
		break;
	}
	case DT_USHORT:
	{
		unsigned short us;
	
		if ((rc = decode_ushort(m, &us)) < 0)
			return rc; 

		if (!(*pf = sct_make_ushort_val(us)))
			return -ENOMEM; 
		break;
	}
	case DT_INT:
	{
		int i;
		if ((rc = decode_int(m, &i)) < 0)
			return rc; 

		if (!(*pf = sct_make_int_val(i)))
			return -ENOMEM; 

		break;
	}
	case DT_UINT:
	{
		unsigned int ui;
		
		if ((rc = decode_uint(m, &ui)) < 0)
			return rc; 

		if (!(*pf = sct_make_uint_val(ui)))
			return -ENOMEM; 

		break;
	}
	case DT_LONG:
	{
		long l;
		
		if ((rc = decode_long(m, &l)) < 0)
			return rc; 
		
		if (!(*pf = sct_make_long_val(l)))
			return -ENOMEM; 
		
		break;
	}
	case DT_ULONG:
	{
		unsigned long ul;
		
		if ((rc = decode_ulong(m, &ul)) < 0)
			return rc; 
		
		if (!(*pf = sct_make_ulong_val(ul)))
			return -ENOMEM; 
		
		break;
	}
	case DT_LONG_LONG:
	{
		long long ll;
		
		if ((rc = decode_long_long(m, &ll)) < 0)
			return rc; 

		if (!(*pf = sct_make_long_long_val(ll)))
			return -ENOMEM; 

		break; 
	}
	case DT_FLOAT:
	{
		float f;
		
		if ((rc = decode_float(m, &f)) < 0)
			return rc; 

		if (!(*pf = sct_make_float_val(f)))
			return -ENOMEM; 

		break;
	}
	case DT_STRING:
	{
		char* str;
		
		if ((rc = decode_string(m, &str)) < 0)
			return rc; 

		if (!(*pf = sct_make_string_val_without_copy(str)))
			return -ENOMEM; 
		break;
	}
	case DT_BUFFER:
	{
		void* in; 
		unsigned int size; 
		
		if ((rc = decode_buffer(m, &in, &size)) < 0)
			return rc; 
		
		if (!(*pf = sct_make_buffer_val_without_copy(in, size))) {
			mem_free(in); 
			return -ENOMEM; 
		}
		break; 
	}
	case DT_PTR:
	{
		void* vp;
		if ((rc = decode_ptr(m, &vp)) < 0)
			return rc; 
		if (!(*pf = sct_make_ptr_val(vp)))
			return -ENOMEM; 
		break;
	}
	default: 
		/* shouldn't happen. */
		rc = -EINVAL; 
		break; 
	}

	return rc;
}

/* serialize a 'compound' filter node, and its sub-nodes. */
	static int sct_serialize_filter_node_compound(sct_filter_node_t* f,
					      struct sct_membuf* m)
{
	int rc; 
	int sum = 0; 

	if ((rc = encode_char(m, NT_COMPOUND)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_char(m, f->data_type)) < 0)
		return rc; 
	
	sum += rc; 
	if ((rc = encode_char(m, f->u.compound.op)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = do_serialize_filter_node(f->u.compound.node1, m)) < 0)
		return rc; 


	sum += rc; 
	if ((rc = do_serialize_filter_node(f->u.compound.node2, m)) < 0)
		return rc; 

	sum += rc; 
	return sum; 
}

/* de-serialize a 'compound' filter node, and its sub-nodes. */
static int sct_deserialize_filter_node_compound(sct_filter_node_t** pf, 
						struct sct_membuf* m)
{
	char data_type;
	char op;
	sct_filter_node_t* f1 = NULL;
	sct_filter_node_t* f2 = NULL;
	int rc; 

	if ((rc = decode_char(m, &data_type)))
		return rc; 

	if ((rc = decode_char(m, &op)))
		return rc; 

	if ((rc = do_deserialize_filter_node(&f1, m)))
		return rc; 

	if ((rc = do_deserialize_filter_node(&f2, m)))
		return rc; 

	if (!(*pf = sct_make_compound_node_with_type(op, f1, f2, data_type)))
		rc = -EINVAL; 

	/* free the sub-nodes, if we failed somewhere along the path. */
	if (rc < 0) {
		if (f1)
			sct_free_filter_node(f1);
		if (f2)
			sct_free_filter_node(f2);
	}
	
	return rc;
}

/* serialize a filter node (and all its sub-nodes) into a */
/* buffer, for export purposes. returns the size of the   */
/* serialized data, -ERRNO on error                       */ 
int do_serialize_filter_node(sct_filter_node_t* f, struct sct_membuf* m)
{
	int rc = 0;
	int sum = 0; 

	if (!f) {
		rc = encode_char(m, NT_NIL); 
		return rc; 
	}

	switch (f->type) {
	case NT_VAR:
		rc = sct_serialize_filter_node_var(f, m);
		break;
	case NT_CONST:
		rc = sct_serialize_filter_node_const(f, m);
		break;
	case NT_COMPOUND:
		rc = sct_serialize_filter_node_compound(f, m);
		break;
	case NT_VECTOR:
		rc = sct_serialize_filter_node_vector(f, m);
		break;
	default:
		/* shouldn't happen. */
		rc = -EINVAL;
		break;
	}

	if (rc < 0)
		return rc; 

	sum += rc; 
	return sum; 
}

/* de-serialize a filter node (and all its sub-nodes)     */
/* from a buffer, for import purposes. return 0 on        */
/* success, -ERRNO on error                               */ 
int do_deserialize_filter_node(sct_filter_node_t** pf,
			       struct sct_membuf* m)
{
	int rc = 0;
	char nt;

	if ((rc = decode_char(m, &nt)))
		return rc; 

	switch (nt) {
	case NT_VAR:
		rc = sct_deserialize_filter_node_var(pf, m);
		break;
	case NT_CONST:
		rc = sct_deserialize_filter_node_const(pf, m);
		break;
	case NT_COMPOUND:
		rc = sct_deserialize_filter_node_compound(pf, m);
		break;
	case NT_VECTOR:
		rc = sct_deserialize_filter_node_vector(pf, m);
		break;
	case NT_NIL:
		/* a NULL filter_node. */
		*pf = NULL;
		rc = 0;
		break;
	default:
		/* shouldn't happen. */
		rc = -EINVAL;
		break;
	}

	return rc;
}

/*
 * Public interface implementation.
 */

/* serialize a filter node (and all its sub-nodes) into a */
/* buffer, for export purposes. returns the size of the   */
/* serialized data, -ERRNO on error.                      */ 
int sct_serialize_filter_node(sct_filter_node_t* f, struct sct_membuf* m)
{
	int rc = do_serialize_filter_node(f, m); 

	return rc; 
}

/* de-serialize a filter node (and all its sub-nodes)     */
/* from a buffer, for import purposes. returns the new    */
/* allocated filter node (the top of the tree), or NULL   */
/* if an error occured.                                   */
sct_filter_node_t* sct_deserialize_filter_node(struct sct_membuf* m)
{
	sct_filter_node_t* f = NULL;
	int rc = 0;

	if (!(rc = do_deserialize_filter_node(&f, m)))
		return f; 

	return NULL; 
}

