/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: filter_vars.c,v 1.19 2003/01/27 17:20:54 mulix Exp $
 */

#include "mem_manager_public.h"
#include "filter_vars_public.h"

/* constants. */
#define MAX_PARAM_NUM 5		/* highest supported parameter. */

typedef enum {
	CB_TYPE_NONE = 0,
	CB_TYPE_VAR,
	CB_TYPE_INDEXED,
	CB_TYPE_INDEXED_FIELD
} callback_type_t;

/* attribute of a var. */
struct var_attr {
	int is_set;
	int val;
};

/* var struct - contains data about each var */
struct var_struct {
	/* is this variable persistent? */ 
	/* (FIXME: what does 'persistent' mean here?) */ 
	/* presumably, should not be freed between requests */ 
	int persistent;
	/* when releasing this var, do we own the data */ 
	/* (and therefore should free it?) */ 
	int free_data; 
	/* the value of this variable (added once the callback was executed) */ 
	void *value;
	/* attributes for this variable */ 
	struct var_attr attrs[VAN_LAST_TYPE];
	/* data fot the callback */ 
	void *data;
	/* data type */ 
	sct_data_type_t dtype;
	/* callback type */ 
	callback_type_t cb_type;
	/* callback func */ 
	union {
		callback_func var;
		callback_indexed_func indexed;
		callback_indexed_field_func indexed_field;
	} f;
};

/* list of variables data. */
struct var_struct vars[VT_LAST_TYPE];
struct var_struct params[MAX_PARAM_NUM];
static int var_initialized = 0;

/*
 * Private functions.
 */

static void init_var_struct(struct var_struct* v)
{
	v->persistent = 0; 
	v->free_data = 0; 
	v->value = NULL; 
	memset(v->attrs, 0, sizeof(v->attrs)); 
	v->data = NULL; 
	v->dtype = DT_LAST_TYPE;
	v->cb_type = CB_TYPE_NONE;
	v->f.var = NULL;
}

/* we use this to clear the var struct from */ 
/* clear_params_list() and reset_all_vars() */ 
static void clear_var_struct(struct var_struct* v)
{
	sct_free_filter_node((sct_filter_node_t*)v->value);
	v->value = NULL;
	memset(v->attrs, 0, sizeof(v->attrs)); 

	if (!v->persistent) {
		v->cb_type = CB_TYPE_NONE;
		v->f.var = NULL;
		if (v->free_data)
			mem_free(v->data); 

		v->data = NULL;
		v->dtype = DT_LAST_TYPE;
		v->free_data = 0; 
	}
}

/* get the value of the given parameter */
static sct_filter_node_t*
get_param(int param_index)
{
	sct_filter_node_t* val = NULL;

	if (var_initialized && (param_index >= 0 && param_index < MAX_PARAM_NUM)) {
		val = params[param_index].value;
		if (val == NULL && params[param_index].cb_type == CB_TYPE_INDEXED &&
		    params[param_index].f.indexed != NULL) {
			val = params[param_index].f.indexed(VT_PARAM,
							    param_index,
							    params[param_index].dtype,
							    params[param_index].data);
			params[param_index].value = val;
			/* persistent callback support for params isn't needed for now */
			params[param_index].f.indexed = NULL; 
			params[param_index].cb_type = CB_TYPE_NONE; 
		}
	}

	return val;
}

/* get the value of the given field of the given parameter */
static sct_filter_node_t*
get_param_field(int param_index, int field_num, int struct_type)
{
	sct_filter_node_t* field = NULL;

	if (var_initialized && (param_index >= 0 && param_index < MAX_PARAM_NUM)) {
		if (params[param_index].cb_type == CB_TYPE_INDEXED_FIELD &&
		    params[param_index].f.indexed_field != NULL) {
			field = params[param_index].f.indexed_field(
				VT_PARAM,
				param_index,
				field_num,
				params[param_index].dtype,
				params[param_index].data, struct_type);
		}
	}

	return field;
}

/* get the value of the given attribute of the given parameter */
/* returns '1' if its set, '0' if its not. the value is set    */
/* into '*p_val'.                                              */
static int
get_param_attr(int param_index, sct_var_attr_name_t name, int* p_val)
{
	int is_set = 0;

	if (var_initialized && (param_index >= 0 && param_index < MAX_PARAM_NUM)) {
		if (params[param_index].attrs[name].is_set) {
			*p_val = params[param_index].attrs[name].val;
			is_set = 1;
		}
	}

	return is_set;
}


/* set the value of the given parameter to the given value. */
static void
set_param(int param_index, sct_filter_node_t* param_val)
{
	/* range checking. */
	if (param_index < 0 || param_index >= MAX_PARAM_NUM)
		return;

	sct_free_filter_node(params[param_index].value);
	params[param_index].value = param_val;
}

/* set the given attribute of the given parameter to the given value. */
static void
set_param_attr(int param_index, sct_var_attr_name_t name, int val)
{
	/* range checking. */
	if (param_index < 0 || param_index >= MAX_PARAM_NUM)
		return;

	params[param_index].attrs[name].is_set = 1;
	params[param_index].attrs[name].val = val;
}

/* set the value of the given parameter to the callback function */
static void
set_param_callback(int param_index,
		   callback_indexed_func func,
		   sct_data_type_t dtype,
		   void *data)
{
	/* range checking. */
	if (param_index < 0 || param_index >= MAX_PARAM_NUM)
		return;

	params[param_index].cb_type = CB_TYPE_INDEXED;
	params[param_index].f.indexed = func;
	params[param_index].data = data;
	params[param_index].dtype = dtype;
	params[param_index].free_data = 0; 
}

/* set the value of the given 2 parameters to the callback function */
static void
set_param_callback2(int param_index,
		    callback_indexed_func func,
		    sct_data_type_t dtype,
		    void *data1, 
		    void *data2)
{
	struct sct_aggregate_data2* d; 

	/* range checking. */
	if (param_index < 0 || param_index >= MAX_PARAM_NUM)
		return;

	if (!(d = mem_alloc(sizeof(*d))))
		return; 
    
	d->v1 = data1; 
	d->v2 = data2; 

	params[param_index].cb_type = CB_TYPE_INDEXED;
	params[param_index].f.indexed = func;
	params[param_index].data = d; 
	params[param_index].dtype = dtype;
	params[param_index].free_data = 1; 
}

/* set the value of the given parameter's field to the callback function */
static void
set_param_field_callback(int param_index,
			 callback_indexed_field_func func,
			 sct_data_type_t dtype,
			 void *data)
{
	/* range checking. */
	if (param_index < 0 || param_index >= MAX_PARAM_NUM)
		return;

	params[param_index].cb_type = CB_TYPE_INDEXED_FIELD;
	params[param_index].f.indexed_field = func;
	params[param_index].data = data;
	params[param_index].dtype = dtype;
}

/* free all allocated parameter values. */
static void
free_params_list(void)
{
	int i;

	for (i=0; i<MAX_PARAM_NUM; i++) 
		clear_var_struct(&params[i]); 
}

/* initialized the parameter array. */
static void
init_params(void)
{
	int i;
	for (i = 0; i < MAX_PARAM_NUM; i++) 
		init_var_struct(&params[i]); 
}

/* initialized the variables array. */
static void
init_vars(void)
{
	int i;
    
	init_params();
	
	for (i=0; i<VT_LAST_TYPE; i++) 
		init_var_struct(&vars[i]); 
	
	var_initialized = 1;
}


/*
 * Kernal variables manipulation code.
 */

/* empty variable values array. */
void reset_all_vars(void)
{
	int i; 

	if (!var_initialized) {
		init_vars();
	}
	
	for (i=0; i<VT_LAST_TYPE; i++) {
		switch (i) {
		case VT_PARAM:
			free_params_list();
			break;
		default:
			clear_var_struct(&vars[i]); 
			break; 
		}
	}
}

/* store a given variable's value. */
void set_var(sct_var_type_t type, sct_filter_node_t* val)
{
	assert(val); 

	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch (type) {
		case VT_PARAM:
			/* this is an indexed variable. */
			break;
		default:
			vars[type].value = val;
			break;
		}
	}
}

/* store a given variable's callback function. */
static void _set_var_callback(sct_var_type_t type,
			      int persistent,
			      callback_func func,
			      void* data)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch (type) {
		case VT_PARAM:
			/* this is an indexed variable. */
			break;
		default:
			vars[type].cb_type = CB_TYPE_VAR;
			vars[type].f.var = func;
			vars[type].data = data;
			vars[type].persistent = persistent;
			vars[type].free_data = 0; 
			break;
		}
	}
}

void set_var_callback(sct_var_type_t type, callback_func func, void* data)
{
	_set_var_callback(type, 0, func, data);
}

void set_persistent_callback(sct_var_type_t type, callback_func func, void* data)
{
	_set_var_callback(type, 1, func, data);
}

void set_var_indexed(sct_var_type_t type, int var_index, sct_filter_node_t* val)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch (type) {
		case VT_PARAM:
			set_param(var_index, val);
			break;
		default:
			/* these are not indexed variables. */
			break;
		}
	}
}

void set_var_indexed_callback(sct_var_type_t type,
			      int var_index,
			      callback_indexed_func func,
			      sct_data_type_t dtype,
			      void *data)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch (type) {
		case VT_PARAM:
			set_param_callback(var_index, func, dtype, data);
			break;
		default:
			/* these are not indexed variables. */
			break;
		}
	}
}

void set_var_indexed_callback2(sct_var_type_t type,
			       int var_index,
			       callback_indexed_func func,
			       sct_data_type_t dtype,
			       void *data1, 
			       void *data2)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch (type) {
		case VT_PARAM:
			set_param_callback2(var_index, func, dtype, data1, data2);
			break;
		default:
			/* these are not indexed variables. */
			break;
		}
	}
}

void set_var_indexed_field_callback(sct_var_type_t type,
				    int var_index,
				    callback_indexed_field_func func,
				    sct_data_type_t dtype,
				    void *data)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch (type) {
		case VT_PARAM:
			set_param_field_callback(var_index, func, dtype, data);
			break;
		default:
			/* these are not indexed variables. */
			break;
		}
	}
}

/* evaluate the value of the given parameter, assuming it was previously */
/* set with set_var/set_var_index.                                       */
sct_filter_node_t* get_var(sct_var_type_t type)
{
	sct_filter_node_t* val = NULL;
	callback_func func; 
	struct var_struct* this; 

	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch(type) {
		case VT_PARAM:
			/* VT_PARAM is an indexed var. */
			break;
		default:
			this = &vars[type]; 

			val = this->value;

			if (!val && this->cb_type == CB_TYPE_VAR &&
			    this->f.var != NULL) {

				func = this->f.var; 
				val = func(type, this->data);

				this->value = val;
				if (this->persistent == 0) {
					this->cb_type = CB_TYPE_NONE;
					this->f.var = NULL;
				}
			}
			break;
		}
	}

	return val;
}

sct_filter_node_t* get_var_indexed(sct_var_type_t type, int var_index)
{
	sct_filter_node_t* val = NULL;

	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch(type) {
		case VT_PARAM:
			val = get_param(var_index);
			break;
		default:
			/* these are not indexed vars */
			break;
		}
	}

	return val;
}

sct_filter_node_t* get_var_indexed_field(sct_var_type_t type,
					 int var_index,
					 int field_num,
					 int struct_type)
{
	sct_filter_node_t* val = NULL;

	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch(type) {
		case VT_PARAM:
			val = get_param_field(var_index, field_num, struct_type);
			break;
		default:
			/* these are not indexed vars */
			break;
		}
	}

	return val;
}

/* set a variable's attribute. gets reset when calling 'rest_all_vars()'. */
void set_var_attr(sct_var_type_t type, sct_var_attr_name_t name, int val)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized &&
	    name >= 0 && name < VAN_LAST_TYPE) {
		switch (type) {
		case VT_PARAM:
			/* this is an indexed variable. */
			break;
		default:
			vars[type].attrs[name].is_set = 1;
			vars[type].attrs[name].val = val;
			break;
		}
	}
}

void set_var_indexed_attr(sct_var_type_t type, int var_index,
			  sct_var_attr_name_t name, int val)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized &&
	    name >= 0 && name <= VAN_LAST_TYPE) {
		switch (type) {
		case VT_PARAM:
			set_param_attr(var_index, name, val);
			break;
		default:
			/* these are not indexed variables. */
			break;
		}
	}
}


/* get a variable's attribute. returns '1' if its defined, '0' if not. */
/* value itself stores in '*p_val'.                                    */
int get_var_attr(sct_var_type_t type, sct_var_attr_name_t name, int* p_val)
{
	int is_set = 0;

	if (type > 0 && type < VT_LAST_TYPE && var_initialized &&
	    name >= 0 && name < VAN_LAST_TYPE && p_val) {
		switch(type) {
		case VT_PARAM:
			/* VT_PARAM is an indexed var. */
			break;
		default:
			if (vars[type].attrs[name].is_set) {
				*p_val = vars[type].attrs[name].val;
				is_set = 1;
			}
			break;
		}
	}

	return is_set;
}

int get_var_indexed_attr(sct_var_type_t type, int var_index,
		     sct_var_attr_name_t name, int* p_val)
{
	int is_set = 0;

	if (type > 0 && type < VT_LAST_TYPE && var_initialized &&
	    name >= 0 && name < VAN_LAST_TYPE) {
		switch(type) {
		case VT_PARAM:
			is_set = get_param_attr(var_index, name, p_val);
			break;
		default:
			/* these are not indexed vars */
			break;
		}
	}

	return is_set;
}
