/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: filter_vector.c,v 1.12 2003/01/31 14:08:46 mulix Exp $
 */
/*
 * Rule engine - implementation of a vector of filter nodes.
 */

#include "rule_engine_private.h"
#include "encode_decode_public.h"
#include "mem_manager_public.h"

#include "archdep.h"
/*
 * Private functions.
 */

/* free the 'vec' part of a filter node. don't free the node itself. */
void free_vec_of_filter_node(struct tracker_filter_node* f)
{
	assert(f);
	/* nothing to do, since a var node contains only basic data types. */
	if (f->u.vec.nodes) {
		int i;
		for (i=0; i<f->u.vec.len; i++) {
			if (f->u.vec.nodes[i])
				sct_free_filter_node(f->u.vec.nodes[i]);
		}
		mem_free(f->u.vec.nodes);
		f->u.vec.len = 0;
		f->u.vec.nodes = NULL;
	}
}

/* print a vector node. */
void sct_print_filter_node_vec(struct tracker_filter_node* f, unsigned int indent)
{
	int i;

	assert(f);

	print_indent(indent);
	PPRINT("vector[%d]:\n", f->u.vec.len);
	print_indent(indent);

	if (f->u.vec.len && !f->u.vec.nodes)
		PPRINT("EEK - len is %d and no nodes!\n", 
		       f->u.vec.len); 

	for (i=0; i<f->u.vec.len; i++) {
		PPRINT("f%d:\n", i);
		if (f->u.vec.nodes[i]) {
			do_print_filter_node(f->u.vec.nodes[i], indent+INDENT_LEVEL);
		}
		else {
			print_indent(indent);
			PPRINT("Empty\n");
		}
		print_indent(indent);
	}
}

/*
 * Implementation of public filter building functions.
 */

/* create a vector filter node, with the given size. */
/* returns NULL if there's not enough memory.        */
sct_filter_node_t* sct_make_filter_vector(int len)
{
	struct tracker_filter_node* node; 

	assert(len > 0); 

	if (len <= 0)
		return NULL; 

	if (!(node = mem_alloc(sizeof(*node))))
		return NULL; 

	node->type = NT_VECTOR;
	node->data_type = DT_MULTI;
	node->u.vec.len = len;
	node->u.vec.nodes = mem_alloc(sizeof(node->u.vec.nodes[0]) * len);
	if (!node->u.vec.nodes) {
		mem_free(node);
		node = NULL;
	} else {
		int i;
		for (i = 0; i < len; i++)
			node->u.vec.nodes[i] = NULL;
	}

	return node;
}

/* inserts the given filter node into the given vector node. */
/* returns '1' on success, '0' if 'vec' is not a vector,     */
/* '-1' if index is out of range.                            */
int sct_insert_filter_node(sct_filter_node_t* vec, int index, sct_filter_node_t* f)
{
	int rc = 1;  /* assume success. */

	assert(vec);
    
	if (vec->type != NT_VECTOR) 
		return 0; 

	if (index < 0 || index >= vec->u.vec.len)
		return -1; 

	if (vec->u.vec.nodes[index]) {
		/* free former value in the vector's cell, if there was any. */
		sct_free_filter_node(vec->u.vec.nodes[index]);
	}
	vec->u.vec.nodes[index] = f;

	return rc;
}

/* return the filter node in the given index of the given vector.    */
/* returns NULL if index is out of range, or vec is not a DT_VECTOR. */
sct_filter_node_t* sct_get_filter_node_at(sct_filter_node_t* vec, int index)
{
	sct_filter_node_t* f = NULL;    /* assume failure. */

	assert(vec);

	if (vec->type == NT_VECTOR && index >= 0 && index < vec->u.vec.len) {
		f = vec->u.vec.nodes[index];
	}

	return f;
}

/* return the length of the given vector node. returns '-1'  */
/* if 'vec' is not a vector.				     */
int sct_get_filter_vector_len(sct_filter_node_t* vec)
{
	assert(vec);

	if (vec->type == NT_VECTOR)
		return vec->u.vec.len;
	else
		return -1;
}

/* serialize a 'vector' filter node. */
int sct_serialize_filter_node_vector(sct_filter_node_t* f,
				     struct sct_membuf* m)
{
	int vec_len = f->u.vec.len; 
	sct_filter_node_t* node;
	int rc = -EINVAL;  /* assume failure */
	int sum = 0; 
	int i; 

	/* not sure at all this is correct - can we legally be asked to */ 
	/* encode an "empty" vector node, with len -1? */ 

	if (vec_len > -1) {
		if ((rc = encode_char(m, NT_VECTOR)) < 0)
			return rc; 

		sum += rc; 

		if ((rc = encode_int(m, vec_len)) < 0)
			return rc; 

		sum += rc; 
		
		for (i=0; rc > 0 && i < vec_len; i++) {
				node = f->u.vec.nodes[i]; 
				if ((rc = do_serialize_filter_node(node, m)) < 0)
					return rc; 

				sum += rc; 
		}
	}

	return sum; 
}

/* de-serialize a 'vector' filter node. */
int sct_deserialize_filter_node_vector(sct_filter_node_t** pf,
				       struct sct_membuf* m)
{
	int vec_len;
	sct_filter_node_t* vec = NULL;
	sct_filter_node_t* node;
	int rc = decode_int(m, &vec_len);
	int i; 

	if (rc)
		goto done; 

	if (!(vec = sct_make_filter_vector(vec_len))) {
		rc = -ENOMEM;  
		goto done; 
	}

	for (i = 0; i < vec_len; i++) {
		if ((rc = do_deserialize_filter_node(&node, m)) < 0)
			break; 

		/* FIXME: what should we do if insertion failed? */ 
		sct_insert_filter_node(vec, i, node);
	}

	if (rc < 0)
		sct_free_filter_node(vec);
	else 
		*pf = vec;

 done:
	return rc;
}
