/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU Lesser General Public License                       */

/*
 * $Id: logging.c,v 1.15 2003/01/29 15:47:30 mulix Exp $
 */

#include "rule_engine_public.h"
#include "rule_engine_private.h"
#include "membuf.h"
#include "filter_vars_public.h"
#include "archdep.h"
#include "loguser.h" 

/* maximum size of a log record. */
#define MAX_LOG_RECORD	1024

/* types of logging macros, used when logging syscall invocations. */
typedef enum {
	LM_PID = 0,
	LM_UID = 0,
	LM_EUID = 0
} log_macro_t;

/* maximum number of supported logger callbacks. */
#define	MAX_LOGGER_CALLBACKS	5

/* logger callbacks array. */
struct logger_callback logger_callbacks[MAX_LOGGER_CALLBACKS];
static int num_logger_callbacks = 0;

sct_loguser_t* loguser_head;

/* find the registered logger calback function, that handles  */
/* logging a given type of spec-char. returns the callback ID */
/* if one was registered, '-1' otherwise.                     */
static int get_logger_callback_index(char spec_char)
{
	int i;

	for (i=0; i < num_logger_callbacks; i++) {
		if (logger_callbacks[i].spec_char == spec_char)
			break;
	}

	return (i < num_logger_callbacks ? i : -1);
}

/* register the given logger calback function, that handles */
/* logging a given type of spec-char                        */
void set_logger_callback(char spec_char, logger_callback_t cb, void* cb_data)
{
	int i;

	for (i=0; i < num_logger_callbacks; i++) {
		if (logger_callbacks[i].spec_char == spec_char)
			break;
	}

	if (i == num_logger_callbacks) {
		if (i >= MAX_LOGGER_CALLBACKS) 
			return; 

		num_logger_callbacks++;
	}
	logger_callbacks[i].spec_char = spec_char;
	logger_callbacks[i].cb = cb;
	logger_callbacks[i].cb_data = cb_data;
}

/* stringify the given string value, into the given buffer.           */
/* return 0 on success, -ERRNO otherwise                              */
static int tracker_stringify_string(const char* string, struct sct_membuf* m, 
				    unsigned int max_rec_len, 
				    int add_quotes)
{
	int rc; 
	char quote = '"'; 
	unsigned int reserved = strlen("...\""); 

	/* we want to output '"string"' */
	if (add_quotes) {
		if ((rc = sct_membuf_add_chunk(m, &quote, sizeof(quote))) < 0)
			return rc; 
	}
	
	if (string) {
		size_t len = strlen(string); 
                /* we want to output the minimal ammount of bytes from these values: */ 
		/* the length of the string, */ 
		/* the ammount available for write (minux '..."'), */ 
		/* and the max_rec_len (minus '..."') (if it's not 0) */ 
		unsigned int printlen = sct_min(len, sct_membuf_free_for_write(m) - reserved); 

		if (max_rec_len > reserved)
			printlen = sct_min(printlen, max_rec_len - reserved); 

		/* do we have space for the entire string? */ 
		if (printlen == len) {
			if ((rc = sct_membuf_add_chunk(m, string, strlen(string))) < 0)
				return rc; 
		} else if (len) { /* out len bytes and then ... */ 
			if ((rc = sct_membuf_add_chunk(m, string, printlen)) < 0)
				return rc; 
			if ((rc = sct_membuf_add_chunk(m, "...", strlen("..."))) < 0)
				return rc; 
		}
	} else { /* the string was null */
		if ((rc = sct_membuf_add_chunk(m, "(nil)", strlen("(nil"))) < 0)
			return rc; 
	}
	/* emit the trailing '"'. */
	if (add_quotes) {
		if ((rc = sct_membuf_add_chunk(m, &quote, 1)) < 0)
			return rc; 
	}

	return 0;
}

/* can't use an inline function, since we need the type */ 
#define stringify_val(val, fmt, m) do {                   \
        int rc;                                           \
	char sbuf[32] = {0};                              \
                                                          \
	sprintf(sbuf, fmt, val);                          \
	rc = sct_membuf_add_chunk(m, sbuf, strlen(sbuf)); \
	                                                  \
	return rc;                                        \
} while (0)

/* stringify the given unsigned numeric value, into the given buffer, */
static int tracker_stringify_unum(unsigned long num, struct sct_membuf* m)
{
	stringify_val(num, "%lu", m); 
}

/* stringify the given pointer value, into the given buffer,   */
static int tracker_stringify_pointer(void* p, struct sct_membuf* m)
{
	stringify_val(p, "%p", m); 
}

/* stringify the given unsigned numeric value, into the given buffer, */
static int tracker_stringify_num(long num, struct sct_membuf* m)
{
	stringify_val(num, "%ld", m); 
}

/* stringify the given unsigned numeric value, into the given buffer, */
static int tracker_stringify_num64(long long num, struct sct_membuf* m)
{
	stringify_val(num, "%lld", m); 
}

/* stringify the given variable into the given buffer.                */
static int tracker_stringify_var_node(struct sct_membuf* m, 
				      unsigned int max_rec_len, 
				      sct_filter_node_t *var_val)
{
	int rc = 0;
	char* string = NULL;

	if (!var_val)
		return tracker_stringify_string("<NULL>", m, 0, 0); 
	if (var_val->type != NT_CONST) {
		PPRINT("eek, internal incosistency, %s got a filter node of type %d\n", 
		       __FUNCTION__, var_val->type); 
		return 0; 
	}
		
	switch (var_val->data_type) {
	case DT_USHORT:
		rc = tracker_stringify_unum(CONST_USHORT(var_val), m);
		break;
	case DT_SHORT:
		rc = tracker_stringify_num(CONST_SHORT(var_val), m);
		break;
	case DT_UINT:
		rc = tracker_stringify_unum(CONST_UINT(var_val), m);
		break;
	case DT_INT:
		rc = tracker_stringify_num(CONST_INT(var_val), m);
		break;
	case DT_ULONG:
		rc = tracker_stringify_unum(CONST_ULONG(var_val), m);
		break;
	case DT_LONG:
		rc = tracker_stringify_num(CONST_LONG(var_val), m);
		break;
	case DT_LONG_LONG: 
		rc = tracker_stringify_num64(CONST_LONG_LONG(var_val), m); 
		break; 
	case DT_STRING:
		string = CONST_STRING(var_val);
		rc = tracker_stringify_string(string, m, max_rec_len, 1);
		break;
	case DT_BUFFER:
		string = sct_buffer_to_string(CONST_BUFFER(var_val), 
					      BUFFER_SIZE(var_val)); 
		rc = tracker_stringify_string(string, m, max_rec_len, 1); 
		mem_free(string); 
		break; 
	case DT_PTR:
		rc = tracker_stringify_pointer(CONST_PTR(var_val), m);
		break;
	default:
		break;
	}
    
	return rc; 
}

static int tracker_stringify_var(struct sct_membuf* m, unsigned int max_rec_len, 
				 sct_var_type_t var)
{
	return tracker_stringify_var_node(m, max_rec_len, get_var(var));
}

static int tracker_stringify_var_indexed(struct sct_membuf* m, 
					 unsigned int max_rec_len, 
					 sct_var_type_t var, 
					 int var_index)
{
	sct_filter_node_t* t = get_var_indexed(var, var_index); 
	return tracker_stringify_var_node(m, max_rec_len, t);
}

static int do_logger_callback(int logger_cb_id, 
			      char this_char, va_list ap, 
			      struct sct_membuf* m, 
			      unsigned int max_rec_len)
{
	void* logee; 
	void* extra_info;
	int rc; 

	struct logger_callback* lb; 

	logee = va_arg(ap, void*);
	extra_info  = va_arg(ap, void*);

	lb = &logger_callbacks[logger_cb_id]; 

	rc = (lb->cb)(this_char, m, max_rec_len, 
		      logee, extra_info, lb->cb_data);

	return rc; 
}

/* stringify the parameters of the syscall. */
static int tracker_stringify_params(struct sct_membuf* m, unsigned int max_rec_len, 
				    const char* param_spec, va_list ap)
{
	char c;
	unsigned char uc;
	long l;
	unsigned long ul;
	long long ll; 
	int rc = 0;
	void* p;

	int next_is_unsigned = 0;
	char* string = NULL; 
	char this_char = 0; 
	char next_char = 0; 
	int is_string; /* strings require a bit of special handling */ 
	int skip_params = 0; /* are we in 'parameters skipping' mode? */
	int logger_cb_id = -1;
	
	if (!param_spec)
		return -EINVAL; 

	while (*param_spec) {
		this_char = *param_spec++;

		/* handle logger callbacks first */ 
		if ((logger_cb_id = get_logger_callback_index(this_char)) > -1) {
			rc = do_logger_callback(logger_cb_id, this_char, 
						ap, m, max_rec_len); 
			goto tracker_stringify_params_next_char; 
		}
		
		is_string = (this_char == 's');
		
		switch (this_char){
		case ' ': break; /* skip spaces */
		case 'u': next_is_unsigned = 1; break;
		case 'c':
			if (next_is_unsigned) {
				uc = (unsigned char)va_arg(ap, unsigned int);
				if (!skip_params) 
					sct_membuf_add_chunk(m, &uc, 1); 
				
				next_is_unsigned = 0;
			} else {
				c = (char)va_arg(ap, int);
				if (!skip_params)
					sct_membuf_add_chunk(m, &c, 1); 
			}
			break;
		case 'h':
			if (next_is_unsigned) {
				ul = (unsigned short)va_arg(ap, unsigned int);
				if (!skip_params)
					rc = tracker_stringify_unum(ul, m); 
				
				next_is_unsigned = 0;
			} else {
				l = (short)va_arg(ap, int);
				if (!skip_params) 
					rc = tracker_stringify_num(l, m);
			}
			break;
		case 'i':
			if (next_is_unsigned) {
				ul = va_arg(ap, unsigned int);
				if (!skip_params) 
					rc = tracker_stringify_unum(ul, m);
				
				next_is_unsigned = 0;
			} else {
				l = va_arg(ap, int);
				if (!skip_params) 
					rc = tracker_stringify_num(l, m);
			}
			break;
		case 'l':
			if (next_is_unsigned) {
				ul = va_arg(ap, unsigned long);
				if (!skip_params) 
					rc = tracker_stringify_unum(ul, m);
				
				next_is_unsigned = 0;
			} else {
				l = va_arg(ap, long);
				if (!skip_params) 
					rc = tracker_stringify_num(l, m);
				
			}
			break;
		case 'L': 
			ll = va_arg(ap, long long); 
			if (!skip_params)
				rc = tracker_stringify_num(ll, m); 
			break; 
		case 'p':
			p = va_arg(ap, void*);
			if (!skip_params) 
				rc = tracker_stringify_pointer(p, m);
			
			break;
		case 'P': /* indexed var */
			l = va_arg(ap, int);
			if (!skip_params) 
				rc = tracker_stringify_var_indexed(m, max_rec_len, 
								   VT_PARAM, l);
			
			break;
			/* TODO: get rid of either 'B', seems identical to 'P'? */ 
		case 'B': /* indexed var */ 
			l = va_arg(ap, int);
			if (!skip_params)
				rc = tracker_stringify_var_indexed(m, max_rec_len, 
								   VT_PARAM, l);
			break; 
		case 'T': /* type-name */
			string = va_arg(ap, char*);
			if (!skip_params)
				rc = tracker_stringify_string(string, m, max_rec_len, 0);
			break;
		case 's': /* string */
			string = va_arg(ap, char*);
			if (!skip_params) 
				rc = tracker_stringify_string(string, m, max_rec_len, 1); 
			break;
		case '{':
			rc = sct_membuf_add_chunk(m, "{", 1); 
			break;
		case '}':
			rc = sct_membuf_add_chunk(m, "}", 1); 
			if (skip_params)
				skip_params = 0;
			break;
		case 'v':
			if (!skip_params) 
				rc = sct_membuf_add_chunk(m, "void", strlen("void")); 
			break;
		default:
			break;
		}

	tracker_stringify_params_next_char:
		/* if we just printed a value, print a comma after it */
		next_char = *param_spec;
		if (logger_cb_id > -1 && next_char && next_char != '}') {
			if (!skip_params) 
				sct_membuf_add_chunk(m, ", ", strlen(", ")); 
			
		} else if ((this_char != ' ') && (this_char != 'u') && 
			   (this_char != 'S') && (this_char != 'T') && 
			   (this_char != '{') && (next_char) && next_char != '}') {  
			if (!skip_params) 
				sct_membuf_add_chunk(m, ", ", strlen(", ")); 
		}
	}
	if (rc < 0)
		return rc; 

	return 0; 
}


/* stringify the parameters of the syscall. */
int sct_tracker_stringify_params(struct sct_membuf* m, 
				 unsigned int max_rec_len, 
				 const char* param_spec,
				 ...)
{
	int rc;
	va_list ap;
	va_start (ap, param_spec);

	rc = tracker_stringify_params(m, max_rec_len, 
				      param_spec, ap);
	va_end(ap);

	return rc;
}

/* perform macro translation into the given buffer with the given size. */
/* returns 0 on success or -ERRNO                   .                   */
static int tracker_translate_macro(sct_loguser_t* logger, 
				   int rule_id, 
				   int syscall_id, 
				   const char* syscall_name, 
				   const char* macro, 
				   struct sct_membuf* m, 
				   const char* param_spec, 
				   va_list ap)
{
	unsigned int len = (logger ? sct_loguser_get_max_record_len(logger) : 0); 
			    
	if (!strcmp(macro, "ruleid"))
		return tracker_stringify_num(rule_id, m);

	if (!strcmp(macro, "sid"))
		return tracker_stringify_num(syscall_id, m);

	if (!strcmp(macro, "sname")) 
		return sct_membuf_add_chunk(m, syscall_name, strlen(syscall_name)); 
		
	if (!strcmp(macro, "params")) 
		return tracker_stringify_params(m, len, param_spec, ap); 
		
	if (!strcmp(macro, "pid"))
		return tracker_stringify_var(m, len, VT_PROC_PID);

	if (!strcmp(macro, "uid"))
		return tracker_stringify_var(m, len, VT_PROC_UID);

	if (!strcmp(macro, "euid")) 
		return tracker_stringify_var(m, len, VT_PROC_EUID);

	if (!strcmp(macro, "suid"))
		return tracker_stringify_var(m, len, VT_PROC_SUID);

	if (!strcmp(macro, "gid"))
		return tracker_stringify_var(m, len, VT_PROC_GID);

	if (!strcmp(macro, "egid"))
		return tracker_stringify_var(m, len, VT_PROC_EGID);
	
	if (!strcmp(macro, "sgid"))
		return tracker_stringify_var(m, len, VT_PROC_SGID);

	if (!strcmp(macro, "comm"))
		return tracker_stringify_var(m, len, VT_PROC_COMM);

	if (!strcmp(macro, "ppid"))
		return tracker_stringify_var(m, len, VT_PROC_PPID);

	if (!strcmp(macro, "pcomm"))
		return tracker_stringify_var(m, len, VT_PROC_PCOMM);

	if (!strcmp(macro, "cwd"))
		return tracker_stringify_var(m, len, VT_PROC_CWD);

	if (!strcmp(macro, "retval"))
		return tracker_stringify_var(m, len, VT_RETVAL);
	
	PPRINT(KERN_NOTICE "%s: unknown macro '%s'\n", __FUNCTION__, macro); 
	return 0; 
}

/* log the syscall that matched one of the rules in 'rules' */
void tracker_do_logging(sct_loguser_t* logger, 
			int rule_id, 
			int syscall_id, 
			const char* syscall_name, 
			const char* log_format, 
			const char* param_spec, 
			va_list ap)
{
	char macro[MAX_LOG_RECORD+1];
	const char* pf = log_format;
	char const* ps; 
	int mlen; 
	struct sct_membuf* mbuf; 
	int using_printk = 0; 

	if (!logger) {
		if (!(mbuf = sct_membuf_create(MAX_LOG_RECORD + 1, NULL, NULL, 
					       SCT_MEMBUF_EXACT_MATCH))) {
			PPRINT(KERN_WARNING "failed to allocate membuf, quitting\n"); 
			return; 
		}
		using_printk = 1; 
	} else
		mbuf = sct_loguser_get_membuf(logger); 
	
	while (*pf) {
		if (*pf != '%') {
			sct_membuf_add_chunk(mbuf, pf, 1); 
			pf++; 
			continue; 
		}
		pf++; 
		if (!*pf)
			break; 

		if (*pf == '%') {
			/* translate '%%' to '%'. */
			sct_membuf_add_chunk(mbuf, "%", 1); 
			pf++; 
			continue; 
		} else {
			/* translate macro name into value. */
			ps = pf; 
			/* parse out the macro name - from the current char
			   until the next non alphabetic character */
			while (isalpha(*ps++))
				;
			ps--;
			mlen = sct_min(ps - pf, sizeof(macro) - 2);
			strncpy(macro, pf, mlen);
			macro[mlen] = '\0';	
			tracker_translate_macro(logger, rule_id, syscall_id,
						syscall_name, macro, mbuf,
						param_spec, ap);
			pf = ps;
			continue;
		}
	}

	if (using_printk) {
		PPRINT(KERN_NOTICE "%s", sct_membuf_get_data(mbuf));
		sct_membuf_destroy(mbuf); 
	}
}

sct_loguser_t* sct_get_loguser_head(void)
{
	return loguser_head; 
}

int sct_register_loguser_head(sct_loguser_t* loguser)
{
	loguser_head = loguser;
	return 0; 
}

int sct_update_loguser_head(sct_loguser_t* loguser)
{
	if (loguser_head)
		loguser_head = loguser;
	/* if not registered (loguser_head is NULL) - do nothing */
	return 0;
}

void sct_deregister_loguser_head(void)
{
	loguser_head = NULL;
}

