/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Itai Segall, spapuk@t2.technion.ac.il    */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: loguser.c,v 1.7 2003/01/29 15:47:32 mulix Exp $
 */

#include "archdep.h"
#include "loguser.h"

/* loguser represents a single node in a chained list. Every node services
   a single log device file reader, holding its ring membuf, its log format, 
   and supplying locking capabilities */

struct sct_loguser {
	struct sct_membuf* ring;
	char* before_log_format;
	char* after_log_format;
	unsigned int max_record_len; 
	sct_sem_t sem;
	struct sct_loguser* next;
};

/* create a new loguser. */
sct_loguser_t* sct_loguser_create(struct sct_membuf* rng, char* before_format,
				  char* after_format)
{
	sct_loguser_t* lb = mem_alloc(sizeof(*lb));
	if(!lb)
		return NULL;

	memset(lb, 0, sizeof(*lb)); 

	/* the loguser doesn't own the log_format, nor the ring */
	lb->ring = rng; 
	lb->before_log_format = before_format;
	lb->after_log_format = after_format;
	lb->max_record_len = 0; 

	lb->next = NULL;

	sct_sema_init(&lb->sem, 1); /* initialize as a mutex */

	return lb;
}

void sct_loguser_destroy(sct_loguser_t* buf)
{
	if (!buf)
		return;

	buf->ring = NULL;
	buf->before_log_format = NULL;
	buf->after_log_format = NULL;

	mem_free (buf);
}

struct sct_membuf* sct_loguser_get_membuf(sct_loguser_t* loguser)
{
	if(!loguser)
		return NULL;
	return loguser->ring;
}

const char* sct_loguser_get_log_format(sct_loguser_t* loguser, 
				       sct_rule_when_t when)
{
	if (!loguser)
		return NULL;
	return (when==RW_BEFORE) ? loguser->before_log_format : 
		loguser->after_log_format;
}

char* sct_loguser_change_log_format(sct_loguser_t* loguser, 
				    sct_rule_when_t when, char* format)
{
	char *old_format;

	if (loguser == NULL)
		return NULL;

	if (when == RW_BEFORE) {
		old_format = loguser->before_log_format;
		loguser->before_log_format = format;
	} else {
		old_format = loguser->after_log_format;
		loguser->after_log_format = format;
	}

	return old_format;
}

sct_loguser_t* sct_loguser_get_next(sct_loguser_t* loguser)
{
	if (loguser == NULL)
		return NULL;

	return loguser->next;
}


unsigned int sct_loguser_set_max_record_len(sct_loguser_t* loguser, 
					    unsigned int val)
{
	unsigned int tmp; 

	if (!loguser)
		return 0; 

	tmp = loguser->max_record_len; 
	loguser->max_record_len = val; 
	return 0; 
}

unsigned int sct_loguser_get_max_record_len(sct_loguser_t* loguser)
{
	if (!loguser)
		return 0; 

	return loguser->max_record_len; 
}

void sct_loguser_set_next(sct_loguser_t* cur, sct_loguser_t* nxt)
{
	if (cur== NULL)
		return;

	cur->next = nxt;
}

int lock_loguser_interruptible(sct_loguser_t* loguser)
{
	return sct_lock_sem_interruptible(&loguser->sem); 
}

void lock_loguser(sct_loguser_t* loguser)
{
	sct_lock_sem(&loguser->sem);
}

void unlock_loguser(sct_loguser_t* loguser)
{
	sct_unlock_sem(&loguser->sem);
}
