/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU Lesser General Public License                       */

/* 
 * $Id: membuf.c,v 1.14 2003/01/29 15:47:33 mulix Exp $
 */

#include "archdep.h"
#include "membuf.h"

struct sct_membuf {
	unsigned char* data; 
	int owns_data;   /* 1 if we malloc'd the data, 0 if it was externally supplied */ 

	unsigned int rh; /* read head */ 
	unsigned int wh; /* write head */ 
	
	size_t size;     /* total data size */ 
	size_t wfree;    /* free for writing space */ 
	size_t ravailable;    /* available for reading */ 

	membuf_copy_fn tofn;    /* to membuf copy fn */ 
	membuf_copy_fn fromfn;  /* from membuf copy fn */ 
	
	enum sct_membuf_add_policy policy; 
}; 

static inline int copy_to(struct sct_membuf* m, unsigned char* whead, 
			  const void* chunk, size_t size)
{
	if (m->tofn) {
		if (m->tofn(whead, chunk, size))
			/* if an error occurs, return it */ 
			return -EFAULT; 
	} else {
		memcpy(whead, chunk, size); 
	}	
	return 0; 
}

static inline int copy_from(struct sct_membuf* m, void* chunk, 
			    const unsigned char* rhead, size_t size)
{
	if (m->fromfn) {
		if (m->fromfn(chunk, rhead, size))
			/* if an error occures, return it */ 
			return -EFAULT; 
	} else {
		memcpy(chunk, rhead, size); 
	}
	return 0; 
}

struct sct_membuf* sct_membuf_create(size_t size, membuf_copy_fn tofn, 
				membuf_copy_fn fromfn, 
				enum sct_membuf_add_policy policy)
{
	struct sct_membuf* m = mem_alloc(sizeof(struct sct_membuf)); 
	if (!m)
		return NULL; 

	m->data = vmem_alloc(size); 
	if (!m->data) {
		mem_free(m); 
		return NULL; 
	}

	memset(m->data, 0, size); 
	
	m->owns_data = 1; 

	m->rh = m->wh = 0; 

	m->wfree = m->size = size; 
	m->ravailable = 0; 

	m->tofn = tofn; 
	m->fromfn = fromfn; 

	m->policy = policy; 

	return m; 
}

struct sct_membuf* sct_membuf_create_with_data(unsigned char* data, size_t size, 
					  membuf_copy_fn tofn, membuf_copy_fn fromfn)
{
	struct sct_membuf* m = mem_alloc(sizeof(struct sct_membuf)); 
	if (!m)
		return NULL; 

	m->owns_data = 0; 
	m->data = data; 
	m->ravailable = m->size = size;
	m->wh = 0; /* wraparound */ 
	m->rh = 0; 
	m->wfree = 0; 

	m->tofn = tofn; 
	m->fromfn = fromfn; 

	m->policy = -1; 

	return m; 
}

void sct_membuf_destroy(struct sct_membuf* m)
{
	assert(m); 
	
	if (!m) 
		return; 

	if (m->data) {
		/* only free it if we own it */ 
		if (m->owns_data)
			vmem_free(m->data);
		/* but set it to NULL anyway */ 
		m->data = NULL; 
	}
	/* set the fields to 0 to catch usage after free() asap */ 
      	m->rh = m->wh = m->wfree = m->ravailable =  m->size = 0;
	mem_free(m); 
}

static ssize_t common_add(struct sct_membuf* m, const void* chunk, size_t sz)
{
	ssize_t ret = sz; 
	while (sz) {
		ssize_t rc; 
		int ammount = m->size - m->wh; 
		if (ammount > sz)
			ammount = sz; 
		
		if ((rc = copy_to(m, m->data + m->wh, chunk, ammount)))
			return rc; 

		m->wh = (m->wh + ammount) % m->size; 
		chunk = ((const unsigned char*)chunk) + ammount; 
		sz -= ammount; 
		m->ravailable += ammount; 
		m->wfree -= ammount; 
	}
	return ret; 
}

static ssize_t add_drop_new(struct sct_membuf* m, const void* chunk, size_t chunk_size)
{

	if (chunk_size > m->wfree)
		chunk_size = m->wfree; 

	return common_add(m, chunk, chunk_size); 
}

static ssize_t add_exact_match(struct sct_membuf* m, const void* chunk, size_t chunk_size)
{
	if (chunk_size > m->wfree)
		return -ENOSPC; 

	return common_add(m, chunk, chunk_size); 
}

static ssize_t add_overwrite_old(struct sct_membuf* m, const void* chunk, size_t chunk_size)
{
	ssize_t ret; 
	int overwritten = 0; 

	if (chunk_size > m->wfree)
		overwritten = 1; 

	ret = common_add(m, chunk, chunk_size); 
	
	if (overwritten && (ret >= 0)) {
		m->rh = m->wh; 
		m->wfree = m->ravailable = m->size; 
	}

	return ret; 
}

ssize_t sct_membuf_add_chunk(struct sct_membuf* m, const void* chunk, size_t chunk_size)
{
	assert(m);
	/* allow a NULL chunk if the size is 0, */ 
	/* allow a 0 size if the chunk is NULL */ 
	if (!m || (!chunk && chunk_size) || (chunk && !chunk_size))
		return -EINVAL; 

	if (!chunk_size) 
		return 0; 

	switch(m->policy) {
	case SCT_MEMBUF_DROP_OLD:
		return add_overwrite_old(m, chunk, chunk_size); 
	case SCT_MEMBUF_DROP_NEW:
		return add_drop_new(m, chunk, chunk_size); 
	case SCT_MEMBUF_EXACT_MATCH:
		return add_exact_match(m, chunk, chunk_size); 
	default: 
		PPRINT("uh oh, invalid drop policy: %d\n", m->policy); 
		return -EINVAL; 
	}
}

ssize_t sct_membuf_get_chunk(struct sct_membuf* m, void* chunk, size_t chunk_size)
{
	size_t ammount; 
	ssize_t ret; 

	assert(m); 

	/* a valid chunk but with size 0 is allowed */ 
	/* think read() semantics - you dont know its eof until you get 0 */ 
	if (!m || (!chunk && chunk_size))
		return -EINVAL; 
	
	if (!chunk_size)
		return 0; 

	if (chunk_size > m->ravailable)
		chunk_size = m->ravailable; 

	ret = chunk_size; 

	while (chunk_size) {
		ssize_t rc; 
		ammount = (m->wh > m->rh) ? m->wh - m->rh : m->size - m->rh; 
		if (ammount > chunk_size)
			ammount = chunk_size; 

		if ((rc = copy_from(m, chunk, m->data + m->rh, ammount)))
			return rc; 

		m->rh = (m->rh + ammount) % m->size; 
		chunk = ((unsigned char*)chunk) + ammount; 
		chunk_size -= ammount; 
		m->ravailable -= ammount; 
		m->wfree = (m->wfree + ammount > m->size ? m->size: m->wfree + ammount); 
	}

	return ret; 
}

unsigned char* sct_membuf_get_data(struct sct_membuf* m)
{
	assert(m); 

	if (!m)
		return NULL; 

	/* might be NULL! */ 
	return m->data; 
}

size_t sct_membuf_total_size(struct sct_membuf* m)
{
	assert(m); 

	if (!m)
		return 0; 

	return m->size; 
}

/* available for reading */ 
size_t sct_membuf_available_for_read(struct sct_membuf* m)
{
	assert(m); 

	if (!m)
		return 0; 

	return m->ravailable; 
}

/* available for writing */ 
size_t sct_membuf_free_for_write(struct sct_membuf* m)
{
	assert(m); 

	if (!m)
		return 0; 

	return (m->policy == SCT_MEMBUF_DROP_OLD ? m->size : m->wfree); 
}

void sct_membuf_clear(struct sct_membuf* m)
{
	assert(m); 

	if (!m) 
		return; 

	m->ravailable = m->rh = m->wh = 0; 
	m->wfree = m->size; 
}

int sct_membuf_enlarge(struct sct_membuf* m, size_t new_size)
{
	/* check that new size is bigger than the current size */ 
	if (new_size < m->size)
		return -EINVAL; 

	return sct_membuf_resize(m, new_size);
}

void sct_membuf_drop_old(struct sct_membuf* m)
{
	m->policy = SCT_MEMBUF_DROP_OLD; 
}

void sct_membuf_drop_new(struct sct_membuf* m)
{
	m->policy = SCT_MEMBUF_DROP_NEW; 
}

int sct_membuf_resize(struct sct_membuf* m, size_t new_size)
{
	unsigned char* new_data;

	assert (m);

	if(!m)
		return -EINVAL;

	/* it's an error to resize a buffer where we dont own the data */ 
	if (!m->owns_data)
		return -EINVAL; 

	if (new_size == m->size)
		return 0; /* nothing to do */

	/* allocate space for new buffer */ 
	if (!(new_data = vmem_alloc(new_size)))
		return -ENOMEM; 

	if (new_size > m->size)	{
                /* when enlarging, we'll preserve old data */

		/* copy old data */ 
		memcpy(new_data, m->data, m->size); 

		/* zero out remainder */ 
		memset(new_data + m->size, 0, new_size - m->size); 
		/* these two are unchanged */ 
		/* unsigned int rh;  read head */ 
		/* unsigned int wh;  write head */ 
	
		m->wfree += (new_size - m->size);     /* free (for writing) space */ 
		m->size = new_size;     /* total data size */ 
	} else {

		/* zero out new buffer */
		memset(new_data, 0, new_size); 

		m->rh = m->wh = 0;
	
		m->wfree = new_size;     /* free (for writing) space */ 
		m->size = new_size;     /* total data size */ 
		m->ravailable = 0;
	}

	vmem_free(m->data); 

	m->data = new_data; 

	return 0; 
	
}
