/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU Lesser General Public License                       */

/* 
 * $Id: membuf.h,v 1.8 2003/01/29 15:47:34 mulix Exp $
 */

#ifndef MEMBUF_H_
#define MEMBUF_H_

#include "archdep.h"
#include "rule_engine_public.h"

/* when adding new data to the membuf, what should we do? */
/* DROP_NEW: if the membuf overflows, discard the new data */ 
/* DROP_OLD: if the membuf overflows, overwrite the old data */ 
/* EXACT_MATCH: if there isn't enough room, return -ENOSPC */ 
enum sct_membuf_add_policy 
{
	SCT_MEMBUF_DROP_NEW = 0, 
	SCT_MEMBUF_DROP_OLD, 
	SCT_MEMBUF_EXACT_MATCH
};

/* we want to allow membuf users to supply their own functions for copying data  */ 
/* from the membuf / to the membuf. this is crucial, for example, for the kernel,*/ 
/* where data needs to be copied over the user space / kernel space divide       */ 
/* the copy function should return != 0 on error or 0 if succesful               */ 
typedef unsigned long(*membuf_copy_fn)(void* to, const void* from, unsigned long sz);  

/* create a membuf of size 'size         */ 
/* returns the membuf or NULL on failure */ 
struct sct_membuf* sct_membuf_create(size_t size, membuf_copy_fn to_membuf, 
				membuf_copy_fn from_membuf, 
				enum sct_membuf_add_policy policy); 

/* create a membuf which holds the preallocated data which is of size 'size' */ 
/* returns the membuf or NULL on failure. Note that the data is not copied,  */
/* and the membuf will NOT release its memory when destructed                */ 
struct sct_membuf* sct_membuf_create_with_data(unsigned char* data, size_t size, 
					  membuf_copy_fn to_membuf, 
					  membuf_copy_fn from_membuf); 

/* destroy the given membuf */ 
void sct_membuf_destroy(struct sct_membuf* m);

/* add the chunk 'chunk' to the membuf 'm'. if m doesn't have enough free     */
/* space, if any parameter is invalid or an error occurs, -ERRNO  is returned.*/ 
/* If the addition is succesful, returns the size of the serialized data -    */ 
/* which might be 0.                                                          */ 
ssize_t sct_membuf_add_chunk(struct sct_membuf* m, const void* chunk, size_t chunk_size);

/* get a chunk from the membuf. the chunk is copied into 'chunk', which is     */
/* of size 'chunk_size'. If any parameter is invalid, -1 is returned. Returns  */
/* the ammount of bytes in 'chunk', (which can be less than chunk_size)        */
ssize_t sct_membuf_get_chunk(struct sct_membuf* m, void* chunk, size_t chunk_size);

/* get a pointer to the data the membuf holds. This returns a pointer to the */
/* entire data buffer, regardless of the location of the read/write heads!   */
/* you must NEVER use this function if the membuf is used as a ring buffer   */
/* only use it if its used as a plan "char*" buffer                          */ 
unsigned char* sct_membuf_get_data(struct sct_membuf* m);

/* get the current total size of this membuf */ 
size_t sct_membuf_total_size(struct sct_membuf* m); 

/* get the ammount of space available for adding a chunk */ 
size_t sct_membuf_free_for_write(struct sct_membuf* m);

/* get ammount of available data for reading from the membuf */ 
size_t sct_membuf_available_for_read(struct sct_membuf* m); 

/* "reset" the membuf */ 
void sct_membuf_clear(struct sct_membuf* m); 

/* enlarge the membuf */ 
int sct_membuf_enlarge(struct sct_membuf* m, size_t new_size); 

void sct_membuf_drop_old(struct sct_membuf* m); 

void sct_membuf_drop_new(struct sct_membuf* m); 

/* resize the membuf. NOTE: all data might be lost */
int sct_membuf_resize(struct sct_membuf* m, size_t new_size);

#endif /* MEMBUF_H_ */ 
