/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 guy keren, choo@actcom.co.il	            */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: rule_engine.c,v 1.65 2003/01/31 14:08:46 mulix Exp $
 */

#include "mem_manager_public.h"
#include "encode_decode_public.h"
#include "rule_engine_private.h"
#include "filter_vars_public.h"
#include "membuf.h"
#include "loguser.h"
#include "archdep.h"

/* a list of tracking rules. */
#define	MAX_NUM_TRACKER_RULES	20

struct rule_list {
	struct tracker_rule* rules[MAX_NUM_TRACKER_RULES];
	int num_rules;
};

/*
 * Private functions.
 */

/* find the location of a rule with the given ID in the rules list. */
/* returns the rule's location, or '-1' if none is found.           */
static int find_tracking_rule(sct_rule_list_t* rules, int rule_id)
{
	int i;

	if (!rules)
		return -1;

	for (i = 0; i < rules->num_rules; i++) {
		if (rules->rules[i] && rules->rules[i]->rule_id == rule_id)
			break;
	}

	return (i < rules->num_rules ? i : -1);
}

/* add the given rule to the rules list.                                     */
/* returns the rule's location (not ID!) on success, -1 if an error occured. */
static int do_add_tracking_rule(sct_rule_list_t* rules,
			        int rule_id,
			        sct_filter_node_t* filter,
			        sct_action_t* action)
{
	int i;
	int j;
	struct tracker_rule* rule;

	/* make sure we have free space. */
	if (!rules || rules->num_rules >= MAX_NUM_TRACKER_RULES) {
		return -1;
	}

	/* allocate a structure for the new rule. */
	rule = mem_alloc(sizeof(struct tracker_rule));
	if (!rule) {
		return -1;
	}

	rule->rule_id = rule_id;
	rule->filter = filter;
	rule->action = action; 

	/* OPTIMIZATION: use something a little saner than this linear search */ 
	/* find the right location for the new rule. */
	for (i = 0; i < rules->num_rules; i++) {
		if (rules->rules[i] && rules->rules[i]->rule_id > rule_id)
			break;
	}
	
	/* make a free space for the new rule. */
	for (j = rules->num_rules; j > i; j--) {
		rules->rules[j] = rules->rules[j-1];
	}
	
	/* insert the new rule. */
	rules->rules[i] = rule;
	rules->num_rules++;
	
	return (i);
}

/* serialize a 'sct_action_t' action node. */
static int do_serialize_action_node(sct_action_t* a, struct sct_membuf* m)
{
	int rc = 0;
	int sum = 0; 
	char sct_action_type; 
	int i; 

	if (!a) {
		rc = encode_char(m, AT_NIL);
		return rc; 
	}

	sct_action_type = a->type;
	if ((rc = encode_char(m, sct_action_type)) < 0)
		return rc; 

	sum += rc; 
	switch (sct_action_type) {
	case AT_LOG:
		if ((rc = encode_int(m, a->u.log.priority)) < 0)
			return rc; 

		sum += rc; 

		if ((rc = encode_string(m, a->u.log.format)) < 0)
			return rc; 

		sum += rc; 
		break;
	case AT_RULE_ENABLE:
		if ((rc = encode_int(m, a->u.rule_id)) < 0)
			return rc; 
		
		sum += rc; 
		break;
	case AT_RULE_DISABLE:
		if ((rc = encode_int(m, a->u.rule_id)) < 0)
			return rc; 

		sum += rc; 
		break;
	case AT_PROCESS_SUSPEND: /* fallthrough */ 
	case AT_PROCESS_KILL:
		if ((rc = encode_int(m, a->u.pid)) < 0)
			return rc; 

		sum += rc; 
		break;
	case AT_SYSCALL_FAIL:
		if ((rc = encode_int(m, a->u.error_code)) < 0)
			return rc; 

		sum += rc; 
		break;
	default:
		rc = -EINVAL;
		break;
	}

	if (rc < 0)
		return rc; 

	/* encode attr set commands. */
	if ((rc = encode_int(m, a->attr_set_cmds_len)) < 0)
		return rc; 

	sum += rc; 
	for (i = 0; rc > 0 && i < a->attr_set_cmds_len; i++) {
		struct tracker_action_attr_set_data* atsd =
			&(a->attr_set_cmds[i]);
		if ((rc = encode_int(m, atsd->var_type)) < 0)
			return rc; 

		sum += rc; 
		if ((rc = encode_int(m, atsd->var_index)) < 0)
			return rc; 

		sum += rc; 
		if ((rc = encode_int(m, atsd->name)) < 0)
			return rc; 

		sum += rc; 
		if ((rc = encode_int(m, atsd->val)) < 0)
			return rc; 

		sum += rc; 
	}
	
	return sum; 
}

/* de-serialize an 'sct_action_t' action node. */
static int do_deserialize_action_node(sct_action_t** pa, struct sct_membuf* m)
{
	char sct_action_type;
	int rc; 

	if ((rc = decode_char(m, &sct_action_type)) < 0)
		return rc; 

	switch (sct_action_type) {
	case AT_LOG:
	{
		int priority;
		char* format;
		if ((rc = decode_int(m, &priority)) < 0)
			return rc; 


		if ((rc = decode_string(m, &format)) < 0)
			return rc; 

		*pa = sct_make_action_log(priority, format);

		if (format)
			mem_free(format);
			
		if (!*pa) {
			PPRINT(" do_deserialize_action_node: failed sct_make_action_log\n");
			rc = -ENOMEM; 
		}
		break;
	}
	case AT_RULE_ENABLE:
	{
		int rule_id;
		if ((rc = decode_int(m, &rule_id)) < 0)
			return rc; 

		if (!(*pa = sct_make_action_enable_rule(rule_id)))
			return -ENOMEM; 

		break;
	}
	case AT_RULE_DISABLE:
	{
		int rule_id;
		if ((rc = decode_int(m, &rule_id)) < 0)
			return rc; 

		if (!(*pa = sct_make_action_disable_rule(rule_id)))
			return -ENOMEM; 
		
		break;
	}
	case AT_PROCESS_SUSPEND:
	{
		int pid; 
		if ((rc = decode_int(m, &pid)) < 0)
			return rc; 
		if (!(*pa = sct_make_action_suspend_process(pid)))
			return -ENOMEM; 

		break;
	}
	case AT_PROCESS_KILL:
	{
		int pid; 
		
		if ((rc = decode_int(m, &pid)) < 0)
			return rc; 

		if (!(*pa = sct_make_action_kill_process(pid)))
			return -ENOMEM; 
		
		break;
	}
	case AT_SYSCALL_FAIL:
	{
		int error_code;
		if ((rc = decode_int(m, &error_code)) < 0)
			return rc; 


		if (!(*pa = sct_make_action_fail_syscall(error_code)))
			return -ENOMEM; 

		break;
	}
	case AT_NIL:
		/* a NULL action_node. */
		*pa = NULL;
		rc = 0;
		break;
	default:
		/* shouldn't happen. */
		rc = -EINVAL;
		break;
	}

	if (pa && *pa && rc == 0) {
		int i;
		int var_type;
		int attr_name;
		
                /* decode attr set commands. */
		if ((rc = decode_int(m, &((*pa)->attr_set_cmds_len))) < 0)
			return rc; 

		if ((*pa)->attr_set_cmds_len > MAX_ATTR_SET_PER_ACTION) {
			/* enforce sane limits, to avoid abuse by user-code */
			(*pa)->attr_set_cmds_len = MAX_ATTR_SET_PER_ACTION;
		}

		for (i=0; rc == 0 && i < (*pa)->attr_set_cmds_len; i++) {
			struct tracker_action_attr_set_data* atsd =
				&((*pa)->attr_set_cmds[i]);
			if ((rc = decode_int(m, &var_type)) < 0)
				return rc; 

			atsd->var_type = var_type;

			if ((rc = decode_int(m, &atsd->var_index)) < 0)
				return rc; 


			if ((rc = decode_int(m, &attr_name)) < 0)
				return rc; 

			atsd->name = attr_name;

			if ((rc = decode_int(m, &atsd->val)) < 0)
				return rc; 

		}
		/* if we failed decoding attr commands - free the action. */
		if (rc < 0) {
			sct_free_action(*pa);
			*pa = NULL;
		}
	}

	return 0;
}


/*
 * Functions for actions manipulation.
 */

/* allocate an action object for logging purposes. */
sct_action_t* sct_make_action_log(int priority, const char* format)
{
	/* OPTIMIZATION: keep static actions and return pointers to them */
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->attr_set_cmds_len = 0;
		action->type = AT_LOG;
		action->u.log.priority = priority;
		if (format) {
			action->u.log.format =
				(char*)mem_alloc(strlen(format)+1);
			if (action->u.log.format)
				strcpy(action->u.log.format, format);
			else {
				mem_free(action);
				action = NULL;
			}
		}
		else {
			action->u.log.format = NULL;
		}
	}

	return action;
}

/* allocate an action object for rule enabling purposes. */
sct_action_t* sct_make_action_enable_rule(int rule_id)
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->attr_set_cmds_len = 0;
		action->type = AT_RULE_ENABLE;
		action->u.rule_id = rule_id;
	}

	return action;
}

/* allocate an action object for rule disabling purposes. */
sct_action_t* sct_make_action_disable_rule(int rule_id)
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->attr_set_cmds_len = 0;
		action->type = AT_RULE_DISABLE;
		action->u.rule_id = rule_id;
	}

	return action;
}

/* allocate an action object for process suspending. */
sct_action_t* sct_make_action_suspend_process(int pid)
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->attr_set_cmds_len = 0;
		action->type = AT_PROCESS_SUSPEND;
		action->u.pid = pid; 
	}

	return action;
}

/* allocate an action object for process killing. */
sct_action_t* sct_make_action_kill_process(int pid)
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->attr_set_cmds_len = 0;
		action->type = AT_PROCESS_KILL;
		action->u.pid = pid; 
	}

	return action;
}

/* allocate an action object for system call failing. */
sct_action_t* sct_make_action_fail_syscall(int error_code)
{
	struct tracker_action* action =
		mem_alloc(sizeof(struct tracker_action));
	if (action) {
		action->attr_set_cmds_len = 0;
		action->type = AT_SYSCALL_FAIL;
		action->u.error_code = error_code;
	}

	return action;
}

/* set action variable attribute set command. */
void set_action_var_attr_set_command(sct_action_t* a,
				     sct_var_type_t var_type,
				     sct_var_attr_name_t name,
				     int val)
{
	if (a && a->attr_set_cmds_len < MAX_ATTR_SET_PER_ACTION) {
		a->attr_set_cmds[a->attr_set_cmds_len].var_type = var_type;
		a->attr_set_cmds[a->attr_set_cmds_len].var_index = -1;
		a->attr_set_cmds[a->attr_set_cmds_len].name = name;
		a->attr_set_cmds[a->attr_set_cmds_len].val = val;
		a->attr_set_cmds_len++;
	}
}

void set_action_var_indexed_attr_set_command(sct_action_t* a,
				 	     sct_var_type_t var_type,
				 	     int var_index,
				 	     sct_var_attr_name_t name,
				 	     int val)
{
	if (a && a->attr_set_cmds_len < MAX_ATTR_SET_PER_ACTION) {
		a->attr_set_cmds[a->attr_set_cmds_len].var_type = var_type;
		a->attr_set_cmds[a->attr_set_cmds_len].var_index = var_index;
		a->attr_set_cmds[a->attr_set_cmds_len].name = name;
		a->attr_set_cmds[a->attr_set_cmds_len].val = val;
		a->attr_set_cmds_len++;
	}
}

void sct_free_action(sct_action_t* a)
{
	if (a) {
		switch(a->type) {
		case AT_LOG:
			if (a->u.log.format)
				mem_free(a->u.log.format);
			break;
		default: break; /* do nothing */
		}
		mem_free(a);
	}
}


/*
 * Action serialization/de-serialization functions.
 */

/* serialize an action node into a buffer, for export     */
/* purposes. returns the size of the serialized data, 0   */
/* if there was an error, -1 if buffer is too small.      */
int sct_serialize_action_node(sct_action_t* a, struct sct_membuf* m)
{
	int rc = do_serialize_action_node(a, m);

	return rc; 
}

/* de-serialize an action node from a buffer, for import  */
/* purposes. returns the new allocated action node or     */
/* NULL if an error occuerd.                              */
sct_action_t* sct_deserialize_action_node(struct sct_membuf* m)
{
	sct_action_t* a = NULL;
	int rc = 0;

	rc = do_deserialize_action_node(&a, m);
	
	return (rc == 0 ? a : NULL); 
}

#define ACTION_AS_STR_BUFLEN 512

/* get a string representation of this action */
char* sct_get_action_as_string(sct_action_t* a)
{
	char* buf; 
	char* lf; 
	int i;

	if (!a)
		return NULL;

	/* should be large enough for everybody */
	buf = mem_alloc(ACTION_AS_STR_BUFLEN); 
	if (!buf)
		return NULL;

	switch (a->type) {
	case AT_LOG:
		/* we need to escape special chars in lf */ 
		lf = a->u.log.format; 
		if (lf) 
			lf = sct_buffer_to_string(lf, strlen(lf)); 

		/* be very careful not to overflow our buffer with the */
		/* user supplied log format */
		if (lf && (strlen(lf) > 200))
			lf = "[truncated]";
		
		/* cant use snprintf, as long as we support 2.2 */
		/* unless we implement it ourselves, of course. */
		sprintf(buf, "type = LOG log_format \"%s\" priority %d", lf, 
			a->u.log.priority);
		mem_free(lf); 
		break;
	case AT_RULE_ENABLE:
		sprintf(buf, "type = ENABLE, rule_id = %d", a->u.rule_id);
		break;
	case AT_RULE_DISABLE:
		sprintf(buf, "type = DISABLE rule_id = %d", a->u.rule_id);
		break;
	case AT_PROCESS_SUSPEND:
		sprintf(buf, "type = SUSPEND pid = %d", a->u.pid);
		break;
	case AT_PROCESS_KILL:
		sprintf(buf, "type = KILL pid = %d", a->u.pid);
		break;
	case AT_SYSCALL_FAIL:
		sprintf(buf, "type = FAIL pid = %d", a->u.error_code);
		break;
	default:
		sprintf(buf, "(unknown)");
		break;
	}
	/* log attr set commands. */
	for (i=0; i < a->attr_set_cmds_len; i++) {
		struct tracker_action_attr_set_data* atsd = &a->attr_set_cmds[i];
		char* pb = buf+strlen(buf);
		if (ACTION_AS_STR_BUFLEN - (pb - buf) > 50) {
			sprintf(pb, "; attr set: var %d(%d), attr %d = %d",
				atsd->var_type, atsd->var_index, atsd->name, atsd->val);
		}
		else {
			sprintf(pb, "; attr set truncated");
		}
	}

	return buf;
}

void sct_get_syscall_id_as_str(int syscall_id, char* buf, size_t sz)
{
	/* ugly hackery, since we dont have snprintf() in all 2.4 kernels */ 
	/* strlen of UINT_MAX == 10 */ 
	const unsigned int minbufsz = 22; 

	assert(buf && sz); 
	
	if (!buf || (sz < minbufsz))
		return; 

	/* if it's a mux syscall */ 
	if (FUNC_PART_OF_SYSCALL_ID(syscall_id)) {
		sprintf(buf, "%d:%d", 
			SYSCALL_PART_OF_SYSCALL_ID(syscall_id), 
			FUNC_PART_OF_SYSCALL_ID(syscall_id)); 
	} else {
		sprintf(buf, "%d", syscall_id); 
	}
}

/* print an action node to a "standard" device */
void sct_print_action_node(sct_action_t* a)
{
	char* desc = sct_get_action_as_string(a);
	PPRINT("%s", desc ? desc : "(nil)");
	mem_free(desc);
}

/*
 * Functions for tracking rules manipulation.
 */

/* serialize a tracking rule into a buffer, for export   */
/* purposes. returns the size of the serialized data, 0  */
/* if there was an error, -1 if the buffer is too small. */
int sct_serialize_tracking_rule(int rule_id, sct_rule_when_t when, int syscall_id,
				sct_action_t* a, sct_filter_node_t* f,
				struct sct_membuf* m)
{
	int rc; 
	int sum = 0; 

	if ((rc = encode_int(m, rule_id)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_int(m, when)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = encode_int(m, syscall_id)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = sct_serialize_action_node(a, m)) < 0)
		return rc; 

	sum += rc; 
	if ((rc = sct_serialize_filter_node(f, m)) < 0)
		return rc; 

	sum += rc; 
	return sum; 
}

/* de-serialize a tracking rule from a buffer, for import */
/* purposes. returns the 1 for success, -1 for failure.   */
/* the given action rule and filter node are filled with  */
/* elements geenrated from the buffer, and should be      */
/* freed by the caller.                                   */
int sct_deserialize_tracking_rule(struct sct_membuf* m, 
				  int* p_rule_id, sct_rule_when_t* p_when,
				  int* p_syscall_id,
				  sct_action_t** pa, sct_filter_node_t** pf)
{
	int rc; 

	if ((rc = decode_int(m, p_rule_id)) < 0)
		return rc; 

	if ((rc = decode_int(m, (int*)p_when)) < 0)
		return rc; 

	if ((rc = decode_int(m, p_syscall_id)) < 0)
		return rc; 

	if ((rc = do_deserialize_action_node(pa, m)) < 0)
		return rc; 

	if ((rc = do_deserialize_filter_node(pf, m)) < 0) {
		sct_free_action(*pa);
		return rc; 
	}

	return 0; 		
}

/* create an empty tracking rules list. */
sct_rule_list_t* sct_create_tracking_rule_list(void)
{
	sct_rule_list_t* rules = mem_alloc(sizeof(struct rule_list));
	if (rules) {
		rules->num_rules = 0;
	}

	return rules;
}

/* delete the given tracking rule list, freeing any rule found in it. */
void sct_delete_tracking_rule_list(sct_rule_list_t* rules)
{
	if (rules) {
		sct_del_all_tracking_rules_in_list(rules);
		mem_free(rules);
	}
}

/* add a new tracking rule. returns the new rule's id     */
/* returns '-1' if an error occured (including if there's */
/* already a rule with the given rule ID).                */
int sct_add_tracking_rule(sct_rule_list_t* rules,
			  int rule_id,
			  sct_filter_node_t* filter,
			  sct_action_t* action)
{
	/* assume failure. */
	int res = -1; 

	if (!filter || !action)
		return res; 

	/* make sure we don't already have a rule with the same ID, */ 
	/* unless the ID is -1 which means "dont care" - put it at  */ 
	/* the end */ 
	if (rule_id == -1) {
		if (!rules) {
			PPRINT("NULL rules in sct_add_tracking_rules!"); 
			rule_id = 0; /* first rule */
		} else 
			rule_id = (rules->num_rules == 0 ? 0 :
				   rules->rules[rules->num_rules - 1]->rule_id + 1); 
	} else {
		int rule_loc = find_tracking_rule(rules, rule_id);
		if (rule_loc >= 0) 
			return -1; 
	}
	
	/* add the rule, sorted by rule ID. */
	res = do_add_tracking_rule(rules, rule_id, filter, action);
	return (res == -1 ? -1 : rule_id);
}

static void do_del_tracking_rule(struct tracker_rule* rule)
{
	if (rule->filter) {
		sct_free_filter_node(rule->filter);
	}
	if (rule->action) {
		sct_free_action(rule->action); 
	}
	mem_free(rule);
}

/* delete a tracking rule. */
void
sct_del_tracking_rule(sct_rule_list_t* rules, int rule_id)
{
	int i;

	if (!rules)
		return; 

	i = find_tracking_rule(rules, rule_id);

	if (i >= 0 && i < rules->num_rules) {
		struct tracker_rule* rule = rules->rules[i];
		int j;

		/* we're removing a rule - close the 'gap' created in the list. */
		for (j = i; j < rules->num_rules-1; j++) {
			rules->rules[j] = rules->rules[j+1];
		}
		/* set to NULL the last rule slot, since it's now in slot last - 1*/ 
		rules->rules[rules->num_rules-1] = NULL; 
		do_del_tracking_rule(rule); 
		rules->num_rules--;
	}
}

/* deletes all tracking rules in the list. */
void sct_del_all_tracking_rules_in_list(sct_rule_list_t* rules)
{
	int i;

	if (!rules)
		return; 

	for (i = 0; i < rules->num_rules; i++) {
		if (rules->rules[i]) {
			do_del_tracking_rule(rules->rules[i]); 
			rules->rules[i] = NULL;
		}
	}
	rules->num_rules = 0;
}

/* check if there is a rule with the given ID. */
/* returns '1' if there is, '0' otehrwise.     */
int sct_has_tracking_rule(sct_rule_list_t* rules, int rule_id)
{
	int i = find_tracking_rule(rules, rule_id);

	return (i >= 0 ? 1 : 0);
}

/* returns the number of rules in the given rule list. */
int sct_get_tracking_rule_count(sct_rule_list_t* rules)
{
	return (rules ? rules->num_rules : 0);
}

/* print out the list of current tracking rules. */
void sct_print_tracking_rules(sct_rule_list_t* rules, unsigned int indent)
{
	int i;
	if (!rules) {
		print_indent(indent); 
		PPRINT("rule list is NULL!\n"); 
		return; 
	}

	for (i=0; i < rules->num_rules; i++) {
		/* skip empty rules */
		if (!rules->rules[i])
			continue;
		print_indent(indent); 
		PPRINT("Rule id: %d\n", rules->rules[i]->rule_id);
		print_indent(indent); 
		PPRINT("action { "); 
		sct_print_action_node(rules->rules[i]->action);
		PPRINT(" }\n"); 
		print_indent(indent); 
		PPRINT("filter_expression {\n"); 
		sct_print_filter_node(rules->rules[i]->filter, indent + INDENT_LEVEL);
		print_indent(indent); 
		PPRINT("}\n"); 
		if (i + 1 < rules->num_rules)
			PPRINT("\n"); 
	}
	PPRINT("\n");
}

/*
 * Functions used by tracking functions to perform filtering.
 */

/* filter an incoming system call.                                          */
/* accepts a system call ID and its (var_args) parameters.                  */
/* returns a tracker_action structure if there was a match, NULL otherwise. */
/* on successfull match, the rule id is set into '*rule_id'.                */
sct_action_t* sct_filter_syscall(sct_rule_list_t* rules, int syscall_id, int* rule_id, ...)
{
	int i;
	sct_action_t* action = NULL;
	sct_filter_node_t* result; 

	if (!rules)
		return NULL; 

	result = sct_make_bool_val(TF_FALSE);

	for (i=0; i<rules->num_rules; i++) {
		if (rules->rules[i] && rules->rules[i]->filter) {
			sct_exp_error_t rc =
				sct_eval_filter_node(rules->rules[i]->filter, result);
			if (rc == ERR_NONE && cast_bool(result)) 
				break;
		}
	}

	if (i < rules->num_rules) {
		action = rules->rules[i]->action;
		if (rule_id)
			*rule_id = rules->rules[i]->rule_id;
	}

	sct_free_filter_node(result);

	return action;
}

static void do_action_log(sct_action_t* action, sct_rule_when_t when, 
			  int rule_id, int syscall_id, 
			  const char* syscall_name, const char* param_spec, 
			  va_list ap)
{
	const char* rule_log_format = action->u.log.format;
	const char* log_format;
	sct_loguser_t* loguser;
	
	if (!rule_log_format || !*rule_log_format)
		rule_log_format =
			"syscall: %pid[%comm]: %sid_%sname(%params) (rule %ruleid)\n";
	
	if (!loguser_head) /* not using log dev */
		tracker_do_logging(NULL, rule_id, syscall_id, syscall_name,
				   rule_log_format, param_spec, ap);
	
	for (loguser = loguser_head; loguser != NULL; 
	     loguser = sct_loguser_get_next(loguser)) {

		lock_loguser(loguser); 
		
		if (!(log_format = sct_loguser_get_log_format (loguser, when)))
			log_format = rule_log_format;

		tracker_do_logging(loguser,
				   rule_id,
				   syscall_id,
				   syscall_name,
				   log_format,
				   param_spec,
				   ap);
		
		unlock_loguser(loguser);
	}
}

static void set_action_attributes(sct_action_t* action)
{
	int i; 

	/* perform set attribute commands. */
	for (i=0; i < action->attr_set_cmds_len; i++) {
		struct tracker_action_attr_set_data* atsd = 
			&action->attr_set_cmds[i];
		if (atsd->var_index == -1)
			set_var_attr(atsd->var_type, atsd->name, atsd->val);
		else
			set_var_indexed_attr(atsd->var_type, atsd->var_index,
					     atsd->name, atsd->val);
	}
}

/*
 * public functions.
 */

/* perform an action in the current system call's context. */
/* accepts a system call ID and its (var_args) parameters. */
/* returns an 'sct_syscall_retval_policy'. if it's         */
/* SCT_OVERRIDE_RETVAL, the new retval to use will be put  */
/* in 'syscall_retval'. otherwise, it's left untouched.    */ 
sct_syscall_retval_policy sct_tracker_perform_action(sct_action_t* action,
        sct_rule_when_t when, int rule_id, int syscall_id, 
        int *syscall_retval, const char* syscall_name, 
        const char* param_spec, ...)
{
	sct_syscall_retval_policy retval = SCT_USE_ORIG_RETVAL; 

	if (!action) 
		return retval; 

	set_action_attributes(action); 

	/* TODO TODO - handle other action types. */
	/* also handle logging with varargs.      */
	switch (action->type) {
	case AT_LOG:
	{
		va_list ap;
		va_start (ap, param_spec);
		do_action_log(action, when, rule_id, syscall_id, 
			      syscall_name, param_spec, ap); 
		va_end(ap);
		break;
	}
	case AT_SYSCALL_FAIL:
	{
		PPRINT(KERN_NOTICE "failing syscall '%s(%d)' with '%d'\n",
		       syscall_name, syscall_id, action->u.error_code);
		retval = SCT_OVERRIDE_RETVAL; 
		if (syscall_retval)
			*syscall_retval = action->u.error_code;
		break;
	}
	case AT_PROCESS_SUSPEND:
	{
		pid_t pid = action->u.pid;
		if (pid == -1)
			pid = sct_getpid();
		
		PPRINT(KERN_NOTICE "suspending process %d\n", pid);
		
		sct_suspend(pid);
		break;
	}
	case AT_PROCESS_KILL:
	{
		pid_t pid = action->u.pid;
		if(pid == -1)
			pid = sct_getpid();
		
		PPRINT(KERN_NOTICE "killing process %d\n", pid);
		sct_kill(pid,SIGKILL);
		
		if (pid == sct_getpid()) {
                        /* If we're in a before rule, and the process
			   we're killing is the current, fail the 
			   system call */
			retval = SCT_OVERRIDE_RETVAL;
			
			/* sct_retval is irrelevant, since the process 
			   (hopefully) will never get to check it */
		}
		break;
	}
	default:
		PPRINT(KERN_WARNING "syscall '%s(%d)' invoked - unknown action %d\n",
		       syscall_name, syscall_id, action->type);
		break;
	}
	return retval;
}

/* serialize a list of tracking rules into a buffer,        */
/* for export purposes. returns the size of the serialized  */
/* data, 0 if there was an error, -1 if the buffer was      */
/* too small.                                               */ 
int sct_serialize_tracking_rule_list(sct_rule_list_t* rules, 
				     struct sct_membuf* mbuf)
{
	int i; 
	int rule_len, rc; 

	if (!rules || !mbuf)
		return -EINVAL; 

	/* first store the number of rules in the list */ 
	if ((rc = sct_membuf_add_chunk(mbuf, &rules->num_rules, 
				       sizeof(rules->num_rules))) < 0) {
		PPRINT("serialize_tracing_rule_list: adding the number "
		       "of rules in the buffer failed with error %d\n", rc); 
		return rc; 
	}
	
	/* for each rule in the list, first store its length, */ 
	/* and then store the rule itself */ 
	for (i = 0; i < rules->num_rules; i++) {
		struct tracker_rule* r;

		/* NULL rule, should never happen */ 
		if (!rules->rules[i]) {
			PPRINT("rule_list_t rules has a NULL rule at location '%d'\n", i); 
			continue; 
		}
		
		r = rules->rules[i]; 
		rule_len = sct_serialize_tracking_rule(r->rule_id, 
						       0, /* bogus when, we never use it */ 
						       0, /* bogus syscall id */ 
						       r->action, 
						       r->filter, 
						       mbuf); 

		/* we have the serialized rule in tempbuf, now copy it to the membuf */ 
		if (rule_len <= 0) {
			PPRINT("sct_serialize_tracking_rule_list: "
			       "rule_len is %d, aborting\n", rule_len); 
			return rule_len; 
		}

	}
	return sct_membuf_available_for_read(mbuf); 
}

/* deserialize a tracking rule_list from a buffer, for import      */
/* purposes. returns 1 for success, -1 for failure. the rule list  */
/* pointed to by p_rules is filled with rules and should be freed  */
/* by the caller (with sct_delete_tracking_rule_list)              */
int sct_deserialize_tracking_rule_list(struct sct_membuf* mbuf, 
				       sct_rule_list_t** p_rules)
{
	int i; 
	int num_rules; 
	sct_rule_list_t* list; 

	int rule_id;
	sct_rule_when_t when; 
	int syscall_id; 
	sct_action_t* a; 
	sct_filter_node_t* f; 

	int rc; 

	if (!p_rules || !mbuf)
		return -EINVAL; 

	/* get the number of rules stored in this list */ 
	if ((rc = sct_membuf_get_chunk(mbuf, &num_rules, sizeof(num_rules))) 
	    != sizeof(num_rules))
		return -ENOSPC; 

	/* create a rule list with this number of rules */ 
	if (!(list = sct_create_tracking_rule_list()))
		return -ENOMEM; 

	/* for each rule, we stored its length first and then the rule itself */ 
	for (i = 0; i < num_rules; i++) {
		if ((rc = sct_deserialize_tracking_rule(mbuf, &rule_id, &when, 
							&syscall_id, &a, &f)) < 0)
			goto free_rule_elements; 

		/* we have succesfully deserialized a rule. now add it to the list of rules */ 
		if ((rc = sct_add_tracking_rule(list, rule_id, f, a)) < 0)
			goto free_rule_elements; 
	}
	/* whee, success! */ 
	*p_rules = list; 
	return 1; 	

 free_rule_elements:
	sct_free_filter_node(f); 
	sct_free_action(a); 
	sct_delete_tracking_rule_list(list);
	return rc; 
}

int sct_fuzzy_cmp_actions(sct_action_t* a, sct_action_t* b)
{
	assert(a); 
	assert(b); 

	return !(a->type == b->type); 
}

