/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2003 Muli Ben-Yehuda, mulix@mulix.org	    */
/* License: GNU General Public License                              */

/*
 * $Id: main.cpp,v 1.17 2003/02/01 19:21:22 mulix Exp $
 */

/* 
 * This is the main entry point into the system call tracker 
 * "strace compatible" utility 
 */

#include <string>
#include <list>

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <libgen.h>
#include <signal.h>
#include <setjmp.h>

/* yucky yucky */ 
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#include <getopt.h>
#undef _GNU_SOURCE
#else 
#include <getopt.h>
#endif /* !_GNU_SOURCE */ 

#include <sys/types.h>
#include <sys/wait.h>

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "sct_rules/rule_engine_public.h"
#include "sct_parselib/parselib.h" 
#include "sct_parselib/syscalls.h"
#include "sct_parselib/logger.h"
#include "sct_parselib/tokens_to_types.h"
#include "sct_parselib/validity_checks.h"

#include "version.h" 

char* progname; 

bool verbose; 

using namespace std;

#ifdef SCT_CONFIG_DEBUG
int debug = 1;
#else /* !defined SCT_CONFIG_DEBUG */ 
int debug = 0;
#endif

struct rule_info { int syscall_id; int rule_id; sct_rule_when_t when; }; 
typedef std::list < struct rule_info > id_list; 

static void usage(const char* name)
{
	char usage[] = 
		"%s (%s) is an strace replacement for the system call tracker.\n"
		"\n"
		"Usage: %s [command [arg ...]]\n"
		"\n"
		"Options: \n"
		"  -v, --verbose                give verbose output.\n"
		"  -p, --process 'pid'          trace the process 'pid'\n"
	        "  -f, --follow-forks           when the process forks, keep\n"
		"                               tracing its children\n"; 
		
        fprintf(stdout, usage, name, VERSION_STRING, name); 
		
        exit(EXIT_FAILURE);
}

void parse_options(char** argv, int argc, int* cmd_index, 
		   pid_t& pid, bool& follow_forks)
{
	int option_index = 0;
	verbose = false;
	pid = 0; 

	while(1) { 
		/* structure with all recognised options for sct_config */
		static struct option long_options[] = {
			{"verbose", 0, NULL, 'v'},  
			{"process", 1, NULL, 'p'}, 
			{"follow-forks", 0, NULL, 'f'}, 
			{0, 0, 0, 0}
		};
		int c;
		const char* optspec = "+vp:f"; 
		c = getopt_long (argc, argv, optspec, long_options, 
				 &option_index);
		
		switch (c) {
		case 0: /* long option specified */ 
			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
			break;
		case 'v':
			verbose = true; 
			break; 
		case 'p': 
			if (is_int(optarg))
				pid = atoi(optarg); 
			else 
				usage(progname); 
			break; 
		case 'f': 
			follow_forks = true; 
			break; 
		case '?': /* unrecognised option */
			/* fall through */
		default:
			c = -1;
			break;
		}
		if(c == -1) 
			break;  /* no more options */
	}

	*cmd_index = optind; 
	argc -= optind; 
	if (argc <= 0 && !pid) { /* no command specified and no -p given */
		usage(progname); 
	}
}


static sct_filter_node_t* make_either_or_match(sct_filter_node_t* head, 
					       sct_filter_node_t* link)
{
	if (!head)
		return link;
	
	if (!link)
		return head;
	
	return sct_make_logical_or(head, link); 
}

static void load_rules(sct_ctrl_t ctrl, pid_t pid, bool follow_forks, 
		       id_list& ids)
{
	int rc; 
	sct_filter_node_t* match = NULL; 

	sct_filter_node_t* pid_match, *forks_match; 
	pid_match = forks_match = NULL; 

	
	sct_filter_node_t* varpid = sct_make_var_val(VT_PROC_PID, DT_INT); 
	sct_filter_node_t* fpid = sct_make_int_val(pid); 
	pid_match = sct_make_equal(varpid, fpid); 

	if (follow_forks) {
		sct_filter_node_t* ancestors = sct_make_var_val(
			VT_PROC_ANCESTORS, DT_MULTI); 
		sct_filter_node_t* fpid = sct_make_int_val(pid); 
		forks_match = sct_make_in_match(fpid, ancestors); 
	}

	match = make_either_or_match(pid_match, forks_match); 

	sct_action_t* act = sct_make_action_log(4, NULL); 
	if (!match || !act) {
		logger::log(logger::LOG_LEVEL_ERROR, "could not create filter "
			    "nodes or action, aborting"); 
		exit(EXIT_FAILURE); 
	}

	logger::log(logger::LOG_LEVEL_DEBUG, "loading rules...\n"); 
	const int rule_id = -1; 
	sct_rule_when_t when; 
	syscalls::syscalls_set::const_iterator it = syscalls::_sys.begin(); 
	for ( ; it != syscalls::_sys.end(); ++it) {
		string sname = (*it).first; 
		int syscall_id = (*it).second->id; 
		when = ((sname == "exit" || sname == "execve") ? RW_BEFORE : RW_AFTER); 
#ifdef SCT_SCTRACE_DEBUG
		if (verbose) {
			char buf[32]; 
			sct_get_syscall_id_as_str(syscall_id, buf, sizeof(buf)); 
			fprintf(stdout, "syscall %s(%s), filter: ", sname.c_str(), buf); 
			sct_print_filter_node(match); 
			fprintf(stdout, "\n"); 
		}
#endif /* SCT_SCTRACE_DEBUG */ 
		rc = sct_ctrl_add_rule(ctrl, rule_id, when, syscall_id,  
				       match, act, 1); 
		if (rc < 0) {
			logger::log(logger::LOG_LEVEL_ERROR, "could not upload "
				    "rule %d, for syscall %s, add_rule returned "
				    "%d\n", rule_id, (*it).first.c_str(), rc); 
		} else {
			struct rule_info info = { syscall_id, rc, when }; 
			ids.push_back(info); 
		}
	}
	/* if we didn't manage to add even one rule, it's time to say buh bye */
	if (!ids.size()) {
		logger::log(logger::LOG_LEVEL_ERROR, "could not upload any "
			    "rules. Make sure you have the necessary "
			    "permissions to access the device file...\n"); 
		exit(EXIT_FAILURE); 
	}
}

static void remove_rules(sct_ctrl_t ctrl, id_list& ids) 
{
	id_list::const_iterator it = ids.begin(); 
	for ( ; it != ids.end(); ++it) {
		int rc = sct_ctrl_del_rule(ctrl, it->rule_id, 
					   it->when, 
					   it->syscall_id); 
		if (!rc) 
			logger::log(logger::LOG_LEVEL_ERROR, "failed to delete "
				    "rule %d for syscall %d (when = %)\n", 
				    it->rule_id, it->syscall_id, it->when); 
	}
}


bool load_syscalls(void)
{
	/* maps from string to enumerated types     */
	/* must be done here since the syscalls.dat */
	/* parser uses them as well                 */
	if (sct_parselib_init()) {
		fprintf(stderr, "could not initialize the parser library.\n"); 
		exit(EXIT_FAILURE); 
	}
	return true; 
}

static void exec_son(int readfd, char* exec, char** argv)
{
	char dummy; 
	/* wait on the pipe for papa to tell us to go on */ 
	int rc = read(readfd, &dummy, sizeof(char)); 
	if (rc != sizeof(char)) {
		if (rc < 0)
			perror("read"); 
		else 
			fprintf(stderr, "got EOF from the pipe?!\n"); 
		exit(EXIT_FAILURE); 
	}
	logger::log(logger::LOG_LEVEL_DEBUG, "executing...\n"); 
	
	execv(exec, argv); 
	
	/* we only get here on failure */ 
	perror("execv"); 
	
	exit(EXIT_FAILURE); 
}

static void wait_for_son(int writefd, pid_t son, sct_ctrl_t ctrl, 
			 bool follow_forks, id_list& ids)
{
	load_rules(ctrl, son, follow_forks, ids); 

	int rc = write(writefd, "go!", sizeof("go!")); 
	if (rc < 0) {
		perror("write"); 
		exit(EXIT_FAILURE); 
	}
	int status; 
	while ((rc = waitpid(son, &status, 0)) != -1) {
		if (WIFEXITED(status) || WIFSIGNALED(status))
			break; 
	}
	if (rc == -1) {
		perror("waitpid"); 
		exit (EXIT_FAILURE); 
	}
	remove_rules(ctrl, ids); 
}

static char* find_in_path(const char* name)
{
	char* ret = NULL;
	char* env = getenv("PATH"); 

	if (!env)
		return NULL; 

	std::string path(env); 
	std::string part; 
	
	/* find the first : character. */
	std::string::size_type end = 0, beg = 0; 
	while (1) {
		end = path.find_first_of(":", beg); 
		part = path.substr(beg, end - beg); 
		part += "/"; part += name; 
		/* check if we can run this executable */ 
		if (!access(part.c_str(), X_OK)) {
			ret = strdup(part.c_str()); 
			break; 
		}
		if (end == path.npos)
			break; 
		++end; 
		beg = end; 
	}
	return ret; 
}

void sctrace_executable(sct_ctrl_t ctrl, char* exec, bool follow_forks, 
			char** argv, int argc, id_list& ids)
{
	char* fullpath; 

	if (*exec == '/')
		fullpath = exec; 
	else if ((*exec == '.') && 
		 ((strlen(exec) > 1) && (exec[1] == '.' || exec[1] == '/')))
		fullpath = exec; 
	else if (!(fullpath = find_in_path(exec))) {
		logger::log(logger::LOG_LEVEL_ERROR, "could not find file "
			    "to execute '%s'\n", exec); 
		exit(EXIT_FAILURE); 
	}

	int pfd[2]; 
	if (pipe(pfd) < 0) {
		perror("pipe"); 
		exit(EXIT_FAILURE);  
	}
	
	int rc = fork(); 
	if (rc < 0) {
		perror("fork"); 
		exit(EXIT_FAILURE); 
	}
	if (rc == 0) {
		/* child */ 
		close(pfd[1]); 
		exec_son(pfd[0], fullpath, argv); 
	} else {
		/* father */ 
		close(pfd[0]); 
		wait_for_son(pfd[1], rc, ctrl, follow_forks, ids); 
	}
}

void sctrace_pid(sct_ctrl_t ctrl, pid_t pid, bool follow_forks, id_list& ids)
{
	int rc; 

	/* does it exist? */ 
	if (kill(pid, 0)) {
		logger::log(logger::LOG_LEVEL_ERROR, "no such process '%d', "
			    "sorry\n", pid); 
		exit(EXIT_FAILURE); 
	}

	load_rules(ctrl, pid, follow_forks, ids); 

	while (1) {
		if (!(rc = kill(pid, 0))) {
			sleep(1); 
			continue; 
		}
		break; 
	}
		
	assert(rc == -1); 
	if (errno != ESRCH) {
		logger::log(logger::LOG_LEVEL_ERROR, "uh oh, sctrace_pid: kill "
			    "returned with error %s\n", strerror(errno)); 
		exit(EXIT_FAILURE); 
	}
	remove_rules(ctrl, ids); 
}

sigjmp_buf env; 

static void sighandler(int signum)
{
	fprintf(stderr, "sighandler: caught signal %d\n", signum); 
	
	siglongjmp(env, 1);
}

static void signal_setup(void)
{
        /* we need to protect from SIGINT, SIGQUIT and SIGTERM */ 
	struct sigaction act; 
	
	memset(&act, 0, sizeof(act)); 

	act.sa_handler = &sighandler; 
	sigemptyset(&act.sa_mask); 

	if (sigaction(SIGINT, &act, NULL)) {
		perror("sigaction"); 
		return; 
	}
	if (sigaction(SIGQUIT, &act, NULL)) {
		perror("sigaction"); 
		return; 
	}
	if (sigaction(SIGTERM, &act, NULL)) {
		perror("sigaction"); 
		return; 
	}
}

int main(int argc, char ** argv)
{
	sct_ctrl_impl_t impl = SCT_CTRL_DEVICE_FILE; 
	int cmd_index = -1; 
	bool follow_forks = false; 
	pid_t pid = 0; 
	id_list ids; 

	progname = argv[0];

	if (argc < 2)
		usage(progname);

	if (verbose)
		logger::set_verbosity_level(logger::LOG_LEVEL_INFO);

	parse_options(argv, argc, &cmd_index, pid, follow_forks); 

	if (!load_syscalls()) 
		return EXIT_FAILURE; 

	/* init the control library */
	sct_ctrl_t sct_ctrl = sct_ctrl_init(impl);
	if (!sct_ctrl){
		logger::log(logger::LOG_LEVEL_ERROR,
			    "sctrace failed to initialize the ctrl library.\n"
			    "are you sure the kernel modules are loaded?\n");
		return EXIT_FAILURE;
	}

	signal_setup(); 

	if (sigsetjmp(env, 1)) {
		/* we had a signal, remove all rules and get out */ 
		remove_rules(sct_ctrl, ids); 
		exit(EXIT_SUCCESS); 
	}

	if (pid) 
		sctrace_pid(sct_ctrl, pid, follow_forks, ids); 
	else 
		sctrace_executable(sct_ctrl, argv[cmd_index], follow_forks, 
				   argv + cmd_index, argc - cmd_index, ids); 
}
