/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/* 
 * $Id: get_rules_test.c,v 1.1 2003/01/31 18:35:32 mulix Exp $
 */ 

/* test the 'get rules' and 'get rule count' functionality, */ 
/* using both sysctl and device file */ 

#include <assert.h>

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "sct_rules/membuf.h"

#include "tester.h"

sct_ctrl_t sysctl, dev; 

static void init_sct_ctrl(void)
{
	/* initialize sct_ctrl_lib for sysctl and device file*/ 
	sct_ctrl_impl_t isysctl = SCT_CTRL_SYSCTL; 
	sct_ctrl_impl_t idev = SCT_CTRL_DEVICE_FILE; 

	if (!(sysctl = sct_ctrl_init(isysctl))) {
		die("failed to init sct_ctrl_lib for sysctl\n");
	}

	if (!(dev = sct_ctrl_init(idev))) {
		die("failed to init sct_ctrl_lib for device file\n");
	}

	if (!sct_ctrl_del_all_rules(sysctl))
		die("failed to delete the module rules\n"); 
		
	if (!sct_ctrl_del_all_rules(dev))
		die("failed to delete the module rules\n"); 
}

#if 0 /* don't print the rule lists for the time being */ 
static void 
print_rule_list(syscall_rules_t* rules)
{
	assert(rules); 

	for ( ; rules; rules = rules->next) {
		char buf[32]; 
		sct_get_syscall_id_as_str(rules->syscall_id, buf, sizeof(buf)); 
		printf(" syscall: %s, %s rules:\n", buf, 
		       (rules->when == RW_BEFORE? "before" : "after")); 
		sct_print_tracking_rules(rules->list); 
	}
}
#endif /* 0 */ 

static int compare_rule_list(syscall_rules_t* r1, syscall_rules_t* r2)
{
	int ret = 0; 
	struct sct_membuf* m1 = sct_membuf_create(1024, NULL, NULL, 
					     SCT_MEMBUF_EXACT_MATCH); 
	struct sct_membuf* m2 = sct_membuf_create(1024, NULL, NULL, 
					     SCT_MEMBUF_EXACT_MATCH); 
	int rc1, rc2; 
	size_t a1, a2; 
	unsigned char* d1, *d2; 

	assert(r1 && r2); 

	if (!m1 || !m2) 
		goto done; 

	for ( ; r1 && r2; r1 = r1->next, r2 = r2->next) {
		if (r1->syscall_id != r2->syscall_id || r1->when != r2->when)
			goto done; 
		/* compare the rules lists themselves */ 
		/* we're pretty sneaky here, doing it by seraializing */ 
		/* them and the comparing the byte stream */ 
		sct_membuf_clear(m1); sct_membuf_clear(m2); 
		rc1 = sct_serialize_tracking_rule_list(r1->list, m1); 
		rc2 = sct_serialize_tracking_rule_list(r2->list, m2); 
		CHECK((rc1 == rc2), 1); 
		if (rc1 != rc2)
			goto done; 
		a1 = sct_membuf_available_for_read(m1); 
		a2 = sct_membuf_available_for_read(m2); 
		CHECK((a1 == a2), 1); 
		if (a1 != a2)
			goto done; 
		d1 = sct_membuf_get_data(m1); 
		d2 = sct_membuf_get_data(m2); 
		CHECK(memcmp(d1, d2, a1), 0); 
		if (memcmp(d1, d2, a1))
			goto done; 
	}
	if (r1 || r2) 
		goto done; 

	ret = 1; 
	
 done: 
	sct_membuf_destroy(m1); 
	sct_membuf_destroy(m2); 
	return ret;
}

int test_get_rules(void)
{
	int rule_count; 
	int rc1, rc2; 

	syscall_rules_t d, s, *di = &d, *si = &s; 

	init_sct_ctrl(); 

	assert(sct_ctrl_get_all_rule_count(sysctl) == 0); 
	assert(sct_ctrl_get_all_rule_count(dev) == 0); 

	/* upload some rules */ 
	rule_count = test_all_syscalls_register_rules(); 
	
	rc1 = sct_ctrl_get_all_rule_count(sysctl); 
	rc2 = sct_ctrl_get_all_rule_count(dev); 
	printf("test_get_rule_count: rc1 is %d, rc2 is %d, rule_count=%d\n", 
	       rc1, rc2, rule_count); 
	assert(rc1 == rule_count); 
	assert(rc1 == rc2); 

	/* now get the rules */ 
	rc1 = sct_ctrl_get_all_rules(sysctl, &s); 
	rc2 = sct_ctrl_get_all_rules(dev, &d); 
	printf("test_get_rule_count: rc1 is %d\n", rc1); 
	assert(rc1 == rc2); 
	
	/* compare the two lists */ 
	/* dont do the loop if there are no rules registered */ 
	while (rc1 && si && di) {
		assert(compare_rule_list(di, si)); 
		si = si->next; 
		di = di->next; 
	}
	if (rc1)
		assert(!si && !di); 

	if (!sct_ctrl_del_all_rules(sysctl))
		die("failed to delete the module rules\n"); 
		
	if (!sct_ctrl_del_all_rules(dev))
		die("failed to delete the module rules\n"); 

	sct_ctrl_cleanup(dev); 
	sct_ctrl_cleanup(sysctl); 

	return 1; 
}
