/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/* 
 * $Id: membuf_test.c,v 1.1 2003/01/31 18:35:32 mulix Exp $
 */

#include <assert.h>
#include <stdio.h>

#include "sct_rules/membuf.h"
#include "tester.h"

int test_ring_membuf(void); 

int test_membuf_enlarge(void)
{
	int ret = 1; 
	char buf[] = "hello, world!"; 
	char buf2[sizeof(buf)]; 

	struct sct_membuf* m; 

	m = sct_membuf_create(10, NULL, NULL, SCT_MEMBUF_EXACT_MATCH); 
	
	if (sct_membuf_enlarge(m, 20) < 0) {
		ret = 0; 
		goto done; 
	}
	
	CHECK(sct_membuf_total_size(m), 20); 
	CHECK(sct_membuf_available_for_read(m), 0); 
	CHECK(sct_membuf_free_for_write(m), 20); 

	CHECK(sct_membuf_add_chunk(m, buf, sizeof(buf)), sizeof(buf)); 
	
	if (sct_membuf_enlarge(m, 40) < 0) {
		ret = 0; 
		goto done; 
	}

	CHECK(sct_membuf_get_chunk(m, buf2, sizeof(buf2)), sizeof(buf2)); 

 done: 
	sct_membuf_destroy(m); 
	return ret; 
}

int test_membuf_resize(void)
{
	int ret = 1; 
	char buf[] = "hello, world!"; 

	struct sct_membuf* m; 

	m = sct_membuf_create(50, NULL, NULL, SCT_MEMBUF_EXACT_MATCH); 
	
	if (sct_membuf_resize(m, 40) < 0) {
		ret = 0; 
		goto done; 
	}
	
	CHECK(sct_membuf_total_size(m), 40); 
	CHECK(sct_membuf_available_for_read(m), 0); 
	CHECK(sct_membuf_free_for_write(m), 40); 

	CHECK(sct_membuf_add_chunk(m, buf, sizeof(buf)), sizeof(buf)); 
	
	if (sct_membuf_resize(m, 30) < 0) {
		ret = 0; 
		goto done; 
	}

	CHECK(sct_membuf_total_size(m), 30); 
	CHECK(sct_membuf_available_for_read(m), 0); 
	CHECK(sct_membuf_free_for_write(m), 30); 

 done: 
	sct_membuf_destroy(m); 
	return ret; 
}

int test_membuf(void)
{
	char buf[] = "hello, world!"; 
	char buf2[sizeof(buf)]; 
	char scratch[1000] = {0}; 
	int rc; 

	struct sct_membuf* m1; 
	struct sct_membuf* m2; 

	printf("\ntesting memory buffers\n"); 

	m1 = sct_membuf_create(100, NULL, NULL, SCT_MEMBUF_EXACT_MATCH); 
	m2 = sct_membuf_create_with_data(buf, sizeof(buf), NULL, NULL); 
	assert(m1 && m2); 

	CHECK(sct_membuf_get_chunk(m2, buf2, sizeof(buf2)), sizeof(buf2)); 
	assert(!memcmp(buf, buf2, sizeof(buf))); 

	sct_membuf_add_chunk(m1, buf, sizeof(buf)-1); /* get rid of the NULL */ 
	sct_membuf_add_chunk(m1, buf, sizeof(buf)); 

	printf("m1 now contains: %s\n", sct_membuf_get_data(m1)); 
	printf("removing a chunk of size 6 from m1...\n"); 
	sct_membuf_get_chunk(m1, buf2, 6);
	buf2[6] = '\0';
	printf("chunk is '%s', m2 is '%s'\n", buf2, sct_membuf_get_data(m1)); 

	printf("trying to get a really big chunk...\n"); 
	rc = sct_membuf_get_chunk(m1, scratch, sizeof(scratch)); 
	printf("get_chunk returned '%d', buffer is now '%s'\n", rc, scratch); 

	printf("trying to get a 0 chunk...\n"); 
	rc = sct_membuf_get_chunk(m2, scratch, 0); 
	printf("get_chunk returned '%d'\n", rc);

	printf("trying to push a big chunk to a small buffer\n");
	rc = sct_membuf_add_chunk(m1, scratch, sizeof(scratch)); 
	printf("add_chunk returned '%d'\n", rc);

	sct_membuf_destroy(m1); 
	sct_membuf_destroy(m2); 

	printf("testing enlarge\n"); 
	rc = test_membuf_enlarge(); 
	printf("enlarge returned: %d\n", rc); 

	printf("testing resize\n");
	rc = test_membuf_resize();
	printf("resize returned: %d\n",rc);

	return rc; 
}

int test_ring_membuf_exact(void)
{
	struct sct_membuf* m1; 
	char buf[32]; 

	printf("\ntesting ring memory buffers\n"); 

	m1 = sct_membuf_create(5, NULL, NULL, SCT_MEMBUF_EXACT_MATCH); 
	CHECK(sct_membuf_add_chunk(m1, "1", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "2", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "3", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "4", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "5", 1), 1); 

	/* this one should fail with -ENOSPC - no free space */ 
	CHECK(sct_membuf_add_chunk(m1, "6", 1), -ENOSPC); 

	CHECK(sct_membuf_get_chunk(m1, buf, 1), 1); 
	CHECK((*buf == '1'), 1); 

	CHECK(sct_membuf_add_chunk(m1, "6", 1), 1); 
	
	CHECK((memcmp(sct_membuf_get_data(m1), "62345", 5)), 0); 

	/* a???? */
	CHECK(sct_membuf_get_chunk(m1, buf, 4), 4); 

	/* 6aaaa */ 
	CHECK(sct_membuf_add_chunk(m1, "aaaa", 4), 4); 
	CHECK((memcmp(sct_membuf_get_data(m1), "6aaaa", 5)), 0); 

	/* 6aaaa */ 
	CHECK(sct_membuf_get_chunk(m1, buf, 5), 5); 
	CHECK((memcmp(buf, "6aaaa", 5)), 0); 

	/* 77777 */ 
	CHECK(sct_membuf_add_chunk(m1, "77777", 5), 5); 
	CHECK((memcmp(sct_membuf_get_data(m1), "77777", 5)), 0); 

	/* 7777 */ 
	buf[5] = 'A'; 
	CHECK(sct_membuf_get_chunk(m1, buf, 5), 5); 
	CHECK((memcmp(buf, "77777A", 6)), 0); 

	printf("\nfinished testing ring memory buffers\n"); 

	sct_membuf_destroy(m1); 
	return 1; 
}

int test_ring_membuf_drop_new(void)
{
	struct sct_membuf* m1; 
	char buf[32]; 

	printf("\ntesting ring memory buffers\n"); 

	m1 = sct_membuf_create(5, NULL, NULL, SCT_MEMBUF_DROP_NEW); 
	CHECK(sct_membuf_add_chunk(m1, "1", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "2", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "3", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "4", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "5", 1), 1); 

	/* this one should fail with 0 - no free space */ 
	CHECK(sct_membuf_add_chunk(m1, "6", 1), 0); 

	CHECK(sct_membuf_get_chunk(m1, buf, 1), 1); 
	CHECK((*buf == '1'), 1); 

	CHECK(sct_membuf_add_chunk(m1, "6", 1), 1); 
	
	CHECK((memcmp(sct_membuf_get_data(m1), "62345", 5)), 0); 

	/* a???? */
	CHECK(sct_membuf_get_chunk(m1, buf, 4), 4); 

	/* 6aaaa */ 
	CHECK(sct_membuf_add_chunk(m1, "aaaaa", 5), 4); 
	CHECK((memcmp(sct_membuf_get_data(m1), "6aaaa", 5)), 0); 

	/* 6aaaa */ 
	CHECK(sct_membuf_get_chunk(m1, buf, 5), 5); 
	CHECK((memcmp(buf, "6aaaa", 5)), 0); 

	/* 77777 */ 
	CHECK(sct_membuf_add_chunk(m1, "7777777777", 10), 5); 
	CHECK((memcmp(sct_membuf_get_data(m1), "77777", 5)), 0); 

	/* 7777 */ 
	buf[5] = 'A'; 
	CHECK(sct_membuf_get_chunk(m1, buf, sizeof(buf)), 5); 
	CHECK((memcmp(buf, "77777A", 6)), 0); 

	printf("\nfinished testing ring memory buffers\n"); 

	sct_membuf_destroy(m1); 

	return 1; 
}

int test_ring_membuf_drop_old(void)
{
	struct sct_membuf* m1; 
	char buf[32]; 

	printf("\ntesting ring memory buffers\n"); 

	m1 = sct_membuf_create(5, NULL, NULL, SCT_MEMBUF_DROP_OLD); 
	CHECK(sct_membuf_add_chunk(m1, "1", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "2", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "3", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "4", 1), 1); 
	CHECK(sct_membuf_add_chunk(m1, "5", 1), 1); 

	/* this one should succeed */ 
	CHECK(sct_membuf_add_chunk(m1, "6", 1), 1); 

	CHECK(sct_membuf_get_chunk(m1, buf, 5), 5); 
	buf[5] = '\0'; 
	CHECK(strcmp(buf, "23456"), 0); 

	CHECK(sct_membuf_add_chunk(m1, "7", 1), 1); 
	CHECK((memcmp(sct_membuf_get_data(m1), "67345", 5)), 0); 

	/* ?7??? */
	CHECK(sct_membuf_get_chunk(m1, buf, 1), 1); 
	CHECK((memcmp(buf, "7", 1)), 0); 

	/* bbbbb */ 
	CHECK(sct_membuf_add_chunk(m1, "aaaaabbbbb", 10), 10); 
	CHECK((memcmp(sct_membuf_get_data(m1), "bbbbb", 5)), 0); 

	/* bbbbb */ 
	CHECK(sct_membuf_get_chunk(m1, buf, 5), 5); 
	CHECK((memcmp(buf, "bbbbb", 5)), 0); 

	/* ????? */ 
	CHECK(sct_membuf_get_chunk(m1, buf, 5), 0); 

	/* 77777 */ 
	CHECK(sct_membuf_add_chunk(m1, "7777777777", 10), 10); 
	CHECK((memcmp(sct_membuf_get_data(m1), "77777", 5)), 0); 

	/* 7777 */ 
	buf[5] = 'A'; 
	CHECK(sct_membuf_get_chunk(m1, buf, sizeof(buf)), 5); 
	CHECK((memcmp(buf, "77777A", 6)), 0); 


	printf("\nfinished testing ring memory buffers\n"); 

	sct_membuf_destroy(m1); 

	return 1; 
}

int test_cyclic_membuf(void)
{
	char msg[] = "syscall 11 at sadfklasdlkfj allkdk 139"; 
	char tst[] = "139syscall 11 at sadfklasdlkfj allkdk "; 
	char tmp[sizeof(msg)]; 
	struct sct_membuf* m = sct_membuf_create(strlen(msg) + 3, NULL, NULL, 
					    SCT_MEMBUF_DROP_OLD); 

	CHECK(sct_membuf_add_chunk(m, msg, strlen(msg)), strlen(msg)); 
	CHECK(sct_membuf_add_chunk(m, msg, strlen(msg)), strlen(msg)); 

	CHECK(sct_membuf_get_chunk(m, tmp, strlen(msg)), strlen(msg)); 
	tmp[strlen(msg)] = '\0'; 
	CHECK(strcmp(tmp, tst), 0); 

	sct_membuf_destroy(m); 

	return 1; 	
}
