/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/* 
 * $Id: rule_engine_test.c,v 1.2 2003/01/31 19:01:31 mulix Exp $
 */

#include <assert.h>
#include <stdio.h>

#include "sct_rules/rule_engine_public.h"
#include "sct_rules/rule_engine_private.h"
#include "sct_rules/membuf.h"

#include "tests/stress/tester.h"

int test_empty_tracking_rule_list(void)
{
	int rc; 
	struct sct_membuf* mbuf; 

	/* create a tracking rule list */ 
	sct_rule_list_t* list = sct_create_tracking_rule_list(); 
	if (!list)
		die("creating an empty tracking rule list failed"); 

	/* create a membuf */ 
	mbuf = sct_membuf_create(1000, NULL, NULL, SCT_MEMBUF_EXACT_MATCH); 
	if (!mbuf)
		die("couldn't create a membuf"); 

	/* serialize the list */ 
	rc = sct_serialize_tracking_rule_list(list, mbuf); 
	if (rc <= 0)
		die("serializing the list failed"); 
	
	/* delete the list */ 
	sct_delete_tracking_rule_list(list); 

	rc = sct_deserialize_tracking_rule_list(mbuf, &list); 
	if (rc <= 0) 
		die("failed to deserialize list"); 
	
	sct_membuf_destroy(mbuf); 
	sct_delete_tracking_rule_list(list); 
	return 1; 
}

int test_tracking_rule_list(void)
{
	int rc; 
	int rid = 0; 
	sct_rule_list_t* list; 
	struct sct_membuf* mbuf; 

	(void)test_empty_tracking_rule_list(); 

	/* create a tracking rule list */ 
	list = sct_create_tracking_rule_list(); 
	if (!list)
		die("creating an empty tracking rule list failed"); 

	/* add a few rules to the list */ 
	while (++rid < 10){
		sct_filter_node_t* f = sct_make_int_val(rid * 3);
		sct_action_t* a = sct_make_action_log(8, "bla bla"); 
		rc = sct_add_tracking_rule(list, rid, f, a); 
		if (rc < 0)
			die("failed to add a tracking rule to the list"); 
	}

	/* create a membuf */ 
	mbuf = sct_membuf_create(1000, NULL, NULL, SCT_MEMBUF_EXACT_MATCH); 
	if (!mbuf)
		die("couldn't create a membuf"); 

	/* serialize the list */ 
	rc = sct_serialize_tracking_rule_list(list, mbuf); 
	if (rc <= 0)
		die("serializing the list failed"); 
	
	/* delete the list */ 
	sct_delete_tracking_rule_list(list); 

	rc = sct_deserialize_tracking_rule_list(mbuf, &list); 
	if (rc <= 0) 
		die("failed to deserialize list"); 

	/* print the list */ 
	sct_print_tracking_rules(list, 0);  
	
	sct_membuf_destroy(mbuf); 
	sct_delete_tracking_rule_list(list); 
	return 1; 
}

int test_long_long(void)
{
	int rc; 
	sct_filter_node_t* f1, *f2, *op, *res; 
	long long l1 = 0xf000000000000; 
	long long l2 = l1 + 1; 
	f1 = sct_make_long_long_val(l1); 
	f2 = sct_make_long_long_val(l2); 
	res = sct_make_bool_val(0); 

	op = sct_make_less_then(f1, f2); 
	
	CHECK(f1 && f2 && op, 1); 
	rc = sct_eval_filter_node(op, res); 
	CHECK(rc, ERR_NONE); 
	fprintf(stdout, "result should be (bool)true: \n"); 
	sct_print_filter_node(res, 8); 

	sct_free_filter_node(op);
	sct_free_filter_node(res);
	return 1; 
}

int test_void_ptr(void)
{
	int rc;
	sct_filter_node_t* f1, *f2, *op, *res;
	void* ptr1 = (void*)0x1234;
	void* ptr2 = (void*)0x1238;
	f1 = sct_make_ptr_val(ptr1);
	f2 = sct_make_ptr_val(ptr2);
	res = sct_make_bool_val(0);

	op = sct_make_less_then(f1, f2);

	CHECK(f1 && f2 && op, 1);
	rc = sct_eval_filter_node(op,res);
	CHECK(rc, ERR_NONE); 
	fprintf(stdout, "result should be (bool)true: \n"); 
	sct_print_filter_node(res, 8); 

	sct_free_filter_node(op);
	sct_free_filter_node(res);
	return 1; 

}
