/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2003 Muli Ben-Yehuda, mulix@mulix.org         */ 
/* License: GNU General Public License                              */

/* 
 * $Id: sct_rules_module_tests.c,v 1.2 2003/02/01 19:21:23 mulix Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/times.h>
#include <sys/mount.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include "tester.h"

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "module/common/module_interface.h"

static unsigned int rule_id = 0;
extern sct_ctrl_t ctrl;
extern char* progname;

static int getgroups_test(void)
{
	getgroups(0, NULL); 
	return 1; 
}

static int rlimit_test(void)
{
	struct rlimit rr; 
	getrlimit(RLIMIT_CPU, &rr);
	setrlimit(RLIMIT_CPU, &rr); 
	return 1; 
}

#ifdef TEST_ME_HARDER
static int hostname_test(void)
{
	sethostname(gethostname(), strlen(gethostname())); 
	return 1; 
}
#endif /* TEST_ME_HARDER */ 

static int getrusage_test(void)
{
	struct rusage r; 
	getrusage(RUSAGE_SELF, &r); 
	return 1; 
}

static int sigpending_test(void)
{
	sigpending(NULL); 
	return 1; 
}

static int sigsuspend_test(void)
{
	sigsuspend(NULL); 
	return 1; 
}

static int sigaction_test(void)
{
	sigaction(0xFFFF, NULL, NULL); 
	return 1; 
}

static int setsid_test(void)
{
	setsid(); 
	return 1; 
}

static int getpgrp_test(void)
{
	getpgrp(); 
	return 1; 
}

static int getppid_test(void)
{
	getppid(); 
	return 1; 
}

static int setuid_test(void)
{
	setuid(getuid()); 
	return 1; 
}

static int mount_test(void)
{
	/* don't care that it fails */ 
	mount("./foo", "./foo", "ext2", 0, NULL); 
	umount("./foo"); 
	return 1;
}

static int lchown_test(void)
{
	/* we dont care that it fails */ 
	lchown("./foo", -1, -1); 
	return 1; 
}

static int setgid_test(void)
{
	setgid(getgid()); 
	return 1; 
}

static int signal_test(void)
{
	signal(SIGURG, SIG_DFL); 
	return 1; 
}

static int geteuid_test(void)
{
	geteuid(); 
	return 1; 
}

static int getegid_test(void)
{
	getegid(); 
	return 1; 
}

#ifdef TEST_ME_HARDER
static int acct_test(void)
{
	acct(NULL); 
	return 1; 
}
#endif /* TEST_ME_HARDER */ 

static int setpgid_test(void)
{
	setpgid(getpid(), getpgrp()); 
	return 1; 
}

static int umask_test(void)
{
	umask(022); 
	return 1; 
}

static int pipe_test(void)
{
	int fd[2]; 
	if (!pipe(fd)) {
		close(fd[0]); 
		close(fd[1]); 
	}
	return 1; 
}

static int times_test(void)
{
	struct tms buf; 
	times(&buf); 
	return 1; 
}

static int sync_test(void)
{
	sync(); 
	return 1; 
}

static int exit_test(void)
{
	int rc = fork(); 
	if (rc == -1)
		perror("fork"); 
	if (rc) /* father */ 
		return 1; 
    
	/* son */ 
	exit(7); 
}

static int open_test(void)
{
	int fd = open("tmp",O_CREAT | O_EXCL,S_IWUSR);
	if (fd != -1) {
		/* dangerous! tmp might be pointing to something else now */ 
		/* is there a way to do open(); unlink(); atomically? */ 
		unlink("tmp"); 
		close(fd);
	}
	return 1;
}

static int read_write_test(void)
{
	char buf[32]; 
	int fd = open("tmp", O_CREAT, S_IWUSR); 
	if (fd == -1)
		return 1; 	
	unlink("tmp"); 
	memset(buf, 'A', sizeof(buf)); 
	write(fd, buf, sizeof(buf)); 
	read(fd, buf, sizeof(buf)); 
	close(fd); 
	return 1; 
}

#if 0 /* waitpid and creat are mapped to something else? */
static int waitpid_test(void)
{
	int rc = fork(); 
	if (rc == -1) {
		perror("fork"); 
		return 0; 
	}
	if (rc == 0) /* child */
		exit(0); 
    
	else /* father */
		waitpid(rc, NULL, 0);

	return 1; 
}

static int creat_test(void)
{
	int fd = creat("tmp", S_IWUSR); /* glibc actually calls open here */
	if (fd != -1) {
		unlink("tmp"); 
		close(fd); 
	}
	return 1; 
}
#endif 

static int unlink_test(void)
{
	(void)link(progname, "tmplink2");
	(void)unlink("tmplink2");
	return 1;
}

static int link_test(void)
{
	(void)link(progname, "tmplink");
	unlink("tmplink"); 
	return 1;
}
    
static int chdir_test(void) 
{
	static char cwd[1024];
	getcwd(cwd, sizeof(cwd));

	if (chdir("tmpdir") == 0)
		(void)chdir(cwd);
    
	return 1; 
}

static int chmod_test(void)
{
	(void) chmod("tmplink", S_IXUSR);
	return 1;
}

static int access_test(void)
{
	(void)access(progname, X_OK);
	return 1;
}

static int mkdir_test(void)
{
	(void)mkdir("tmpdir", S_IRWXU);
	(void)rmdir("tmpdir");
	(void)mkdir("tmpdir", S_IRWXU);
	(void)rmdir("tmpdir");
	return 1;
}

static int rmdir_test(void)
{
	const char location[] = "tmpdir";
	(void)rmdir(location);
	return 1;
}

static int getpid_test(void)
{
	(void)getpid();
	return 1;
}

static int chroot_test(void)
{
	(void)chroot("/");
	return 1; 
}

static int rename_test(void)
{
	(void)creat("tmprename", S_IRUSR | S_IWUSR);
	(void)rename("tmprename","tmp");
	(void)unlink("tmp"); 
	return 1;
}

static int nice_test(void)
{
	(void)nice(1);
	return 1;
}           

static int dup_test(void)            
{
	int fd = dup(1);
	if (fd != -1)
		close(fd);

	return 1;
}

static int dup2_test(void)
{
	int fd = fileno(stdout);
	if ((fd = dup2(fd, fileno(stderr))) != -1) {
		close(fd);
	}
	return 1; 
}



static int symlink_test(void)
{
	(void)symlink(progname, "tmpsymlink");
	unlink("tmpsymlink"); 
	return 1;
}

static int getpriority_test(void)
{
	(void)getpriority(PRIO_PROCESS, getpid());
	return 1;
}

static int setpriority_test(void)
{
	(void)setpriority(PRIO_PROCESS,getpid(),1);
	return 1;
}

static int lseek_test(void)
{
	int fd;
	if ((fd = open("tmp", O_CREAT, S_IWUSR)) != -1){
		unlink("tmp"); 
		(void)lseek(fd,3,SEEK_CUR);
		close(fd);
	}
	return 1;
}

static int socket_test(void)
{
	int fd;
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) != -1){
		close(fd);
	}
	return 1;
}

static int bind_test(void)
{
	int fd;
	struct sockaddr_in sa;
	struct hostent* hen;

	hen = gethostbyname("localhost");
	if (!hen)
		return 0;

	/* initiate machine's Internet address structure */
	/* first clear out the struct, to avoid garbage  */
	memset(&sa, 0, sizeof(sa));
	/* Using Internet address family */
	sa.sin_family = AF_INET;
	/* copy port number in network byte order */
	sa.sin_port = htons(35775);
	/* copy IP address into address struct */
	memcpy(&sa.sin_addr.s_addr, hen->h_addr_list[0], hen->h_length);

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) != -1){
		bind(fd, (struct sockaddr *)&sa, sizeof(sa));
		close(fd);
	}
	return 1;
}

static int listen_test(void)
{
	int fd;
	struct sockaddr_in sa;
	struct hostent* hen;

	hen = gethostbyname("localhost");
	if (!hen)
		return 0;

	/* initiate machine's Internet address structure */
	/* first clear out the struct, to avoid garbage  */
	memset(&sa, 0, sizeof(sa));
	/* Using Internet address family */
	sa.sin_family = AF_INET;
	/* copy port number in network byte order */
	sa.sin_port = htons(35776);
	/* copy IP address into address struct */
	memcpy(&sa.sin_addr.s_addr, hen->h_addr_list[0], hen->h_length);

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) != -1){
		if (bind(fd, (struct sockaddr *)&sa, sizeof(sa)) == 0){
			listen(fd, 5);
		}
		close(fd);
	}
	return 1;
}

#define SOCKET_CONNECT_ACCEPT_PORT 35778

static int accept_test(void)
{
	int fd;
	struct sockaddr_in sa;
	fd_set accept_set;

	/* initiate machine's Internet address structure */
	/* first clear out the struct, to avoid garbage  */
	memset(&sa, 0, sizeof(sa));
	/* Using Internet address family */
	sa.sin_family = AF_INET;
	/* copy port number in network byte order */
	sa.sin_port = htons(SOCKET_CONNECT_ACCEPT_PORT);
	/* copy IP address into address struct */
	sa.sin_addr.s_addr = INADDR_ANY;

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) != -1){
		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, NULL, 0);
		if (bind(fd, (struct sockaddr *)&sa, sizeof(sa)) == 0){
			if (listen(fd, 5) == 0){
				struct sockaddr_in client_addr;
				socklen_t client_addr_len;
				int cs;
				int rc;
				struct timeval timeout;

				memset(&client_addr, 0, sizeof(client_addr));
				client_addr_len = sizeof(client_addr);

				/* check if we've got a new pending connection */
				FD_ZERO(&accept_set);
				FD_SET(fd, &accept_set);
				timeout.tv_sec = 10;
				timeout.tv_usec = 0;
				rc = select(fd+1, &accept_set, NULL, NULL, &timeout);
				if (rc == 1 && FD_ISSET(fd, &accept_set)) {
					cs = accept(fd, (struct sockaddr *)&client_addr, &client_addr_len);
					if (cs >= 0)
						close(cs);
				}
			}
		}
		close(fd);
	}
	return 1;
}

static int connect_test(void)
{
	int fd;
	struct sockaddr_in sa;

	/* initiate machine's Internet address structure */
	/* first clear out the struct, to avoid garbage  */
	memset(&sa, 0, sizeof(sa));
	/* Using Internet address family */
	sa.sin_family = AF_INET;
	/* copy port number in network byte order */
	sa.sin_port = htons(rand() % 2 == 0 ? SOCKET_CONNECT_ACCEPT_PORT
			    : SOCKET_CONNECT_ACCEPT_PORT+1);
	/* copy IP address into address struct */
	inet_aton("127.0.0.1", &sa.sin_addr);

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) != -1){
		int rc;

		rc = connect(fd, (struct sockaddr *)&sa, sizeof(sa));
		if (rc == -1) {
			/*perror("connect_test: connect:");*/
		}
		shutdown(fd, 1);
		close(fd);
	}
	else {
		/*perror("connect_test: socket:");*/
	}
	return 1;
}

static void
register_module_passwd_rule(unsigned int rid)
{
	sct_filter_node_t* f0 = sct_make_var_indexed_val(VT_PARAM, 0, DT_STRING);
	sct_filter_node_t* f1 = sct_make_string_val("passwd");
	sct_filter_node_t* f2 = sct_make_pattern_match(f0, f1);
	sct_action_t* a1 = sct_make_action_log(4, NULL);
    
	if (sct_ctrl_add_rule(ctrl, rid, RW_BEFORE, 10, f2, a1, 1) < 0)
		die ("failed to upload rule '%d'\n", rid);
}

int
test_simple_syscall_usage(void)
{
	unsigned int num_tests = 100;
	int rc = fork();
	if (rc == -1) {
		perror("test_simple_syscall_usage: fork failed:");
		die("'%s': fork failed\n", __FUNCTION__);
	}

	/* both father and child execute this */ 
	while (--num_tests){ 
		int delay = 1+(int) (10.0 * rand() /(RAND_MAX+1.0));
		creat("passwd", S_IRUSR | S_IWUSR);
		usleep(delay);
		remove("passwd");
	}
    
	if (rc == 0) /* child */
		_exit(0);

	return 1;
}

int
test_settimeofday_usage(void)
{
	unsigned int num_tests = 100;
	int rc;
	int rid = ++rule_id;

	sct_filter_node_t* timeval_match = sct_make_bool_val(1);
	if (sct_ctrl_add_rule(ctrl, rid, RW_BEFORE, 79, timeval_match, 
			      sct_make_action_log(4, NULL), 1) < 0)
		die("failed to upload rule '%d'\n", rid);

	rc = fork();
	if (rc == -1) {
		perror("test_settimeofday_usage: fork failed:");
		die("'%s': fork failed\n", __FUNCTION__);
	}

	/* both father and child execute this */ 
	while (--num_tests){ 
		struct timeval tv = {10, 1000 - num_tests};
		int delay = 1+(int) (10.0 * rand() /(RAND_MAX+1.0));
		settimeofday(&tv, NULL);
		usleep(delay);
	}
    
	if (rc == 0) /* child */
		_exit(0);

	if (!sct_ctrl_del_rule(ctrl, rid, RW_BEFORE, 79))
		die("failed to delete rule '%d'\n", rid);
    
	return 1;
}


int 
test_rule_register_deregister(void)
{
	unsigned int num_tests = 100;
	int rc;

	if (!sct_ctrl_del_all_rules(ctrl))
		die("'%s': failed to delete all module rules\n");

	rc = fork();
	if (rc == -1) {
		perror("test_rule_register_deregister: fork failed:");
		die("'%s': fork failed\n", __FUNCTION__);
	}

	if (rc == 0){ /* child code */
		++rule_id;
	}
    
	/* both father and child execute this */ 
	while (--num_tests){ 
		int delay = 1+(int) (5.0 * rand() /(RAND_MAX+1.0));
		unsigned int rid = (rule_id += 2);
		register_module_passwd_rule(rid);
		usleep(delay);
		if (!sct_ctrl_del_rule(ctrl, rid, RW_BEFORE, 10))
			die ("'%s': failed to delete rule\n", __FUNCTION__);
	}
    
	if (rc == 0) /* child */
		_exit(0);

	printf("\n"); fflush(stdout); 

	return 1;    
}

static int
upload_rules(sct_ctrl_t ctrl, int syscall_id, sct_filter_node_t* f,
	     sct_var_attr_name_t attr_name, int attr_val, int param_index)
{
	int rule_count = 0; 

	sct_action_t* a1 = sct_make_action_log(4, NULL);
	sct_action_t* a2; 
	a2 = sct_make_action_log(4, "after syscall: %pid[%comm]: "
				 "%sid_%sname(%params) = %retval (rule %ruleid)\n");

	if (a1 && a2) {
		if (attr_name != VAN_LAST_TYPE) {
			set_action_var_indexed_attr_set_command(a1,
								VT_PARAM,
								param_index,
								attr_name,
								attr_val);
			set_action_var_indexed_attr_set_command(a2,
								VT_PARAM,
								param_index,
								attr_name,
								attr_val);
		}
		if (sct_ctrl_add_rule(ctrl, ++rule_id, RW_BEFORE, syscall_id, f, a1, 1) < 0)
			die ("failed to upload rule '%d'\n", rule_id);

		if (sct_ctrl_add_rule(ctrl, ++rule_id, RW_AFTER, syscall_id, f, a2, 1) < 0)
			die ("failed to upload rule '%d'\n", rule_id);

		rule_count += 2; 
	
	}
	sct_free_action(a1);    
	sct_free_action(a2);

	return rule_count; 
}

int test_all_syscalls_register_rules(void)
{
	int rule_count = 0; 

	sct_filter_node_t* p0 = sct_make_var_indexed_val(VT_PARAM, 0, DT_STRING);
	sct_filter_node_t* p1 = sct_make_var_indexed_val(VT_PARAM, 1, DT_STRING);

	sct_filter_node_t* prog = sct_make_string_val(progname);
	sct_filter_node_t* progmatch0 = sct_make_pattern_match(p0, prog);
	/* sct_filter_node_t* progmatch1 = sct_make_pattern_match(p1, prog); */

	sct_filter_node_t* tmp = sct_make_string_val("tmp");
	sct_filter_node_t* tmpmatch0 = sct_make_pattern_match(p0, tmp);
	sct_filter_node_t* tmpmatch1 = sct_make_pattern_match(p1, tmp);

	sct_filter_node_t* test = sct_make_string_val("tests");
	sct_filter_node_t* testmatch0 = sct_make_pattern_match(p0, test);

	sct_filter_node_t* truematch = sct_make_bool_val(1);

	sct_filter_node_t* intmatch0 = sct_make_var_indexed_val(VT_PARAM, 0, DT_INT);
	sct_filter_node_t* intmatch1 = sct_make_int_val(1);
	sct_filter_node_t* intmatch2 = sct_make_equal(intmatch0, intmatch1);

	sct_filter_node_t* matchnum0 = sct_make_var_indexed_val(VT_PARAM, 1, DT_INT);
	sct_filter_node_t* matchnum1 = sct_make_int_val(3);
	sct_filter_node_t* matchnum2 = sct_make_equal(matchnum0, matchnum1);
	sct_filter_node_t* true_match =sct_make_bool_val(1);

	sct_filter_node_t* timeval_match = sct_make_bool_val(1);

	sct_filter_node_t* match_num_2 = sct_make_int_val(2);
	sct_filter_node_t* match_var_sa = sct_make_var_indexed_field_val(VT_PARAM, 1, 0, DT_USHORT, -1);
	sct_filter_node_t* sa_inet_match = sct_make_equal(match_var_sa, match_num_2);
	sct_filter_node_t* match_sin_port_num = sct_make_ushort_val(htons(SOCKET_CONNECT_ACCEPT_PORT));
	sct_filter_node_t* match_var_sin_port = 
		sct_make_var_indexed_field_val(VT_PARAM, 1, 1, DT_USHORT, 4 /* struct sockaddr_in */);
	sct_filter_node_t* sin_port_match = sct_make_equal(match_sin_port_num, match_var_sin_port);
	sct_filter_node_t* connect_match = sct_make_logical_and(sa_inet_match, sin_port_match);

	sct_filter_node_t* smallbuf1 = 
		sct_make_less_then(sct_make_var_indexed_val(VT_PARAM, 2, DT_UINT), 
				   sct_make_uint_val(2)); 

	sct_filter_node_t* smallbuf2 = 
		sct_make_less_then(sct_make_var_indexed_val(VT_PARAM, 2, DT_UINT), 
				   sct_make_uint_val(2)); 

	sct_filter_node_t* noklogdmatch = sct_make_logical_not(
		sct_make_pattern_match(sct_make_string_val("klogd"), 
				       sct_make_var_val(VT_PROC_COMM, DT_STRING))); 

	sct_filter_node_t* noklogdsmallbuf = sct_make_logical_and(noklogdmatch, smallbuf1); 

	/* special handling for exit */ 
	sct_action_t* a_exit = sct_make_action_log(4, NULL);
	if (sct_ctrl_add_rule(ctrl, ++rule_id, RW_BEFORE, 1, truematch, a_exit, 1) < 0)
		die ("failed to upload rule '%d'\n", rule_id);

	rule_count++; 
	sct_free_action(a_exit);

	rule_count += upload_rules(ctrl, 2, truematch, VAN_LAST_TYPE, -1, -1); 
	rule_count += upload_rules(ctrl, 3, smallbuf2, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 4, noklogdsmallbuf, VAN_LAST_TYPE, -1, -1); 
	rule_count += upload_rules(ctrl, 5, tmpmatch0, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 6, truematch, VAN_LAST_TYPE, -1, -1); 
	rule_count += upload_rules(ctrl, 7, truematch, VAN_LAST_TYPE, -1, -1); 
	rule_count += upload_rules(ctrl, 8, tmpmatch0, VAN_LAST_TYPE, -1, -1); 
	rule_count += upload_rules(ctrl, 9, tmpmatch1, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 10, tmpmatch0, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 12, testmatch0, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 15, tmpmatch0, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 19, matchnum2, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 20, truematch, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 33, progmatch0, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 34, intmatch2, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 36, truematch, VAN_LAST_TYPE, -1, -1);


	/* TODO: add 'kill' rule */

	rule_count += upload_rules(ctrl, 38, tmpmatch1, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 39, tmpmatch0, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 40, tmpmatch0, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 41, intmatch2, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 42, truematch, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 43, truematch, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 61, tmpmatch0, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 63, intmatch2, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 79, timeval_match, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 83, tmpmatch1, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 96, intmatch2, VAN_LAST_TYPE, -1, -1);
	rule_count += upload_rules(ctrl, 97, intmatch2, VAN_LAST_TYPE, -1, -1);

	/* test 'socket' call. */
	rule_count += upload_rules(ctrl, 65638, true_match, VAN_LAST_TYPE, -1, -1);

	/* test 'bind' call. */
	rule_count += upload_rules(ctrl, 131174, true_match, VAN_LAST_TYPE, -1, -1);

	/* test 'listen' call. */
	rule_count += upload_rules(ctrl, 262246, true_match, VAN_LAST_TYPE, -1, -1);

	/* test 'accept' call. */
	rule_count += upload_rules(ctrl, 327782, true_match, VAN_LAST_TYPE, -1, -1);

	/* test 'connect' call. note: the order of these rules is crucial. */
	rule_count += upload_rules(ctrl, 196710, connect_match, VAN_DYN_TYPE, 4, 1);
	rule_count += upload_rules(ctrl, 196710, noklogdmatch, VAN_LAST_TYPE, -1, -1);

	/* test 'shutdown' call. */
	rule_count += upload_rules(ctrl, 852070, true_match, VAN_LAST_TYPE, -1, -1);

	sleep(3);

	return rule_count; 
}


static void 
all_syscalls_deregister_rules(void)
{
	if (!sct_ctrl_del_all_rules(ctrl))
		die("failed to delete all module rules");
}

static void cleanup(void)
{
	(void)unlink("tmplink");
	(void)unlink("tmpsymlink");
	(void)unlink("tmplink");
	(void)unlink("tmplink2");
	(void)unlink("tmprename"); 
	(void)unlink("tmp");
}    

int
test_all_syscalls_usage(void)
{
	typedef struct syscall_struct_t {
		unsigned int syscall_num;
		int (*ptest)(void);
	} syscall_struct;

	syscall_struct syscall_test[] =
		{
			{1, exit_test},
			{3, read_write_test}, 
			{4, read_write_test},
			{5, open_test}, 
			{6, open_test}, /* open and close use the same test */ 
/*    	{7, waitpid_test},*/ /* doesnt appear to get called. need to investigate */
/*	{8, creat_test},  */ /* glibc actually calls open */ 
			{9, link_test}, 
			{10, unlink_test},
			{12, chdir_test},
			{15, chmod_test}, 
			{16, lchown_test}, 
			{19, lseek_test}, 
			{20, getpid_test},
			{21, mount_test}, 
			{22, mount_test}, 
			{23, setuid_test}, 
			{24, setuid_test}, 
			{33, access_test}, 
			{34, nice_test},
			{36, sync_test}, 
			{38, rename_test}, 
			{39, mkdir_test}, 
			{40, rmdir_test}, 
			{41, dup_test},
			{43, pipe_test},
			{44, times_test}, 
			{46, setgid_test}, 
			{47, setgid_test}, 
			{48, signal_test}, 
			{49, geteuid_test}, 
			{50, getegid_test}, 
#ifdef TEST_ME_HARDER
			{51, acct_test}, 
#endif /* TEST_ME_HARDER */ 
			{52, mount_test}, 
			{57, setpgid_test}, 
			{60, umask_test}, 
			{61, chroot_test}, 
			{63, dup2_test}, 
			{64, getppid_test}, 
			{65, getpgrp_test}, 
			{66, setsid_test}, 
			{67, sigaction_test}, 
			{72, sigsuspend_test}, 
			{73, sigpending_test}, 
#ifdef TEST_ME_HARDER
			{74, hostname_test}, 
#endif /* TEST_ME_HARDER */ 
			{75, rlimit_test}, 
			{76, rlimit_test}, 
			{77, getrusage_test}, 
			{80, getgroups_test}, 
			{83, symlink_test}, 
			{96, getpriority_test}, 
			{97, setpriority_test},
			{65638, socket_test}, 
			{131174, bind_test}, 
			{262246, listen_test},
			{327782, accept_test}, 
			{196710, connect_test}
		};

	const unsigned int num_syscalls = sizeof(syscall_test)/sizeof(syscall_struct) ;
	unsigned int times_to_run = 15; 
	unsigned int i;
	int rc = 0;  /* silence the compiler */ 

	test_all_syscalls_register_rules();

	for (i = 0; i < num_syscalls; ++i){
		rc = fork();
		if (rc == -1) {
			perror("test_all_syscalls_usage: fork failed:");
			die("'%s': fork failed\n", __FUNCTION__);
		}
		if (rc == 0)
			break;
	}
	/* parent code */
	if (rc != 0){
		/* we have as many children as there are syscalls
		   a relic from the days when each child perform 
		   only one test */
		unsigned int children = num_syscalls;
		while (children--)
			(void)wait(NULL);
	
	} else { /* child code */
        
		while (--times_to_run){
			int delay = 1+(int) (10.0 * rand() /(RAND_MAX+1.0));
			usleep(delay);

			for (i = 0; i < num_syscalls; ++i) 
			{
				rc = syscall_test[i].ptest();
			}
			if (times_to_run % 10 == 0) {
				printf("."); fflush(stdout);
			}
		}	
		fprintf(stderr, "\nchild '%d' finished\n", i);
		_exit(0); 
	}
    
	printf("\n"); fflush(stdout); 

	all_syscalls_deregister_rules();

	cleanup(); 

	return 1;
}

/* test_kill_suspend_current: tests killing or suspending of the current
 * process.
 * The child's process creates a rule to match its own pid, and a 
 * corresponding action which kills/suspends itself (using pid -1).
 * If the child wasn't killed/suspended after issuing the system call,
 * it notifies the parent via a pipe that the test has failed. */
int test_kill_suspend_current(int is_kill)
{
	int filedes[2];
	int success = 0;
	int rc;

        /* prepare the pipe for the parent-child communication */
	if (pipe(filedes) < 0) {
		perror("test_kill_suspend_process: pipe failed:");
		die("'%s': pipe failed\n", __FUNCTION__);
	}
	if (fcntl (filedes[0], F_SETFL, O_NONBLOCK) < 0) {
		perror("test_kill_suspend_process: fcntl failed:");
		die("'%s': fcntl failed\n", __FUNCTION__);
	}

	rc = fork();
	if (rc == -1) {
		perror("test_kill_suspend_process: fork failed:");
		die("'%s': fork failed\n", __FUNCTION__);
	}

	/* Child code */
	if (rc == 0) {
		sct_action_t* a = NULL;

                /* Create the rule */
		sct_filter_node_t* pid_val = sct_make_int_val( getpid() );
		sct_filter_node_t* pid = sct_make_var_val(VT_PROC_PID,DT_LONG);
		sct_filter_node_t* pid_match = sct_make_equal(pid_val, pid);

		/* Create the action which kills/suspends 'current' (pid -1) */
		if (is_kill)
			a = sct_make_action_kill_process(-1);
		else
			a = sct_make_action_suspend_process(-1);

		if (sct_ctrl_add_rule(ctrl, -1, RW_BEFORE, 20 /* getpid() */,
				 pid_match, a, 1) < 0)
			die ("test_kill_suspend_process: failed to upload rule\n");

		getpid(); /* this should kill/suspend me */

                /* if we got this far - the test failed. notify the parent. */
		while (write(filedes[1],"f",1) < 1)
		{
                    /* if failed because of an interrupting signal - try again */
			if (errno != EINTR) {
				perror ("test_kill_suspend_process: unable to write to pipe\n");
				kill(getppid(),SIGTERM);
                                /* we have to kill the parent, otherwise it 
                                   will think the test has succeeded */
			}
		}

		while (1)
			pause(); /* Wait for the parent to kill me */
	}
	else { /* Parent code */
		int status,rd;
		char buf;
		pid_t p;

		while (1) {
			/* Wait for the child to either notify of failure,
			   or be killed/suspended */
			   
			rd = read(filedes[0],&buf,1);
			if ((rd < 0) && (errno != EAGAIN) && (errno != EINTR)) {
				die("parent failed to read from pipe\n");
			}
			else if (rd > 0) { /* test failed */
				success = 0;
				break;
			}

			p = waitpid(rc, &status, WNOHANG | WUNTRACED);
			if (p > 0) { /* child killed/suspended - test succeeded */
				success = 1;
				break;
			}
		}
	}

	kill(rc,SIGKILL); /* if the child is  not dead, kill it */

	if (!sct_ctrl_del_all_rules(ctrl))
		die("failed to delete all module rules\n");

	return success;
}

/* test_kill_suspend_other: tests killing/suspending a process, other than
 * the current one.
 * The parent creates a rule which will match its own pid, with a 
 * corresponding action which kills/suspends the child.
 * Then a single char is sent from the parent to the child via a pipe. If the 
 * child got that char, it means it wasn't killed/suspended, and the child 
 * sends a single char back to notify the parent of that fact. */
int test_kill_suspend_other(int is_kill)
{
	int p_c[2];
	int c_p[2];
	int rc;
	int success = 0;

        /* prepare the pipes for the parent-child communication */
	if (pipe(p_c) < 0) {
		perror("test_kill_suspend_process: pipe failed:");
		die("'%s': pipe failed\n", __FUNCTION__);
	}
	if (pipe(c_p) < 0) {
		perror("test_kill_suspend_process: pipe failed:");
		die("'%s': pipe failed\n", __FUNCTION__);
	}
	if (fcntl (c_p[0], F_SETFL, O_NONBLOCK) < 0) {
		perror("test_kill_suspend_process: fcntl failed:");
		die("'%s': fcntl failed\n", __FUNCTION__);
	}

	rc = fork();
	if (rc == -1) {
		perror("test_kill_suspend_process: fork failed:");
		die("'%s': fork failed\n", __FUNCTION__);
	}

        /* Child code */
	if (rc == 0) {
		char buf;

		/* Wait to be killed/suspended, or for the test to fail */
		while (read(p_c[0],&buf,1) < 0) {
			if (errno != EINTR)
			{
				perror ("child failed to read from pipe\n");
				break;
			}
		}
					
		/* If we got so far, the test failed. */
		while (write(c_p[1],"f",1) < 0) {
			if (errno != EINTR)
			{
				perror ("child failed to write to pipe\n");
				kill (getppid(),SIGTERM);
                                /* kill the parent, otherwise it will think 
                                   the test succeeded */
			}
		}

		while(1) 
			pause(); /* Wait to be killed. */
	}
	else { /* Parent code */
		int status;
		sct_action_t* a = NULL;
		int rd;
		pid_t p;
		char buf[2];

                /* Create a rule which will match the parent's getpid() */
		sct_filter_node_t* pid_val = sct_make_int_val( getpid() );
		sct_filter_node_t* pid = sct_make_var_val(VT_PROC_PID,DT_LONG);
		sct_filter_node_t* pid_match = sct_make_equal(pid_val, pid);

		/* Create an action which kills the child (child's pid = rc) */
		if (is_kill)
			a = sct_make_action_kill_process(rc);
		else
			a = sct_make_action_suspend_process(rc);

		if (sct_ctrl_add_rule(ctrl, -1, RW_BEFORE, 20 /* getpid() */,
				 pid_match, a, 1) < 0)
			die ("test_kill_suspend_process: failed to upload rule\n");

		getpid(); /* this should kill the child. */

		/* Notify the child it should have been killed/suspended by now */
		while (write(p_c[1],"a",1) < 0) {
			if (errno != EINTR)
				die ("parent failed to write to pipe\n");
		}


		/* wait for the child's response */
		while (1) {
			rd = read(c_p[0],&buf,1);
			if ((rd < 0) && (errno != EAGAIN) && (errno != EINTR)) {
				die("parent failed to read from pipe\n");
			}
			else if (rd > 0) { /* child replied - test failed */
				success = 0;
				break;
			}

			p = waitpid(rc, &status, WNOHANG | WUNTRACED);
			if (p > 0) { /* child killed/suspended - test succeeded */
				success = 1;
				break;
			}
		}
	}

	kill(rc,SIGKILL);  /* in case the child is not dead, kill it */

	if (!sct_ctrl_del_all_rules(ctrl))
		die("failed to delete all module rules\n");

	return success;
}

int test_kill_suspend_process(void)
{
	return test_kill_suspend_current(1) /* kill current process*/
		&& test_kill_suspend_other(1) /* kill a different process */
		&& test_kill_suspend_current(0) /* suspend current process */
		&& test_kill_suspend_other(0); /* suspend a different process */
}

int read_log(int fd, char* rd, int len)
{
	int res, pos = 0;
	while (pos < len)
	{
		res = read(fd, rd + pos, len - pos);
		if (res < 0)
			return res;
		if (res == 0)
			break;
		pos += res;
	}
	return pos;
}

int test_loguser(void)
{
	int fd1, fd2;
	char rd1[300], rd2[300];
	const char* frmt_sname = "logging. sname = %sname";
	const char* expected_sname = "logging. sname = getpid";
	const int len_sname = strlen (expected_sname);
	const char* frmt_sid = "second format. sid = %sid";
	const char* expected_sid = "second format. sid = 20";
	const int len_sid = strlen (expected_sid);
	int ret1,ret2;
	int test_num = 1;

	/* create a rule for getpid() */
	sct_filter_node_t* pid_val = sct_make_int_val( getpid() );
	sct_filter_node_t* pid = sct_make_var_val(VT_PROC_PID,DT_LONG);
	sct_filter_node_t* pid_match = sct_make_equal(pid_val, pid);
	sct_action_t* a = sct_make_action_log(4, "action format");

	if (sct_ctrl_add_rule (ctrl, -1, RW_BEFORE, 20 /* getpid() */,
			       pid_match, a, 1) < 0)
		die ("loguser_test failed to upload rule\n");

	printf("Test %d: two device files, same log format...  ", test_num++);

        /* open two file descriptors and enable device */
	fd1 = open ("/dev/sct_log", O_RDONLY);
	if (fd1 == -1)
		die ("Error openning device file 1\n");
	fd2 = open ("/dev/sct_log", O_RDONLY);
	if (fd2 == -1)
		die ("Error openning device file 2\n");
	if (ioctl (fd1, CTL_IOC_RESIZE, 300) != 0) 
		die ("Error resizing device file 1\n");
	if (ioctl (fd2, CTL_IOC_RESIZE, 300) != 0) 
		die ("Error resizing device file 2\n");
        /* flush both device files to make sure no previously logged
           stuff is accidentally caught */
	if (ioctl (fd1, CTL_IOC_FLUSH) != 0)
		die ("Error flushing device file 1\n");
	if (ioctl (fd2, CTL_IOC_FLUSH) != 0)
		die ("Error flushing device file 2\n");
        /* both files have the same log format */
	if (ioctl (fd1, CTL_IOC_BEFORE_LOG_FORMAT, frmt_sname))
		die("Error changing log format for device file 1\n");
	if (ioctl (fd2, CTL_IOC_BEFORE_LOG_FORMAT, frmt_sname))
		die("Error changing log format for device file 2\n");

	/* this should be logged to both device files */
	getpid();

	if ((ret1 = read_log (fd1, rd1, len_sname)) < 0)
		die ("Error reading from device file 1\n");
	if ((ret2 = read_log (fd2, rd2, len_sname)) < 0)
		die ("Error reading from device file 2\n");

	rd1[ret1] = rd2[ret2] = '\0';

	CHECK ((strcmp (rd1, rd2) || strcmp (rd1, expected_sname)), 0);

	printf ("passed.\n");
	printf ("Test %d: Change log format for device file 1...  ", test_num++);

	/* change log format for file 1 */
	if (ioctl (fd1, CTL_IOC_BEFORE_LOG_FORMAT, frmt_sid))
		die("Error changing log format for device file 1\n");

	/* this should be logged */
	getpid();

	if ((ret1 = read_log (fd1, rd1, len_sid)) < 0)
		die ("Error reading from device file 1\n");
	if ((ret2 = read_log (fd2, rd2, len_sname)) < 0)
		die ("Error reading from device file 2\n");

	rd1[ret1] = rd2[ret2] = '\0';

	CHECK ((strcmp (rd1, expected_sid) || strcmp (rd2, expected_sname)),0);

	printf ("passed.\n");
	printf ("Test %d: close device file 1...  ", test_num++);

	/* close fd1. next action should be logged only to fd2 */
	if (close(fd1))
		die ("Error closing device file 1\n");

	/* this should be logged only to fd2 */
	getpid();

	if ((ret2 = read_log (fd2, rd2, len_sname)) < 0)
		die ("Error reading from device file 2\n");

	rd2[ret2] = '\0';

	CHECK ((strcmp (rd2, expected_sname)), 0);

	printf("passed.\n");
	printf ("Test %d: add after log format...  ", test_num++);

	if (ioctl (fd2, CTL_IOC_AFTER_LOG_FORMAT, frmt_sid))
		die("Error changing log format for device file 2\n");

	/* add the same rule as an after rule */
	if (sct_ctrl_add_rule (ctrl, -1, RW_AFTER, 20 /* getpid() */,
			       pid_match, a, 1) < 0)
		die ("loguser_test failed to upload rule\n");

	/* this should be logged both before and after */
	getpid();

	if ((ret2 = read_log (fd2, rd2, len_sname)) < 0)
		die ("Error reading from device file 2\n");
	rd2[ret2] = '\0';
	CHECK ((strcmp (rd2, expected_sname)), 0);

	if ((ret2 = read_log (fd2, rd2, len_sid)) < 0)
		die ("Error reading from device file 2\n");

	rd2[ret2] = '\0';

	CHECK ((strcmp (rd2, expected_sid)), 0);

	printf ("passed.\n");

	if (!sct_ctrl_del_all_rules(ctrl))
		die("failed to delete all module rules\n");

	if (close(fd2))
		die ("Error closing device file 2\n");

	return 1;
}
