/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/* 
 * $Id: struct_test.c,v 1.2 2003/01/31 19:01:31 mulix Exp $
 */ 

/* test the module's struct handling mechanism */ 

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/* some ugliness to allow including kernel headers */ 
#undef printk
#define printk printf
#undef BUG
#define BUG() do { assert(0); } while (0)
#include "../../module/rules/struct_handling.h"
#undef printk

#include "sct_rules/rule_engine_public.h"
#include "sct_rules/filter_vars_public.h"

#include "tester.h"

/* big parts of this file are copied verbatim from the module. If changes */
/* are made to these functions in the module code, it is necessary to sync */ 
/* them here as well */ 

/* array of struct fields data (offsets, loggers...) - to be used for */
/* dynamic struct (fields) handling.                                  */
struct struct_data struct_type_data[2];
int struct_type_data_sz = 2;

/* calculate offsets of fields for struct 'timeval'. */
static size_t*
sct_make_struct_offsets_vec_timeval(long* num_fields, size_t* struct_size)
{
	struct timeval s;

        size_t* offsets_vec =
		(size_t*)malloc(sizeof(size_t) * 2); 
  
	if (offsets_vec) {
		offsets_vec[0] =  offsetof(struct timeval, tv_sec);
		offsets_vec[1] =  offsetof(struct timeval, tv_usec);

		*num_fields = 2;
	}

	*struct_size = sizeof(s);

	return offsets_vec;
}

/* calculate types of fields for struct 'timeval'. */
static sct_data_type_t*
sct_make_struct_types_vec_timeval(long* num_fields)
{
        sct_data_type_t* types_vec =
		(sct_data_type_t*)malloc(sizeof(sct_data_type_t) * 2); 
  
	if (types_vec) {
		types_vec[0] = DT_LONG;
		types_vec[1] = DT_LONG;

	*num_fields = 2;
	}

	return types_vec;
}
/* log fields of struct 'timeval' into the given buffer. */
static int sct_log_struct_fields_timeval(void* void_s, struct sct_membuf* mbuf, 
					 unsigned int max_rec_len)
{
	int rc = 0; 
	struct timeval* s = (struct timeval*)void_s;
  
	if (s) {
		rc = sct_tracker_stringify_params(mbuf, max_rec_len,
						  "T{l l}", 
						  "timeval", 
						  s->tv_sec, s->tv_usec);
	}
	else {
		rc = sct_tracker_stringify_params(mbuf, max_rec_len, "T{T}", 
						  "timeval", 
                                                  "(nil)")	     ;
	}					  
	
	return rc;
}


/* calculate offsets of fields for struct 'timezone'. */
static size_t*
sct_make_struct_offsets_vec_timezone(long* num_fields, size_t* struct_size)
{
	struct timezone s;

        size_t* offsets_vec =
		(size_t*)malloc(sizeof(size_t) * 2); 
  
	if (offsets_vec) {
		offsets_vec[0] =  offsetof(struct timezone, tz_minuteswest);
		offsets_vec[1] =  offsetof(struct timezone, tz_dsttime);

		*num_fields = 2;
	}

	*struct_size = sizeof(s);

	return offsets_vec;
}

/* calculate types of fields for struct 'timezone'. */
static sct_data_type_t*
sct_make_struct_types_vec_timezone(long* num_fields)
{
        sct_data_type_t* types_vec =
	    (sct_data_type_t*)malloc(sizeof(sct_data_type_t) * 2); 
  
	if (types_vec) {
		types_vec[0] = DT_INT;
		types_vec[1] = DT_INT;

	*num_fields = 2;
	}

	return types_vec;
}
/* log fields of struct 'timezone' into the given buffer. */
static int sct_log_struct_fields_timezone(void* void_s, struct sct_membuf* mbuf, 
					  unsigned int max_rec_len)
{
	int rc = 0;
	struct timezone* s = (struct timezone*)void_s;
  
	if (s) {
		rc = sct_tracker_stringify_params(mbuf, max_rec_len, 
						  "T{i i}", 
						  "timezone", 
						  s->tz_minuteswest, s->tz_dsttime);
	}
	else {
		rc = sct_tracker_stringify_params(mbuf, max_rec_len, 
						  "T{T}", 
						  "timezone", 
                                                  "(nil)")	     ;
	}					  
	
	return rc;
}

/* function used to perform logging of structs. */
int struct_logger_callback(char spec_char,
			   struct sct_membuf* mbuf, 
			   unsigned int max_rec_len, 
			   void* s,
			   void* s_data,
			   void* cb_data)
{
	int rc = 0;
	int dyn_struct_type;
	struct struct_logger_data* pdata = (struct struct_logger_data*)s_data;
	struct struct_data* struct_data =
		(pdata ? get_struct_data(pdata->struct_type) : NULL);
	
	if (get_var_indexed_attr(VT_PARAM, pdata->param_index,
				 VAN_DYN_TYPE, &dyn_struct_type)) {
		struct struct_data* dyn_struct_data =
			get_struct_data(dyn_struct_type);
		if (dyn_struct_data) {
			if (dyn_struct_data->struct_size > struct_data->struct_size) {
				printf("sct: struct_logger_callback: bad cast "
				       "from struct type '%d' (size %d) to "
				       "struct type '%d' (size %d). using "
				       "static struct type.\n",
				       pdata->struct_type,
				       struct_data->struct_size,
				       dyn_struct_type,
				       dyn_struct_data->struct_size);
			}
			else {
				struct_data = dyn_struct_data;
			}
		}
	}
	
	if (struct_data) {
		rc = (*struct_data->logger)(s, mbuf, max_rec_len);
	}
	
	return rc;
}

void init_struct_data(void)
{
	long num_fields;
	int struct_size;
        size_t* offsets_vec;
	sct_data_type_t* types_vec;

	offsets_vec = sct_make_struct_offsets_vec_timeval(&num_fields, &struct_size);
	types_vec = sct_make_struct_types_vec_timeval(&num_fields);
	struct_type_data[0].offsets_vec = offsets_vec;
	struct_type_data[0].types_vec = types_vec;
	struct_type_data[0].num_fields = num_fields;
	struct_type_data[0].struct_size = struct_size;
	struct_type_data[0].logger = sct_log_struct_fields_timeval;
	offsets_vec = sct_make_struct_offsets_vec_timezone(&num_fields, &struct_size);
	types_vec = sct_make_struct_types_vec_timezone(&num_fields);
	struct_type_data[1].offsets_vec = offsets_vec;
	struct_type_data[1].types_vec = types_vec;
	struct_type_data[1].num_fields = num_fields;
	struct_type_data[1].struct_size = struct_size;
	struct_type_data[1].logger = sct_log_struct_fields_timezone;

	set_logger_callback('S', struct_logger_callback, NULL);
}

void cleanup_struct_data(void)
{
	free(struct_type_data[0].offsets_vec); 
	free(struct_type_data[0].types_vec); 

	free(struct_type_data[1].offsets_vec); 
	free(struct_type_data[1].types_vec); 
}


sct_filter_node_t* global_param_callback(sct_var_type_t type, int param_index, 
					 sct_data_type_t dtype, void *data) 
{
	sct_filter_node_t *f = NULL;
    
	switch (dtype) {
	case DT_CHAR:
		f = sct_make_char_val(*((char*)data));
		break;
	case DT_BOOL:
	case DT_INT:
		f = sct_make_int_val(*((int*)data));
		break;
	case DT_UINT:
		f = sct_make_uint_val(*((unsigned int*)data));
		break;
	case DT_SHORT:
		f = sct_make_short_val(*((short*)data));
		break;
	case DT_USHORT:
		f = sct_make_ushort_val(*((unsigned short*)data));
		break;
	case DT_LONG:
		f = sct_make_long_val(*((long*)data));
		break;
	case DT_ULONG:
		f = sct_make_ulong_val(*((unsigned long*)data));
		break;
	case DT_FLOAT:
		f = sct_make_ulong_val(*((float*)data));
		break;
	case DT_STRING: 
	{
		char *string = strdup(data); 
		f = sct_make_string_val_without_copy(string);
		break;
	}
	case DT_BUFFER:
	{
		struct sct_aggregate_data2* d = (struct sct_aggregate_data2*)data; 
		size_t count = (size_t)d->v2; 
		char* buf = malloc(count); 
		if (!buf)
			return NULL; 

		memcpy(buf, d->v1, count); 
		f = sct_make_buffer_val_without_copy(buf, count); 
		break; 
	}
		
	default:
		f = NULL;
		break;
	}
    
	return f;
}

/* get a filter node with a struct's field parameter value */
sct_filter_node_t* global_struct_field_param_callback(sct_var_type_t type,
						      int param_index,
						      int field_num,
						      sct_data_type_t dtype,
						      void *data,
						      int struct_type)
{
	sct_filter_node_t* field = NULL;
	struct pdata {
		char* data;
		int struct_type;
	} *p_data;

	/* the type according to which the struct should be accessed. */
	int real_struct_type = struct_type;
	struct struct_data* struct_data = NULL;
	struct struct_data* static_struct_data = NULL;
	int do_type_cast = 0;

    	p_data = (struct pdata*)data;
    	if (struct_type == -1) {
		real_struct_type = p_data->struct_type;
    	} else {
		do_type_cast = 1;
		static_struct_data = get_struct_data(p_data->struct_type);
		if (!static_struct_data) {
	    		printf("sct: global_struct_field_param_callback: "
			       "internal error: struct type '%d' unknown.\n", 
			       p_data->struct_type);
			goto global_struct_field_param_callback_fail;
		}
	}

	struct_data = get_struct_data(real_struct_type);
	if (!struct_data) {
	    	printf("sct: global_struct_field_param_callback: error: "
		       "struct type '%d' unknown.\n", real_struct_type);
		goto global_struct_field_param_callback_fail;
	}
	if (do_type_cast && struct_data->struct_size > static_struct_data->struct_size) {
	    	printf("sct: global_struct_field_param_callback: bad cast "
		       "from struct type '%d' (size %d) to struct type '%d' (size %d).\n",
		       p_data->struct_type,
		       static_struct_data->struct_size,
		       real_struct_type,
		       struct_data->struct_size);
		goto global_struct_field_param_callback_fail;
	}
	/* offsets_vec is a size_t, so cast to long to get the -1 properly */ 
	if ((long)struct_data->offsets_vec[field_num] >= 0) {
		void* loc = p_data->data + struct_data->offsets_vec[field_num];
		sct_data_type_t field_dtype = struct_data->types_vec[field_num];

		field = global_param_callback(type, param_index, field_dtype, loc);
	}
	else {
	    printf("sct: global_struct_field_param_callback: offset for field "
		   "'%d' (%ld) not supported (unmatchable field?)\n",
		   field_num, (unsigned long)struct_data->offsets_vec[field_num]);
	}
	
 global_struct_field_param_callback_fail:
	return field;
}

static void test_struct_syscall(sct_rule_list_t* rules, sct_action_t* act_to_match)
{
	sct_action_t* action; 
	int rule_id = -1;
	int dummy; 

	struct timeval kern_tv = {3, 4}, *tv = &kern_tv; 
	struct {
		char* data;
		int struct_type;
	} tv_struct_data;
	struct struct_logger_data tv_logger_data;
	struct timezone kern_tz = {5, 6}, *tz = &kern_tz; 
	struct {
		char* data;
		int struct_type;
	} tz_struct_data;
	struct struct_logger_data tz_logger_data;

	tv_struct_data.data = (char*)&kern_tv;
	tv_struct_data.struct_type = 0;
	tv_logger_data.param_index = 0;
	tv_logger_data.struct_type = 0;

	tz_struct_data.data = (char*)&kern_tz;
	tz_struct_data.struct_type = 1;
	tz_logger_data.param_index = 1;
	tz_logger_data.struct_type = 1;

	reset_all_vars();
	set_var_indexed_field_callback(VT_PARAM, 0, global_struct_field_param_callback, DT_INT, (void*)&tv_struct_data);
	set_var_indexed_field_callback(VT_PARAM, 1, global_struct_field_param_callback, DT_INT, (void*)&tz_struct_data);
	
	action = sct_filter_syscall(rules, 78, &rule_id);

	CHECK(action == act_to_match, 1); 

	if (action) {
		sct_tracker_perform_action(
			action, RW_BEFORE, rule_id, 78, &dummy, "gettimeofday",
			"S S" , tv, (void*)&tv_logger_data, tz, (void*)&tz_logger_data);
	}	
}

int test_struct(void)
{
	sct_filter_node_t* f1, *f2, *m; 
	sct_action_t* act;
	sct_rule_list_t* rules = sct_create_tracking_rule_list(); 

	init_struct_data(); 
	
	f1 = sct_make_int_val(3); 
	f2 = sct_make_var_indexed_field_val(VT_PARAM, 0, 0, DT_INT, 0); 
	m = sct_make_equal(f1, f2); 

	act = sct_make_action_log(4, "params: '%params'\n"); 
	CHECK(sct_add_tracking_rule(rules, -1, m, act), 0); 

	test_struct_syscall(rules, act); 
	test_struct_syscall(rules, act); 
	test_struct_syscall(rules, act); 
	test_struct_syscall(rules, act); 
	test_struct_syscall(rules, act); 

	sct_delete_tracking_rule_list(rules); 

	cleanup_struct_data(); 

	return 1; 
}
