/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2003 Muli Ben-Yehuda, mulix@mulix.org         */
/* License: GNU General Public License                              */

/* 
 * $Id: tester.c,v 1.1 2003/01/31 18:35:32 mulix Exp $
 */ 

/* 
 * a testing program for the kernel module. to add more tests, write a
 * test function and add a call to it in main. your funtion should return
 * 0 if the test was succesful and anything else otherwise. 
 */

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>
#include <assert.h>
#include <signal.h>
#include <setjmp.h>
#include <getopt.h>

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "sct_rules/rule_engine_public.h"

#include "tester.h"
#include "version.h" 

const char* progname;
static sct_ctrl_impl_t impl; 
sct_ctrl_t ctrl;
static char* dirname; 
static char* topdir; 
static int dir_did_cleanup; 
static int comm_initialized; 
int debug = 0; 

static void dir_setup(void), dir_cleanup(void); 
static void comm_setup(void); 
static void comm_cleanup(sct_ctrl_t ctrl); 
void cleanup(void); 

/* we need to protect from SIGINT, SIGQUIT and SIGTERM */ 
struct sigsave {
	int signum; 
	struct sigaction sigact; 
} oldsigs[3]; 

sigjmp_buf env; 
static void sighandler(int signum)
{
	fprintf(stderr, "sighandler: caught signal %d\n", signum); 
	
	siglongjmp(env, 1);
}

static void signals_setup(void)
{
        /* we need to protect from SIGINT, SIGQUIT and SIGTERM */ 
	struct sigaction act; 
	
	memset(&act, 0, sizeof(act)); 
	memset(&oldsigs, 0, sizeof(oldsigs)); 

	act.sa_handler = &sighandler; 
	sigemptyset(&act.sa_mask); 

	if (sigaction(SIGINT, &act, &oldsigs[0].sigact)) {
		perror("sigaction"); 
		return; 
	}
	oldsigs[0].signum = SIGINT; 

	if (sigaction(SIGQUIT, &act, &oldsigs[1].sigact)) {
		perror("sigaction"); 
		return; 
	}
	oldsigs[1].signum = SIGQUIT; 
	
	if (sigaction(SIGTERM, &act, &oldsigs[2].sigact)) {
		perror("sigaction"); 
		return; 
	}
	oldsigs[2].signum = SIGTERM; 
}

static void signals_cleanup(void)
{
	int i; 
	for (i = 0; i < sizeof(oldsigs)/sizeof(oldsigs[0]); ++i) {
		if (oldsigs[i].signum) {
			if (sigaction(oldsigs[i].signum, 
				      &oldsigs[0].sigact, 
				      NULL)) {
				perror("sigaction"); 
			}
		}
	}
}


void fatal(const char* func, const char* file,
	   unsigned int line, const char* fmt, ...)
{
	static char FAILED[] = "TEST FAILED"; 
	va_list ap; 

	fprintf(stdout, "[%s:%d:%s] ",file, line, func);
    
	va_start(ap, fmt);
	vfprintf(stdout, fmt, ap);
	va_end(ap);

	cleanup(); 

	assert(!FAILED); 
}

void logerror(const char* func, const char* file,
	      unsigned int line, const char* fmt, ...)
{
	va_list ap; 
    
	printf("[%s:%d:%s] ",file, line, func);

	va_start(ap, fmt);
	vfprintf(stdout, fmt, ap);
	va_end(ap);
}

static void comm_setup(void)
{
	/* setup module */
	ctrl = sct_ctrl_init(impl);
	if (!ctrl){
		die("failed to init sct_ctrl_lib\n");
	}
	
	/* del module rules */
	if (!sct_ctrl_del_all_rules(ctrl)){
		die("failed to delete all rules\n");
	}

	comm_initialized = 1; 
}

static void comm_cleanup(sct_ctrl_t ctrl)
{
	if (comm_initialized)
		sct_ctrl_cleanup(ctrl); 
}

static void dir_setup(void)
{
	struct stat buf; 
	char tmpl[32];
	int tmpfd; 

	topdir = getcwd(NULL, 0); 
	if (!topdir) 
		die("getcwd failed"); 

	snprintf(tmpl, sizeof(tmpl), "%sXXXXXX", "testdir");

	/* can't use mkdtemp here, unfortunately, due to being too new */ 
	if ((tmpfd = mkstemp(tmpl)) < 0)
		die("mkstemp: %s", strerror(errno));
	/* save the file name */ 
	if (!(dirname = strdup(tmpl)))
		die("strdup"); 
	/* close and unlink the file fd mkstemp returne to us -  */
	/* we need a directory, not a file. */ 
	close(tmpfd); 
	unlink(dirname); 

	if (mkdir(tmpl, 0700))
		die("mkdir: %s", strerror(errno));
	if (stat(dirname, &buf))
		die("stat: %s", strerror(errno)); 
	if ((!S_ISDIR(buf.st_mode)) || (buf.st_mode & (S_IRWXG | S_IRWXO)))
		die("%s is not a dir or wrong permissions"); 

	if (chdir(dirname))
		die("chdir: %s", strerror(errno)); 
}

static void dir_cleanup(void)
{
	DIR* d; 
	struct dirent* e; 

	if (dir_did_cleanup)
		return; 

	if (!(d = opendir("./"))) {
		perror("opendir"); 
		return; 
	}
	
	/* we ignore errors */ 
	while ((e = readdir(d))) {
		if (!strcmp(e->d_name, ".") || !strcmp(e->d_name, ".."))
			continue; 

		if (remove(e->d_name)) {
			perror("remove"); 
			continue; 
		}

		/* we removed an entry, so rewind the dir stream */ 
		rewinddir(d); 
	}

	closedir(d); 

	/* back to our starting directory */ 
	chdir(topdir); 

	if (rmdir(dirname))
		perror("rmdir"); 

	free(dirname); 
	dirname = NULL; 
	free(topdir); 
	topdir = NULL; 

	dir_did_cleanup = 1; 
}

void cleanup(void)
{
	dir_cleanup(); 
	
	comm_cleanup(ctrl); 
	
	signals_cleanup(); 
}

static int usage(const char* progname)
{
	static char usagestr[] = 
		"%s (%s) is a stress test program for the system call tracker\n"
		"Usage: %s [OPTIONS]\n"
		"\n"
		"Options: \n"
		"  -d, --device_file           use the device file for control\n"
                "  -s, --sysctl                use sysctl for control\n"
		"  -c, --control device_file   use the device file for control\n"
                "  -c, --control sysctl        use sysctl for control\n"
		"  -t, --test 'test_name'      run only the test 'test_name'\n"
		"  -m, --subsystem module      run only the kernel module tests\n"
		"  -m, --subsystem rules       run only the rules library tests\n"
		"  -l, --list                  print the list of tests\n"; 

	fprintf(stdout, usagestr, progname, VERSION_STRING, progname); 
	exit(EXIT_FAILURE); 
}

static void run_test(struct test* p)
{
	unsigned int subsys = (p->flags & SCT_SUBSYS_BITS); 
	unsigned int repeatable = (p->flags & SCT_TEST_REPEATABLE_BITS); 
	
	fprintf(stdout, "starting test '%s' (subsystem %s(%d)) [%s]\n", 
		p->name, subsystem_as_str(subsys), subsys, 
		repeatable ? "REPEATABLE" : ""); 


	/* initialize comm to the module - if necessary */ 
	if (!comm_initialized && ((subsys & SCT_SUBSYS_MODULE)))
		comm_setup(); 

	assert(p->name); 
	assert(p->test_fn); 

	if (!p->test_fn()) 
		die("test %s failed\n!", p->name); 

	fprintf(stdout, "test '%s' succeeded\n", p->name);
	fflush(stdout); 
}

static void run_one_test(const char* test_name)
{
	struct test* p; 
	unsigned int found; 

	assert(test_name); 

	found = find_test(test_name, &p); 

	switch(found) {
	case 0: 
		fprintf(stderr, "no such test '%s'\n", test_name); 
		break; 
	case 1: 
		run_test(p); 
		break; 
	default: 
		fprintf(stderr, "%d tests start with '%s'. You're going to "
			"have to be more specific...\n", found, test_name); 
		break; 
	} 
}

void run_all_tests(void)
{
	struct test* p; 
	
	/* run all of the tests */ 
	for (p = &tests[0]; p && p->test_fn; ++p) {
		run_test(p); 
	}
}

void run_subsystem_tests(const char* subsys_str)
{
	struct test* p; 
	unsigned int target = str_to_subsystem(subsys_str); 
	unsigned int subsys; 

	/* run all of the tests */ 
	for (p = &tests[0]; p && p->test_fn; ++p) {
		subsys = p->flags & SCT_SUBSYS_BITS; 

		if ((subsys & target) == subsys)
			run_test(p); 
	}
}

static int parse_cmdline(int argc, char* argv[], sct_ctrl_impl_t* impl, 
			 char** test_string, char** subsystem_string, 
			 int* list)
{
	while (1) {
		int c;
		int option_index = 0;
		static struct option long_options[] = {
			{"device_file", 0, NULL, 'd'}, 
			{"sysctl", 0, NULL, 's'},
			{"control", 1, NULL, 'c'},
			{"test", 1, NULL, 't'}, 
			{"subsystem", 1, NULL, 'm'}, 
			{"list", 0, NULL, 'l'}, 
			{0, 0, 0, 0}
		};
		
		c = getopt_long (argc, argv, "dsc:t:m:l",
				 long_options, &option_index);
		if (c == -1)
			break;
		
		switch (c) {
		case 0:
			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
			break;
		   
		case 'd':
			*impl = SCT_CTRL_DEVICE_FILE; 
			break;
		case 's':
			*impl = SCT_CTRL_SYSCTL; 
			break; 
		case 'c':
			if (!strcmp(optarg, "device_file"))
				*impl = SCT_CTRL_DEVICE_FILE; 
			else if (!strcmp(optarg, "sysctl"))
				*impl = SCT_CTRL_SYSCTL;
			else 
				usage(progname); 
			break; 
		case 't':
			*test_string = strdup(optarg); 
			break;
		case 'm': 
			*subsystem_string = strdup(optarg); 
			break; 
		case 'l':
			*list = 1; 
			break; 
		case '?':
			usage(progname); 
			break;
		default:
			printf ("?? getopt returned character code 0%o ??\n", c);
		}
	}
	
	if (optind < argc) 
		usage(progname);
	
	return 0; 
}

int main(int argc, char* argv[])
{
	char* test_string = NULL; 
	char* subsystem_string = NULL; 
	int list = 0; 

	impl = SCT_CTRL_DEFAULT; 

	progname = argv[0];

	parse_cmdline(argc, argv, &impl, &test_string, &subsystem_string, &list); 

	signals_setup(); 

	if (sigsetjmp(env, 1))
		goto shutdown;

	dir_setup(); 

	if (list)
		list_tests(stdout); 

	else if (test_string)
		run_one_test(test_string); 

	else if (subsystem_string)
		run_subsystem_tests(subsystem_string); 

	else 
		run_all_tests(); 
   
 shutdown:
	
	cleanup(); 

	free(test_string); 

	free(subsystem_string); 

	return 0;
}

