/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 * $Id: sct_logctrl.c,v 1.14 2003/01/31 14:08:47 mulix Exp $
 */

#include <unistd.h>
#include <getopt.h> 
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <limits.h> 
#include <errno.h> 

#include "module/common/module_interface.h"

#include "version.h" 

const char* prog; 

static void usage(const char* name)
{
	char usage[] = 
		"%s is a log device file configuration and control utility for "
		"the system call tracker.\n"
		"\n"
		"Usage: %s [OPTIONS]... [COMMAND]\n"
		"\n"
		"Options: \n"
		"  -v, --verbose                     verbose messages.\n"
		"  -f, --fd fdnum                    use this (pre opened)file descriptor.\n"
		"  -l, --logformat log_format        events logged to the device file should \n"
		"                                    use this format.\n"
		"  -b, --beforelogformat log_format  before events logged to the device file\n"
		"                                    should use this format.\n"
		"  -a, --afterlogformat log_format   after events logged to the device file\n"
		"                                    should use this format.\n"
		"  -s, --size sz                     use a buffer of this size in the log device.\n"
                "  -m, --max-rec-len len             when printing long strings, only print this much.\n"
		"  -h, --help                        print this usage information.\n"
		"\n"
		"Commands: \n"
		"  drop_new             on buffer overflow, drop new data.\n"
		"  drop_old             on buffer overflow, drop old data.\n"
		"  flush                flush all data from the buffer.\n"
		"  enable               enable the log device.\n"
		"  disable              disable the log device.\n"; 

	fprintf(stdout, usage, name, name); 
		
	exit(EXIT_FAILURE);
}

static void parse_options(int argc, char** argv, int* fd, size_t* sz, 
			  unsigned int* max_rec_len, char** bef_logformat, 
			  char **aft_logformat, int* cmd, int* verbose)
{
	int option_index = 0;
	unsigned long res; 
	char* scmd; 

	if (argc < 2) 
		usage(prog); 

	while(1){ 
		/* structure with all recognised options for sct_config */
		static struct option long_options[] = {
			{"verbose", 0, NULL, 'v'},
			{"fd", 1, NULL, 'f'}, 
			{"size", 1, NULL, 's'}, 
			{"logformat", 1, NULL, 'l'},
			{"beforelogformat", 1, NULL, 'b'},
			{"afterlogformat", 1, NULL, 'a'},
			{"max-rec-len", 1, NULL, 'm'}, 
			{"help", 0, NULL, 'h'}, 
			{0, 0, NULL, 0}
		};
		int c;
		c = getopt_long (argc, argv, "+vf:s:l:b:a:m:h", long_options, &option_index);

		switch (c) {
		case 0: /* long option specified */ 
			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
			break;
		case 'v': 
			*verbose = 1; 
			break; 
		case 'f':
			if ((res = strtoul(optarg, NULL, 10)) == ULONG_MAX) {
				fprintf(stderr, "invalid fd specified\n"); 
				usage(prog); 
			}
			*fd = res; 
			break; 
		case 's':
			if ((res = strtoul(optarg, NULL, 10)) == ULONG_MAX) {
				fprintf(stderr, "invalid buffer size specified\n"); 
				usage(prog); 
			}
			*sz = (size_t)res; 
			break; 
		case 'm': 
			if ((res = strtoul(optarg, NULL, 10)) == ULONG_MAX) {
				fprintf(stderr, "invalid max record length specified\n"); 
				usage(prog); 
			}
			*max_rec_len = (unsigned int) res; 
			break; 
		case 'l':
			if (!(*bef_logformat = strdup(optarg))) {
				fprintf(stderr, "out of memory!\n"); 
				exit(EXIT_FAILURE); 
			}
			if (!(*aft_logformat = strdup(optarg))) {
				fprintf(stderr, "out of memory!\n"); 
				exit(EXIT_FAILURE); 
			}
			break; 
		case 'b':
			if (!(*bef_logformat = strdup(optarg))) {
				fprintf(stderr, "out of memory!\n"); 
				exit(EXIT_FAILURE); 
			}
			break; 
		case 'a':
			if (!(*aft_logformat = strdup(optarg))) {
				fprintf(stderr, "out of memory!\n"); 
				exit(EXIT_FAILURE); 
			}
			break; 
		case 'h': 
			usage(prog); 
			break; /* not reached */ 
		case '?': /* unrecognised option */
			/* fall through */
		default:
			c = -1;
			break;
		}
		if(c == -1) 
			break;  /* no more options */
	}

	argc -= optind; 
	if (argc > 1)  /* there should be one or no commands remaining */ 
		usage(prog); 
	if (argc == 1) {

		scmd = argv[optind]; 
	
		if (!strcasecmp(scmd, "drop_old")) 
			*cmd = CTL_IOC_DROP_OLD;
		else if (!strcasecmp(scmd, "drop_new")) 
			*cmd = CTL_IOC_DROP_NEW; 
		else if (!strcasecmp(scmd, "flush"))
			*cmd = CTL_IOC_FLUSH; 
		else if (!strcasecmp(scmd, "enable"))
			*cmd = CTL_IOC_ENABLE_DEV; 
		else if (!strcasecmp(scmd, "disable"))
			*cmd = CTL_IOC_DISABLE_DEV; 
		else
			usage(prog); 
	}

}

static const char* cmd_to_str(int cmd)
{
	switch(cmd) {
	case CTL_IOC_DROP_OLD:
		return "drop old"; 
	case CTL_IOC_DROP_NEW: 
		return "drop new"; 
	case CTL_IOC_FLUSH: 
		return "flush"; 
	case CTL_IOC_ENABLE_DEV:
		return "enable"; 
	case CTL_IOC_DISABLE_DEV:
		return "disable"; 
	case CTL_IOC_RESIZE:
		return "resize"; 
	case CTL_IOC_BEFORE_LOG_FORMAT: 
		return "before log format"; 
	case CTL_IOC_AFTER_LOG_FORMAT: 
		return "after log format"; 
	case CTL_IOC_MAX_REC_SIZE:
		return "max record length"; 
	default:
		return "(unknown)"; 
	}
}

static void translate_special_chars(char** lf)
{
	/* translate literal \r, \n, \t to their special meanings */ 
	/* surely there's a library function somwhere to do this... */ 
	char* pos = *lf; 
	char* new = malloc(strlen(*lf) + 1); 
	char* w = new; 
	if (!new)
		return; 

	memset(new, 0, strlen(*lf) + 1); 

	while (*pos) {
		if (*pos != '\\') {
			*w++ = *pos++;
			continue; 
		}
		switch (*++pos) {
		case 't': 
			*w++ = '\t'; 
			break; 
		case 'n': 
			*w++ = '\n'; 
			break; 
		case 'r': 
			*w++ = '\r'; 
			break; 
		case '0':
			*w++ = '\0';
			break; 
		}
		++pos;
	}
	*w = '\0'; 
	free(*lf); 
	*lf = new; 
}

static int do_cmd(int fd, int cmd, void* param, int verbose)
{	
	int ret;
	char fmt[] = "cmd(%s): %d(%s)\n";
	
	if (param)
		ret = ioctl(fd, cmd, param);
	else
		ret = ioctl(fd, cmd);
	
	if (ret || verbose)                         
		fprintf(stdout, fmt, cmd_to_str(cmd), 
			ret, ret ? strerror(errno) : "success");
	return ret; 
}

int main(int argc, char** argv)
{
	int cmd = -1; 
	int fd = -1; 
	size_t sz = 0; 
	char* blf = NULL; 
	char* alf = NULL;
	int ret = 0; 
	int my_fd = 0; 
       /* we can't use 0 because 0 is a valid max_rec_len (unlimited) */ 
	unsigned int max_rec_len = (~0UL); 
	int verbose = 0; 

	prog = argv[0]; 

	parse_options(argc, argv, &fd, &sz, &max_rec_len, &blf, &alf, &cmd, &verbose); 

	if (fd == -1) {
		my_fd = 1; 
		if ((fd = open("/dev/sct_log", 0)) < 0) {
			perror("open /dev/sct_log"); 
			fprintf(stderr, "are the kernel modules loaded?\n"); 
			exit(EXIT_FAILURE); 
		}
	}

	/* ok, we have an fd. First deal with the 'size' and 'log format' commands, */
	/* because we might have 'enable' in cmd, which should be done last */ 

	/* resize */ 
	if (sz != 0) {
		if (verbose)
			fprintf(stdout, "resize - %d\n", (int)sz); 
		ret |= do_cmd(fd, CTL_IOC_RESIZE, (void*)sz, verbose); 
	}

	/* max record len */ 
	if (max_rec_len != (~0UL)) {
		if (verbose)
			fprintf(stdout, "max record len - %u\n", max_rec_len); 
		ret |= do_cmd(fd, CTL_IOC_MAX_REC_SIZE, 
			      (void*)max_rec_len, verbose); 
	}
	
	/* change before log format */ 
	if (blf) {
		if (verbose)
			fprintf(stdout, "before log format - '%s\n", blf); 
		translate_special_chars(&blf); 
		ret |= do_cmd(fd, CTL_IOC_BEFORE_LOG_FORMAT, blf, verbose); 
			      
		free(blf);
	}

	/* change after log format */ 
	if (alf) {
		if (verbose)
			fprintf(stdout, "after log format - '%s'\n", alf); 
		translate_special_chars(&alf); 
		ret |= do_cmd(fd, CTL_IOC_AFTER_LOG_FORMAT, alf, verbose); 
		free(alf);
	}

	/* other commands */ 
	if (cmd != -1)
		ret |= do_cmd(fd, cmd, NULL, verbose); 
	
	if (my_fd) 
		close(fd); 

	return (ret ? EXIT_FAILURE : EXIT_SUCCESS); 
}
